/*******************************************************************************
+
+  LEDA 5.0  
+
+
+  isug_gw_demo.c
+
+
+  Copyright (c) 1995-2004
+  by Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


#include <LEDA/graphics/graphwin.h>

using namespace leda;

#if defined(LEDA_STD_IO_HEADERS)
using std::cout;
using std::cin;
using std::cerr;
using std::ostream;
using std::istream;
using std::endl;
#endif

#include <AGD/isugiyama.h>
#include <AGD/GraphCopy.h>
#include <AGD/Hierarchy.h>
#include <AGD/GraphWinInterface.h>
#include <AGD/GraphCopyLI.h>
#include <AGD/FastHierarchyLayout.h>


#define DEBUG

/*-----------------------------------------------------------------------------*/

//Variables are global because of graphwin interface
	
graph G;
InteractiveSugiyama IS;

/*-----------------------------------------------------------------------------*/

void set_positions_in_hierarchy
(const graph& G, const GraphCopy& GC, Hierarchy& H, const InteractiveSugiyama& IS)
{	
	node_map<int> gc_node_positions(GC,-1);
	node v;forall_nodes(v,G) {
		gc_node_positions[GC.copy(v)]=IS.node_pos(v);
	}

	edge e;forall_edges(e,G) {
		list<edge> chain=GC.chain(e);
		edge ec;forall(ec,chain) {
			node s=GC.source(ec);node t=GC.target(ec);
			list<two_tuple<int,int> > lp=IS.edge_positions(e);
			two_tuple<int,int> p;forall(p,lp) {
				if (p.first()==H.rank(s)) gc_node_positions[s]=p.second();
				if (p.first()==H.rank(t)) gc_node_positions[t]=p.second();
			}
		}
	}

	for (int r=0;r<H.number_of_layers();r++) {H[r].set_pos(gc_node_positions);}
}

/*-----------------------------------------------------------------------------*/

void update_drawing
(const graph& G, const InteractiveSugiyama& IS, GraphWin& gw)
{	
	GraphCopy GC_local(G);
	Hierarchy H_local(GC_local,IS.node_rank());
	set_positions_in_hierarchy(G,GC_local,H_local,IS);

	GraphWinInterface gwi_local(gw);
	GraphCopyLI gcli_local(GC_local,gwi_local);

	FastHierarchyLayout L_local;
	L_local.node_distance(IS.node_dist());
	L_local.layer_distance(IS.level_dist());
	L_local.fixed_layer_distance(IS.fixed_level_dist());

	L_local.call(GC_local,H_local,gcli_local);
	gcli_local.update();
	gw.zoom_graph();
}

/*-----------------------------------------------------------------------------*/

typedef int (*draw_alg) (const graph& G);

void call_draw_alg(GraphWin& gw, draw_alg alg)
{
	if (!alg(G)) return;
	update_drawing(G,IS,gw);
	cout << "crossings=" << IS.number_of_crossings() << endl;
}

/*-----------------------------------------------------------------------------*/

void sugoptions (GraphWin& gw) {
	int ranking=IS.ranking();
	int cross_min=IS.cross_min();
	int cycle_removal=IS.cycle_removal();
	int width=IS.width();
	int  runs=IS.runs();
	int  fails=IS.fails();
	bool init_order=IS.init_order();
	bool transpose=IS.transpose();
	double node_dist=IS.node_dist();
	double level_dist=IS.level_dist();
    bool fixed_level_dist=IS.fixed_level_dist();

	window w(500,200,"Sugiyama Options");
	w.text_item("\\bf\\blue Rank Assignment");
	w.choice_item ("method:", ranking, "Longest-Path", "Coffman-Graham", "DFS");
	w.int_item ("width:", width);
	w.choice_item ("cycle removal:", cycle_removal, "DFS", "Greedy");

	w.text_item("\\bf\\blue Crossing Minimization");
	w.choice_item ("method:", cross_min, "Barycenter", "Median", "Weighted-Median", "Split", "Sifting");
	w.int_item ("runs:", runs);
	w.int_item ("fails:", fails);
	w.bool_item ("initial order:", init_order);
	w.bool_item ("transpose:", transpose);

	w.text_item("\\bf\\blue Computation of Coordinates");
	w.real_item ("node distance:", node_dist);
	w.real_item ("level distance:", level_dist);
	w.bool_item ("fixed level distance:", fixed_level_dist);

	w.button ("Ok");
	w.display();
	w.read_mouse();
	IS.set_options(ranking,cross_min,cycle_removal,width,runs,
		fails,init_order,transpose,node_dist,level_dist,fixed_level_dist);
}

/*-----------------------------------------------------------------------------*/

int redraw ( const graph& G ) {return IS.redraw(G);}
int update ( const graph& G ) {return IS.update(G);}

int main()
{
	GraphWin gw(G);
	int draw_menu=gw.add_menu("Draw Menu");
	gw_add_call(gw,redraw,call_draw_alg,"Sugiyama",draw_menu);
	gw.add_simple_call(sugoptions,"Sugiyama Options",draw_menu);
	gw_add_call(gw,update,call_draw_alg,"Update Drawing",draw_menu);
	gw.open();

	return 0;
}
