/*******************************************************************************
+
+  LEDA 5.0  
+
+
+  spqr_viewer.c
+
+
+  Copyright (c) 1995-2004
+  by Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



#include <LEDA/graphics/graphwin.h>
#include <LEDA/graphs/edge_set.h>
#include <LEDA/graphs/node_set.h>
#include <LEDA/system/assert.h>


using namespace leda;

#if defined(LEDA_STD_IO_HEADERS)
using std::cout;
using std::cin;
using std::cerr;
using std::ostream;
using std::istream;
using std::endl;
#endif

#include <AGD/PlaneGraphCopy.h>
#include <AGD/StPlanarLayout.h>
#include <AGD/generators.h>
#include <AGD/GraphWinInterface.h>
#include <AGD/GraphWinAnimInterface.h>
#include <AGD/PlanarSPQRTree.h>





/*
class PlanarSPQRTree;
class Skeleton;
*/

class GraphObserver;


struct SkeletonOpts
{
	SkeletonOpts();

	color colorRefEdge, colorRealEdge, colorVirtualEdge;
	int widthRefEdge, widthRealEdge, widthVirtualEdge;
	gw_edge_style styleRefEdge, styleRealEdge, styleVirtualEdge;
	int animSteps;
};


class StPlaneGraph
{
	graph G;
	PlaneGraphCopy PG;
	PlanarSPQRTree *spqr;
	node_array<int> stnum;
	edge theOldRefEdge;

	node_set *markedNodes;
	edge_set *markedEdges;

	StPlanarLayout stPlanarLayouter;
	SkeletonOpts _skeletonOpts;

	list<GraphObserver *> observers;

public:
	enum ChangeStatus { topologyChanged, topologyEdited, embeddingChanged,
		referenceEdgeChanged, stLayoutOptionsChanged,
		skeletonOptsChanged, permMarkGraphChanged, permMarkGraphCleared };

	StPlaneGraph() : spqr(nil), theOldRefEdge(nil), 
                         markedNodes(nil), markedEdges(nil) { }
	~StPlaneGraph();

	const graph &getGraph() const {
		return G;
	}

	graph &getGraph() {
		return G;
	}

	PlaneGraphCopy &getPlaneGraph() {
		return PG;
	}

	const PlaneGraphCopy &getPlaneGraph() const {
		return PG;
	}

	const PlanarSPQRTree &spqrTree() const {
		return *spqr;
	}

	PlanarSPQRTree &spqrTree() {
		return *spqr;
	}

	edge refEdge() const;
	edge oldRefEdge() const {
		return theOldRefEdge;
	}

	int stNum(node v) const {
		return stnum[v];
	}

	bool isGraphValid() const {
		return spqr != nil;
	}

	const node_set *permanentMarkedNodesGraph() {
		return markedNodes;
	}

	const edge_set *permanentMarkedEdgesGraph() {
		return markedEdges;
	}

	bool isPermanentMarkedGraph() const;

	void callStLayout(GraphCopy &GC, edge e_st, LayoutInterface &gwi);
	void setStLayoutOptions(StPlanarGridLayout::CompactionType compaction,
		double separation);
	void getStLayoutOptions(StPlanarGridLayout::CompactionType &compaction,
		double &separation);

	void skeletonOpts(const SkeletonOpts &opts);
	const SkeletonOpts &skeletonOpts() const {
		return _skeletonOpts;
	}

	void subscribe(GraphObserver &observer);
	void unsubscribe(GraphObserver &observer);

	void topologyHasChanged(edge eRef, const LayoutInterface *li);
	void topologyHasChanged(edge eRef, const LayoutInterface *li, bool valid);
	void reverse(node vT);
	void swap(node vT, edge e1, edge e2);
	void embeddingHasChanged();
	void changeReferenceEdge(edge newRefEdge);
	void reverseRefEdge();
	void setPermanentMarkGraph(list<node> &nodes, list<edge> &edges);
	void unsetPermanentMarkGraph();

protected:
	edge find_e_st(edge eRef);
	void make_st_graph();
	void notifyAll(ChangeStatus status);
};

class GraphObserver
{
	friend void StPlaneGraph::subscribe(GraphObserver&);
	friend void StPlaneGraph::unsubscribe(GraphObserver&);

	StPlaneGraph *observable;
	list_item pos;

public:
	GraphObserver() : observable(nil), pos(nil) { }
	virtual ~GraphObserver() {
		if (observable)
			observable->unsubscribe(*this);
	}

	virtual void notify(StPlaneGraph::ChangeStatus status) { }
};

#include <LEDA/graphwin.h>
#include <AGD/StPlanarLayout.h>


class SPQRTreeWindow;


struct GraphOpts
{
	GraphOpts();

	color edgeColor, refEdgeColor, markedEdgeColor, unmarkedEdgeColor;
	color nodeColor, markedNodeColor, unmarkedNodeColor;
	int edgeWidth, refEdgeWidth, markedEdgeWidth, unmarkedEdgeWidth;
};


class GraphWindow : public GraphObserver, protected GraphWin
{
	StPlaneGraph &planeGraph;
	SPQRTreeWindow *spqrWindow;

	bool _isOpen;
	bool genMultiEdges;

	GraphOpts opts;

public:
	GraphWindow(StPlaneGraph &theGraph);
	~GraphWindow();

	bool eventHandler();

	void display();

protected:
	void notify(StPlaneGraph::ChangeStatus status);
	void draw_graph(bool anim);
	void disableInput();
	void enableInput();

	void setupNodesEdges();
	void setupRefEdge();
	void resetupRefEdge();

	void doOpen();
	void doSave();
	void doGenerate();
	void doEdit();
	void doExit();
	void doSelectRef();
	void doUpsideDown();
	void doSetup();
	void doLayoutOpt();
	void doAnimOpt();

	static void actionSelectRef(GraphWin& gw, const point& p);
	static void showEdge(GraphWin& gw, const point& p);

	static void deleteWindowHandler(window *win);
	static void openCaller (GraphWin &gw);
	static void saveCaller (GraphWin &gw);
	static void generateCaller (GraphWin &gw);
	static void editCaller (GraphWin &gw);
	static void exitCaller (GraphWin &gw);
	static void selectRefCaller (GraphWin &gw);
	static void upsideDownCaller (GraphWin &gw);
	static void unmarkCaller (GraphWin &gw);
	static void setupCaller (GraphWin &gw);
	static void layoutOptCaller (GraphWin &gw);
	static void animOptCaller (GraphWin &gw);
	static void helpMouseCaller (GraphWin &gw);
	static void aboutCaller (GraphWin &gw);
};



#include <LEDA/graphwin.h>

class SPQRTreeWindow;


class SkeletonWindow : public GraphObserver, protected GraphWin
{
	SPQRTreeWindow &parent;
	StPlaneGraph &planeGraph;
	graph S;
	node_array<node> vMtoS, vStoM;
	edge_array<edge> eMtoS, eStoM;

	node vT;
	bool isOpen;
	edge oldRefEdge, oldRefEdgeM;

public:
	list_item pos;
	node vCaller;

	SkeletonWindow(SPQRTreeWindow &spqrWin, StPlaneGraph &theGraph, node vT);
	~SkeletonWindow() { }

	node treeNode() const {
		return vT;
	}

	bool eventHandler(int event, window *wp, int but, double x, double y);
	void display(int x, int y);

	void setFocus() {
		get_window().set_focus();
	}

	void disableInput();
	void enableInput();

protected:
	void notify(StPlaneGraph::ChangeStatus status);

	string winTitle(const StPlaneGraph &stGraph, node v);
	void buildGraph();
	bool adaptDirections();
	void drawGraph(bool anim);

	void setup();
	void setupRefEdge(edge eS);
	void setupRealEdge(edge eS);
	void setupVirtualEdge(edge eS);
	void adaptRefEdge();

	void doSave();
	void doReverse();
	void doSwap();
	void doClose();

	static void showSkeleton(GraphWin& gw, const point& p);
	static void showExpansion(GraphWin& gw, const point& p);
	static void showVertex(GraphWin& gw, const point& p);

	static void deleteWindowHandler(window *win);
	static void saveCaller (GraphWin &gw);
	static void reverseCaller (GraphWin &gw);
	static void swapCaller (GraphWin &gw);
	static void showPertinent (GraphWin &gw);
	static void closeCaller (GraphWin &gw);
	static void helpMouseCaller (GraphWin &gw);

	edge read_edge(const string msg);
};


//---------------------------------------------------------
// GraphWindow
//
// author: Carsten Gutwenger
//---------------------------------------------------------

#include <LEDA/graphwin.h>
#include <AGD/StPlanarLayout.h>
#include <AGD/TreeLayout.h>

class SPQRTree;


class SPQRTreeWindow : public GraphObserver, protected GraphWin
{
	friend void SkeletonWindow::setup();
	friend void SkeletonWindow::setupRefEdge(edge eS);
	friend void SkeletonWindow::setupRealEdge(edge eS);
	friend void SkeletonWindow::setupVirtualEdge(edge eS);

	StPlaneGraph &planeGraph;
	GraphCopy GC;

	node_map<SkeletonWindow *> skeletonWin;
	list<SkeletonWindow *> skeletonWindows;
	int skCount;

	// options for SPQR-tree
	gw_node_shape shapeSNode, shapePNode, shapeRNode;
	color colorSNode, colorPNode, colorRNode;
	TreeLayout treeLayouter;

public:
	SPQRTreeWindow(StPlaneGraph &theGraph);
	~SPQRTreeWindow();

	void eventHandler(int event, window *wp, int but, double x, double y);

	void display(int x, int y);
	
	void disableInput();
	void enableInput();

	void openSkeleton(node v);
	void showPertinent(node v);

protected:
	static void buildPertinent(const SPQRTree &T, node v, list<node> &nodes,
		list<edge> &edges);

	static void actionOpenSkeleton(GraphWin& gw, const point& p);
	static void actionShowPertinent(GraphWin& gw, const point& p);

	void doSave();
	void doCloseAll();
	void doAllToFront();
	void doInfo();
	void doSetup();
	void doSkeletonOpt();
	void doAnimOpt();

	void notify(StPlaneGraph::ChangeStatus status);

	void setupTreeGraph();
	void draw_tree(bool anim);
	void adaptDirections();
	void closeAllSkeletons();
	void closeSkeletonWindow(SkeletonWindow *win);

	static void randomEmbed (GraphWin &gw);
	static void saveCaller (GraphWin &gw);
	static void infoCaller (GraphWin &gw);
	static void allToFrontCaller (GraphWin &gw);
	static void closeAllCaller (GraphWin &gw);
	static void setupCaller (GraphWin &gw);
	static void skeletonOptCaller (GraphWin &gw);
	static void animOptCaller (GraphWin &gw);
	static void helpMouseCaller (GraphWin &gw);
};


#include <LEDA/graphwin.h>
#include <AGD/GraphCopy.h>
#include <AGD/AgdKey.h>


class LayoutModule;
class AugmentationModule;
class SubgraphModule;

class StPlanarGridLayout;
class StPlanarLayout;
class SugiyamaLayout;
class FastHierarchyLayout;

class QuasiOrthogonalLayout;
class MixedModelLayout;
class PlanarSubgraph;


class EditWindow : protected GraphWin
{
	GraphCopy gc;
	GraphWin &gw;

	edge refEdge;

	color edgeColor, refEdgeColor;
	int edgeWidth, refEdgeWidth;

	bool rbg_multi, rbg_map, rst_multi;
	double rst_stprob;

	int sug_cross_min, sug_cycle_removal;

	StPlanarGridLayout *stPlanarGrid;
	StPlanarLayout *stPlanar;
	SugiyamaLayout *sugiyama;
	FastHierarchyLayout *fhl;

	QuasiOrthogonalLayout *quod;
	MixedModelLayout *mixed;
	PlanarSubgraph *planSub;

public:
	EditWindow(GraphWin &GW, edge eRef, color ec, int ew, color rec, int rew);
	~EditWindow();

	bool open();
	edge assignGraph();

protected:
	void setupRefEdge(edge e);
	void unsetupRefEdge(edge e);

	void doSetRefEdge();

	static bool del_edge_handler(GraphWin &gw, edge e);
	static void newGraph(GraphWin &GW);
	static void setRefEdgeHandler(GraphWin &GW);
	static void gw_open_node_menu(GraphWin& gw, const point& p);
	static void gw_open_edge_menu(GraphWin& gw, const point& p);
	static list_item min_bend_dist(const list<point>& P, const point& p);

	static string get_error_message(AgdKey p);
	static bool callAugment (EditWindow& GW, AugmentationModule& A);
	static bool callSubgraph (EditWindow& GW, SubgraphModule& S);
	static bool callLayout (EditWindow& GW, LayoutModule& L);

	static void embed_random(const graph& G, node_array<double>& x,
		node_array<double>& y);
	void gen_pre(int n, int m);
	void gen_post();
	static void genRandomGraph(GraphWin& gw1, void (*f)(graph&,int,int), string s);
	static void randomPlanarHandler (GraphWin& GW);
	static void randomPlanarBicHandler(GraphWin &gw1);
	static void randomPlanarTricHandler (GraphWin& GW);
	static void randomStPlanarHandler(GraphWin &gw1);
	static void randomSPHandler(GraphWin &gw1);

	static void make_yes_panel(panel& P, string s, bool proof=false);
	static void make_no_panel(panel& P, string s, bool proof=false);
	static void gw_test_simplicity(GraphWin& gw);
	static void gw_test_connectivity(GraphWin& gw);
	static void gw_test_biconnectivity(GraphWin& gw);
	static void gw_test_triconnectivity(GraphWin& gw);
	static void gw_test_planarity(GraphWin& gw);
	static void gw_test_spness(GraphWin& gw);
	static void gw_test_acyclicity(GraphWin& gw);
	static void gw_test_loopfree(GraphWin& gw);
	static void gw_test_st_graph(GraphWin& gw);

	static void gw_del_loops(GraphWin& gw);
	static void gw_del_isolated(GraphWin& gw);

	static void callPlanAug (GraphWin& GW);
	static void callMakeBic (GraphWin& GW);
	static void callPlanSub (GraphWin& GW);
	static void callMakeAcyclic (GraphWin& GW);
	static void callGreedyCycleRemoval (GraphWin& GW);
	static void callReducedDigraph (GraphWin& GW);

	static void callStPlanarGrid (GraphWin& GW);
	static void callStPlanar (GraphWin& GW);
	static void callSugiyama(GraphWin& GW);
	static void callMixedModel (GraphWin& GW);
	static void callQuod (GraphWin& GW);
	static void callPlanarization (GraphWin& GW);
};


enum { SAVE_MF,
       LOAD_GW, SAVE_GW, LOAD_GML, SAVE_GML, SAVE_PS, SAVE_TEX, SAVE_SCREEN,
                                                      FILE_ALL, FILE_CANCEL };

void gw_zoom_graph(GraphWin& gw, const point& p);
void gw_zoom_up(GraphWin& gw, const point& p);
void gw_zoom_down(GraphWin& gw, const point& p);

void del_selection(GraphWin& gw);


ostream &operator<<(ostream &os, edge e)
{
	if (e)
		os << "(" << index(source(e)) << "," << index(target(e)) << ")";
	else
		os << "nil";
	return os;
}


//---------------------------------------------------------
// Constructor / Destructor
//---------------------------------------------------------

GraphOpts::GraphOpts()
{
	edgeColor = black; refEdgeColor = red;
	markedEdgeColor = blue; unmarkedEdgeColor = grey2;

	nodeColor = ivory;
	markedNodeColor = blue; unmarkedNodeColor = white;

	edgeWidth = 1; refEdgeWidth = 2;
	markedEdgeWidth = 2; unmarkedEdgeWidth = 1;
}


GraphWindow::GraphWindow(StPlaneGraph &theGraph) :
	GraphWin(theGraph.getGraph(),window::screen_width()/2-6,window::screen_height()-55,
	"Graph Window"), planeGraph(theGraph),
	_isOpen(false), genMultiEdges(false)
{
	// setup GraphWin
	set_default_menu(M_NONE);
	clear_actions();
	set_show_status(false);

	// create menu
	int fileMenuId = add_menu("&File");
	add_simple_call(openCaller, "&Open", fileMenuId);
	add_simple_call(saveCaller, "&Save", fileMenuId);
	add_separator(fileMenuId);
	add_simple_call(generateCaller, "&Generate", fileMenuId);
	add_simple_call(editCaller, "&Edit", fileMenuId);
	add_separator(fileMenuId);
	add_simple_call(exitCaller, "E&xit", fileMenuId);

	int refEdgeMenuId = add_menu("&Reference Edge");
	add_simple_call(selectRefCaller, "&select", refEdgeMenuId);
	add_simple_call(upsideDownCaller, "&reverse", refEdgeMenuId);

    int windowMenuId = add_menu("&Window");
    add_simple_call(unmarkCaller, "unmark",       windowMenuId);
    add_separator(windowMenuId);
    add_simple_call(gw_center_graph, "center graph",       windowMenuId);
    add_simple_call(gw_zoom_graph,   "zoom to graph",      windowMenuId);
    add_simple_call(gw_zoom_area,    "zoom to area",       windowMenuId);
    add_separator(windowMenuId);
    add_simple_call(gw_zoom_up,      "zoom +",             windowMenuId);
    add_simple_call(gw_zoom_down,    "zoom -",             windowMenuId);

    int optionsMenuId = add_menu("&Options");
    add_simple_call(setupCaller, "&Nodes and Edges", optionsMenuId);
    add_simple_call(layoutOptCaller, "&Layout", optionsMenuId);
    add_simple_call(animOptCaller, "&Animation", optionsMenuId);

    int helpMenuId = add_menu("&Help");
    add_simple_call(helpMouseCaller, "&Mouse Buttons", helpMenuId);
	add_separator(helpMenuId);
    add_simple_call(aboutCaller, "&About", helpMenuId);

	// mouse action handlers
	set_action(A_LEFT  | A_EDGE, showEdge);
	set_action(A_LEFT  | A_EDGE | A_CTRL, actionSelectRef);
	set_action(A_LEFT  | A_DRAG, leda::gw_scroll_graph);
	set_action(A_LEFT  | A_SHIFT | A_IMMEDIATE, leda::gw_zoom_area);
	set_action(A_LEFT  | A_CTRL  | A_IMMEDIATE, ::gw_zoom_graph);
	set_action(A_LEFT,  ::gw_zoom_up);
	set_action(A_RIGHT, ::gw_zoom_down);

	setupNodesEdges();

	// install observer
	theGraph.subscribe(*this);

	// create SPQRTreeWindow child
	spqrWindow = new SPQRTreeWindow(planeGraph);
}


GraphWindow::~GraphWindow()
{
	delete spqrWindow;
}


//---------------------------------------------------------
// handling of notification messages
//---------------------------------------------------------

void GraphWindow::notify(StPlaneGraph::ChangeStatus status)
{
	switch (status) {
	case StPlaneGraph::topologyChanged:
		update_graph();
		if (planeGraph.isGraphValid()) {
			set_flush(false);
			setupNodesEdges();
			draw_graph(false);
		} else {
			acknowledge("The current graph posseses no SPQR-tree!");
		}
		break;

	case StPlaneGraph::topologyEdited:
		if (planeGraph.isGraphValid()) {
			set_flush(false);
			update_graph();
			setupNodesEdges();
			draw_graph(true);
			zoom_graph();
		} else {
			acknowledge("The current graph posseses no SPQR-tree!");
		}
		break;

	case StPlaneGraph::referenceEdgeChanged:
		update_graph();
		resetupRefEdge();
		setupRefEdge();
		draw_graph(true);
		break;

	case StPlaneGraph::embeddingChanged:
		update_graph();
		draw_graph(true);
		break;

	case StPlaneGraph::stLayoutOptionsChanged:
		draw_graph(true);
		break;

	case StPlaneGraph::permMarkGraphChanged:
	case StPlaneGraph::permMarkGraphCleared:
		set_flush(false);
		setupNodesEdges();
		set_flush(true);
		redraw();

        default: assert(0);
                 break;
	}
}


//---------------------------------------------------------
// drawing the graph
//---------------------------------------------------------

void GraphWindow::draw_graph(bool anim)
{
	graph &G = planeGraph.getGraph();

	PlaneGraphCopy &PG = planeGraph.getPlaneGraph();
	GraphCopy GC(G);

	list<edge> edges = GC.all_edges();
	edge e;
	forall(e,edges)
		GC.set_reversal(e,GC.new_edge(target(e),source(e)));

	list<edge> e_list;
	node v;
	forall_nodes(v,PG) {
		edge e;
		forall_adj_edges(e,v) {
			edge eG = PG.original(e) ? PG.original(e) : PG.original(PG.reversal(e));
			edge eGC = GC.chain(eG).head();
			if (GC.original(source(eGC)) == PG.original(source(e)))
				e_list.append(eGC);
			else
				e_list.append(GC.reversal(eGC));
		}
	}

	GC.sort_edges(e_list);

	GC.compute_faces();

	edge e_st = planeGraph.refEdge();

	e_st = GC.chain(e_st).head();

	if (anim) {
		GraphWinAnimInterface gwi(*this);
		planeGraph.callStLayout(GC,e_st,gwi);
	} else {
		GraphWinInterface gwi(*this);
		gwi.flush(false);
		planeGraph.callStLayout(GC,e_st,gwi);
		zoom_graph();
	}
	set_flush(true);
}


//---------------------------------------------------------
// settting node/edge attributes
//---------------------------------------------------------

void GraphWindow::setupNodesEdges()
{
	const graph &G = planeGraph.getGraph();
	const edge_set *markedEdges = planeGraph.permanentMarkedEdgesGraph();
	const node_set *markedNodes = planeGraph.permanentMarkedNodesGraph();

	if (!planeGraph.isPermanentMarkedGraph()) {
		set_edge_color(opts.edgeColor,true);
		set_edge_thickness(1.5*opts.edgeWidth,true);
		set_node_color(opts.nodeColor,true);
		setupRefEdge();

	} else {

		edge e;
		forall_edges(e,G)
			if (markedEdges->member(e)) {
				set_color(e, opts.markedEdgeColor);
				set_thickness(e, 1.5*opts.markedEdgeWidth);
			} else {
				set_color(e, opts.unmarkedEdgeColor);
				set_thickness(e, 1.5*opts.unmarkedEdgeWidth);
			}

		node v;
		forall_nodes(v,G)
			if (markedNodes->member(v)) {
				set_color(v, opts.markedNodeColor);
			} else {
				set_color(v, opts.unmarkedNodeColor);
			}
	}
}

void GraphWindow::setupRefEdge()
{
	edge eRef = planeGraph.refEdge();

	if (!eRef) return;
	set_color(eRef,opts.refEdgeColor);
	set_thickness(eRef,1.5*opts.refEdgeWidth);
}

void GraphWindow::resetupRefEdge()
{
	edge eRef = planeGraph.oldRefEdge();

	if (!eRef) return;

	set_color(eRef,opts.edgeColor);
	set_thickness(eRef,1.5*opts.edgeWidth);
}


//---------------------------------------------------------
// handling of events
//---------------------------------------------------------

bool GraphWindow::eventHandler()
{
	window *wp;
	double x, y;
	int but;
	int event = read_event(wp,but,x,y);
	edit_loop(event,wp,but,x,y);

	spqrWindow->eventHandler(event,wp,but,x,y);

	return _isOpen;
}

void GraphWindow::display()
{
	GraphWin::display(0,0);
	_isOpen = true;
	get_window().set_window_delete_handler(deleteWindowHandler);

	spqrWindow->display(window::screen_width()/2,0);
}


//---------------------------------------------------------
// callback's for zooming
//---------------------------------------------------------

void gw_zoom_graph(GraphWin& gw, const point& p) 
{
	gw_zoom_graph(gw);
}

void gw_zoom_up(GraphWin& gw, const point& p) 
{
	gw_zoom_up(gw);
}

void gw_zoom_down(GraphWin& gw, const point& p) 
{
	gw_zoom_down(gw);
}


//---------------------------------------------------------
// callback's for deleting
//---------------------------------------------------------

void del_selection(GraphWin& gw)
{
	bool oldFlush = gw.set_flush(false);

	list<node> selNodes = gw.get_selected_nodes();
	node v;
	forall(v,selNodes)
		gw_del_node(gw,v);

	list<edge> selEdges = gw.get_selected_edges();
	edge e;
	forall(e,selEdges)
		gw_del_edge(gw,e);

	gw.redraw();
	gw.set_flush(oldFlush);
}


//---------------------------------------------------------
// mouse action functions
//---------------------------------------------------------

void GraphWindow::actionSelectRef(GraphWin& gw, const point& p) 
{
	edge eRef = gw.get_found_edge();
	GraphWindow &graphWindow = (GraphWindow &) gw;
	StPlaneGraph &stGraph = graphWindow.planeGraph;
	const PlaneGraphCopy &PG = stGraph.getPlaneGraph();
	stGraph.changeReferenceEdge(PG.chain(eRef).head());
}


//---------------------------------------------------------
// functions for menu entries
//---------------------------------------------------------

void GraphWindow::doOpen()
{
	file_handler(LOAD_GW);
	planeGraph.topologyHasChanged(nil,nil);
	undo_clear();
}

void GraphWindow::doSave()
{
	file_handler(SAVE_GW);
}

void GraphWindow::doGenerate()
{
	if (!check_init_graph_handler()) return;

	panel P;
	int n = get_gen_nodes(),
		m = get_gen_edges();
	bool multiEdges = genMultiEdges;

	P.text_item("\\bf\\blue Generate Graph");
	P.int_item ("nodes", n);
	P.int_item ("edges", m);
	P.bool_item ("multi-edges", multiEdges);
	P.button ("ok", 1);
	P.button ("cancel", 0);

	if (P.open(get_window()) == 0 || n < 0 || m < 0) return;

	set_gen_nodes (n);
	set_gen_edges (m);
	genMultiEdges = multiEdges;

	graph& G = get_graph();

	G.clear();
	update_graph();

	planar_biconnected_graph (G, n, m, multiEdges, false);

	update_graph();
	redraw();
	call_init_graph_handler();

	planeGraph.topologyHasChanged(nil,nil,true);
}

void GraphWindow::doEdit()
{
	EditWindow editWin(*this, planeGraph.refEdge(), opts.edgeColor, opts.edgeWidth, opts.refEdgeColor, opts.refEdgeWidth);
	disableInput();
	if (editWin.open()) {
		edge eRef = editWin.assignGraph();
		const GraphWinInterface gwi(*this);
		planeGraph.topologyHasChanged(eRef,&gwi);
	}
	enableInput();
}

void GraphWindow::disableInput()
{
	get_window().disable_panel();
	spqrWindow->disableInput();
}

void GraphWindow::enableInput()
{
	get_window().enable_panel();
	spqrWindow->enableInput();
}

void GraphWindow::doExit()
{
	_isOpen = false;
}

void GraphWindow::doSelectRef()
{
	edge eRef = read_edge();
	const PlaneGraphCopy &PG = planeGraph.getPlaneGraph();
	planeGraph.changeReferenceEdge(PG.chain(eRef).head());
}

void GraphWindow::doUpsideDown()
{
	planeGraph.reverseRefEdge();
}


void GraphWindow::doSetup()
{
	GraphOpts graphOpts = opts;

	panel P("Node and Edge Options");
	P.text_item("\\bf\\blue nodes");
	P.color_item("color", graphOpts.nodeColor);
	P.color_item("marked color", graphOpts.markedNodeColor);
	P.color_item("unmarked color", graphOpts.unmarkedNodeColor);

	P.text_item("\\bf\\blue edges");
	P.color_item ("color", graphOpts.edgeColor);
	P.lwidth_item("width", graphOpts.edgeWidth);
	P.color_item ("marked color", graphOpts.markedEdgeColor);
	P.lwidth_item("marked width", graphOpts.markedEdgeWidth);
	P.color_item ("unmarked color", graphOpts.unmarkedEdgeColor);
	P.lwidth_item("unmarked width", graphOpts.unmarkedEdgeWidth);

	P.text_item("\\bf\\blue reference edge");
	P.color_item ("color", graphOpts.refEdgeColor);
	P.lwidth_item("width", graphOpts.refEdgeWidth);

	P.button ("Ok");
	int cancel = P.button ("Cancel");

	int result = P.open();
	if (result != cancel) {
		opts = graphOpts;

		setupNodesEdges();
		redraw();
	}

}


void GraphWindow::doLayoutOpt()
{
	StPlanarGridLayout::CompactionType compaction;// = (int)stPlanarLayouter.compaction();
	double separation;// = stPlanarLayouter.separation();
	
	planeGraph.getStLayoutOptions(compaction, separation);

	panel P("Layout Options");
	P.text_item("\\bf\\blue StPlanarLayout:");
	P.choice_item("compaction", (int &)compaction, "none", "symmetry", "min. area");
	P.double_item("separation", separation);

	P.button ("Ok");
	int cancel = P.button ("Cancel");

	int result = P.open();
	if (result != cancel) {
		planeGraph.setStLayoutOptions(compaction, separation);
		//stPlanarLayouter.compaction((StPlanarGridLayout::CompactionType)compaction);
		//stPlanarLayouter.separation(separation);
	}
}

void GraphWindow::doAnimOpt()
{
	int animSteps  = get_animation_steps();
	
	panel P("Animation Options");
	P.int_item("steps",animSteps,0,50);

	P.button ("Ok");
	int cancel = P.button ("Cancel");

	int result = P.open();
	if (result != cancel) {
		set_animation_steps(animSteps);
	}
}


void GraphWindow::showEdge(GraphWin& gw, const point& p)
{
	GraphWindow &graphWin = (GraphWindow &) gw;
	const SPQRTree &T = graphWin.planeGraph.spqrTree();

	edge eG = graphWin.get_found_edge();
	if (eG == nil) return;

	edge ePG = graphWin.planeGraph.getPlaneGraph().chain(eG).head();
	node v = T.skeleton_of_real(ePG).tree_node();

	graphWin.spqrWindow->openSkeleton(v);
}


//---------------------------------------------------------
// static caller functions for menu entries
// usually delegate job to member functions
//---------------------------------------------------------

void GraphWindow::deleteWindowHandler(window *win)
{
	GraphWindow *graphWindow = (GraphWindow *)win->get_graphwin();
	graphWindow->doExit();
}


void GraphWindow::openCaller (GraphWin &gw)
{
	GraphWindow &graphWindow = (GraphWindow &) gw;
	graphWindow.doOpen();
}

void GraphWindow::saveCaller (GraphWin &gw)
{
	GraphWindow &graphWindow = (GraphWindow &) gw;
	graphWindow.doSave();
}

void GraphWindow::generateCaller (GraphWin &gw)
{
	GraphWindow &graphWindow = (GraphWindow &) gw;
	graphWindow.doGenerate();
}

void GraphWindow::editCaller (GraphWin &gw)
{
	GraphWindow &graphWindow = (GraphWindow &) gw;
	graphWindow.doEdit();
}

void GraphWindow::exitCaller (GraphWin &gw)
{
	GraphWindow &graphWindow = (GraphWindow &) gw;
	graphWindow.doExit();
}

void GraphWindow::selectRefCaller (GraphWin &gw)
{
	GraphWindow &graphWindow = (GraphWindow &) gw;
	graphWindow.doSelectRef();
}

void GraphWindow::upsideDownCaller (GraphWin &gw)
{
	GraphWindow &graphWindow = (GraphWindow &) gw;
	graphWindow.doUpsideDown();
}

void GraphWindow::unmarkCaller (GraphWin &gw)
{
	GraphWindow &graphWindow = (GraphWindow &) gw;
	graphWindow.planeGraph.unsetPermanentMarkGraph();
}

void GraphWindow::setupCaller (GraphWin &gw)
{
	GraphWindow &graphWindow = (GraphWindow &) gw;
	graphWindow.doSetup();
}

void GraphWindow::layoutOptCaller (GraphWin &gw)
{
	GraphWindow &graphWindow = (GraphWindow &) gw;
	graphWindow.doLayoutOpt();
}

void GraphWindow::animOptCaller (GraphWin &gw)
{
	GraphWindow &graphWindow = (GraphWindow &) gw;
	graphWindow.doAnimOpt();
}

void GraphWindow::aboutCaller (GraphWin &gw)
{
	panel P("About SPQR-Tree Viewer");

	P.text_item("\\bf\\blue SPQR-Tree Viewer");
	P.text_item(""); P.text_item("");
	P.text_item("This program visualizes the SPQRTree \
data structure of AGD.");
	P.text_item(""); P.text_item("");
	P.text_item(string("AGD version: %s", AGD::version_string()));
	P.text_item("");
	P.text_item("SPQR-Tree Viewer version 1.0");
	P.text_item(""); P.text_item("");
	P.text_item("Copyright Carsten Gutwenger");
	P.text_item("");
	P.text_item("Max-Planck-Institut fr Informatik, Saarbrcken");

	P.button("Ok");
	P.open();
}

void GraphWindow::helpMouseCaller (GraphWin &gw)
{
	panel P("Mouse Buttons");
	P.text_item("\\bf\\blue Clicking on an edge:"); P.text_item("");
	P.text_item("\\red left\\black mouse button: \\orange open corresponding skeleton"); P.text_item("");
	P.text_item("\\red left\\black mouse button + \\red <CTRL>\\black key: \\orange select reference edge");
	P.text_item(""); P.text_item(""); P.text_item("");

	P.text_item("\\bf\\blue Clicking on the background:"); P.text_item("");
	P.text_item("\\red left\\black mouse button: \\orange zoom-up"); P.text_item("");
	P.text_item("\\red right\\black mouse button: \\orange zoom-down"); P.text_item("");
	P.text_item("");
	P.text_item("\\red left\\black mouse button and \\red dragging\\black : \\orange scroll graph"); P.text_item("");
	P.text_item("");
	P.text_item("\\red left\\black mouse button + \\red <CTRL>\\black key: \\orange zoom graph"); P.text_item("");
	P.text_item("\\red left\\black mouse button + \\red <SHIFT>\\black key: \\orange zoom area");

	P.button("Ok");
	P.open();
}


//---------------------------------------------------------
// SkeletonWindow
//
// author: Carsten Gutwenger
//---------------------------------------------------------

#include <AGD/PlanarSPQRTree.h>
#include <AGD/GraphWinInterface.h>
#include <AGD/GraphWinAnimInterface.h>


//---------------------------------------------------------
// Constructor / Destructor
//---------------------------------------------------------

SkeletonWindow::SkeletonWindow(SPQRTreeWindow &spqrWin, StPlaneGraph &theGraph, node v) :
	GraphWin(200,200,winTitle(theGraph,v)),
	parent(spqrWin),
	planeGraph(theGraph),
	vT(v), isOpen(true)
{
	const PlanarSPQRTree &T = planeGraph.spqrTree();

	// setup GraphWin
	set_default_menu(M_NONE);
	clear_actions();
	set_show_status(false);

	// create menu
    int fileMenuId = add_menu("&File");
	add_simple_call(saveCaller, "&Save", fileMenuId);
	if (T.type(v) == SPQRTree::RNode)
		add_simple_call(reverseCaller, "&Reverse", fileMenuId);
	else if (T.type(v) == SPQRTree::PNode)
		add_simple_call(swapCaller, "S&wap", fileMenuId);
	add_simple_call(showPertinent, "Show &Pertinent", fileMenuId);
    add_separator(fileMenuId);
	add_simple_call(closeCaller, "&Close", fileMenuId);

    int windowMenuId = add_menu("&Window");
    add_simple_call(gw_center_graph, "center graph",       windowMenuId);
    add_simple_call(gw_zoom_graph,   "zoom to graph",      windowMenuId);
    add_simple_call(gw_zoom_area,    "zoom to area",       windowMenuId);
    add_separator(windowMenuId);
    add_simple_call(gw_zoom_up,      "zoom +",             windowMenuId);
    add_simple_call(gw_zoom_down,    "zoom -",             windowMenuId);

    int helpMenuId = add_menu("&Help");
    add_simple_call(helpMouseCaller, "&Mouse Buttons", helpMenuId);

	// mouse action handlers
	set_action(A_LEFT  | A_NODE,  showVertex);
	set_action(A_LEFT  | A_FRAME, showVertex);
	set_action(A_LEFT  | A_EDGE,  showExpansion);
	set_action(A_LEFT  | A_EDGE | A_CTRL, showSkeleton);
	set_action(A_LEFT  | A_DRAG, leda::gw_scroll_graph);
	set_action(A_LEFT  | A_SHIFT | A_IMMEDIATE, leda::gw_zoom_area);
	set_action(A_LEFT  | A_CTRL  | A_IMMEDIATE, ::gw_zoom_graph);
	set_action(A_LEFT,  ::gw_zoom_up);
	set_action(A_RIGHT, ::gw_zoom_down);

	// install observer
	theGraph.subscribe(*this);

	// build graph
	buildGraph();
}


//---------------------------------------------------------
// handling of notification messages
//---------------------------------------------------------

void SkeletonWindow::notify(StPlaneGraph::ChangeStatus status)
{
	switch (status) {
	case StPlaneGraph::topologyEdited:
	case StPlaneGraph::topologyChanged:
		break;

	case StPlaneGraph::referenceEdgeChanged:
		if (adaptDirections()) {
			update_graph();
			adaptRefEdge();
			drawGraph(true);
		}
		break;

	case StPlaneGraph::stLayoutOptionsChanged:
	case StPlaneGraph::embeddingChanged:
		drawGraph(true);
		break;

	case StPlaneGraph::skeletonOptsChanged:
		set_flush(false);
		setup();
		redraw();
		break;

        default: assert(0);
                 break;
	}
}


//---------------------------------------------------------
// build the graph to be displayed
//---------------------------------------------------------

/*ostream &operator<<(ostream &os, edge e)
{
	if (e)
		os << "(" << index(source(e)) << "," << index(target(e)) << ")";
	else
		os << "nil";
	return os;
}*/

void SkeletonWindow::buildGraph()
{
	const PlanarSPQRTree &T = planeGraph.spqrTree();
	const PlaneGraphCopy &PG = planeGraph.getPlaneGraph();
	const Skeleton &Sk = T.skeleton(vT);
	const graph &M = Sk.get_graph();

	vMtoS = node_array<node>(M);
	vStoM = node_array<node>(S, M.number_of_nodes(), nil);
	node u;
	forall_nodes(u,M)
		vStoM[vMtoS[u] = S.new_node()] = u;

	eMtoS = edge_array<edge>(M);
	eStoM = edge_array<edge>(S, M.number_of_edges()/2, nil);
	edge e;
	forall_edges(e,M)
		if (planeGraph.stNum(PG.original(Sk.original(source(e)))) <  planeGraph.stNum(PG.original(Sk.original(target(e)))))
			eStoM[eMtoS[e] = S.new_edge(vMtoS[source(e)],vMtoS[target(e)])] = e;
		else
			eMtoS[e] = nil;

	set_graph(S);

	set_node_label_type(user_label);
	forall_nodes(u,M)
		set_user_label(vMtoS[u],string("%d",index(PG.original(Sk.original(u)))));

	setup();
}


//---------------------------------------------------------
// setting node / edge attributes
//---------------------------------------------------------

void SkeletonWindow::setup()
{
	const PlanarSPQRTree &T = planeGraph.spqrTree();
	const Skeleton &Sk = T.skeleton(vT);
	const graph &M = Sk.get_graph();
	//window &W = get_window();
	const SkeletonOpts &opts = planeGraph.skeletonOpts();

	set_animation_steps(opts.animSteps);

	edge e;
	forall_edges(e,M) {
		edge eS = eMtoS[e];
		if (eS == nil) continue;

		if (e == Sk.reference_edge() || M.reversal(e) == Sk.reference_edge()) {
			oldRefEdge  = eS;
			oldRefEdgeM = e;
			setupRefEdge(eS);
		} else if (Sk.is_virt(e)) {
			setupVirtualEdge(eS);
		} else {
			setupRealEdge(eS);
		}
	}
}

void SkeletonWindow::setupRefEdge(edge eS)
{
	const SkeletonOpts &opts = planeGraph.skeletonOpts();
	set_style(eS, opts.styleRefEdge);
	set_color(eS, opts.colorRefEdge);
	set_thickness(eS, 1.5*opts.widthRefEdge);
}

void SkeletonWindow::setupRealEdge(edge eS)
{
	const SkeletonOpts &opts = planeGraph.skeletonOpts();
	set_style(eS, opts.styleRealEdge);
	set_color(eS, opts.colorRealEdge);
	set_thickness(eS, 1.5*opts.widthRealEdge);
}

void SkeletonWindow::setupVirtualEdge(edge eS)
{
	const SkeletonOpts &opts = planeGraph.skeletonOpts();
	set_style(eS, opts.styleVirtualEdge);
	set_color(eS, opts.colorVirtualEdge);
	set_thickness(eS, 1.5*opts.widthVirtualEdge);
}



//---------------------------------------------------------
// adapt the graph, which is displayed, to changes
//---------------------------------------------------------

void SkeletonWindow::adaptRefEdge()
{
	const PlanarSPQRTree &T = planeGraph.spqrTree();
	const Skeleton &Sk = T.skeleton(vT);
	const graph &M = Sk.get_graph();

	edge eM = Sk.reference_edge();
	if (eMtoS[eM] == nil) eM = M.reversal(eM);
	edge refEdge = eMtoS[eM];

	if (refEdge != oldRefEdge) {
		if (Sk.is_virt(oldRefEdgeM))
			setupVirtualEdge(oldRefEdge);
		else
			setupRealEdge(oldRefEdge);
		oldRefEdge  = refEdge;
		oldRefEdgeM = eM;
		setupRefEdge(oldRefEdge);
	}
}

bool SkeletonWindow::adaptDirections()
{
	const PlanarSPQRTree &T = planeGraph.spqrTree();
	const PlaneGraphCopy &PG = planeGraph.getPlaneGraph();
	const Skeleton &Sk = T.skeleton(vT);
	const graph &M = Sk.get_graph();

	bool changed = false;

	edge e;
	forall_edges(e,M) {
		edge eS = eMtoS[e];
		if (eS == nil) continue;
		bool MOk = planeGraph.stNum(PG.original(Sk.original(source(e)))) <  planeGraph.stNum(PG.original(Sk.original(target(e))));
		bool SOk = vMtoS[source(e)] == source(eS);
		if ((MOk && !SOk) || (!MOk && SOk)) {
			S.rev_edge(eS);
			changed = true;
		}
	}

	return changed;
}


//---------------------------------------------------------
// the window title
//---------------------------------------------------------

string SkeletonWindow::winTitle(const StPlaneGraph &stGraph, node v)
{
	const PlanarSPQRTree &T = stGraph.spqrTree();
	char type = ' ';

	switch(T.type(v)) {
	case SPQRTree::SNode:
		type = 'S'; break;

	case SPQRTree::PNode:
		type = 'P'; break;

	case SPQRTree::RNode:
		type = 'R'; break;
	}

	return string("skeleton(%c %d)", type, index(v));
}


//---------------------------------------------------------
// handling of events
//---------------------------------------------------------

bool SkeletonWindow::eventHandler(int event, window *wp, int but,
	double x, double y)
{
	edit_loop(event,wp,but,x,y);
	return isOpen;
}

void SkeletonWindow::display(int x, int y)
{
	GraphWin::display(x,y);
	get_window().set_window_delete_handler(deleteWindowHandler);

	drawGraph(false);
}

void SkeletonWindow::disableInput()
{
	get_window().disable_panel();
}

void SkeletonWindow::enableInput()
{
	get_window().enable_panel();
}


//---------------------------------------------------------
// drawing the graph
//---------------------------------------------------------

void SkeletonWindow::drawGraph(bool anim)
{
	const PlanarSPQRTree &T = planeGraph.spqrTree();
	//const PlaneGraphCopy &PG = planeGraph.getPlaneGraph();
	const Skeleton &Sk = T.skeleton(vT);
	const graph &M = Sk.get_graph();

	GraphCopy GC(S);
	list<edge> R;
	GC.make_map(R);

	list<edge> eList;

	node v;
	forall_nodes(v,M) {
		edge e;
		forall_adj_edges(e,v) {
			edge eS = eMtoS[e] ? eMtoS[e] : eMtoS[M.reversal(e)];
			edge eGC = GC.chain(eS).head();
			if (vMtoS[source(e)] == source(eS))
				eList.append(eGC);
			else
				eList.append(GC.reversal(eGC));
		}
	}

	GC.sort_edges(eList);
	GC.compute_faces();

	edge e = Sk.reference_edge();
	edge eS = eMtoS[e] ? eMtoS[e] : eMtoS[M.reversal(e)];
	edge e_st = GC.chain(eS).head();

	if (anim) {
		GraphWinAnimInterface gwi(*this);
		planeGraph.callStLayout(GC,e_st,gwi);
	} else {
		GraphWinInterface gwi(*this);
		gwi.flush(false);
		planeGraph.callStLayout(GC,e_st,gwi);
		zoom_graph();
	}
	set_flush(true);
}


//---------------------------------------------------------
// functions for menu entries
//---------------------------------------------------------

void SkeletonWindow::doSave()
{
	file_handler(SAVE_GW);
}


void SkeletonWindow::doReverse()
{
	planeGraph.reverse(vT);
	drawGraph(true);
}


void SkeletonWindow::doSwap()
{
	const graph &M = planeGraph.spqrTree().skeleton(vT).get_graph();

	edge e1M=0, e1;
	edge e2M=0, e2;

	if ((e1 = read_edge("select first edge")) == nil) return;
	if ((e2 = read_edge("select second edge")) == nil || e1 == e2) return;

	edge e;
	forall_edges(e,M)
		if (eMtoS[e] == e1) e1M = e;
		else if (eMtoS[e] == e2) e2M = e;

	planeGraph.swap(vT,e1M,e2M);
	drawGraph(true);
}


edge SkeletonWindow::read_edge(const string msg)
{
	graph &G = get_graph();
	window &win = get_window();

	if (G.number_of_edges() == 0) return nil;
	point p;
	edge e = nil;
	message(msg);
	do {
		if (win.read_mouse(p) == MOUSE_BUTTON(3)) break;
		e=find_edge(p);
	} while (e == nil);
	message("");
	return e;
}


void SkeletonWindow::doClose()
{
	isOpen = false;
}


void SkeletonWindow::showSkeleton(GraphWin& gw, const point& p)
{
	SkeletonWindow &skWin = (SkeletonWindow &) gw;

	const PlanarSPQRTree &T = skWin.planeGraph.spqrTree();
	const Skeleton &Sk = T.skeleton(skWin.vT);

	edge eS = gw.get_found_edge();
	if (eS == nil) return;
	edge eM = skWin.eStoM[eS];

	edge eT = Sk.tree_edge(eM);
	if (eT == nil) return;

	node v = (target(eT) != skWin.vT) ? target(eT) : source(eT);

	skWin.parent.openSkeleton(v);
}

void SkeletonWindow::showExpansion(GraphWin& gw, const point& p)
{
	SkeletonWindow &skWin = (SkeletonWindow &) gw;

	const PlanarSPQRTree &T = skWin.planeGraph.spqrTree();
	const Skeleton &Sk = T.skeleton(skWin.vT);
	const graph &M = Sk.get_graph();

	edge eS = gw.get_found_edge();
	if (eS == nil) return;
	edge eM = skWin.eStoM[eS];

	if (Sk.is_virt(eM)) {
		edge eT = Sk.tree_edge(eM);

		node v = target(eT);
		if (v == skWin.vT) return;

		skWin.parent.showPertinent(v);

	} else {
		list<node> nodes;
		list<edge> edges;

		edges.append(Sk.real_edge(eM));
		edges.append(Sk.real_edge(M.reversal(eM)));
		nodes.append(Sk.original(source(eM)));
		nodes.append(Sk.original(target(eM)));
		skWin.planeGraph.setPermanentMarkGraph(nodes,edges);
	}
}


void SkeletonWindow::showVertex(GraphWin& gw, const point& p)
{
	SkeletonWindow &skWin = (SkeletonWindow &) gw;

	const PlanarSPQRTree &T = skWin.planeGraph.spqrTree();
	const Skeleton &Sk = T.skeleton(skWin.vT);

	node vS = gw.get_found_node();
	if (vS == nil) return;

	list<node> nodes;
	list<edge> edges;

	nodes.append(Sk.original(skWin.vStoM[vS]));
	skWin.planeGraph.setPermanentMarkGraph(nodes,edges);
}


//---------------------------------------------------------
// static caller functions for menu entries
// usually delegate job to member functions
//---------------------------------------------------------

void SkeletonWindow::deleteWindowHandler(window *win)
{
	SkeletonWindow *skWindow = (SkeletonWindow *)win->get_graphwin();
	skWindow->doClose();
}


void SkeletonWindow::showPertinent (GraphWin &gw)
{
	SkeletonWindow &skWindow = (SkeletonWindow &) gw;
	skWindow.parent.showPertinent(skWindow.vT);
}

void SkeletonWindow::reverseCaller (GraphWin &gw)
{
	SkeletonWindow &skWindow = (SkeletonWindow &) gw;
	skWindow.doReverse();
}

void SkeletonWindow::swapCaller (GraphWin &gw)
{
	SkeletonWindow &skWindow = (SkeletonWindow &) gw;
	skWindow.doSwap();
}

void SkeletonWindow::closeCaller (GraphWin &gw)
{
	SkeletonWindow &skWindow = (SkeletonWindow &) gw;
	skWindow.doClose();
}

void SkeletonWindow::saveCaller (GraphWin &gw)
{
	SkeletonWindow &skWindow = (SkeletonWindow &) gw;
	skWindow.doSave();
}

void SkeletonWindow::helpMouseCaller (GraphWin &gw)
{
	panel P("Mouse Buttons");
	P.text_item("\\bf\\blue Clicking on a node:"); P.text_item("");
	P.text_item("\\red left\\black mouse button: \\orange show original node");
	P.text_item(""); P.text_item(""); P.text_item("");

	P.text_item("\\bf\\blue Clicking on an edge:"); P.text_item("");
	P.text_item("\\red left\\black mouse button: \\orange show expansion graph"); P.text_item("");
	P.text_item("\\red left\\black mouse button + \\red <CTRL>\\black key:   \\orange open corresponding skeleton");
	P.text_item(""); P.text_item(""); P.text_item("");

	P.text_item("\\bf\\blue Clicking on the background:"); P.text_item("");
	P.text_item("\\red left\\black mouse button: \\orange zoom-up"); P.text_item("");
	P.text_item("\\red right\\black mouse button: \\orange zoom-down"); P.text_item("");
	P.text_item("");
	P.text_item("\\red left\\black mouse button and \\red dragging\\black : \\orange scroll graph"); P.text_item("");
	P.text_item("");
	P.text_item("\\red left\\black mouse button + \\red <CTRL>\\black key: \\orange zoom graph"); P.text_item("");
	P.text_item("\\red left\\black mouse button + \\red <SHIFT>\\black key: \\orange zoom area");

	P.button("Ok");
	P.open();
}


//---------------------------------------------------------
// SPQRTreeWindow
//
// author: Carsten Gutwenger
//---------------------------------------------------------

#include <AGD/PlanarSPQRTree.h>
#include <AGD/GraphWinInterface.h>
#include <AGD/GraphWinAnimInterface.h>

#include "bitmaps.h"

#define nshape_width	rectangle_width
#define nshape_height	rectangle_height

static unsigned char* nshape_bits[]= { circle_bits, ellipse_bits, square_bits, rectangle_bits, roundrect_bits, ovalrect_bits, rhombus_bits };


//---------------------------------------------------------
// Constructor / Destructor
//---------------------------------------------------------

SPQRTreeWindow::SPQRTreeWindow(StPlaneGraph &theGraph) :
	GraphWin(window::screen_width()/2-6,window::screen_height()-55,
	"SPQR-Tree Window"), planeGraph(theGraph), GC(),
	shapeSNode(ellipse_node), shapePNode(roundrect_node), shapeRNode(rectangle_node),
	colorSNode(yellow), colorPNode(green), colorRNode(red)
{
	// setup GraphWin
	set_default_menu(M_NONE);
	clear_actions();
	set_show_status(false);
	set_flush(false);
	set_node_label_type(user_label);

    int fileMenuId = add_menu("&File");
	add_simple_call(saveCaller, "&Save", fileMenuId);
    add_separator(fileMenuId);
	add_simple_call(infoCaller, "&Info", fileMenuId);

    int skeletonMenuId = add_menu("&Skeletons");
	add_simple_call(randomEmbed, "&embed randomly", skeletonMenuId);
    add_separator(skeletonMenuId);
	add_simple_call(allToFrontCaller, "All to &front", skeletonMenuId);
	add_simple_call(closeAllCaller, "&Close all", skeletonMenuId);

    int windowMenuId = add_menu("&Window");
    add_simple_call(gw_center_graph, "center graph",       windowMenuId);
    add_simple_call(gw_zoom_graph,   "zoom to graph",      windowMenuId);
    add_simple_call(gw_zoom_area,    "zoom to area",       windowMenuId);
    add_separator(windowMenuId);
    add_simple_call(gw_zoom_up,      "zoom +",             windowMenuId);
    add_simple_call(gw_zoom_down,    "zoom -",             windowMenuId);

    int optionsMenuId = add_menu("&Options");
    add_simple_call(setupCaller, "&Nodes", optionsMenuId);
    add_simple_call(animOptCaller, "&Animation", optionsMenuId);
    add_separator(optionsMenuId);
    add_simple_call(skeletonOptCaller, "&Skeletons", optionsMenuId);

    int helpMenuId = add_menu("&Help");
    add_simple_call(helpMouseCaller, "&Mouse Buttons", helpMenuId);

	set_action(A_LEFT  | A_NODE, actionOpenSkeleton);
	set_action(A_LEFT  | A_NODE | A_CTRL, actionShowPertinent);
	set_action(A_LEFT  | A_DRAG, leda::gw_scroll_graph);
	set_action(A_LEFT  | A_SHIFT | A_IMMEDIATE, leda::gw_zoom_area);
	set_action(A_LEFT  | A_CTRL  | A_IMMEDIATE, ::gw_zoom_graph);
	set_action(A_LEFT,  ::gw_zoom_up);
	set_action(A_RIGHT, ::gw_zoom_down);

	// install observer
	theGraph.subscribe(*this);

	set_graph(GC);
	skeletonWin.init(GC,nil);
	update_graph();

	skCount = 0;
}


SPQRTreeWindow::~SPQRTreeWindow()
{
	closeAllSkeletons();
}


//---------------------------------------------------------
// setting node / edge attributes
//---------------------------------------------------------

void SPQRTreeWindow::setupTreeGraph()
{
	if (!planeGraph.isGraphValid()) return;

	PlanarSPQRTree &T = planeGraph.spqrTree();
	const graph &G = T.tree();

	set_flush(false);
	node v;
	forall_nodes(v,G) {
		node vGC = GC.copy(v);
		switch(T.type(v)) {
		case SPQRTree::SNode:
			set_shape(vGC,shapeSNode);
			set_color(vGC,colorSNode);
			set_user_label(vGC,string("S %d",index(v)));
			break;

		case SPQRTree::PNode:
			set_shape(vGC,shapePNode);
			set_color(vGC,colorPNode);
			set_user_label(vGC,string("P %d",index(v)));
			break;

		case SPQRTree::RNode:
			set_shape(vGC,shapeRNode);
			set_color(vGC,colorRNode);
			set_user_label(vGC,string("R %d",index(v)));
			break;
		}
	}
	set_flush(true);
}


//---------------------------------------------------------
// handling of notification messages
//---------------------------------------------------------

void SPQRTreeWindow::notify(StPlaneGraph::ChangeStatus status)
{
	switch (status) {
	case StPlaneGraph::topologyEdited:
	case StPlaneGraph::topologyChanged:
		closeAllSkeletons();
		if (planeGraph.isGraphValid()) {
			GC = GraphCopy(planeGraph.spqrTree().tree());
			set_flush(false);
			update_graph();
			setupTreeGraph();
			draw_tree(false);
		} else {
			GC = GraphCopy();
			update_graph();
		}
		break;

	case StPlaneGraph::referenceEdgeChanged:
		adaptDirections();
		update_graph();
		draw_tree(true);
		break;

	case StPlaneGraph::embeddingChanged:
		break;

        default: assert(0);
                 break;
	}
}


//---------------------------------------------------------
// adapt edge directions
//---------------------------------------------------------

void SPQRTreeWindow::adaptDirections()
{
	edge e;
	forall_edges(e,GC) {
		if (GC.is_reversed(GC.original(e)))
			GC.rev_edge(e);
	}
}


//---------------------------------------------------------
// drawing the tree
//---------------------------------------------------------

void SPQRTreeWindow::draw_tree(bool anim)
{
	treeLayouter.sib_separation(15.0);
	treeLayouter.subtree_separation(15.0);
	treeLayouter.level_separation(20.0);

	if (anim) {
		GraphWinAnimInterface gwi(*this);
		treeLayouter.call((const graph &)GC,gwi);
	} else {
		GraphWinInterface gwi(*this);
		gwi.flush(false);
		treeLayouter.call((const graph &)GC,gwi);
		zoom_graph();
	}
}


//---------------------------------------------------------
// handling of events
//---------------------------------------------------------

void SPQRTreeWindow::eventHandler(int event, window *wp, int but,
	double x, double y)
{
	edit_loop(event,wp,but,x,y);

	SkeletonWindow *win;
	forall(win,skeletonWindows) {
		if (win->eventHandler(event,wp,but,x,y) == false)
			closeSkeletonWindow(win);
	}
}

void SPQRTreeWindow::display(int x, int y)
{
	GraphWin::display(x,y);
}


void SPQRTreeWindow::disableInput()
{
	get_window().disable_panel();
	SkeletonWindow *win;
	forall(win,skeletonWindows)
		win->disableInput();
}

void SPQRTreeWindow::enableInput()
{
	get_window().enable_panel();
	SkeletonWindow *win;
	forall(win,skeletonWindows)
		win->enableInput();
}


//---------------------------------------------------------
// skeleton windows
//---------------------------------------------------------

void SPQRTreeWindow::closeSkeletonWindow(SkeletonWindow *win)
{
	skeletonWindows.del_item(win->pos);
	skeletonWin[win->vCaller] = nil;
	delete win;
}

void SPQRTreeWindow::closeAllSkeletons()
{
	SkeletonWindow *win;
	forall(win,skeletonWindows)
		closeSkeletonWindow(win);
	skCount = 0;
}


//---------------------------------------------------------
// functions for menu entries
//---------------------------------------------------------

void SPQRTreeWindow::doSave()
{
	file_handler(SAVE_GW);
}


static string optionalS(int val, string str)
{
	if (val == 1) return string("%d ", val) + str;
	else return string("%d ", val) + str + "s";
}

void SPQRTreeWindow::doInfo()
{
	const graph &G = planeGraph.getGraph();
	const PlanarSPQRTree &T = planeGraph.spqrTree();

	panel P("Info");

	if (planeGraph.isGraphValid()) {
		P.text_item("\\bf\\blue Graph:");
		P.text_item("");
		P.text_item(string("%d nodes and %d edges", G.number_of_nodes(), G.number_of_edges()));
		P.text_item("");
		P.text_item("\\bf\\blue SPQR-Tree:");
		P.text_item("");
		P.text_item(string("\\red ") + optionalS(T.number_of_S_nodes(),"\\black S-node") +
			", \\red " + optionalS(T.number_of_P_nodes(),"\\black P-node") + " and \\red " +
			optionalS(T.number_of_R_nodes(),"\\black R-node"));
		P.text_item("");
		integer num = T.number_of_embeddings();
		if (num == 1)
			P.text_item("1 embedding");
		else
			P.text_item(num.to_string() + " embeddings");

	} else {
		P.text_item("Currently, there is no SPQR-tree available.");
	}
	P.button("Ok");
	P.open();
}


void SPQRTreeWindow::doCloseAll()
{
	closeAllSkeletons();
}


void SPQRTreeWindow::doAllToFront()
{
	SkeletonWindow *win;

	forall(win, skeletonWindows)
		win->setFocus();
}


void SPQRTreeWindow::doSetup()
{
	int theShapeSNode = (int)shapeSNode;
	int theShapePNode = (int)shapePNode;
	int theShapeRNode = (int)shapeRNode;
	color theColorSNode = colorSNode;
	color theColorPNode = colorPNode;
	color theColorRNode = colorRNode;

	panel P("Node Options");
	P.text_item("\\bf\\blue S-Node");
	P.color_item("color", theColorSNode);
	P.choice_item("shape",theShapeSNode,7,nshape_width,nshape_height,nshape_bits);
	P.text_item("\\bf\\blue p-Node");
	P.color_item("color", theColorPNode);
	P.choice_item("shape",theShapePNode,7,nshape_width,nshape_height,nshape_bits);
	P.text_item("\\bf\\blue R-Node");
	P.color_item("color", theColorRNode);
	P.choice_item("shape",theShapeRNode,7,nshape_width,nshape_height,nshape_bits);

	P.button ("Ok");
	int cancel = P.button ("Cancel");

	int result = P.open();
	if (result != cancel) {
		shapeSNode = (gw_node_shape)theShapeSNode;
		shapePNode = (gw_node_shape)theShapePNode;
		shapeRNode = (gw_node_shape)theShapeRNode;
		colorSNode = theColorSNode;
		colorPNode = theColorPNode;
		colorRNode = theColorRNode;

		setupTreeGraph();
		redraw();
	}
}


void SPQRTreeWindow::doSkeletonOpt()
{
	SkeletonOpts opts = planeGraph.skeletonOpts();

	panel P("Skeleton Options");
	P.text_item("\\bf\\blue Reference Edge");
	P.color_item("color", opts.colorRefEdge);
	P.lstyle_item("style", (line_style &)opts.styleRefEdge);
	P.lwidth_item("width", opts.widthRefEdge);
	P.text_item("\\bf\\blue Real Edge");
	P.color_item("color", opts.colorRealEdge);
	P.lstyle_item("style", (line_style &)opts.styleRealEdge);
	P.lwidth_item("width", opts.widthRealEdge);
	P.text_item("\\bf\\blue Virtual Edge");
	P.color_item("color", opts.colorVirtualEdge);
	P.lstyle_item("style", (line_style &)opts.styleVirtualEdge);
	P.lwidth_item("width", opts.widthVirtualEdge);
	P.text_item("\\bf\\blue Animation");
	P.int_item("steps",opts.animSteps,0,50);

	P.button ("Ok");
	int cancel = P.button ("Cancel");

	int result = P.open();
	if (result != cancel) {
		planeGraph.skeletonOpts(opts);
	}
}


void SPQRTreeWindow::doAnimOpt()
{
	int treeAnimSteps  = get_animation_steps();
	
	panel P("Animation Options");
	P.text_item("SPQR-Teee Window");
	P.int_item("steps",treeAnimSteps,0,50);

	P.button ("Ok");
	int cancel = P.button ("Cancel");

	int result = P.open();
	if (result != cancel) {
		set_animation_steps(treeAnimSteps);
	}
}


//---------------------------------------------------------
// mouse action functions
//---------------------------------------------------------

void SPQRTreeWindow::actionOpenSkeleton(GraphWin& gw, const point& p) 
{
	node v = gw.get_found_node();
	SPQRTreeWindow &spqrWindow = (SPQRTreeWindow &) gw;

	spqrWindow.openSkeleton(spqrWindow.GC.original(v));
}

void SPQRTreeWindow::buildPertinent(const SPQRTree &T, node v, list<node> &nodes,
	list<edge> &edges)
{
	const Skeleton &Sk = T.skeleton(v);
	const graph &M = Sk.get_graph();

	node vM;
	forall_nodes(vM,M)
		nodes.append(Sk.original(vM));

	edge eM;
	forall_edges(eM,M) {
		edge ePG = Sk.real_edge(eM);
		if (ePG != nil) {
			edges.append(ePG);
		}
	}

	node w;
	forall_adj_nodes(w,v)
		buildPertinent(T,w,nodes,edges);
}

void SPQRTreeWindow::actionShowPertinent(GraphWin& gw, const point& p) 
{
	node v = gw.get_found_node();
	SPQRTreeWindow &spqrWindow = (SPQRTreeWindow &) gw;

	spqrWindow.showPertinent(spqrWindow.GC.original(v));
}

void SPQRTreeWindow::showPertinent(node v)
{
	list<node> nodes;
	list<edge> edges;

	const SPQRTree &T = planeGraph.spqrTree();
	buildPertinent(T, v, nodes, edges);

	planeGraph.setPermanentMarkGraph(nodes, edges);
}

void SPQRTreeWindow::openSkeleton(node v)
{
	node vGC = GC.copy(v);

	if (skeletonWin[vGC]) {
		skeletonWin[vGC]->setFocus();

	} else {
		SkeletonWindow *win = new SkeletonWindow(*this,planeGraph,v);
		skeletonWin[vGC] = win;
		win->pos = skeletonWindows.append(win);
		win->vCaller = vGC;

		const int mx = 25, my = 50, mBot = 60, step = 30;
		if (my+skCount*step+200 > window::screen_height()-mBot) skCount = 0;
		win->display(mx+skCount*step,my+skCount*step);
		skCount++;
	}
}

void SPQRTreeWindow::randomEmbed (GraphWin &gw)
{
	SPQRTreeWindow &spqrWindow = (SPQRTreeWindow &) gw;
	StPlaneGraph &P = spqrWindow.planeGraph;

	P.spqrTree().random_embed(P.getPlaneGraph());
	P.embeddingHasChanged();
}


//---------------------------------------------------------
// static caller functions for menu entries
// usually delegate job to member functions
//---------------------------------------------------------

void SPQRTreeWindow::saveCaller (GraphWin &gw)
{
	SPQRTreeWindow &spqrWindow = (SPQRTreeWindow &) gw;
	spqrWindow.doSave();
}

void SPQRTreeWindow::infoCaller (GraphWin &gw)
{
	SPQRTreeWindow &spqrWindow = (SPQRTreeWindow &) gw;
	spqrWindow.doInfo();
}

void SPQRTreeWindow::allToFrontCaller (GraphWin &gw)
{
	SPQRTreeWindow &spqrWindow = (SPQRTreeWindow &) gw;
	spqrWindow.doAllToFront();
}

void SPQRTreeWindow::closeAllCaller (GraphWin &gw)
{
	SPQRTreeWindow &spqrWindow = (SPQRTreeWindow &) gw;
	spqrWindow.doCloseAll();
}

void SPQRTreeWindow::setupCaller (GraphWin &gw)
{
	SPQRTreeWindow &spqrWindow = (SPQRTreeWindow &) gw;
	spqrWindow.doSetup();
}

void SPQRTreeWindow::skeletonOptCaller (GraphWin &gw)
{
	SPQRTreeWindow &spqrWindow = (SPQRTreeWindow &) gw;
	spqrWindow.doSkeletonOpt();
}

void SPQRTreeWindow::animOptCaller (GraphWin &gw)
{
	SPQRTreeWindow &spqrWindow = (SPQRTreeWindow &) gw;
	spqrWindow.doAnimOpt();
}

void SPQRTreeWindow::helpMouseCaller (GraphWin &gw)
{
	panel P("Mouse Buttons");
	P.text_item("\\bf\\blue Clicking on a node:"); P.text_item("");
	P.text_item("\\red left\\black mouse button: \\orange open skeleton"); P.text_item("");
	P.text_item("\\red left\\black mouse button + \\red <CTRL>\\black key: \\orange show pertinent graph");
	P.text_item(""); P.text_item(""); P.text_item("");

	P.text_item("\\bf\\blue Clicking on the background:"); P.text_item("");
	P.text_item("\\red left\\black mouse button: \\orange zoom-up"); P.text_item("");
	P.text_item("\\red right\\black mouse button: \\orange zoom-down"); P.text_item("");
	P.text_item("");
	P.text_item("\\red left\\black mouse button and \\red dragging\\black : \\orange scroll graph"); P.text_item("");
	P.text_item("");
	P.text_item("\\red left\\black mouse button + \\red <CTRL>\\black key: \\orange zoom graph"); P.text_item("");
	P.text_item("\\red left\\black mouse button + \\red <SHIFT>\\black key: \\orange zoom area");

	P.button("Ok");
	P.open();
}

//---------------------------------------------------------
// StPlaneGraph
//
// author: Carsten Gutwenger
//---------------------------------------------------------

//#include <LEDA/plane_graph_alg.h>
#include <AGD/GraphCopy.h>
#include <AGD/AGD.h>
#include <AGD/PlanarSPQRTree.h>
#include <AGD/TopologyModule.h>
#include <AGD/LongestPathRanking.h>
#include <LEDA/node_set.h>
#include <LEDA/edge_set.h>


//---------------------------------------------------------
// SkeletonOpts
//---------------------------------------------------------

SkeletonOpts::SkeletonOpts() :
	colorRefEdge(red), colorRealEdge(black), colorVirtualEdge(blue),
	widthRefEdge(1), widthRealEdge(1), widthVirtualEdge(1),
	styleRefEdge(solid_edge), styleRealEdge(solid_edge), styleVirtualEdge(dotted_edge),
	animSteps(16) { }


//---------------------------------------------------------
// Constructor / Destructor
//---------------------------------------------------------

StPlaneGraph::~StPlaneGraph()
{
	delete markedEdges;
	delete markedNodes;
	delete spqr;
}


//---------------------------------------------------------
// subscribe / unsubscribe
//---------------------------------------------------------

void StPlaneGraph::subscribe(GraphObserver &observer)
{
	if (observer.observable) {
		observer.observable->unsubscribe(observer);
	}
	observer.pos = observers.append(&observer);
	observer.observable = this;
}


void StPlaneGraph::unsubscribe(GraphObserver &observer)
{
	observers.del_item(observer.pos);
	observer.pos = nil;
	observer.observable = nil;
}


//---------------------------------------------------------
// Access methods
//---------------------------------------------------------

edge StPlaneGraph::refEdge() const
{
	if (spqr == nil) return nil;
	edge e_st = spqrTree().root_edge();
	return PG.original(e_st) ? PG.original(e_st) : PG.original(PG.reversal(e_st));
}

void StPlaneGraph::callStLayout(GraphCopy &GC, edge e_st, LayoutInterface &gwi)
{
	stPlanarLayouter.call_plane(GC,e_st,gwi);
}

bool StPlaneGraph::isPermanentMarkedGraph() const
{
	return (markedEdges && !(markedEdges->empty() && markedNodes->empty()));
}


//---------------------------------------------------------
// Update operations
// -> create notifiaction messages
//---------------------------------------------------------

void StPlaneGraph::notifyAll(ChangeStatus status)
{
	list_item it;
	for (it = observers.first(); it != nil; it = observers.succ(it))
		observers[it]->notify(status);
}

void StPlaneGraph::topologyHasChanged(edge eRef, const LayoutInterface *li)
{
	const int n = G.number_of_nodes();
	const int m = G.number_of_edges();
	bool valid = ((n == 2 && m >= 3) || (n >= 3 && Is_Biconnected(G) &&
		AGD::is_planar(G)));

	topologyHasChanged(eRef, li, valid);
}


void StPlaneGraph::topologyHasChanged(edge eRef, const LayoutInterface *li, bool valid)
{
	delete spqr;
	delete markedNodes;
	delete markedEdges;
	spqr = nil;

	if (valid) {
		if (li != nil) {
			PlaneGraphCopy *ppg = &TopologyModule().call(G,*li,false);
			PG = *ppg;
			delete ppg;
		} else
			PG = PlaneGraphCopy(G,false);

		markedNodes = new node_set(G);
		markedEdges = new edge_set(G);

		edge e_st = find_e_st(eRef);
		if (eRef == nil) eRef = e_st;

		if (e_st && e_st == eRef) {
			spqr = new PlanarSPQRTree(PG,PG.chain(eRef).head(),true);
			stnum = node_array<int>(G);
			LongestPathRanking().call(G,stnum);

		} else if (eRef != nil) {
			spqr = new PlanarSPQRTree(PG,PG.chain(eRef).head(),true);
			make_st_graph();

		} else {
			spqr = new PlanarSPQRTree(PG,true);
			make_st_graph();
		}

		theOldRefEdge = nil;
		spqr->embed(PG);
	}

	notifyAll((li == nil) ? topologyChanged : topologyEdited);
}


void StPlaneGraph::reverse(node vT)
{
	spqr->reverse(vT);
	embeddingHasChanged();
}

void StPlaneGraph::swap(node vT, edge e1, edge e2)
{
	spqr->swap(vT,e1,e2);
	embeddingHasChanged();
}

void StPlaneGraph::embeddingHasChanged()
{
	spqr->embed(PG);

	notifyAll(embeddingChanged);
}


void StPlaneGraph::changeReferenceEdge(edge newRefEdge)
{
	unsetPermanentMarkGraph();
	if (spqr->root_edge() != newRefEdge) {
		theOldRefEdge = refEdge();
		spqr->root_tree_at(newRefEdge);

		make_st_graph();

		notifyAll(referenceEdgeChanged);
	}
}


void StPlaneGraph::reverseRefEdge()
{
	G.rev_edge(refEdge());
	make_st_graph();

	notifyAll(referenceEdgeChanged);
}


void StPlaneGraph::setStLayoutOptions(StPlanarGridLayout::CompactionType compaction,
	double separation)
{
	stPlanarLayouter.compaction(compaction);
	stPlanarLayouter.separation(separation);
	notifyAll(stLayoutOptionsChanged);
}

void StPlaneGraph::getStLayoutOptions(StPlanarGridLayout::CompactionType &compaction,
	double &separation)
{
	compaction = stPlanarLayouter.compaction();
	separation = stPlanarLayouter.separation();
}


void StPlaneGraph::skeletonOpts(const SkeletonOpts &opts)
{
	_skeletonOpts = opts;
	notifyAll(skeletonOptsChanged);
}


void StPlaneGraph::setPermanentMarkGraph(list<node> &nodes, list<edge> &edges)
{
	markedNodes->clear();
	markedEdges->clear();

	node v;
	forall(v,nodes) {
		node vG = PG.original(v);
		if (vG != nil) markedNodes->insert(vG);
	}

	edge e;
	forall(e,edges) {
		edge eG = PG.original(e);
		if (eG != nil) markedEdges->insert(eG);
	}

	notifyAll(permMarkGraphChanged);
}

void StPlaneGraph::unsetPermanentMarkGraph()
{
	if (!isPermanentMarkedGraph())
		return;

	markedNodes->clear();
	markedEdges->clear();
	notifyAll(permMarkGraphCleared);
}


//---------------------------------------------------------
// direct edges as st-planar graph
//---------------------------------------------------------

edge StPlaneGraph::find_e_st(edge eRef)
{
	node s, t;

	if (!AGD::is_st_graph(G,s,t)) return nil;

	if (eRef && source(eRef) == s && target(eRef) == t)
		return eRef;

	edge e;
	forall_out_edges(e,s)
		if (target(e) == t) return e;

	return nil;
}


#include <AGD/generators.h>

void StPlaneGraph::make_st_graph()
{
	stnum = node_array<int>(G);
	edge e_ref = refEdge();

	list<node> stlist;

	ST_NUMBERING1(G,stnum,stlist,e_ref);

	edge e;
	forall_edges(e, G) {
		if (stnum[source(e)] > stnum[target(e)])
			G.rev_edge(e);
	}
}

#include <AGD/generators.h>
#include <AGD/GraphWinInterface.h>
#include <AGD/GraphWinAnimInterface.h>
#include <AGD/StPlanarLayout.h>
#include <AGD/PlanAug.h>
#include <AGD/LEDAMakeBiconnected.h>
#include <AGD/PlanarSubgraph.h>
#include <AGD/LEDAMakeAcyclic.h>
#include <AGD/GreedyCycleRemoval.h>
#include <AGD/ReducedDigraph.h>
#include <LEDA/graph_gen.h>

#include <AGD/StPlanarGridLayout.h>
#include <AGD/StPlanarLayout.h>
#include <AGD/SugiyamaLayout.h>
#include <AGD/FastHierarchyLayout.h>

#include <AGD/LongestPathRanking.h>
#include <AGD/BarycenterHeuristic.h>
#include <AGD/MedianHeuristic.h>
#include <AGD/SiftingHeuristic.h>

#include <AGD/QuasiOrthogonalLayout.h>
#include <AGD/MixedModelLayout.h>
#include <AGD/PlanarizationLayout.h>
#include <AGD/SubgraphPlanarizer.h>


EditWindow::EditWindow(GraphWin &GW, edge eRef, color ec, int ew, color rec, int rew) : GraphWin("Edit Window"),
	gc(GW.get_graph()), gw(GW),
	edgeColor(ec), refEdgeColor(rec), edgeWidth(ew), refEdgeWidth(rew),
	rbg_multi(false), rbg_map(false), rst_multi(false), rst_stprob(1.0)
{
	stPlanarGrid = new StPlanarGridLayout;
	stPlanar     = new StPlanarLayout;
	sugiyama     = new SugiyamaLayout;
	fhl          = new FastHierarchyLayout;
	fhl->node_distance(20);
	fhl->layer_distance(20);
	sug_cross_min = 0;
	sug_cycle_removal = 0;

	quod = new QuasiOrthogonalLayout;
	mixed = new MixedModelLayout;
	planSub = new PlanarSubgraph;

	// setup GraphWin
	set_default_menu(M_UNDO);
	clear_actions();

	// create menu
	int fileMenuId = add_menu("&File");
    add_simple_call(newGraph, "&New", fileMenuId);
    add_separator(fileMenuId);
    add_simple_call(gw_load_handler,    "&Load GW-Graph...",fileMenuId);
    add_simple_call(gw_save_handler,    "&Save GW-Graph...",fileMenuId);
    add_separator(fileMenuId);
    add_simple_call(gw_load_gml_handler,"L&oad GML-Graph...",fileMenuId);
    add_simple_call(gw_save_gml_handler,"S&ave GML-Graph...",fileMenuId);
    add_separator(fileMenuId);
    add_simple_call(gw_save_ps_handler, "&Write Postscript...",fileMenuId);
    add_simple_call(gw_print_ps_handler,"&Print...",           fileMenuId);

	int editMenuId = add_menu("&Edit");
    add_member_call(&GraphWin::select_all_nodes, "Select all nodes", editMenuId);
    add_member_call(&GraphWin::deselect_all_nodes, "Deselect all nodes", editMenuId);
    add_separator(editMenuId);
    add_member_call(&GraphWin::select_all_edges,"Select all edges", editMenuId);
    add_member_call(&GraphWin::deselect_all_edges,"Deselect all edges", editMenuId);
    add_separator(editMenuId);
#ifdef __win32__
	add_simple_call(gw_clipboard_handler,"Copy", editMenuId);
#endif
    add_simple_call(del_selection,"Delete", editMenuId);
    add_separator(editMenuId);
	add_simple_call(setRefEdgeHandler, "Set reference edge", editMenuId);

    int graphMenuId = add_menu("&Graph");
	int createMenuId = add_menu("&Create", graphMenuId);
	add_simple_call(randomPlanarHandler, "planar...", createMenuId);
	add_simple_call(randomPlanarBicHandler, "planar biconnected...", createMenuId);
	add_simple_call(randomPlanarTricHandler, "planar triconnected...", createMenuId);
    add_separator(createMenuId);
	add_simple_call(randomStPlanarHandler, "st-planar...", createMenuId);
	add_simple_call(randomSPHandler, "series-parallel...", createMenuId);

    int testMenuId = add_menu("&Test",graphMenuId);
	add_simple_call(gw_test_planarity,      "planar",          testMenuId);
    add_separator(testMenuId);
	add_simple_call(gw_test_simplicity,     "simple",          testMenuId);
	add_simple_call(gw_test_loopfree,       "self-loops",      testMenuId);
    add_separator(testMenuId);
	add_simple_call(gw_test_connectivity,   "connected",       testMenuId);
	add_simple_call(gw_test_biconnectivity, "biconnected",     testMenuId);
	add_simple_call(gw_test_triconnectivity,"triconnected",    testMenuId);
    add_separator(testMenuId);
	add_simple_call(gw_test_acyclicity,     "acyclic",         testMenuId);
	add_simple_call(gw_test_st_graph,       "st-graph",        testMenuId);
	add_simple_call(gw_test_spness,         "series-parallel", testMenuId);

    int removeMenuId = add_menu("&Remove",graphMenuId);
	add_simple_call(gw_del_loops,           "self-loops",      removeMenuId);
	add_simple_call(gw_del_isolated,        "isolated nodes",  removeMenuId);

    int makeMenuId = add_menu("&Make",graphMenuId);
	add_simple_call(callPlanSub,            "planar",             makeMenuId);
    add_separator(makeMenuId);
	add_simple_call(callPlanAug,            "planar biconnected", makeMenuId);
	add_simple_call(callMakeBic,            "biconnected",        makeMenuId);
    add_separator(makeMenuId);
	add_simple_call(callMakeAcyclic,        "acyclic (LEDA)",     makeMenuId);
	add_simple_call(callGreedyCycleRemoval, "acyclic (greedy)",   makeMenuId);
	add_simple_call(callReducedDigraph,     "reduced",            makeMenuId);

	int layoutMenuId = add_menu("&Layout");
	add_simple_call(callStPlanarGrid, "dominance...",     layoutMenuId);
	add_simple_call(callStPlanar,     "st-planar...",     layoutMenuId);
	add_simple_call(callSugiyama,     "Sugiyama...",      layoutMenuId);
    add_separator(layoutMenuId);
	add_simple_call(callPlanarization,"Planarization...", layoutMenuId);
	add_simple_call(callQuod,         "Quasi-Orthogonal", layoutMenuId);
	add_simple_call(callMixedModel,   "Mixed-Model...",   layoutMenuId);

    int windowMenuId = add_menu("&Window");
	add_simple_call(gw_zoom_default, "default window...", windowMenuId);
    add_simple_call(gw_center_graph, "center graph",      windowMenuId);
    add_separator(windowMenuId);
    add_simple_call(gw_zoom_graph,   "zoom to graph",     windowMenuId);
    add_simple_call(gw_zoom_area,    "zoom to area",      windowMenuId);
    add_separator(windowMenuId);
    add_simple_call(gw_zoom_up,      "zoom +",             windowMenuId);
    add_simple_call(gw_zoom_down,    "zoom -",             windowMenuId);
    add_separator(windowMenuId);
    add_simple_call(gw_unzoom,       "zoom undo",          windowMenuId);


	add_simple_call(done_handler, "ok");
	add_simple_call(exit_handler, "cancel");

	set_action( A_LEFT                               , leda::gw_new_node);
	set_action( A_LEFT   |                    A_NODE , leda::gw_new_edge);
	set_action( A_LEFT   |                    A_FRAME, leda::gw_new_edge);
	set_action( A_LEFT   | A_CTRL           | A_EDGE , leda::gw_zoom_edge);
	set_action( A_LEFT   |           A_DRAG | A_FRAME, leda::gw_move_node);
	set_action( A_LEFT   |           A_DRAG | A_NODE , leda::gw_move_node);
	set_action( A_LEFT   |           A_DRAG | A_EDGE , leda::gw_move_edge);
	set_action( A_LEFT   |           A_DRAG          , leda::gw_scroll_graph);
	set_action( A_LEFT   | A_SHIFT | A_DRAG | A_NODE , leda::gw_move_component);
	set_action( A_LEFT   | A_DOUBLE| A_DRAG | A_NODE , leda::gw_move_component);
	set_action( A_LEFT   | A_SHIFT | A_IMMEDIATE     , leda::gw_zoom_area);
	set_action( A_MIDDLE |                    A_NODE , leda::gw_select_node_toggle);
	set_action( A_MIDDLE |                    A_FRAME, leda::gw_select_node_toggle);
	set_action( A_MIDDLE |                    A_EDGE , leda::gw_select_edge_toggle);
	set_action( A_MIDDLE |           A_DRAG          , leda::gw_select_area);
	set_action( A_MIDDLE |           A_DRAG | A_NODE , leda::gw_select_area);
	set_action( A_MIDDLE |           A_DRAG | A_EDGE , leda::gw_select_area);
	set_action( A_RIGHT  |      A_IMMEDIATE | A_NODE , gw_open_node_menu);
	set_action( A_RIGHT  |      A_IMMEDIATE | A_FRAME, gw_open_node_menu);
	set_action( A_RIGHT  |      A_IMMEDIATE | A_EDGE , gw_open_edge_menu);

	set_del_edge_handler(del_edge_handler);

	set_graph(gc);

	node v;
	forall_nodes(v,gc) {
		set_position(v, gw.get_position(gc.original(v)));
	}

	edge e;
	forall_edges(e,gc) {
		set_bends(e, gw.get_bends(gc.original(e)));
	}

	set_edge_color(edgeColor,true);
	set_edge_thickness(edgeWidth,true);

	if (eRef != nil) {
		refEdge = gc.chain(eRef).head();
		setupRefEdge(refEdge);
	} else refEdge = nil;
}

EditWindow::~EditWindow()
{
	delete planSub;
	delete mixed;
	delete quod;

	delete sugiyama;
	delete stPlanar;
	delete stPlanarGrid;
	delete fhl;
}


bool EditWindow::open()
{
	set_flush(false);
	GraphWin::display();
	zoom_graph();
	return GraphWin::edit();
}


edge EditWindow::assignGraph()
{
	set_flush(false);
	graph &g = gw.get_graph();

	g.clear();

	node_array<node> mapV(gc);
	edge_array<edge> mapE(gc);

	node v;
	forall_nodes(v,gc) {
		mapV[v] = g.new_node();
	}

	edge e;
	forall_edges(e,gc) {
		mapE[e] = g.new_edge(mapV[source(e)],mapV[target(e)]);
	}

	gw.set_flush(false);
	gw.update_graph();

	forall_nodes(v,gc) {
		gw.set_position(mapV[v],get_position(v));
	}

	forall_edges(e,gc) {
		gw.set_bends(mapE[e],get_bends(e));
	}

	gw.zoom_graph();

	return (refEdge != nil) ? mapE[refEdge] : nil;
}


void EditWindow::setupRefEdge(edge e)
{
	set_color(e, refEdgeColor);
	set_thickness(e, 1.5*refEdgeWidth);
}


void EditWindow::unsetupRefEdge(edge e)
{
	set_color(e, edgeColor);
	set_thickness(e, 1.5*edgeWidth);
}


void EditWindow::doSetRefEdge()
{
	edge e = read_edge();

	if (e != nil) {
		if (refEdge)
			unsetupRefEdge(refEdge);

		refEdge = e;

		setupRefEdge(refEdge);
	}
}


bool EditWindow::del_edge_handler(GraphWin &gw, edge e)
{
	EditWindow &editWin = (EditWindow &)gw;

	if (editWin.refEdge == e)
		editWin.refEdge = nil;

	return true;
}


void EditWindow::newGraph(GraphWin &GW)
{
	EditWindow &editWin = (EditWindow &)GW;

	editWin.clear_graph();
	editWin.refEdge = nil;
}


void EditWindow::setRefEdgeHandler(GraphWin &GW)
{
	EditWindow &editWin = (EditWindow &)GW;
	editWin.doSetRefEdge();
}


void EditWindow::gw_open_node_menu(GraphWin& gw, const point& p)
{
	window& W = gw.get_window();

	node v = gw.get_found_node();
	if (v==nil) return;

	point q = gw.get_position(v);

	int x = W.xpix(q.xcoord() - gw.get_radius1(v)) - 1;
	int y = W.ypix(q.ycoord() - gw.get_radius2(v)) + 2;

	int dy = W.height() - (y+80);
	if (dy < 0) y += dy;

	menu  M;
	M.button("delete", 1);
	M.button((gw.is_selected(v)) ? "deselect" : "select", 2);

	int but = M.open(W,x,y);

	bool f = gw.set_flush(true);

	if (but != NO_BUTTON)
	switch (but) {
	case 1:
		gw_del_node(gw,v);
		break;
	case 2:
		gw.set_select(v,!gw.is_selected(v));
		break;
	}

	gw.set_flush(f);
}

list_item EditWindow::min_bend_dist(const list<point>& P, const point& p) 
{
	if ( P.size() == 2) return nil;

	list_item min_it = P.succ(P.first());
	double    min_d  = p.distance(P[min_it]);

	list_item it = P.succ(min_it);

	while (it != P.last()) {
		double d = p.distance(P[it]);
		if (d < min_d) {
			min_d  = d;
			min_it = it;
		}
		it = P.succ(it);
	}
	return min_it;
}

void EditWindow::gw_open_edge_menu(GraphWin& gw, const point& p)
{
	window& W = gw.get_window();

	edge e = gw.get_found_edge();
	if (e == nil) return;

	int x = W.xpix(p.xcoord());
	int y = W.ypix(p.ycoord());

	int dy = W.height() - (y+80);
	if (dy < 0) y += dy;

	list<point> P = gw.get_edge_points(e);
	list_item it = min_bend_dist(P,p);
	if (it != nil && p.distance(P[it]) > W.pix_to_real(10)) it = nil;

	bool s = gw.set_select(e,true);

	menu  M;
	M.button("delete", 6);
	if (it) M.button("rm bend",4);
	if ((P.length() == 3 && it == nil) || P.length() > 3) 
		M.button("rm bends",5);
	M.button((s) ? "deselect" : "select", 3);

	int but = M.open(W,x,y);
	gw.set_select(e,s);
	bool f = gw.set_flush(true);

	if (but != NO_BUTTON)
	switch (but) {
	case 3:
		gw_select_edge_toggle(gw,p);
		break;

	case 4:
		P.del_item(it); P.pop_back(); P.pop_front();
		gw.changing_layout(e);
		gw.set_bends(e,P);
		break;

	case 5:
		gw.changing_layout(e);
		gw.remove_bends(e);
		break;

	case 6:
		gw_del_edge(gw,p);
		break;
	}

	gw.set_flush(f);
}


// graph generators

void EditWindow::embed_random(const graph& G, node_array<double>& x,
	node_array<double>& y) 
{
	random_source ran;
	node v;
	forall_nodes(v,G)  ran >> x[v] >> y[v];
}

void EditWindow::gen_pre(int n, int m)
{
	set_gen_nodes (n);
	set_gen_edges (m);
	graph& G = get_graph();

	G.clear();
	update_graph();
}

void EditWindow::gen_post()
{
	graph& G=get_graph();
	bool old_flush = set_flush(false); 

	update_graph();

	node_array<double> x(G);
	node_array<double> y(G);
	embed_random(G,x,y);
	update_graph();
	adjust_coords_to_win(x,y);
	set_layout(x,y);
	embed_edges();
	set_flush(old_flush);
	redraw();
	call_init_graph_handler();
}

void EditWindow::genRandomGraph(GraphWin& gw1, void (*f)(graph&,int,int), string s)
{
	EditWindow& gw = (EditWindow&)gw1;
	if (!gw.check_init_graph_handler()) return;

	panel P;
	P.text_item("\\bf\\blue " + s);
	P.text_item("");

	int n = gw.get_gen_nodes(),
		m = gw.get_gen_edges();

	P.int_item ("nodes", n);
	P.int_item ("edges", m);
	P.button ("ok", 1);
	P.button ("cancel", 0);

	if (gw.open_panel(P) == 0 || n < 0 || m < 0) return;

	gw.gen_pre(n,m);
	f (gw.get_graph(), n, m);
	gw.gen_post();
}

void EditWindow::randomPlanarHandler (GraphWin& GW)
{
	genRandomGraph (GW, random_planar_graph,
		"Create Random Planar Graph");
}

void EditWindow::randomPlanarBicHandler(GraphWin &gw1)
{
	EditWindow& gw = (EditWindow&)gw1;
	if (!gw.check_init_graph_handler()) return;

	panel P;
	P.text_item("\\bf\\blue Create Biconnected Planar Graph");
	P.text_item("");

	int n = gw.get_gen_nodes(),
		m = gw.get_gen_edges();
	
	P.int_item ("nodes", n);
	P.int_item ("edges", m);
	P.bool_item ("multi-edges", gw.rbg_multi);
	P.bool_item ("planar map", gw.rbg_map);

	P.button ("ok", 1);
	P.button ("cancel", 0);

	if (gw.open_panel(P) == 0 || n < 0 || m < 0) return;

	gw.gen_pre(n,m);
	planar_biconnected_graph (gw.get_graph(), n, m, gw.rbg_multi, gw.rbg_map);
	gw.gen_post();
}

void EditWindow::randomPlanarTricHandler (GraphWin& GW)
{
	genRandomGraph (GW, planar_triconnected_graph,
		"Create Triconnected Planar Graph");
}

void EditWindow::randomStPlanarHandler(GraphWin &gw1)
{
	EditWindow& gw = (EditWindow&)gw1;
	if (!gw.check_init_graph_handler()) return;

	panel P;
	P.text_item("\\bf\\blue Create st-Planar Graph");
	P.text_item("");

	int n = gw.get_gen_nodes(),
		m = gw.get_gen_edges();
	
	P.int_item ("nodes", n);
	P.int_item ("edges", m);
	P.bool_item ("multi-edges", gw.rst_multi);
	P.double_item ("st probability", gw.rst_stprob);

	P.button ("ok", 1);
	P.button ("cancel", 0);

	if (gw.open_panel(P) == 0 || n < 0 || m < 0) return;

	bool old_flush=gw.set_flush(false); 
	gw.gen_pre(n,m);
	st_planar_graph (gw.get_graph(), n, m, gw.rst_multi, gw.rst_stprob);

	gw.update_graph();
	GraphWinInterface A (gw);
	A.flush(false);
	StPlanarLayout st_planar;
	st_planar.call(gw.get_graph(),A);
	gw.fill_window();

	gw.set_flush(old_flush);
	gw.redraw();
}

void EditWindow::randomSPHandler(GraphWin &gw1)
{
	EditWindow& gw = (EditWindow&)gw1;
	if (!gw.check_init_graph_handler()) return;

	panel P;
	P.text_item("\\bf\\blue Create Series-Parallel Graph");
	P.text_item("");

	int n = gw.get_gen_nodes(),
		m = gw.get_gen_edges();
	
	P.int_item ("nodes", n);
	P.int_item ("edges", m);

	P.button ("ok", 1);
	P.button ("cancel", 0);

	if (gw.open_panel(P) == 0 || n < 0 || m < 0) return;

	bool old_flush=gw.set_flush(false); 
	gw.gen_pre(n,m);
	random_sp_graph (gw.get_graph(), n, m);

	gw.update_graph();
	GraphWinInterface A (gw);
	A.flush(false);
	StPlanarLayout st_planar;
	st_planar.call(gw.get_graph(),A);
	gw.fill_window();

	gw.set_flush(old_flush);
	gw.redraw();
}


// testing

void EditWindow::make_yes_panel(panel& P, string s, bool proof)
{
	P.text_item("");
	P.text_item("\\bf This graph is \\blue " + s + "\\black .");
	P.button("ok");
	if (proof) P.fbutton("proof",1);
}


void EditWindow::make_no_panel(panel& P, string s, bool proof)
{
	P.text_item("");
	P.text_item("\\bf This graph is \\red not " + s + "\\black .");
	P.button("ok",0);
	if (proof) P.fbutton("proof",1);
}


void EditWindow::gw_test_simplicity(GraphWin& gw)
{ 
  panel P;
  
  if (Is_Simple(gw.get_graph()))
     make_yes_panel(P,"simple");
  else
     make_no_panel(P,"simple");

  gw.open_panel(P);
}



void EditWindow::gw_test_connectivity(GraphWin& gw)
{
	graph& G = gw.get_graph();
	panel P;
	if (Is_Connected(G))
		make_yes_panel(P,"connected",false);
	else
		make_no_panel(P,"connected",false);

	gw.open_panel(P);
}



void EditWindow::gw_test_biconnectivity(GraphWin& gw)
{
	graph& G = gw.get_graph();
	panel P;
	if (Is_Biconnected(G))
		make_yes_panel(P,"biconnected");
	else
		make_no_panel(P,"biconnected",false);

	gw.open_panel(P);
}


void EditWindow::gw_test_triconnectivity(GraphWin& gw)
{
	graph& G = gw.get_graph();
	panel P;
	if (Is_Triconnected(G))
		make_yes_panel(P,"triconnected");
	else
		make_no_panel(P,"triconnected",false);

	gw.open_panel(P);
}


void EditWindow::gw_test_planarity(GraphWin& gw)
{
	graph& G = gw.get_graph();
	panel P;
	if (Is_Planar(G))
		make_yes_panel(P,"planar",false);
	else
		make_no_panel(P,"planar",false);

	gw.open_panel(P);
}


void EditWindow::gw_test_spness(GraphWin& gw)
{
	graph& G = gw.get_graph();
	panel P;
	if (Is_Series_Parallel(G))
		make_yes_panel(P,"series-parallel",false);
	else
		make_no_panel(P,"series-parallel",false);

	gw.open_panel(P);
}

void EditWindow::gw_test_acyclicity(GraphWin& gw)
{
	graph& G = gw.get_graph();
	panel P;
	if (Is_Acyclic(G))
		make_yes_panel(P,"acyclic",false);
	else
		make_no_panel(P,"acyclic",false);

	gw.open_panel(P);
}

void EditWindow::gw_test_loopfree(GraphWin& gw)
{
	graph& G = gw.get_graph();
	panel P;
	P.text_item("");

	if (AGD::has_no_self_loops(G))
		P.text_item("\\bf This graph contains \\blue no self-loops\\black.");
	else
		P.text_item("\\bf This graph contains \\blue self-loops\\black.");

	P.button("ok");
	gw.open_panel(P);
}

void EditWindow::gw_test_st_graph(GraphWin& gw)
{
	graph& G = gw.get_graph();
	panel P;
	P.text_item("");

	node s, t;
	if (AGD::is_st_graph(G,s,t))
		P.text_item("\\bf This graph is an \\blue st-graph\\black.");
	else
		P.text_item("\\bf This graph is not an \\blue st-graph\\black.");

	P.button("ok");
	gw.open_panel(P);
}

void EditWindow::gw_del_loops(GraphWin& gw)
{
	if (!Delete_Loops(gw.get_graph()).empty()) {
		gw.update_edges();
		if (!gw.get_flush()) {
			gw.embed_edges();
			gw.redraw();
		}
	}
}

void EditWindow::gw_del_isolated(GraphWin& gw)
{
	graph &G = gw.get_graph();
	bool nodesDeleted = false;

	node v;
	forall_nodes(v,G)
		if (G.degree(v) == 0) {
			G.del_node(v);
			nodesDeleted = true;
		}

	if (nodesDeleted) {
		gw.update_nodes();
		gw.redraw();
	}
}


void EditWindow::callPlanAug (GraphWin& GW)
{
	EditWindow &gwAGD = (EditWindow&)GW;
	PlanAug planAug;
	callAugment (gwAGD, planAug);
}

void EditWindow::callMakeBic (GraphWin& GW)
{
	EditWindow &gwAGD = (EditWindow&)GW;
	LEDAMakeBiconnected makeBiconnected;
	callAugment (gwAGD, makeBiconnected);
}

void EditWindow::callPlanSub (GraphWin& GW)
{
	EditWindow &gwAGD = (EditWindow&)GW;
	callSubgraph (gwAGD, *gwAGD.planSub);
}

void EditWindow::callMakeAcyclic (GraphWin& GW)
{
	EditWindow &gwAGD = (EditWindow&)GW;
	LEDAMakeAcyclic makeAcyclic;
	callSubgraph (gwAGD, makeAcyclic);
}

void EditWindow::callGreedyCycleRemoval (GraphWin& GW)
{
	EditWindow &gwAGD = (EditWindow&)GW;
	GreedyCycleRemoval greedyCycleRemoval;
	callSubgraph (gwAGD, greedyCycleRemoval);
}

void EditWindow::callReducedDigraph (GraphWin& GW)
{
	EditWindow &gwAGD = (EditWindow&)GW;
	ReducedDigraph reducedDigraph;
	callSubgraph (gwAGD, reducedDigraph);
}


void EditWindow::callStPlanarGrid (GraphWin& GW)
{
	EditWindow &gwAGD = (EditWindow&)GW;
	
	int compaction = (int)gwAGD.stPlanarGrid->compaction();

	panel P;

	P.text_item("\\bf\\blue Call Dominance Layout");
	P.text_item("");
	P.choice_item ("compaction:", compaction, "none", "symmetry", "min. area");

	P.button ("Ok");
	int cancel = P.button ("Cancel");

	if (gwAGD.open_panel(P) != cancel) {
		gwAGD.stPlanarGrid->compaction((StPlanarGridLayout::CompactionType)compaction);
		callLayout (gwAGD, *gwAGD.stPlanarGrid);
	}
}

void EditWindow::callStPlanar (GraphWin& GW)
{
	EditWindow &gwAGD = (EditWindow&)GW;
	
	int compaction    = (int)gwAGD.stPlanar->compaction();
	double separation = gwAGD.stPlanar->separation();

	panel P;

	P.text_item("\\bf\\blue Call st-Planar Layout");
	P.text_item("");
	P.choice_item ("compaction:", compaction, "none", "symmetry", "min. area");
	P.real_item("separation:", separation);

	P.button ("Ok");
	int cancel = P.button ("Cancel");

	if (gwAGD.open_panel(P) != cancel) {
		gwAGD.stPlanar->compaction((StPlanarGridLayout::CompactionType)compaction);
		gwAGD.stPlanar->separation(separation);

		callLayout (gwAGD, *gwAGD.stPlanar);
	}
}

void EditWindow::callSugiyama (GraphWin& GW)
{
	EditWindow &gwAGD = (EditWindow&)GW;
	
	int cross_min     = gwAGD.sug_cross_min;
	int cycle_removal = gwAGD.sug_cycle_removal;
	int runs          = gwAGD.sugiyama->runs();

	double node_dist      = gwAGD.fhl->node_distance();
	double level_dist     = gwAGD.fhl->layer_distance();
	bool fixed_level_dist = gwAGD.fhl->fixed_layer_distance();

	panel P;

	P.text_item("\\bf\\blue Call Sugiyama-Layout");
	P.text_item("");
	P.choice_item ("cycle removal:", cycle_removal, "DFS", "Greedy");
	P.text_item("");
	P.choice_item ("heuristic:", cross_min, "Barycenter", "Median", "Sifting");
	P.int_item ("runs:", runs);
	P.text_item("");
	P.real_item ("node distance:", node_dist);
	P.real_item ("level distance:", level_dist);
	P.bool_item ("fixed level distance:", fixed_level_dist);

	P.button ("Ok");
	int cancel = P.button ("Cancel");

	if (gwAGD.open_panel(P) != cancel) {
		gwAGD.sugiyama->runs(runs);

		gwAGD.fhl->node_distance(node_dist);
		gwAGD.fhl->layer_distance(level_dist);
		gwAGD.fhl->fixed_layer_distance(fixed_level_dist);
		gwAGD.sugiyama->set_comp_coord(*gwAGD.fhl);

		LongestPathRanking ranking;
		gwAGD.sug_cycle_removal = cycle_removal;
		switch(cycle_removal) {
		case 0:
			ranking.set_subgraph(LEDAMakeAcyclic());
			break;
		case 1:
			ranking.set_subgraph(GreedyCycleRemoval());
			break;
		}
		
		gwAGD.sugiyama->set_ranking(ranking);

		gwAGD.sug_cross_min = cross_min;
		switch (cross_min) {
		case 0:
			gwAGD.sugiyama->set_cross_min(BarycenterHeuristic());
			break;
		case 1:
			gwAGD.sugiyama->set_cross_min(MedianHeuristic());
			break;
		case 2:
			gwAGD.sugiyama->set_cross_min(SiftingHeuristic());
			break;
		}
		callLayout (gwAGD, *gwAGD.sugiyama);
	}
}


void EditWindow::callPlanarization (GraphWin& GW)
{
	EditWindow &gwAGD = (EditWindow&)GW;
	
	int runs = gwAGD.planSub->runs();

	panel P;
	
	P.text_item("Call Planarization Layout");
	P.text_item("");
	P.int_item ("runs:", runs);

	P.button ("Ok");
	int cancel = P.button ("Cancel");

	if (gwAGD.open_panel(P) != cancel) {
		PlanarizationLayout planarization;
		gwAGD.planSub->runs (runs);
		SubgraphPlanarizer P;
		P.set_subgraph(*gwAGD.planSub);
		planarization.set_planarizer(P);
		callLayout (gwAGD, planarization);
	}
}

void EditWindow::callMixedModel (GraphWin& GW)
{
	EditWindow &gwAGD = (EditWindow&)GW;
	
	double base_ratio   = gwAGD.mixed->base_ratio();

	panel P;
	
	P.text_item("Call Mixed-Model Layout");
	P.text_item("");
	P.real_item ("base ratio:", base_ratio);

	P.button ("Ok");
	int cancel = P.button ("Cancel");

	if (gwAGD.open_panel(P) != cancel) {
		gwAGD.mixed->base_ratio (base_ratio);
		callLayout (gwAGD, *gwAGD.mixed);
	}
}

void EditWindow::callQuod (GraphWin& GW)
{
	EditWindow &gwAGD = (EditWindow&)GW;
	
	//panel P;
	
	//P.text_item("Call Quasi-Orthogonal Layout");
	//P.text_item("");

	//P.button ("Ok");
	//int cancel = P.button ("Cancel");

	//if (gwAGD.open_panel(P) != cancel) {
		callLayout (gwAGD, *gwAGD.quod);
	//}
}


string EditWindow::get_error_message(AgdKey p)
{
	if (p == key::planar)
		return "The graph is not planar.";
	if (p == key::simple)
		return "The graph ist not simple.";
	if (p == key::connected)
		return "The graph is not connected.";
	if (p == key::biconnected)
		return "The graph is not biconnected.";
	if (p == key::triconnected)
		return "The graph is not triconnected.";
	if (p == key::four_graph)
		return "The graph is not a four-graph.";
	if (p == key::no_self_loops)
		return "The graph has self loops.";
	if (p == key::directed)
		return "The graph is not directed.";
	if (p == key::undirected)
		return "The graph is not undirected.";
	if (p == key::no_reversals)
		return "The graph has reversal edges.";
	if (p == key::acyclic)
		return "The graph is not acyclic.";
	if (p == key::tree)
		return "The graph does not represent a tree.";
	if (p == key::forest)
		return "The graph does not represent a forest.";
	if (p == key::single_source)
		return "The graph has not a single source.";
	if (p == key::single_sink)
		return "The graph has not a single sink.";
	if (p == key::bipartite)
		return "The graph is not bipartite.";
	if (p == key::st_graph)
		return "The graph is not an st-graph.";
	if (p == key::st_planar)
		return "The graph is not st-planar.";
	return "property <" + p.info() + "> is not satisfied.";
}

bool EditWindow::callLayout (EditWindow& GW, LayoutModule& L)
{
	AgdKey p;
	GraphWinAnimInterface A (GW); // an interface for our GraphWin
	const graph &G = GW.get_graph();

	// can L draw the graph displayed in GW ?
	if (L.check(G, p)) {
		L.call (G, A); // yes, we can call L
		GW.fill_window();

		return true;
	} else {
		GW.acknowledge (get_error_message(p)); // no, so we display the error message
	}

	return false;
}

bool EditWindow::callAugment (EditWindow& GW, AugmentationModule& A)
{
	AgdKey p;

	// can A be called for the graph displayed in GW ?
	if (A.check(GW.get_graph(), p)) {
		list<edge> e_list;
		edge e;

		A.call (GW.get_graph(), e_list); // yes, we can call A

		GW.update_graph();

		GW.deselect_all_edges();
		forall (e, e_list)
			GW.select(e);

		return true;

	} else {
		GW.acknowledge (get_error_message(p)); // no, so we display the error message
	}

	return false;
}


bool EditWindow::callSubgraph (EditWindow& GW, SubgraphModule& S)
{
	graph &G = GW.get_graph();
	AgdKey p;

	// can S be called for the graph displayed in GW ?
	if (S.check(G, p)) {
		S.call (G); // yes, we can call S
		GW.update_graph();

		return true;

	} else {
		GW.acknowledge (get_error_message(p)); // no, so we display the error message
	}

	return false;
}

//---------------------------------------------------------
// main() for SPQR-Tree Viewer
//
// author: Carsten Gutwenger
//---------------------------------------------------------



int main()
{
	StPlaneGraph planeGraph;
	GraphWindow graphWindow(planeGraph);


	graphWindow.display();


	while(graphWindow.eventHandler()) { }


	return 0;
}
