/*******************************************************************************
+
+  AGD 1.2
+
+
+  AGD.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         AGD.h
//
//
//    author: Carsten Gutwenger
//    date:   decembre 1999
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_AGD_H
#define _AGD_AGD_H


#include <AGD/AgdKey.h>
#include <AGD/ClusterGraph.h>


class DefaultKeymapper;
class DefaultRuleAdmin;
class GraphChecker;
class SubgraphChecker;
class ClusterGraphChecker;
class KeySet;
class PreCond;
class PostCond;
class CondBase;

LEDA_BEGIN_NAMESPACE
class random_source;
LEDA_END_NAMESPACE

/*{\Manpage {AGD} {} {AGD Base} {agd} }*/

class __exportAGD AGD {


/*{\Mdefinition
The class |\Mname| provides a global key mapper, that maps strings
to key items of type |AgdKey|. Furthermore, it defines static methods
for testing and establishing properties of graphs.

We use key items to represent properties of instances of a particular type,
e.g., ``planar'' for instances of type |graph|. We represent such
properties by arbitrary strings, which are stored in a global hash table by
the default keymapper of |\Mname|. Strings in this hash table are accessed via keys
of type |AgdKey|. We use $\<s\>$ to denote a key for string $s$.

\headerline{Predefined Properties}

The class |key| defines an enumeration type |Predefined|. Constants
of this type (e.g., |key|::|directed|) can be converted into |AgdKey|. The
following constants are defined by |key::Predefined|:

\begin{agdproptab}{Properties of subgraphs}
constant & string value & meaning\\\hline\hline
\avar{maximal\_planar} & ``maximal planar'' & Inserting any edge of $E'$ makes $G'$ non-planar.\\
\avar{maximum\_planar} & ``maximum planar'' & $G'$ is a maximum planar subgraph of $G$ (there
is no planar subgraph of $G$ with more edges than $G'$).\\
\avar{maximal\_acyclic} & ``maximal acyclic'' & Inserting any edge of $E'$ into $G'$ creates a cycle.\\
\avar{maximum\_acyclic} & ``maximum acyclic'' & $G'$ is a maximum acyclic subgraph of $G$ (there
is no acyclic subgraph of $G$ with more edges than $G'$).\\\hline
\end{agdproptab}

\begin{agdproptab}{Properties of graphs}
constant & string value & meaning\\\hline\hline
\avar{directed}        & ``directed''        & $G$ is directed.\\
\avar{undirected}      & ``undirected''      & $G$ is undirected.\\
\avar{planar}          & ``planar''          & $G$ is planar.\\
\avar{simple}          & ``simple''          & $G$ has no parallel edges.\\
\avar{no\_self\_loops}   & ``no self loops''   & $G$ has no self loops.\\
\avar{no\_reversals}    & ``no reversals''    & $(v,w)\in E\Rightarrow (w,v)\notin E$
	for all $v,w\in V$\\
\avar{connected}       & ``connected''       & $G$ is connected.\\
\avar{biconnected}     & ``biconnected''     & $G$ is biconnected.\\
\avar{triconnected}    & ``triconnected''    & $G$ is triconnected.\\
\avar{bipartite}       & ``bipartite''       & $G$ is bipartite.\\
\avar{four\_graph}      & ``four-graph''      & $|degree(v)|
	\leq 4$ for all $v \in V$\\
\avar{acyclic}         & ``acyclic''         & $G$ is a digraph without cycle \\
\avar{forest}          & ``forest''          & $G$ represents a forest.\\
\avar{tree}            & ``tree''            & $G$ represents a tree.\\
\avar{single\_source}   & ``single source''   & $\exists!\: v\in G : |indeg|(v)=0$\\
\avar{single\_sink}     & ``single sink''     & $\exists!\: v\in G : |outdeg|(v)=0$\\
\avar{st\_graph}     & ``st-graph''     & $G$ is an acyclic digraph with exactly
one source and exactly one sink \\
\avar{st\_planar}     & ``st-planar''     & $G$ is an st-graph an $G\cup(s,t)$ is planar \\
\avar{reduced\_acyclic} & ``reduced acyclic'' & $G$ is a acyclic digraph
	without transitive edges\\\hline
\end{agdproptab}

\nopagebreak where $G = (V,E)$.

\begin{agdproptab}{Properties of clustergraphs}
constant & string value & meaning\\\hline\hline
\avar{c\_planar} & ``c-planar'' & Clustergraph $CG$ is c-planar. \\\hline  
\end{agdproptab}

\nopagebreak where $CG = (G,T)$.

\begin{agdproptab}{Properties of layouts}
constant & string value & meaning\\\hline\hline
\avar{straight\_line} & ``straight-line'' & All edge polylines have no bend points.\\
\avar{orthogonal}    & ``orthogonal'' & All edge segments are either horizontal
or vertical.\\
\avar{no\_crossings}  & ``no crossings'' & The edge polylines do not intersect.\\
\avar{upward\_planar} & ``upward planar'' & All edges are directed upwards (downwards).\\
\avar{convex} & ``convex'' & All faces are convex.\\\hline
\end{agdproptab}
\nopagebreak where $G'=(V,E\setminus E')$ is subgraph of $G=(V,E)$.


\headerline{Predefined Rules on Properties}

We define the following dependencies between predefined properties.
These rules are maintained by the standard rule administrator of \avar{\Mname}.
\begin{eqnarray*}
\avar{tree} & \leftrightarrow & \avar{forest} \wedge \avar{single source}\\
\avar{tree} & \leftrightarrow & \avar{connected} \wedge \avar{forest}\\
\avar{st-graph} & \leftrightarrow & \avar{directed} \wedge \avar{acyclic} \wedge \avar{single source} \wedge \avar{single sink}\\
\avar{st-planar} & \rightarrow & \avar{st-graph} \wedge \avar{planar}\\
\avar{forest} & \rightarrow & \avar{acyclic} \wedge \avar{directed} \wedge \avar{planar}\\
\avar{acyclic} & \rightarrow & \avar{directed} \wedge \avar{no self loops} \wedge \avar{no reversals}\\
\avar{undirected} & \rightarrow & \avar{no reversals}\\
\avar{single source} & \rightarrow & \avar{directed}\\
\avar{single sink} & \rightarrow & \avar{directed}\\
\avar{triconnected} & \rightarrow & \avar{biconnected}\\
\avar{biconnected} & \rightarrow & \avar{connected}\\
\avar{acyclic} \wedge \avar{single source} & \rightarrow & \avar{connected}\\
\avar{reduced acyclic} & \rightarrow & \avar{directed} \wedge \avar{acyclic}\\
\avar{maximum planar} & \rightarrow & \avar{maximal planar}\\
\avar{maximal planar} & \rightarrow & \avar{planar}\\
\avar{maximum acyclic} & \rightarrow & \avar{maximal acyclic}\\
\avar{maximal acyclic} & \rightarrow & \avar{acyclic}
\end{eqnarray*}
}*/


public:

	AGD () { }

	virtual ~AGD () { }


	/*{\Moperations 2 3.2}*/

	static double version() {
		return (double)__AGD__/100;
	}
	/*{\Mstatic returns the version number of the library.}*/

	static leda::string version_string() {
		return version2string(__AGD__);
	}
	/*{\Mstatic returns the version number as string (e.g., 1.1.2).}*/

	static leda::string version2string(int v) {
		return leda::string("%d.%d.%d", v/100, (v/10)%10, v%10);
	}


public:

	/*{\Mtext       
	\headerline{a) Testing Properties of Graphs}

	Some of the following functions simply call LEDA functions. }*/

	static bool is_simple (const leda::graph& G);
	/*{\Mstatic returns |true| if $G$ is simple, e.g. has no parallel
	edges, |false| otherwise.}*/

	static bool is_connected (const leda::graph& G);
	/*{\Mstatic returns |true| if $G$ is connected, |false| otherwise.}*/

	static bool is_biconnected  (const leda::graph& G);
	/*{\Mstatic returns |true| if $G$ is biconnected, |false| otherwise.}*/

	static bool is_triconnected (const leda::graph& G);
	/*{\Mstatic returns |true| if $G$ is triconnected, |false| otherwise.}*/

	static bool is_cplanar      (const ClusterGraph& CG);
	/*{\Mstatic returns |true| if $CG$ is c-planar, |false| otherwise.}*/

	static bool is_planar       (const leda::graph& G);
	/*{\Mstatic returns |true| if $G$ is planar, |false| otherwise.}*/

	static bool is_bipartite    (const leda::graph& G);
	/*{\Mstatic returns |true| if $G$ is bipartite, |false| otherwise.}*/

	static bool is_four_graph   (const leda::graph& G);
	/*{\Mstatic returns |true| if $G$ is a four-graph, e.g. |degree|$(v)
	\leq 4$ holds for every node $v$ of $G$, |false| otherwise.}*/

	static bool has_no_self_loops (const leda::graph& G);
	/*{\Mstatic returns |true| if $G$ contains no self loops, |false|
	otherwise.}*/

	static bool has_no_reversals (const leda::graph& G);
	/*{\Mstatic returns |true| if $G$ is undirected, or if $G$ contains
	no reversal edges, |false| otherwise.}*/

	static bool is_acyclic (const leda::graph& G);
	/*{\Mstatic returns |true| if $G$ is acyclic, |false| otherwise.}*/

	static bool is_st_graph (const leda::graph& G, leda::node& s, leda::node& t);
	/*{\Mstatic returns |true| and sets $s$ to the source and $t$ to the
	sink of $G$ if $G$ an st-graph, |false| otherwise.}*/

	static bool is_st_graph (const leda::graph& G) {
		leda::node s, t;
		return is_st_graph(G,s,t);
	}
	/*{\Mstatic returns |true| if $G$ an st-graph, |false| otherwise.}*/

	static bool is_st_planar (const leda::graph& G, leda::node &s, leda::node &t);
	/*{\Mstatic returns |true| and sets $s$ to the source and $t$ to the
	sink of $G$ if $G$ is st-planar, |false| otherwise.}*/

	static bool is_st_planar (const leda::graph& G) {
		leda::node s, t;
		return is_st_planar(G,s,t);
	}
	/*{\Mstatic returns |true| if $G$ is st-planar, |false| otherwise.}*/

	static bool is_forest (const leda::graph& G);
	/*{\Mstatic returns |true| if $G$ represents a forest, |false|
	otherwise.}*/

	static bool is_tree (const leda::graph& G);
	/*{\Mstatic returns |true| if $G$ represents a tree, |false|
	otherwise.}*/

	static bool has_single_source (const leda::graph& G, leda::node& s);
	/*{\Mstatic returns |true| if $G$ is directed and contains exactly
	one source (a node with no incoming edges) and sets $s$ to this node,
	|false| otherwise.}*/

	static bool has_single_source (const leda::graph& G) {
		leda::node s;
		return has_single_source(G,s);
	}
	/*{\Mstatic returns |true| if $G$ is directed and contains exactly
	one source, |false| otherwise.}*/

	static bool has_single_sink (const leda::graph& G, leda::node& t);
	/*{\Mstatic returns |true| if $G$ is directed and contains exactly
	one sink (a node with no outgoing edges) and sets $t$ to this node,
	|false| otherwise.}*/

	static bool has_single_sink (const leda::graph& G) {
		leda::node t;
		return has_single_sink(G,t);
	}
	/*{\Mstatic returns |true| if $G$ is directed and contains exactly
	one sink, |false| otherwise.}*/


	/*{\Mtext
	\medskip
	\headerline{b) Establishing Properties of Graphs} }*/

	static leda::list<leda::edge> triangulate (leda::graph& G);
	/*{\Mstatic triangulates a directed graph $G$ representing a planar map.
	Returns the list of inserted edges.}*/

	static leda::list<leda::edge> make_bidirected (leda::graph& G);
	/*{\Mstatic makes $G$ bidirected by inserting missing reversal edges.
	Returns the list of inserted edges.}*/

	static leda::list<leda::edge> make_biconnected (leda::graph& G);
	/*{\Mstatic makes $G$ biconnected.
	Returns the list of inserted edges.}*/

	static leda::list<leda::edge> augment_planar_biconnected (leda::graph& G);
	/*{\Mstatic augments a planar graph $G$ to a planar and biconnected graph.
	Returns the list of inserted edges.}*/

	static leda::list<leda::edge> make_map (leda::graph& G);
	/*{\Mstatic makes $G$ bidirected and turns it into a map. Returns the
	list of inserted edges.}*/

	static bool embed_planar_graph (leda::graph& G);
	/*{\Mstatic takes as input a bidirected graph $G$. Returns |true| and
	turns $G$ into a planar map if $G$ is planar, |false| otherwise.}*/

	static bool embed_cplanar_graph (ClusterGraph& CG, leda::list<leda::edge>& L);
	/*{\Mstatic takes as input a clustergraph $CG$. Returns |true| and
	turns $CG$ into a planar map if $CG$ is c-planar, |false| otherwise.
	$L$ contains added reversal edges.}*/


	/*{\Mtext
	\medskip
	\headerline{c) Access to the Default Keymapper and Rule System} }*/

	static const DefaultKeymapper &get_keymapper() {
		return *keymapper;
	}
	/*{\Mstatic returns the default keymapper of |\Mname|.}*/

	static int number_of_keys();
	/*{\Mstatic returns the number of defined keys.}*/

	static AgdKey ask_key (const leda::string& s);
	/*{\Mstatic returns the key $<s>$ if $<s>$ is already
	defined, |nil| otherwise.}*/

	static AgdKey get_key (const leda::string& s);
	/*{\Mstatic returns the key $<s>$.}*/

	static AgdKey predefined (key::Predefined k);
	/*{\Mstatic converts |k| into a key of type |AgdKey|.}*/

	static KeySet get_properties (const leda::graph& G);
	/*{\Mstatic returns the set of all properties that |G| satisfies.}*/

	static KeySet get_properties (const ClusterGraph& G);
	/*{\Mstatic returns the set of all properties that |CG| satisfies.}*/


	static DefaultRuleAdmin &get_rule_admin() {
		return *admin;
	}
	/*{\Mstatic returns the default rule administrator of |\Mname|.}*/

	static KeySet closure(const KeySet &S);
	/*{\Mstatic returns the closure of set |S| according to the rule
	system of |\Mname|.}*/

	static PreCond closure(const PreCond &P);
	/*{\Mstatic returns the closure of precondition |P| according to
	the rule system of |\Mname|.}*/

	static PostCond closure(const PostCond &P);
	/*{\Mstatic returns the closure of postcondition |P| according to
	the rule system of |\Mname|.}*/

	static bool implies(const KeySet &P, const KeySet &Q);
	/*{\Mstatic returns |true| iff $P\rightarrow Q$, i.e., if
	$P^*\supseteq Q$.}*/

	static bool implies(const CondBase &P, const CondBase &Q);
	/*{\Mstatic returns |true| iff $P\rightarrow Q$.}*/


	/*{\Mtext
	\medskip
	\headerline{d) Access to Checkers} }*/

	static GraphChecker &get_graph_checker() {
		return *graph_checker;
	}
	/*{\Mstatic returns a reference to the global graph checker.}*/

	static ClusterGraphChecker &get_clustergraph_checker() {
	        return *clustergraph_checker;
	}
	/*{\Mstatic returns a reference to the global clustergraph checker.}*/

	static SubgraphChecker &get_subgraph_checker() {
		return *subgraph_checker;
	}
	/*{\Mstatic returns a reference to the global subgraph checker.}*/

	//typedef LEDA_NAMESPACE_NAME::random_source random_source;
	typedef leda::random_source random_source;
	static random_source& get_random_source() {
		return *Random_Source;
	}
	/*{\Mstatic returns a reference to the global random source.}*/

#ifdef AGD_DLL
public:
	static void init();
	static void cleanup();
private:

#else
	friend class AGDInit;
private:
	static void init();
	static void cleanup();
#endif

	static bool dfs_forest (leda::node v,
		leda::node_array<bool>& visited, leda::node_array<bool>& mark);
	static bool is_forest (const leda::graph& G, int &num_sources);

	static void closure(CondBase &Pc, const CondBase &P);

	// the keymapper maps strings to keys (items)
	// there is exactly one keymapper for all AGD modules
	static DefaultKeymapper *keymapper;
	static DefaultRuleAdmin *admin;
	static GraphChecker     *graph_checker;
	static ClusterGraphChecker *clustergraph_checker;
	static SubgraphChecker  *subgraph_checker;
	static random_source	*Random_Source;
};



/*{\Mimplementation
Most functions for testing and establishing graph properties take
time $\BigO(n+m)$, except for is\_four\_graph(), has\_single\_source()
and has\_single\_sink() which take time $\BigO(n)$. In each case,
$n$ is the number of nodes and $m$ is the number of edges of
the input graph $G$.
}*/


#endif
