/*******************************************************************************
+
+  AGD 1.2
+
+
+  AGDModule.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         AGDModule.h
//
//
//    author: Carsten Gutwenger
//    date:   octobre 1998
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_AGDMODULE_H
#define _AGD_AGDMODULE_H

#include <AGD/AGD.h>

class PreCond;
class PostCond;
class PostCondRule;
class KeySet;
class TaggedRule;


/*{\Manpage {AGDModule} {} {AGD Modules} {M} }*/

/*{\Mbaseclasses

\begin{hierarchy}{6.5}{2.5}
\hdclass{0}{1.5}{5}{AGD}
\arhook{5}{2}{6}
\hdclass{1.5}{0}{5}{AGDModule}
\end{hierarchy}
}*/

class __exportAGD AGDModule : public AGD {

/*{\Mdefinition
The class |\Mname| is the base class for modules in the AGD--Library.
It provides the handling of pre- and postconditions, and defines
a uniform access to common informations about a module.

\headerline{The Type of a Module}

We can characterize the type of a module by defining the type of its
input $I_1,\ldots,I_n$, the type of its output $O_1,\ldots,O_m$, and
the associated functionality, e.g., augmentation or layout algorithm.
We express the different types of modules by abstract base classes
derived from |\Mname|, which define the input and output types,
and the functionality.

The input consists
of $n\geq0$ and the output of $m\geq0$ parameters of
an arbitrary abstract (not necessarily \CC) type. The call() method of a particular
module uses appropriate classes to represent these types.
It is not necessary to represent each abstract type by exactly
one class, for example
\begin{itemize}
\item the type ``grid layout'' can be represented
by ``|node_map<int>|\& |x|, |node_map<int>|\& |y|,
|edge_map<IPolyline>|\& |bends|'',
\item an input and output parameter ``graph'' can be represented
by a single |graph| reference.
\end{itemize}

\headerline{Precondition and Postcondition}

|\Mname| maintains a set |pre(i)| of properties for each input parameter
$I_i$, $1\leq i\leq n$, which must be satisfied by $I_i$ if the module is
called. Properties
are expressed by key items (cf. Sect.~\ref{AGD Base}), and sets of
properties by the data type |KeySet|. We call
|pre(i)| the precondition of $I_i$, and $|pre|(1),\ldots,|pre|(n)$ the
precondition of the module.

The postcondition of the output parameters are sets of properties that the output
will always satisfy. We allow that they depend
on the properties $P_1,\ldots,P_n$ that the input of the module satisfies. Clearly, these
properties must be more general than the required precondition, i.e., 
$P_i\supseteq |pre|(i)$ must hold for all $1\leq i\leq n$.
We associate a function
$|post|(i):{\calP}(|KEYS|)^n \to {\calP}(|KEYS|)$ with each output parameter
$O_i$, where |KEYS| is the set of all properties.
We also reserve a special postcondition with index 0 for general properties
of the module.

The function |post(i)| is defined by a set of rules |post_rules(i)|, where each
rule is of the form
\[ (p_1,i_1) \wedge \ldots \wedge (p_k,i_k) \rightarrow p\mbox{ , }k\geq 0. \]
Such a rule means that if, for each $1\leq j\leq k$, input parameter $I_{i_j}$
satisfies property $p_j$, then output parameter $O_i$ satisfies property $p$.

|\Mname| provides protected methods to insert rules into |post_rules(i)|, and
to remove rules from |post_rules(i)|. Rules in |post_rules(i)| can be accessed
via the item type |pr_item|, e.g., to remove a rule. The function defined by
|post_rules(i)| is given by
\begin{eqnarray*}
\lefteqn{|post(i)|(Q_1,\ldots,Q_n) =} \\
& & \{
p \mid \exists (p_1,i_1) \wedge \ldots \wedge (p_k,i_k) \to p \in |post_rules|(i)\mbox{ with }
p_j\in Q_{i_j}\forall 1\leq j\leq k\}
\end{eqnarray*}

Note, that the so defined functions are monotonous in the following way. If
$P_j\subseteq Q_j$ holds for all $1\leq j\leq n$, then
$|post(i)|(P_1,\ldots,P_n)\subseteq |post(i)|(Q_1,\ldots,Q_n)$ holds for each
$1\leq i\leq m$.

We represent a precondition $P_1,\ldots,P_n$ by the data type |PreCond|,
a postcondition $Q_0,\ldots,Q_m$ by the data type |PostCond|, and a
postcondition rule $R_0,\ldots,R_m$ by the data type |PostCondRule|,
where $P_i,Q_i$ are sets of keys and $R_i$ are sets of rules.

\headerline{Common Informations}

Every AGD module contains
\begin{itemize}
\item a name in a long (|long_name|) and a short (|name|) version.
\item an author (|author|).
\item the name of the author (|impl_author|), the version (|impl_version|),
and the date (|impl_date|) of the implementation.
\end{itemize}

The default implementations for |name()|, |author()|, |impl_author()|,
|impl_version()| and |impl_date()| return an empty string ``'' indicating
that this information is unknown. |long_name()| returns |name()| by default.
\bigskip
\headerline{Exchangeable Modules and Optional Parameters}

We recommend to provide a standard interface for exchangeable modules and
optional para\-meters. These recommendation is obeyed by the module
implementations in the library.

Exchangeable modules are implemented as module options (see
Sect.~\ref{Module Options}) with an additional method
set\_|name|(|type| |M|), which sets the module option
|name| to |M| if possible.

An optional parameter |name| of type |type| provides two methods
\begin{quote}\tt\begin{tabular}{ll}
void & \var{name}(\var{type} new\_setting) \\
type & \var{name}()
\end{tabular}\end{quote}
for setting and reading the current value of the option. The documentation
describes optional parameters in the definition section of its class in
the following way
\begin{itemize}
\item \var{type} {\bf name} = \var{initial\_value}\\\var{description}
\end{itemize}
where \var{initial\_value} is the default setting of the option.
}*/

public:

	typedef leda::list_item pr_item;

	/*{\Minitialization}*/

	AGDModule (int n, int m);
	/*{\Mcreate initializes a module with |n| input and |m| output parameters.
	Initializes |pre(i)| and |post(j)| with the empty set for all $1\leq i\leq n$,
	and $0\leq j\leq m$.}*/

	virtual ~AGDModule();


	/*{\Moperations 2 3.2}*/

	/*{\Mtext       
	\headerline{a) Queries on Preconditions and Postconditions}
	}*/

	int num_in_param () const;
	/*{\Mop returns the number of input parameters $n$.}*/

	int num_out_param () const;
	/*{\Mop returns the number of output parameters $m$.}*/

	const PreCond& precondition () const;
	/*{\Mop returns |pre|.}*/

	const KeySet& precondition (int i) const;
	/*{\Mop returns |pre(i)|. \precond $1\leq i\leq n$.}*/

	PostCond postcondition (const PreCond &PRE) const;
	/*{\Mop returns $|post|(|pre|(1),\ldots,|pre|(n))$.}*/

	const KeySet &postcondition (int i = 1);
	/*{\Mop returns $|post|(i)(|pre|(1),\ldots,|pre|(n))$. \precond $0\leq i\leq m$.}*/

	KeySet postcondition (const PreCond& PRE, int i) const;
	/*{\Mop returns $|post|(i)(|PRE|)$. \precond $0\leq i\leq m$, and
	|PRE|.num\_param() = n.}*/

	bool applicable (const PreCond& PRE, const PostCond& POST) const;
	/*{\Mop returns |true| iff |\Mvar| can be applied as module for guaranteed precondition
	|PRE| and required postcondition |POST|, i.e., $|PRE|_i\supseteq |pre|(i)\:\forall\,
	1\leq i\leq n$ and $|post|(j)(|PRE|)\subseteq |POST|_j\:\forall\,0\leq j\leq m$.}*/

	virtual AGDModule *clone () const
	/*{\Moppure returns a pointer to a new instance with the same option
	settings as |\Mvar|. Must be overridden by module implementations.}*/
	= 0;

	/*{\Mtext
	\bigskip
	\headerline{b) General Information}
	}*/

	virtual leda::string name () const {
		return leda::string ("");
	}
	/*{\Mop returns the name of the algorithm.}*/

	virtual leda::string long_name () const {
		return name ();
	}
	/*{\Mop returns the name in an extended version.}*/

	virtual leda::string author () const {
		return leda::string ("");
	}
	/*{\Mop returns the author of the algorithm.}*/

	virtual leda::string impl_author () const {
		return leda::string ("");
	}
	/*{\Mop returns the author of the implementation.}*/

	virtual leda::string impl_version () const {
		return leda::string ("");
	}
	/*{\Mop returns the version of the implementation.}*/

	virtual leda::string impl_date () const {
		return leda::string ("");
	}
	/*{\Mop returns the date of the last change of the implementation.}*/


protected:
	/*{\Mtext       
	\bigskip
	\headerline{c) Protected Methods for Maintaining Preconditions} }*/

	void add_precondition (AgdKey k, int i = 1);
	/*{\Mop adds the key |k| to |pre(i)|. \precond $1\leq i\leq n$.}*/

	void add_precondition (AgdKey k, bool status, int i = 1);
	/*{\Mop adds the key |k| to |pre(i)| if
	|status| = |true|, removes it otherwise. \precond $1\leq i\leq n$.}*/

	void add_preconditions (const leda::string& ps, char delim = ',', int i = 1);
	/*{\Mop interprets |ps| as a list of strings seperated by |delim|, and adds
	the key for each string to |pre(i)|. \precond $1\leq i\leq n$.}*/

	void rem_precondition (AgdKey k, int i = 1);
	/*{\Mop removes the key |k| from |pre(i)|. \precond $1\leq i\leq n$.}*/


	/*{\Mtext       
	\bigskip
	\headerline{d) Protected Methods for Maintaining Postconditions} }*/

	pr_item add_post_rule (AgdKey k, int i = 1);
	/*{\Mop adds a new rule item $<\rightarrow k>$ to |post_rules(i)|
	and returns it. \precond $0\leq i\leq m$.}*/

	pr_item add_post_rule (AgdKey p1, int i1, AgdKey p, int i = 1);
	/*{\Mop adds a new rule item $<(p_1,i_1)\rightarrow p>$ to |post_rules(i)|
	and returns it. \precond $0\leq i\leq m$.}*/

	pr_item add_post_rule (const TaggedRule& r, int i = 1);
	/*{\Mop adds a new rule item $<r>$ to |post_rules(i)| and returns it.
	\precond $0\leq i\leq m$.}*/

	void rem_post_rule (pr_item it, int i = 1);
	/*{\Mop removes the rule item |it| from |post_rules(i)|.
	\precond $0\leq i\leq m$ and |it| is a rule item in |post_rules(i)|.}*/


protected:

	/*{\Mtext       
	\bigskip
	\headerline{Checking the Precondition of a Module} }*/

	/*{\Mtext
	Particular module types, for example |LayoutModule|, provide a method
	check() that checks if a given input satisfies the precondition of the
	module. |\Mname| implements extendable default implementations to
	check a precondition for specific types, for example graphs. These
	default implementations can be used by the base class of a particular
	module type (e.g., |LayoutModule|) to implement its check() method.

	The method check\_graph() can test if a given input graph satisfies
	the following keys:
	|directed|, |undirected|, |planar|, |simple|, |no_self_loops|,
	|no_reversals|, |connected|, |biconnected|, |triconnected|,
	|bipartite|, |four_graph|, |acyclic|, |forest|, |tree|, |single_source|,
	|single_sink|. If check\_graph() finds another key it calls
	check\_graph\_key().
	}*/

	bool check_graph (int i, const leda::graph& G, AgdKey &p) const;
	/*{\Mop returns |true| if $G$ satisfies the preconditions |pre(i)|.
	Otherwise, |false| is returned, and |p| contains a property that
	is not satisfied.}*/

	virtual bool check_graph_key (AgdKey p, const leda::graph& G) const;
	/*{\Mop overridable to extend check\_graph. Must return |true| if |G|
	satisfies |p|, |false| otherwise. The default implementation returns false.}*/

	bool check_graph (int i, const ClusterGraph& CG, AgdKey &p) const;
	/*{\Mop returns |true| if $CG$ satisfies the preconditions |pre(i)|.
	Otherwise, |false| is returned, and |p| contains a property that
	is not satisfied.}*/

	virtual bool check_graph_key (AgdKey p, const ClusterGraph& CG) const;
	/*{\Mop overridable to extend check\_graph. Must return |true| if |CG|
	satisfies |p|, |false| otherwise. The default implementation returns false.}*/

private:

	PreCond      *_pre;
	PostCondRule *_post;
	PostCond     *_pc;
	bool prePostChanged;

	static void extract_list (const leda::string& str, char delim,
		leda::list<leda::string>& L);
};



#endif
