/*******************************************************************************
+
+  AGD 1.2
+
+
+  BarycenterHeuristic.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                       BarycenterHeuristic.h
//
//
//    author: Carsten Gutwenger
//    date:   february 1998
//-------------------------------------------------------------------


#ifndef _AGD_BARYCENTER_HEURISTIC_H
#define _AGD_BARYCENTER_HEURISTIC_H


#include <AGD/TwoLayerCrossMin.h>


/*{\Manpage {BarycenterHeuristic} {} {Barycenter Heuristic} {M} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{0.9}{1.5}{6.1}{TwoLayerCrossMin}
\arhook{7}{2}{8}
\hdclass{2}{0}{6.5}{BarycenterHeuristic}
\end{hierarchy}
}*/

class __exportAGD BarycenterHeuristic : public TwoLayerCrossMin {

/*{\Mdefinition
The class |\Mname| implements the two layer crossing minimization heuristic
{\it Barycenter}. It reorders the nodes on the changeable layer according
to the barycenter weight~\cite{stt-mvuhs-81}
\[ Bary(v) = \left\{\begin{array}{cl}
\frac{\sum_{\D w\in adj(v)}\D pos(w)}{\D\mid adj(v)\mid} & \mbox{ if }adj(v)\neq\emptyset\\
0.0 & \mbox{ otherwise}
\end{array}\right. \]
where $v$ is a node on the changeable layer, $|adj|(v)$ is the set of
neighbors of $v$ on the fixed layer, and $|pos|(w)$ is the position
of $w$ on the fixed layer.

\horizline
\commoninfo{Barycenter}{Barycenter (Sugiyama et al.)}{K. Sugiyama, S. Tagawa, M. Toda}{C. Gutwenger}
{February 1998}{1.0}

\begin{preposttab}
precondition & $\emptyset$ \\
postcondition(|PRE|) & $\emptyset$
\end{preposttab}

\horizline
}*/

public:

	/*{\Mcreation M 5}*/

	BarycenterHeuristic() { }
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname|.}*/

	~BarycenterHeuristic() { }


	/*{\Moperations}*/

	leda::string name () const {
		return "Barycenter";
	}

	leda::string long_name () const {
		return "Barycenter (Sugiyama et al.)";
	}

	leda::string author () const {
		return "K. Sugiyama, S. Tagawa, M. Toda";
	}

	leda::string impl_author () const {
		return "C. Gutwenger";
	}

	leda::string impl_version () const {
		return "1.0";
	}

	leda::string impl_date () const {
		return "February 1998";
	}

	AGDModule *clone() const {
		return new BarycenterHeuristic();
	}

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}

	The detailed description of these methods can be found in the
	manual entries of the base class (|TwoLayerCrossMin|).
	}*/

	/* inherited
	bool check (const leda::graph& G, const Layer& L, AgdKey &p) const;
	*/
	/*{\Mop}*/

	void init (const leda::graph& G, const Hierarchy& H) {
		weight.init(G);
	}
	/*{\Mop}*/

	void call (const leda::graph& G, Layer& L);
	/*{\Mop}*/

	void cleanup () {
		weight.init();
	}
	/*{\Mop}*/

	/*{\Mtext \stdend}*/

private:

	leda::node_array<double> weight;
};


/*{\Mimplementation
We use a randomized quicksort for sorting the nodes on
the layer according to the barycenter weight.
The init() function takes time $\BigO(n)$, and the call()
function takes expected time $\BigO(m_L+n_L\log n_L)$,
where $n$ is the number of nodes in the input graph, $n_L$
is the number of nodes on the changeable layer and $m_L$
is the number of edges between the fixed and the changeable layer.
}*/


#endif
