/*******************************************************************************
+
+  AGD 1.2
+
+
+  ClusterGraph.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         ClusterGraph.h
//
//    author: Dirk Luetke-Huettmann
//    date:   july 1999
//-------------------------------------------------------------------

#ifndef AGD_CLUSTER_GRAPH_H 
#define AGD_CLUSTER_GRAPH_H 1

#include <LEDA/graph/graph.h>
#include <LEDA/core/map.h>
#include <AGD/Cluster.h>

class CGraphWin;

/*{\Manpage {ClusterGraph} {} {Cluster Graphs} {CG} }*/

/*{\Mbaseclasses
\begin{hierarchy}{6.5}{2.5}
\hdclass{0}{1.5}{5}{graph}
\arhook{5}{2}{6}
\hdclass{1.5}{0}{5}{ClusterGraph}
\end{hierarchy}
}*/

class __exportAGD ClusterGraph : public leda::graph {

/*{\Mdefinition
An instance |\Mvar| of type |\Mname| suplies additional functions 
to the LEDA graph class for the creation of a hierarchical cluster structure.
The root of this hierarchy is called rootcluster. Below in the hierarchy,
clusters can be nested inside other clusters. Clusters can also contain
nodes. Each node can only be a member of one cluster. By default, nodes are
created as members of the rootcluster unless other specified.}*/

  friend class CGraphWin;
public:


/*{\Mcreation CG 5.5}*/

  ClusterGraph();
  /*{\Mcreate creates an instance |\Mvar| of type |\Mname| initialized
        to an empty graph with an empty cluster hierarchy.}*/

  ~ClusterGraph();
  
  /*{\Moperations}*/

  Cluster* get_rootcluster(void) const {
    return rc_p; 
  };
  /*{\Mop returns the rootcluster of the cluster hierarchy.}*/

  leda::list<Cluster *> all_clusters() const {
	  leda::list<Cluster *> all;
	  _all_clusters(rc_p,all);
	  return all;
  }
  /*{\Mop returns the list of all clusters in |\Mvar|.}*/

  leda::list<Cluster*> all_subcluster(Cluster* cl) const {
    leda::list<Cluster *> subcl;
    _all_clusters(cl, subcl);
    return subcl;
  }
  /*{\Mop returns a list of all subcluster of $T(c)$.}*/

  bool is_member(Cluster* cl, leda::node v) const; 
  /*{\Mop returns |true| if node $v$ is a member of a subcluster of |cl|,
    otherwise |false|.}*/

  Cluster* new_cluster(void) ;
  /*{\Mop creates a new empty cluster |c| as a child of the rootcluster and
    returns |c|.}*/ 
  
  Cluster* new_cluster(Cluster* cl);
  /*{\Mop creates a new empty cluster |c| as a child of the existing 
    cluster |cl| and returns |c|.}*/
  
  leda::node new_node(Cluster* cl);
  /*{\Mop creates a new node |v| and make it a member of cluster |cl|.}*/

  Cluster* cluster_of(leda::node v) const;
  /*{\Mop returns the cluster |c| where |v| belongs too.}*/

  void clear_clustertree(void);
  /*{\Mop removes all clusters from the cluster hierarchy. All nodes of 
    |\Mvar| are made members of the rootcluster.}*/

  void move_node(Cluster* cl, leda::node v) {
    Cluster* cl_old = cluster_of(v); 
    if(cl_old != nil) cl_old->remove_node(v); 
    cl->add_node(v);
    cl_node_map[v] = cl;
  }
  /*{\Mop makes an existing node |v| a new member of cluster |cl| and deletes 
    it from its previous cluster.}*/


  void remove_node(leda::node v) {
    Cluster* cl = cluster_of(v);
    if(cl != nil) {
      cl->remove_node(v);
      cl_node_map[v]=nil; 
    }
  }

protected:
  void post_new_node_handler(leda::node);
  void pre_del_node_handler(leda::node);

private:
  
  Cluster* rc_p;

  leda::map<leda::node, Cluster*> cl_node_map;

  void empty_cluster_tree(Cluster* cl_parent); 

  void _all_clusters(Cluster *c, leda::list<Cluster *> &all) const;

/*{\Mimplementation
All operations for graphs are available for instances of
type |\Mname|, too. |clear_clustertree()| takes time $\BigO(c)$
where |c| is the number of clusters in the cluster hierarchy. 
All other operations take constant time.}*/
};	

#endif
