/*******************************************************************************
+
+  AGD 1.2
+
+
+  ClusterGraphChecker.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         ClusterGraphChecker.h
//
//    author: Dirk Luetke-Huettmann
//    date:   september 1999
//-------------------------------------------------------------------


#ifndef _AGD_CG_CHECKER_H
#define _AGD_CG_CHECKER_H

#include <AGD/ClusterGraph.h>

typedef AgdKey (*check_clustergraph_handler)(const ClusterGraph &CG, 
					     KeySet &props,
					     KeySet &satisfied, 
					     bool check_all);

// avoid a warning with g++
inline istream &operator>> (istream &is, check_clustergraph_handler &) { 
  return is; 
}

/*{\Manpage {ClusterGraphChecker} {} {Cluster Graph Checker} {C} }*/

class __exportAGD ClusterGraphChecker {

/*{\Mdefinition
The class |\Mname| maintains a list of check-cluster-graph handler functions
|Checkers|, which are used in order to test if a cluster graph satsifies a given set of
properties, or to compute the set of properties a given cluster graph satisfies.
The type |check_clustergraph_handler| is a pointer to a check-cluster-graph handler
function. Such a function, e.g. f, has to be declared as
\begin{quote}
|AgdKey| f(const |ClusterGraph|\& |CG|, |KeySet|\& |props|, |KeySet|\&
|satisfied|, |bool| |check_all|);
\end{quote}
It has to test if input cluster graph |CG| satisfies the properties in |props|.
Each property that has been checked must be removed from |props|.
If |check_all| is |true|, each satisfied property is added to |satsified|.
If |check_all| is |false|, f returns an unsatisfied property or a |nil| key
(cf. Sect.~\ref{Keys}) if it cannot find an unsatisfied property.

|\Mname| allows to add and remove functions from |Checkers|. New handler
functions are added at the beginning of |Checkers|, i.e., a newly
added handler is called before any other handler. Finally, a graph checker
(cf. Sect.~\ref{Graph Checker}) is called, which has been passed to
the constructor of |\Mname|.
}*/

	leda::list<check_clustergraph_handler> checkers;

public:

	/*{\Mcreation}*/

	ClusterGraphChecker(GraphChecker &GC);
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

	~ClusterGraphChecker() { }

	/*{\Moperations}*/

	bool check(const ClusterGraph &CG, KeySet &props) const;
	/*{\Mop returns |true| iff |CG| satisfies all properties in |props|.}*/

	bool check(const ClusterGraph &G, KeySet &props, AgdKey &not_satisfied) const;
	/*{\Mop returns |true| if |CG| satisfies all properties in |props|,
	otherwise |false| is returned and |not_satsified| is a property in
	|props| that is not satisfied by |CG|.}*/

	KeySet get_properties(const ClusterGraph &CG, KeySet &props) const;
	/*{\Mop returns the set of all properties in |props| which are
	satisfied by |CG|.}*/

	leda::list_item add_checker(check_clustergraph_handler f);
	/*{\Mop appends a new handler function |f| at the beginning of |Checkers|.
	Returns the corresponding list item.}*/

	void rem_checker(leda::list_item it);
	/*{\Mop removes the check-handler function at position |it|
	in |Checkers|. \precond |it| is a list item in |Checkers|.}*/
	
private:
	GraphChecker *gcp;

	static AgdKey default_checker(const ClusterGraph &CG, KeySet &props,
		KeySet &satisfied, bool check_all);
};


#endif
