/*******************************************************************************
+
+  AGD 1.2
+
+
+  ClusterOrthogonalLayout.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


#ifndef _AGD_CLUSTER_ORTHOGONAL_LAYOUT_H
#define _AGD_CLUSTER_ORTHOGONAL_LAYOUT_H

  
#include <AGD/CGPlaneGraphCopyGridConcept.h>
#include <AGD/CGGridLayoutModuleImpl.h>
#include <AGD/Ovid.h>
#include <AGD/AGDModuleOption.h>


/*{\Manpage {ClusterOrthogonalLayout} {} {The Cluster-Orthogonal Algorithm} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{9.5}{5.5}
\hdclass{0}{4.5}{5}{AGDModule}
\arhook{5}{5}{6}
\hdclass{1}{3}{5.5}{CGLayoutModule}
\arhook{6.5}{3.5}{7.5}
\hdclass{1}{1.5}{7}{CGGridLayoutModule}
\arhook{8}{2}{9}
\hdclass{1.5}{0}{8}{ClusterOrthogonalLayout}
\end{hierarchy}
\hfill\agdeps{COrtho.epsi}{4.5cm}
}*/

class __exportAGD ClusterOrthogonalLayout
  : public CGGridLayoutModuleImpl<CGPlaneGraphCopyGridConcept> {

/*{\Mdefinition
The class |\Mname| represents the drawing algorithms for c-planar cluster graphs
by L\"utke-H\"uttmann and Mutzel. The input graph has to be c-planar
and simple. The produced layout depends on the underlying drawing
algorithm for planar graphs, which can be set with the module option
|base_algorithm|. Possible settings are |OrthogonalLayout| and |QuasiOrthogonalLayout|.

\horizline
\commoninfo{Cluster-Orthogonal}{Cluster-Orthogonal (L\"utke-H\"uttmann/Mutzel)}{D. L\"utke-H\"uttmann, P. Mutzel}{D. L\"utke-H\"uttmann}
{September 1999}{1.0}

\begin{preposttab}
precondition & $\{\, \emph{c-planar, simple, no self loops} \,\} $ \\
postcondition(|PRE|) & $\{\, \emph{no crossings} \,\}$
\end{preposttab}

\headerline{Exchangeable Modules}
Instances of type |\Mname| provide the following module options:
\begin{options}
\item {\bf base\_algorithm}\\
a planar grid layout algorithm of type |Ovid| (e.g., |QuasiOrthogonalLayout|
or |OrthogonalLayout|) used for the final drawing step.

\begin{tabular}{ll}
guaranteed precondition: & $\{\emph{ planar, simple, no self loops }\}$\\
required postcondition: & $\{\emph{ no crossings }\}$\\
initial module: & |QuaisOrthogonalLayout|
\end{tabular}
\end{options}

\horizline
}*/

public:

	/*{\Mcreation L 6}*/

  ClusterOrthogonalLayout();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

  ~ClusterOrthogonalLayout() { }
  

	/*{\Moperations 1.5 2.5}*/

  leda::string name() const { 
    return leda::string("Cluster-Orthogonal"); }
  
  leda::string long_name() const { 
    return leda::string("Cluster-Orthogonal (Ltke-Httmann/Mutzel)"); }

  leda::string author() const { 
    return leda::string("D. Ltke-Httmann, P. Mutzel"); }
  
  leda::string impl_author() const { 
    return leda::string("D. Ltke-Httmann"); }

  leda::string impl_version() const { 
    return leda::string("1.0"); }

  leda::string impl_date() const { 
    return leda::string("September 1999"); }


  AGDModule* clone() const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base classes (|CGLayoutModule|, |CGGridLayoutModule|).}*/

	/* inherited
	bool check (const ClusterGraph& CG, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/* inherited
	void call (const ClusterGraph& CG, CGLayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	void call (const ClusterGraph& CG, leda::face f_0, CGLayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (const ClusterGraph& CG, CGGridLayout &cgl);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (const ClusterGraph& CG, leda::face f, CGGridLayout &cgl);
	*/
	/*{\Mop}*/

	/* inherited
	IRect bounding_box ();
	*/
	/*{\Mop}*/

	/* inherited
	void set_mapper (const GridCoordinateMapper& gcm);
	*/
	/*{\Mop}*/
	
	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	/*
	bool set_base_algorithm (const Ovid& M);
	*/
	/*{\Mop}*/


protected:

  void do_call(CGPlaneGraphCopyGridConcept& C);

//  AGD_DECLARE_MODULE(Ovid,base_algorithm)
protected:
	AGDModuleOption<Ovid> base_algorithm;
public:
	bool set_base_algorithm (const Ovid& M);

private:
  //COL_base_algorithm base_algo;

  //OrthogonalLayout pure_ortho;
  //QuasiOrthogonalLayout quod;

  leda::map<Cluster*, leda::list<leda::node> > cl_boundary_nodes;
  leda::map<Cluster*, leda::list<leda::edge> > cl_boundary_edges;

  leda::face outer_face;

  bool is_boundary_member(Cluster* cl_node, 
                          leda::node w);
  
};

#endif

