/*******************************************************************************
+
+  AGD 1.2
+
+
+  GraphWinInterface.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                      GraphWinInterface.h
//
//    author: Carsten Gutwenger, revised by David Alberts
//    date:   july 1999
//-------------------------------------------------------------------


#include <AGD/LayoutInterface.h>
#include <LEDA/graphics/graphwin.h>


#ifndef _AGD_GRAPH_WIN_INTERFACE_H
#define _AGD_GRAPH_WIN_INTERFACE_H


/*{\Manpage {GraphWinInterface} {} {Interface to GraphWin} {A} }*/

/*{\Mbaseclasses

\begin{hierarchy}{7}{2.5}
\hdclass{0}{1.5}{5.5}{LayoutInterface}
\arhook{5.5}{2}{6.5}
\hdclass{1}{0}{6}{GraphWinInterface}
\end{hierarchy}
}*/

class __exportAGD GraphWinInterface : public LayoutInterface {

/*{\Mdefinition
	An instance |\Mvar| of type |\Mname| is an interface to the
	graphical attributes of a |GraphWin| object. |\Mname| posses
	an option flush, which determines if the update() method
	redraws the graph. If flush is set to |false|, the graph
	is not redrawn. This can be useful, if a subsequent operation,
	e.g. zoom\_graph(), shall be called first, before the
	changes are displayed.
}*/

public:

	/*{\Mcreation}*/

	GraphWinInterface (leda::GraphWin& GW);
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname| initialized
	to an interface to |GW|.}*/

	GraphWinInterface (const GraphWinInterface &gwi) : LayoutInterface(gwi),
		gw_p(gwi.gw_p), _flush(gwi._flush), old_flush(gwi.old_flush) { }
	/*{\Mcreate copy constructor.}*/

	~GraphWinInterface();


	/*{\Moperations}*/

	void init (leda::GraphWin &GW);
	/*{\Mop initializes |\Mvar| for |GW|.}*/

	void flush(bool b) {
		_flush = b;
	}
	/*{\Mop sets current flush to |b|, i.e., |b| is |false| calling update()
	will not redraw the graph.}*/

	bool flush() const {
		return _flush;
	}
	/*{\Mop returns the current value of flush.}*/


	/*{\Mtext\medskip
	|\Mname| inherits the operations from its base class |LayoutInterface|
	(see Sect.~\ref{Layout Interfaces}).
	}*/

	double get_x (leda::node v) const {
		return gw_p->get_position(v).xcoord();
	}

	void set_x (leda::node v, double new_x) {
		gw_p->set_position(v,leda::point(new_x,get_y(v)));
	}

	double get_y (leda::node v) const {
		return gw_p->get_position(v).ycoord();
	}

	void set_y (leda::node v, double new_y) {
		gw_p->set_position(v,leda::point(get_x(v),new_y));
	}

	DPoint get_position (leda::node v) const {
		leda::point p = gw_p->get_position(v);
		return DPoint(p.xcoord(),p.ycoord());
	}

	void set_position (leda::node v, const DPoint& p) {
		gw_p->set_position(v,leda::point(p.xcoord(),p.ycoord()));
	}

	DPolyline get_bend_points (leda::edge e) const {
		leda::list<leda::point> points = gw_p->get_bends(e);
		DPolyline l;
		leda::point p;
		forall(p,points) l.append(DPoint(p.xcoord(),p.ycoord()));
		return l;
	}

	void set_bend_points (leda::edge e, const DPolyline& l) {
		leda::list<leda::point> points;
		DPoint dp;
		forall(dp,l) points.append(leda::point(dp.xcoord(),dp.ycoord()));
		gw_p->set_bends(e,points);
	}

	double get_width (leda::node v) const {
		return gw_p->get_radius(v)*2;
	}

	void set_width (leda::node v, double new_w) {
		gw_p->set_radius (v, new_w/2);
	}

	double get_height (leda::node v) const {
		return gw_p->get_radius2(v)*2;
	}

	void set_height (leda::node v, double new_h) {
		gw_p->set_radius2 (v, new_h/2);
	}


	DPoint get_source_anchor (leda::edge e) const {
		leda::point p = gw_p->get_source_anchor(e);
		return DPoint(p.xcoord(),p.ycoord());
	}

	void set_source_anchor (leda::edge e, const DPoint& pos) {
		gw_p->set_source_anchor (e, leda::point(pos.xcoord(),pos.ycoord()));
	}

	DPoint get_target_anchor (leda::edge e) const {
		leda::point p = gw_p->get_target_anchor(e);
		return DPoint(p.xcoord(),p.ycoord());
	}

	void set_target_anchor (leda::edge e, const DPoint& pos) {
		gw_p->set_target_anchor (e, leda::point(pos.xcoord(),pos.ycoord()));
	}

	void update();
	void refresh() {
		gw_p->update_graph();
	}


private:

	leda::GraphWin *gw_p;

	bool _flush, old_flush;
};



/*{\Mexample
We use a |GraphWinInterface| to apply an AGD layout algorithm
to a graph displayed in a |GraphWin|.

\small\tt
\begin{verbatim}
#include <LEDA/graphics/graphwin.h>
#include <AGD/GraphWinInterface.h>
#include <AGD/PlanarizationLayout.h>

int main()
{
  // create a GraphWin and let user edit a graph
  GraphWin GW;
  GW.open();

  // create an interface to GW for AGD alg's
  GraphWinInterface   GWI(GW);
  PlanarizationLayout L;

  // apply L and display the drawing in GW
  L.call(GW.get_graph(),GWI);

  // let the user edit the result
  GW.edit();
}
\end{verbatim}\normalsize\rm
}*/


#endif
