/*******************************************************************************
+
+  AGD 1.2
+
+
+  GreedyCycleRemoval.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         GreedyCycleRemoval.h
//
//
// author: Carsten Gutwenger
// date:   decembre 1998
//-------------------------------------------------------------------


#ifndef _AGD_G_CYCLE_R_H
#define _AGD_G_CYCLE_R_H


#include <AGD/SubgraphModule.h>
#include <LEDA/core/array.h>



/*{\Manpage {GreedyCycleRemoval} {} {Greedy Cycle Removal} {S} }*/

/*{\Mbaseclasses
\begin{hierarchy}{10}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{1}{1.5}{6}{SubgraphModule}
\arhook{7}{2}{8}
\hdclass{1.5}{0}{7}{GreedyCycleRemoval}
\end{hierarchy}
}*/


class __exportAGD GreedyCycleRemoval : public SubgraphModule {

/*{\Mdefinition
The module |\Mname| computes a maximal acyclic subgraph $G'=(V,E)$ of a
directed input graph $G=(V,E\cup F)$. Or in other words, it computes
a feedback set $F$ of $G$. If $G$ contains no self-loops, then we
can obtain an acyclic graph from $G$ by reversing all the edges in $F$.

The algorithm~\cite{EL95} is also described in~\cite{DETT99}.
If $G$ is a connected digraph
with $n$ vertices, $m$ edges and no two-cycles,
the algorithms guarantees that $\Labs{E}\geq m/2-n/6$.

\horizline
\commoninfo{Greedy Cycle Removal}{Greedy Cycle Removal}{}{C. Gutwenger}{Decembre 1998}{1.0}

\begin{preposttab}
precondition & $\{ \emph{directed} \} $ \\
postcondition(|PRE|) & $\{ \emph{maximal acyclic, directed} \} $
\end{preposttab}

\horizline
}*/

public:

    /*{\Mcreation S 5}*/
    
    GreedyCycleRemoval();
    /*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/
    
    
    ~GreedyCycleRemoval() { }
    
    /*{\Moperations 1.2 3.5}*/

    leda::string name () const {
	return "Greedy Cycle Removal";
    }

    leda::string impl_author () const {
	return "Carsten Gutwenger";
    }

	leda::string impl_version () const {
		return "1.0";
	}

	leda::string impl_date () const {
		return "Decembre 1998";
	}


    AGDModule *clone() const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|SubgraphModule|).
	}*/

	/* inherited
	bool check (const graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/* inherited
	bool call (const graph& G, list<edge>& L);
	*/
	/*{\Mop}*/

	/* inherited
	bool call (graph& G);
	*/
	/*{\Mop}*/

	/* inherited
	bool call (GraphCopy& G, list<edge>& L);
	*/
	/*{\Mop}*/

	/* inherited
	int num_del_edges() const;
	*/
	/*{\Mop}*/

	/*{\Mtext \stdend}*/
	
protected:
	void do_call (const leda::graph& G, leda::list<leda::edge>& feedback);

private:
	void dfs (leda::node v, const leda::graph &G);

	int min, max, counter;

	leda::node_array<int> in, out, index;
	leda::array<leda::list<leda::node> > B;
	leda::node_array<leda::list_item> item;
	leda::node_array<bool> visited;
};

/*{\Mimplementation
The computation of the subgraph takes time $\BigO(n+m)$, where
$G$ is a graph with $n$ vertices and $m$ edges.
}*/

#endif
