/*******************************************************************************
+
+  AGD 1.2
+
+
+  GridLayout.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         GridLayout.h
//
//
//    author: Carsten Gutwenger
//    date:   july 1999
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _GRID_LAYOUT_H
#define _GRID_LAYOUT_H

#include <LEDA/graph/graph.h>
#include <AGD/IPolyline.h>
#include <AGD/IRect.h>


/*{\Manpage {GridLayout} {} {Grid Layouts} {gl} }*/

class __exportAGD GridLayout {

/*{\Mdefinition
A grid layout |\Mvar| of a graph $G$ stores for each node of $G$ an
integer x- and y-coordinate and an integer width and height, and for
each edge of $G$ a list of bend points of type |IPolyline|.

A node at position $(x,y)$ with width $w$ and height $h$ covers the
rectangular region $|rect|(v)$ defined by the two grid points $(x,y)$ and $(x+w,y+h)$.
An edge $e=(v,w)$ with bend point list $p_1,\ldots,p_k$ is represented by
a polyline $|line|(e)$ defined by the points $p_1,\ldots,p_k$ plus point $p_v$
at the beginning if $p_1\notin|rect|(v)$, and plus point $p_w$ at the end
if $p_k\notin|rect|(w)$, where $p_v$ and $p_w$ are the positions of node $v$ and
$w$, respectively.

Class |\Mname| provides also utility functions for measuring aesthetic
criteria like area, total number of bends, or total edge length, and
for removing unnecessary bends from integer polylines.
}*/

	leda::node_map<int> _x, _y, _width, _height;
	leda::edge_map<IPolyline> _bends;
	const leda::graph *gp;

public:

	/*{\Mcreation}*/

	GridLayout() { gp = nil; }
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname| initialized to
	the empty layout with no associated graph.}*/

	GridLayout(const leda::graph &G) : _x(G,0), _y(G,0), _width(G,0),
		_height(G,0), _bends(G)
	{
		gp = &G;
	}
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname| initialized to
	a grid layout for graph $G$. Initially, all x- and y-coordinates are $0$,
	all nodes have width and height $0$, and all bend point lists are empty.}*/

	GridLayout(const GridLayout &gl) : _x(gl._x), _y(gl._y),
		_width(gl._width), _height(gl._height), _bends(gl._bends), gp(gl.gp)
		{ }
	/*{\Mcreate copy constructor.}*/


	/*{\Moperations}*/

	void init_node_size(int w, int h);
	/*{\Mop sets the width of all nodes to $w$ and the height to $h$.}*/

	GridLayout &operator=(const GridLayout &gl1);
	/*{\Mbinop assign |gl1| to |\Mvar|.}*/

	
	/*{\Mtext       
	\bigskip
	\headerline{a) Access operations}
	}*/

	const leda::graph &get_graph() const {
		return *gp;
	}
	/*{\Mop returns the associated graph.}*/

	const int& x(leda::node v) const {
		return _x[v];
	}

	int& x(leda::node v) {
		return _x[v];
	}
	/*{\Mop returns a reference to the x-coordinate of node $v$.}*/

	const int& y(leda::node v) const {
		return _y[v];
	}

	int& y(leda::node v) {
		return _y[v];
	}
	/*{\Mop returns a reference to the y-coordinate of node $v$.}*/

	int width(leda::node v) const {
		return _width[v];
	}
	/*{\Mop returns the width of node $v$.}*/

	void width(leda::node v, int width) {
		_width[v] = width;
	}
	/*{\Mop sets the width of node $v$ to |width|.}*/

	int height(leda::node v) const {
		return _height[v];
	}
	/*{\Mop returns the height of node $v$.}*/

	void height(leda::node v, int height) {
		_height[v] = height;
	}
	/*{\Mop sets the height of node $v$ to |height|.}*/

	bool in_rect(IPoint ip, leda::node v) const;
	/*{\Mop returns |true| iff $|ip|\in|rect|(v)$.}*/

	const IPolyline &bends(leda::edge e) const {
		return _bends[e];
	}

	IPolyline &bends(leda::edge e) {
		return _bends[e];
	}
	/*{\Mop returns a reference to the bend point list of edge $e$.}*/

	void get_line (leda::edge e, IPolyline &ipl) const;
	/*{\Mop computes the whole line corresponding to edge $e$ and
	returns it in |ipl|, i.e., the point connecting $e$ at its source
	and target are appended (at the start or end, resp.) to the
	bend point list of $e$.}*/

	leda::node_map<int> &x() {
		return _x;
	}
	/*{\Mop returns a reference to the node map storing the
	x-coordinates of the nodes.}*/

	leda::node_map<int> &y() {
		return _y;
	}
	/*{\Mop returns a reference to the node map storing the
	y-coordinates of the nodes.}*/

	leda::edge_map<IPolyline> &bends() {
		return _bends;
	}
	/*{\Mop returns a reference to the edge map storing the
	bend point lists of the edges.}*/

	
	/*{\Mtext       
	\bigskip
	\headerline{b) Utility functions}
	}*/

	IRect bounding_box() const;
	/*{\Mop returns a tight bounding box of the grid layout.}*/

	int area() const {
		return bounding_box().area();
	}
	/*{\Mop returns the area.}*/

	int number_of_bends() const;
	/*{\Mop returns the total number of bend points.}*/

	int total_manhattan_edge_length() const;
	/*{\Mop returns the sum of the length of all edge segments
	using manhattan distance, that is, a segment $(x_1,y_1),(x_2,y_2)$
	has length $\Labs{x_2-x_1}+\Labs{y_2-y_1}$.}*/

	double total_edge_length() const;
	/*{\Mop returns the sum of the length of all edge segments
	using euclidean distance, that is, a segment $(x_1,y_1),(x_2,y_2)$
	has length $\sqrt{\Labs{x_2-x_1}^2+\Labs{y_2-y_1}^2}$.}*/

	IPolyline get_compact_bends(leda::edge e) const;
	/*{\Mop returns the bend point list for edge $e$ without
	unnecessary bends.}*/

	void compact_all_bends();
	/*{\Mop removes all unnecessary bends.}*/

	void supplement_map_layout();

	static void compact_bends(IPolyline &ip) {
		_compact(ip);
	}
	/*{\Mstatic removes unnecesary bends from integer polyline |ip|,}*/

private:
	void copy_gl(const GridLayout &gl);

	static IPoint _get_intersection (const IPoint &p, const IPoint &q,
		int w, int h);
	static bool _is_redundant(IPoint &p1, IPoint &p2, IPoint &p3);
	static void _compact(IPolyline& ip);

	static int    _manhattan_dist(const IPoint &ip1, const IPoint &ip2);
	static double _euclidean_dist(const IPoint &ip1, const IPoint &ip2);

};

/*{\Mimplementation
All access operations take constant time, number\_of\_bends() takes time
$\BigO(m)$, total\_edge\_length() and total\_manhattan\_edge\_length()
take time $\BigO(m+b)$, and bounding\_box() and area() take time
$\BigO(n+m+b)$, where $n$ is the number of nodes in $G$, $m$ is
the number of edges in $G$, and $b$ is the total number of edge bends.
}*/

#endif
