/*******************************************************************************
+
+  AGD 1.2
+
+
+  KeySet.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         KeySet.h
//
//
//    author: Carsten Gutwenger
//    date:   octobre 1998
//-------------------------------------------------------------------



#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_KEYSET_H
#define _AGD_KEYSET_H


#include <AGD/Keymapper.h>

#include <LEDA/core/int_set.h>
#include <LEDA/internal/std/iostream.h>



/*{\Manpage {KeySet} {} {Key Sets} {S} }*/

class __exportAGD KeySet {

/*{\Mdefinition
An instance |\Mvar| of the data type |\Mname| is a collection
of elements of the data type |AgdKey| with an associated
keymapper $mapper$. All keys stored in |\Mvar| must be allocated
by |mapper|.
}*/

	static const int min_table_size;

public:

	class const_iterator;
	friend class const_iterator;

	/*{\Mcreation}*/

	KeySet();
	/*{\Mcreate creates an instances |\Mvar| of type |\Mtype|
	initialized to the empty set. Sets the associated keymapper
	to the default keymapper of |AGD|.}*/

	KeySet(const Keymapper &M);
	/*{\Mcreate creates an instances |\Mvar| of type |\Mtype|
	with associated keymapper |M| initialized to the empty set.}*/

	KeySet (const KeySet& T);
	/*{\Mcreate creates an instances |\Mvar| of type |\Mtype|
	initialized to a copy of |T|.}*/

	~KeySet();


	/*{\Moperations}*/

	const Keymapper *get_keymapper() const {
		return _mapper;
	}
	/*{\Mop returns the associated keymapper.}*/

	/*{\Mtext
	\headerline{a) Common Set Operations}
	}*/

	void insert (const AgdKey &k);
	/*{\Mop adds |k| to |\Mvar|.}*/

	void del (const AgdKey &k);
	/*{\Mop deletes |k| from |\Mvar|.}*/

	bool member (const AgdKey &k) const;
	/*{\Mop returns |true| iff $k\in \Mvar$.}*/

	KeySet join (const KeySet& T) const {
		KeySet tmp (*this);
		return tmp += T;
	}
	/*{\Mop returns $\Mvar\cup T$.}*/

	KeySet diff (const KeySet& T) const {
		KeySet tmp (*this);
		return tmp -= T;
	}
	/*{\Mop returns $\Mvar\setminus T$.}*/

	KeySet intersect (const KeySet& T) const {
		KeySet tmp (*this);
		return tmp &= T;
	}
	/*{\Mop returns $\Mvar\cap T$.}*/

	KeySet symdiff (const KeySet& T) const {
		KeySet tmp (*this);
		return tmp %= T;
	}
	/*{\Mop returns $(\Mvar\cup T) \setminus (\Mvar\cap T)$.}*/

	bool empty() const;
	/*{\Mop returns |true| iff $\Mvar = \emptyset$.}*/

	void clear();
	/*{\Mop makes |\Mvar| the empty set.}*/


	/*{\Mtext
	\bigskip
	\headerline{b) Operators}
	The following operators return a new key set instead of changing
	|\Mvar|.\nopagebreak
	}*/

	KeySet operator+ (const KeySet& T) const {
		return join(T);
	}
	/*{\Mbinop returns |\Mvar|.join(|T|).}*/

	KeySet operator- (const KeySet& T) const {
		return diff(T);
	}
	/*{\Mbinop returns |\Mvar|.diff(|T|).}*/

	KeySet operator& (const KeySet& T) const {
		return intersect(T);
	}
	/*{\Mbinop returns |\Mvar|.intersect(|T|).}*/

	KeySet operator% (const KeySet& T) const {
		return symdiff(T);
	}
	/*{\Mbinop returns |\Mvar|.symdiff(|T|).}*/

	/*{\Mtext
	\medskip
	The following operators manipulate |\Mvar| directly.\nopagebreak
	}*/

	KeySet& operator+= (const KeySet& T);
	/*{\Mbinop assigns |\Mvar|.join(|T|) to |\Mvar| and returns |\Mvar|.}*/

	KeySet& operator-= (const KeySet& T);
	/*{\Mbinop assigns |\Mvar|.diff(|T|) to |\Mvar| and returns |\Mvar|.}*/

	KeySet& operator&= (const KeySet& T);
	/*{\Mbinop assigns |\Mvar|.intersect(|T|) to |\Mvar| and returns |\Mvar|.}*/

	KeySet& operator%= (const KeySet& T);
	/*{\Mbinop assigns |\Mvar|.symdiff(|T|) to |\Mvar| and returns |\Mvar|.}*/


	KeySet& operator<< (const AgdKey &k) {
		insert(k);
		return *this;
	}
	/*{\Mbinop inserts |k| into |\Mvar| and returns |\Mvar|.}*/


	/*{\Mtext
	\medskip
	Relational and assignment operators:\nopagebreak
	}*/

	bool operator== (const KeySet &T) const;
	/*{\Mbinop return |true| iff $\Mvar=T$.}*/

	bool operator!= (const KeySet &T) const {
		return !operator==(T);
	}
	/*{\Mbinop return |true| iff $\Mvar\neq T$.}*/

	bool operator<= (const KeySet& T) const;
	/*{\Mbinop returns |true| iff $\Mvar\subseteq T$.}*/

	bool operator>= (const KeySet& T) const {
		return T <= (*this);
	}
	/*{\Mbinop returns |true| iff $\Mvar\supseteq T$.}*/

	KeySet& operator= (const KeySet& T);
	/*{\Mbinop assigns |T| to |\Mvar|. Returns |\Mvar|.}*/


	/*{\Mtext
	\medskip
	\headerline{c) Iteration}
	The following functions return iterators of type
	|KeySet::const_iterator|.
	}*/

	const_iterator begin() const;
	/*{\Mop returns an iterator pointing to first element in |\Mvar|.}*/

	const_iterator end() const;
	/*{\Mop returns an iterator pointing to one-past-last element in |\Mvar|.}*/

	const_iterator rbegin() const;
	/*{\Mop returns an iterator pointing to last element in |\Mvar|.}*/

	const_iterator rend() const;
	/*{\Mop returns an iterator pointing to one-before-first element in |\Mvar|.}*/


	/*{\Mtext
	\medskip
	\headerline{d) Input and Output}
	}*/

	void read (istream& is, int delim = EOF);
	/*{\Mop reads a sequence of strings $s_1,\ldots,s_k$ terminated by
	|delim| from the input stream |is|, and assigns the set $\left\{
	\<s_1\>,\ldots,\<s_k\>\right\}$ to |\Mvar|.}*/

	void read (char delim = '\n') {
		read (cin, delim);
	}
	/*{\Mop calls |\Mvar|.read(|cin,delim|) to read |\Mvar| from the
	standard input stream |cin|.}*/

	void print (ostream& os, char space = ',') const;
	/*{\Mop prints the string representations of the elements of |\Mvar|
	to the output stream |os|. The elements are separated by |space|.}*/

	void print (char space = ' ') const {
		print (cout, space);
	}
	/*{\Mop calls |\Mvar|.print(|cout,space|) to print |\Mvar| to the
	standard output stream |cout|.}*/


private:
	void _alloc_table (int sz);
	void _expand_table (int sz);

	leda::int_set *_table;
	int _table_size;
	const Keymapper *_mapper;
};



__exportAGD istream& operator>> (istream& is, KeySet& S);
/*{\Mbinopfunc reads the key set |S| from the input stream |is|.}*/

__exportAGD ostream& operator<< (ostream& os, const KeySet& S);
/*{\Mbinopfunc writes the key set |S| to the output stream |os|.}*/

/*
inline istream& operator>> (istream& is, KeySet& S)
{
	S.read (is);
	return is;
}


inline ostream& operator<< (ostream& os, const KeySet& S)
{
	S.print (os);
	return os;
}
*/


/*{\Mimplementation
Key sets are implemented by bit vectors using an internal numbering of
keys. All keys stored in a key set must be allocated by the same
keymapper.
}*/



/*{\Manpage {KeySet::const_iterator} {} {Key Set Iterators} {it} }*/

class __exportAGD KeySet::const_iterator {

/*{\Mdefinition
Instances of type |\Mvar| iterate over the elements of a key set. They
behave similar to const-pointers to |AgdKey|. The elements of a key set
are always in the same order as the keys in its keymapper. A key set $S$
provides functions to retrieve iterators pointing to the first or
last element of $S$. We say an iterator of key set |S| is valid iff it
points to an element of |S|. Derefferencing is only allowed for
valid iterators.
}*/

public:

	/*{\Mcreation}*/

	const_iterator() : SS(nil) { }
	/*{\Mcreate creates an invalid iterator |\Mvar|.}*/

	const_iterator(const KeySet &S) : SS((KeySet*)&S), _it(get_first()) { }
	/*{\Mcreate creates an iterator |\Mvar| pointing to the first
	element of |S|.}*/

	const_iterator(const KeySet &S, Keymapper::const_iterator it) :
	SS((KeySet*)&S), _it(get_next(it)) { }

	const_iterator(const const_iterator &iter) : SS(iter.SS), _it(iter._it) { }
	/*{\Mcreate creates an iterator identical to |iter|.}*/


	/*{\Moperations}*/

	bool valid() const { return _it.valid(); }
	/*{\Mop returns |true| iff |\Mvar| is valid, i.e., points to an
	element.}*/

	bool operator==(const const_iterator& iter) const {
		return (_it == iter._it);
	}
	/*{\Mbinop returns |true| iff |\Mvar| and |iter| point to the
	same element.}*/

	bool operator!=(const const_iterator& iter) const {
		return (_it != iter._it);
	}
	/*{\Mbinop returns |true| iff |\Mvar| and |iter| point to
	different elements.}*/

	const_iterator &operator=(const const_iterator &iter) {
		SS = iter.SS;
		_it = iter._it;
		return *this;
	}
	/*{\Mbinop assigns |iter| to |\Mvar|.}*/

	const AgdKey &operator*() { return *_it; }
	/*{\Munop returns the element to which |\Mvar| points.}*/

	const AgdKey *operator->() { return _it.operator->(); }
	/*{\Munop returns the address of the element to which |\Mvar| points.}*/

	const_iterator &operator++() {
		_it = (_it.valid()) ? get_next(++_it) : get_first();
		return *this;
	}
	/*{\Munop increments and returns |\Mvar|.}*/

	const_iterator operator++(int) {
		const_iterator t = *this;
		_it = (_it.valid()) ? get_next(++_it) : get_first();
		return t;
	}
	/*{\Munop increments |\Mvar| and returns its previous value.}*/

	const_iterator &operator--() {
		_it = (_it.valid()) ? get_prev(--_it) : get_last();
		return *this;
	}
	/*{\Munop decrements and returns |\Mvar|.}*/

	const_iterator operator--(int) {
		const_iterator t = *this;
		_it = (_it.valid()) ? get_prev(--_it) : get_last();
		return t;
	}
	/*{\Munop decrements |\Mvar| and returns its previous value.}*/

protected:
	Keymapper::const_iterator get_first() {
		return get_next(SS->_mapper->begin());
	}

	Keymapper::const_iterator get_last() {
		return get_prev(SS->_mapper->rbegin());
	}

	Keymapper::const_iterator get_next(Keymapper::const_iterator it) {
		while (it.valid() && !SS->member(*it))
			++it;
		return it;
	}

	Keymapper::const_iterator get_prev(Keymapper::const_iterator it) {
		while (it.valid() && !SS->member(*it))
			--it;
		return it;
	}

	KeySet *SS;
	Keymapper::const_iterator _it;
};


inline KeySet::const_iterator KeySet::begin() const
{
	return const_iterator(*this,_mapper->begin());
}

inline KeySet::const_iterator KeySet::end() const
{
	return const_iterator(*this,_mapper->end());
}

inline KeySet::const_iterator KeySet::rbegin() const
{
	return const_iterator(*this,_mapper->rbegin());
}

inline KeySet::const_iterator KeySet::rend() const
{
	return const_iterator(*this,_mapper->rend());
}


/*{\Mexample
The following examples shows, how to use key set iterators
in order to iterate over all elements of a key set. The
second loop iterates in reverse order and demonstrates the
use of {\tt operator->}.

\small\tt
\begin{verbatim}
void write_set (const KeySet &S)
{
  KeySet::const_iterator it;

  // iteration over all elements of S
  for (it = S.begin(); it != S.end(); ++it) {
    cout << *it << endl;
  }

  // iteration in reverse order
  for (it = S.rbegin(); it != S.rend(); --it) {
    cout << it->index() << endl;
  }
}
\end{verbatim}\normalsize\rm
}*/


#endif
