/*******************************************************************************
+
+  AGD 1.2
+
+
+  LayoutBuffer.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         LayoutBuffer.h
//
//    author: Carsten Gutwenger
//    date:   july 1999
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_LAYOUT_BUFFER_H
#define _AGD_LAYOUT_BUFFER_H


#include <AGD/LayoutInterface.h>


/*{\Manpage {LayoutBuffer} {} {Layout Buffers} {A} }*/

/*{\Mbaseclasses

\begin{hierarchy}{6.5}{2.5}
\hdclass{0}{1.5}{5}{LayoutInterface}
\arhook{5}{2}{6}
\hdclass{1.5}{0}{5}{LayoutBuffer}
\end{hierarchy}
}*/

class __exportAGD LayoutBuffer : public LayoutInterface {

/*{\Mdefinition
A layout buffer for graph $G$ is a layout interface that stores the
following graphical attributes for graph $G$:
\begin{itemize}
\item the position of each node.
\item the bend points of each edge.
\item the source and target anchor of each edge.
\end{itemize}
The class |\Mname| overrides the pure virtual methods of
|LayoutInterface| for setting and getting these attributes. The
methods set\_width(), get\_width(), set\_height(), and get\_height() must
be overridden by derived classes. Layout buffers are mainly used as
base class for |GraphCopyLI| and |AGDLayout|.
}*/

public:

	/*{\Minitialization A 5.5}*/

	LayoutBuffer (const leda::graph& G) : LayoutInterface(G),
		_pos(G), _bends(G), _s_anchor(G), _t_anchor(G) { }
	/*{\Mcreate initializes a layout buffer for graph $G$.}*/

	LayoutBuffer(const LayoutBuffer& lb) : LayoutInterface(lb),
		_pos(lb._pos), _bends(lb._bends), _s_anchor(lb._s_anchor),
		_t_anchor(lb._t_anchor) { }
	/*{\Mcreate copy constructor.}*/

	~LayoutBuffer() { }


	/*{\Moperations}*/
	/*{\Mtext
	|\Mname| inherits the operations from its base class |LayoutInterface|
	(see Sect.~\ref{Layout Interfaces})
	}*/

	double get_x (leda::node v) const {
		return _pos[v].xcoord();
	}

	void set_x (leda::node v, double new_x) {
		_pos[v].xcoord(new_x);
	}

	double get_y (leda::node v) const {
		return _pos[v].ycoord();
	}

	void set_y (leda::node v, double new_y) {
		_pos[v].ycoord(new_y);
	}

	DPoint get_position (leda::node v) const {
		return _pos[v];
	}

	void set_position (leda::node v, const DPoint& pos) {
		_pos[v] = pos;
	}
	
	DPolyline get_bend_points (leda::edge e) const {
		return _bends[e];
	}

	void set_bend_points (leda::edge e, const DPolyline& l) {
		_bends[e] = l;
	}

	DPoint get_source_anchor (leda::edge e) const {
		return _s_anchor[e];
	}

	void set_source_anchor (leda::edge e, const DPoint& pos) {
		_s_anchor[e] = pos;
	}

	DPoint get_target_anchor (leda::edge e) const {
		return _t_anchor[e];
	}

	void set_target_anchor (leda::edge e, const DPoint& pos) {
		_t_anchor[e] = pos;
	}


private:
	leda::node_map<DPoint> _pos;
	leda::edge_map<DPolyline> _bends;
	leda::edge_map<DPoint> _s_anchor, _t_anchor;
};


#endif
