/*******************************************************************************
+
+  AGD 1.2
+
+
+  LayoutModuleImpl.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                        LayoutModuleImpl.h
//
//    author: Carsten Gutwenger
//    date:   octobre 1998
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_LM_IMPL_H
#define _AGD_LM_IMPL_H


#include <AGD/LayoutModule.h>
#include <AGD/KeySet.h>


/*{\Manpage {LayoutModuleImpl} {CONCEPT} {Implementation of Layout Algorithms} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{2.2}{3}{5}{AGDModule}
\arhook{7.2}{3.5}{8.2}
\hdclass{3.7}{1.5}{5}{LayoutModule}
\arhook{8.7}{2}{9.7}
\hdclass{0}{0}{10.2}{LayoutModuleImpl{\tt<}CONCEPT{\tt>}}
\end{hierarchy}
}*/

template <class CONCEPT>
class __exportAGD LayoutModuleImpl : public LayoutModule {

/*{\Mdefinition
The class |\Mname| represents a generic method for implementing layout
algorithms. The template argument |CONCEPT| is the class
which represents the used implementation concept. It is responsible
for allocating the data structures used by the layout algorithm
and for translating the layout computed by the algorithm into
a layout for the input graph. A class used as template
argument |CONCEPT| must provide the following constructors:
\begin{quote}
\begin{tabular}{p{0.5cm}p{12cm}}
\multicolumn{2}{l}{|CONCEPT|(const |graph| \&|G|,|LayoutInterface| \&|A|)}\\
& allocate data structures for graph $G$ with layout interface $A$.\\[2mm]
\multicolumn{2}{l}{|CONCEPT|(const |graph| \&|G|,|face| |f|,|LayoutInterface| \&|A|)} \\
& allocate data structures for planar map $G$ with exterior face $f$ and layout interface $A$.\\[2mm]
\multicolumn{2}{l}{|CONCEPT|(|PlaneGraphCopy| \&|PG|,|face| |f|,|LayoutInterface| \&|A|)} \\
& allocate data structures for plane graph copy |PG| with exterior face $f$ and layout interface $A$.
\end{tabular}
\end{quote}
These constructors correspond to call() methods inherited
by |LayoutModule|. If an exterior face $f$ is
passed, it may be |nil| indicating that the algorithm chooses an
appropriate exterior face. Moreover, the following operations are required:
\begin{quote}
\begin{tabular}{p{0.5cm}p{12cm}}
\multicolumn{2}{l}{void update()} \\
& transforms the computed layout into a layout for the input graph.\\
\multicolumn{2}{l}{static void set\_postcondition(|list<AgdKey>| \&|post|)} \\
& assigns the general postcondition to |post|, i.e., adds |key::planar_map_callable|
to post if calls for planar maps (second constructor) are supported, and adds
|key::PlaneGraphCopy_callable| to |post| if calls for |PlaneGraphCopy|'s (third
constructor) are supported.
\end{tabular}
\end{quote}
|\Mname| implements all call() methods inherited by |LayoutModule|.
The computation of a layout (call() methods) works as follows:
\begin{enumerate}
\item An object |C| of type |CONCEPT| is constructed using the
constructor corresponding to the call() parameter list.
\item The actual computation of the grid layout is performed by
calling pure virtual method do\_call$(C)$.
\item The computed layout is transformed into a layout for
the input graph by calling update$()$ of $C$.
\end{enumerate}
Thus, it is sufficient to implement do\_call$(C)$. All necessary
transformations are performed by the implementation concept class.
}*/

public:

	LayoutModuleImpl() {
		leda::list<AgdKey> post;
		CONCEPT::set_postcondition(post);
		AgdKey p;
		forall(p,post) add_post_rule(p,0);
	}

	~LayoutModuleImpl() { }

	/*{\Moperations 1.5 5}*/
	/*{\Mtext |\Mname| implements the call() methods
	of its base class. Furthermore, there is the following protected
	overridable:}*/

	virtual void call (const leda::graph& G, LayoutInterface& A) {
		CONCEPT C(G,A);
		do_call(C);
		C.update();
	}

	virtual void call (const leda::graph& G, leda::face f, LayoutInterface& A) {
		if (!postcondition(0).member(key::planar_map_callable))
			leda::error_handler(1, "LayoutModuleImpl: algorithm not callable for planar maps!");
		CONCEPT C(G,f,A);
		do_call(C);
		C.update();
	}

	virtual void call (PlaneGraphCopy& PG, leda::face f, LayoutInterface& A) {
		if (!postcondition(0).member(key::planar_map_callable))
			leda::error_handler(1, "LayoutModuleImpl: algorithm not callable for PlaneGraphCopy's!");
		CONCEPT C(PG,f,A);
		do_call(C);
		C.update();
	}

protected:

	virtual void do_call (CONCEPT &C) = 0;
	/*{\Moppure implements the algorithm.}*/

};



#endif
