/*******************************************************************************
+
+  AGD 1.2
+
+
+  OptNodeRanking.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                       OptNodeRanking.h
//
//
//    author: Carsten Gutwenger
//    date:   june 1999
//-------------------------------------------------------------------


#ifndef _AGD_O_N_RANKING_H
#define _AGD_O_N_RANKING_H


#include <AGD/RankAssignment.h>
#include <AGD/AGDModuleOption.h>
#include <AGD/SubgraphModule.h>


/*{\Manpage {OptNodeRanking} {} {Ranking to Minimize Edge Lengths} {R} }*/

/*{\Mbaseclasses
\abacus
\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{1.5}{1.5}{5.5}{RankAssignment}
\arhook{7}{2}{8}
\hdclass{2}{0}{6.5}{OptNodeRanking}
\end{hierarchy}
}*/

class __exportAGDopt OptNodeRanking : public RankAssignment {

/*{\Mdefinition
The |\Mname| module computes the following ranking of a graph $G=(V,E)$:
First, a feedback set $R\subset E$ is computed, that is, the graph
$G'=/V,E')$ resulting from $G$ by reversing all edges in $R$ is acyclic.
In this step, self-loops are simply ignored. For this purpose, we use
a subgraph module that computes a maximal acyclic subgraph (that also
implies a feedback set $R$). Then, we compute a ranking that is defined
as the optimal solution of the following optimization problem
\begin{center}
\begin{tabular}{lrclp{4.5cm}}
\multicolumn{5}{l}{$\min \sum\limits_{(v,w)\in E'}(y(w)-y(v))$\hspace{1.5cm}s.t.}\\[2.5ex]
\hspace*{0.5cm} & $y(w)-y(v)$ & $\geq$ & $1$ & for all $(v,w)\in E'$\\
& \multicolumn{3}{c}{$x(v)\in\{0,1\}$} & for all $v\in V$
\end{tabular}
\end{center}
where $y(v)$ is the rank of vertex $v$. The resulting ranking minimizes
the total edge lengths (the sum of the spans of all edges) and therefore
the number of dummy vertices, which have to be created in order to
build a proper hierarchy when used with the Sugiyama algorithm. This
method for computing a ranking has been suggested by Gansner et al. in~\cite{GKNV93},
where is also shown, how to solve the integer linear program in
polynomial time.

\horizline
\commoninfo{Optimal Node Ranking}{Optimal Node Ranking}
{E. R. Gansner, E. Koutsofios, S. C. North, K. P. Vo}{C. Gutwenger}{June 1999}{1.0}

\begin{preposttab}
precondition & $\{\, |directed| \,\}$ \\
postcondition(|PRE|) & $\emptyset$
\end{preposttab}

\headerline{Exchangeable Modules}
Instances of type |\Mname| provide the following module options:
\begin{options}
\item {\bf subgraph}\\
an |SubgraphModule| option used for computing a maximal acyclic subgraph.

\begin{tabular}{ll}
guaranteed precondition: & $\{\emph{ directed }\}$\\
required postcondition: & $\{\emph{ directed, maximal\_acyclic }\}$\\
initial module: & |LEDAMakeAcyclic|
\end{tabular}
\end{options}
}*/

public:

	/*{\Mcreation}*/

	OptNodeRanking();
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname|.}*/

	~OptNodeRanking() { }


	/*{\Moperations}*/

	leda::string name () const {
		return "Optimal Node Ranking";
	}

	leda::string author () const {
		return "E. R. Gansner, E. Koutsofios, S. C. North, K. P. Vo";
	}

	leda::string impl_author () const {
		return "C. Gutwenger";
	}

	leda::string impl_version () const {
		return "1.0";
	}

	leda::string impl_date () const {
		return "June 1999";
	}

	AGDModule *clone() const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|RankAssignment|).
	}*/

	/* inherited
	bool check (const leda::graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	void call (const leda::graph& G, leda::node_array<int>& rank);
	/*{\Mop}*/

	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	/*
	bool set_subgraph (const SubgraphModule& M);
	*/
	/*{\Mop}*/

	AGD_DECLARE_MODULE (SubgraphModule, subgraph)

	/*{\Mtext \stdend}*/


private:
};

/*{\Mimplementation
The integrality condition in the integer linear program (ILP) can
be omitted if it is solved with the simplex
algorithm, because the constraint matrix is unimodular
(see~\cite{GKNV93}). Therefore, we solve the relaxation
of the ILP with an LP-solver, what gives us the optimal
solution of the ILP.
}*/


#endif
