/*******************************************************************************
+
+  AGD 1.2
+
+
+  OptPlanarSubgraph.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         OptPlanarSubgraph.h
//
//
// author: Sergej Fialko
// date:   march 1998
//-------------------------------------------------------------------


#ifndef _AGD_OPT_PLANAR_SUBGRAPH_H
#define _AGD_OPT_PLANAR_SUBGRAPH_H


#include <AGD/SubgraphModule.h>


/*{\Manpage {OptPlanarSubgraph} {} {Maximum Planar Subgraphs} {S} }*/

/*{\Mbaseclasses
\abacus
\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{1}{1.5}{6}{SubgraphModule}
\arhook{7}{2}{8}
\hdclass{2}{0}{6.5}{OptPlanarSubgraph}
\end{hierarchy}
}*/

class __exportAGDopt OptPlanarSubgraph : public SubgraphModule {

/*{\Mdefinition
The module |\Mname| computes a planar subgraph of an
input graph $G$ by deleting a minimum number of edges.
Or in other words, the module |\Mname| computes a maximum planar 
subgraph of an input graph $G$ (\cite{JM96a,Mut94}).

Since this problem
is an NP-hard optimization problem, the algorithm may take exponential
time. On average, instances with around 60 edges can be solved
to provable optimality within a few minutes. 

The algorithm is based on branch-and-cut using ABACUS~\cite{JT97}.
Branch-and-cut algorithms combine integer programming techniques
like polyhedral combinatorics with a branch-and-bound scheme.
The advantage of a branch-and-cut algorithm lies in improved
bounding. This can be achieved by investigating the mathematical
structure of the underlying polytope. For the maximum planar subgraph problem
the mathematical investigation has been done in the
articles by J\"unger and Mutzel~\cite{JM96a} and Mutzel~\cite{Mut94}.

\horizline
\commoninfo{Maximum Planar Subgraph}{Maximum Planar Subgraph}{P. Mutzel}{S. Fialko}
{March 1998}{}

\headerline{Pre- and Postcondition}
\begin{tabular}{lcl}
precondition & $=$ & $\emptyset$ \\
postcondition(|PRE|) & $=$ & $\{\, \emph{maximum planar} \,\}$\\
& & $\quad\cup\, \Big( \{\, \emph{simple, no self loops, no reversals} \,\} \cap |PRE| \Big)$
\end{tabular}
\horizline
}*/

private:
  int nHeuristicCalls_;// 

public:

	/*{\Mcreation S 5.5}*/

	OptPlanarSubgraph();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

	~OptPlanarSubgraph() { }


	/*{\Moperations 1.2 3.5}*/

	leda::string name () const {
		return "Maximum Planar Subgraph";
	}

	leda::string author () const {
		return "P. Mutzel";
	}

	leda::string impl_author () const {
		return "S. Fialko";
	}

	leda::string impl_date () const {
		return "March 1998";
	}

	AGDModule *clone() const;
	
	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|SubgraphModule|).
	}*/

	/* inherited
	bool check (const graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/* inherited
	bool call (const graph& G, list<edge>& L);
	*/
	/*{\Mop}*/

	/* inherited
	bool call (graph& G);
	*/
	/*{\Mop}*/

	/* inherited
	bool call (GraphCopy& G, list<edge>& L);
	*/
	/*{\Mop}*/

	/* inherited
	int num_del_edges() const;
	*/
	/*{\Mop}*/

	/*{\Mtext \stdend}*/


	inline void nHeuristicCalls (int n) { nHeuristicCalls_ = n; }

protected:
	void do_call (const leda::graph& G, leda::list<leda::edge>& L);
};


#endif
