/*******************************************************************************
+
+  AGD 1.2
+
+
+  OrthogonalRepresentation.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



//#line 72 "Lweb/OrthoRep.lw"
// OrthogonalRepresentation.h
// by Gunnar W. Klau
// Version 1.2
// November 1999

#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_ORTHOSHAPE_H
#define _AGD_ORTHOSHAPE_H

#include <AGD/PlaneGraphCopy.h>
//#include <AGD/impl/node_map.h>
#include <AGD/impl/Ortho_common.h>
#include <LEDA/internal/std/iostream.h>

/*{\Moptions
outfile=OrthogonalRepresentation.man
}*/

/*{\Manpage {OrthogonalRepresentation} {} {Orthogonal Representations} {H} }*/

/*{\Mbaseclasses

\begin{hierarchy}{10}{6}
\hdclass{0}{4.5}{5}{graph}
\arhook{5}{5}{6}
\hdclass{1.5}{3}{5}{GraphCopy}
\arhook{6.5}{3.5}{7.5}
\hdclass{2.5}{1.5}{5.5}{PlaneGraphCopy}
\arhook{8}{2}{9}
\hdclass{3.5}{0}{8}{OrthogonalRepresentation}
\end{hierarchy}
}*/

class __exportAGD OrthogonalRepresentation : public PlaneGraphCopy {

  friend ostream &operator<<(ostream&, const OrthogonalRepresentation&);

  /*{\Mdefinition
    The class |\Mname| implements the concept of orthogonal representation as introduced in \cite{t-eggmn-87}. Such a representation can be thought of as an embedded graph $G$ with additional information for each edge concerning the shape of any drawing of $G$. For each edge $e \in E(G)$ the following information can be accessed:
    \begin{itemize}
    \item A value $a$ denoting the angle formed with the following edge.
    \item A bit string $s$ describing the bends in the order they appear on $e$. A {\tt 0} stands for a convex and a {\tt 1} stands for a reflex bend.
    \end{itemize} }*/



//#line 119 "Lweb/OrthoRep.lw"
public:

/*{\Mcreation H 6}*/

OrthogonalRepresentation() : PlaneGraphCopy() { }

  OrthogonalRepresentation(const leda::graph &G, leda::face f);
  /*{\Mcreate creates an instance |\Mvar| of type |\Mname| and 
     initializes the underlying graph with |G|. The external face
     is set to |f| unless |f| is the empty face. In this case,
     the external face is set to the biggest face.
  }*/

  OrthogonalRepresentation(const PlaneGraphCopy &G, leda::face f);

  OrthogonalRepresentation(const OrthogonalRepresentation &H);
  OrthogonalRepresentation(const OrthogonalRepresentation &H, bool c);
  
  ~OrthogonalRepresentation() { }

  /*{\Moperations}*/
  /*{\Mtext \headerline{Access Operations}}*/
  
  leda::face get_external_face() const { return f_0_; }
  /*{\Mop returns the external face. }*/

  int get_angle(leda::edge e) const { return a_[e]; }
  /*{\Mop returns the angle edge |e| forms with its successor.
  }*/

  leda::string get_string(leda::edge e) const { return s_[e]; }
  /*{\Mop returns the bit string for edge $e$. }*/

  direction dir(leda::edge e) const { return dir_[e]; }
  /*{\Mop returns the direction of edge $e$. }*/

  bool hor(leda::edge e) const { return (dir_[e] == N_); }
  /*{\Mop returns true if edge $e$ is horizontal (direction north). }*/
  
  bool ver(leda::edge e) const { return (dir_[e] == W_); }
  /*{\Mop returns true if edge $e$ is vertical (direction west). }*/

  int number_of_bends() const { return number_of_bends_; }
  /*{\Mop returns the number of bends}*/
  
  bool simple() const { return number_of_bends_ == 0; }
  /*{\Mop returns true if |H| is simple, i.e., the number of bends is zero, and     false otherwise. }*/

  bool is_cage(leda::face f) const { return L_[first_face_edge(f)] != nil; }
  /*{\Mop returns true, if |f| is an artificial face, i.e., a face corresponding to a vertex in the original graph. }*/
  
  int output_level() const { return output_level_; } 
  /*{\Mop returns the output level}*/

 /*{\Mtext \headerline{Update Operations}}*/
  void set_external_face(leda::face f) { f_0_ = f; }
  /*{\Mop sets the external face to $f$. }*/

  void set_angle(leda::edge e, int a) { a_[e] = a; }
  /*{\Mop sets the angle between $e$ and the following edge to $a$. }*/

  void set_string(leda::edge e, leda::string s) { s_[e] = s; }
  /*{\Mop sets the bit string for edge $e$ to $s$. }*/

  void set_string(leda::edge e, char c, int n);
  /*{\Mop sets the bit string for edge $e$ to $c^n$. }*/

  void add_convex_bend(leda::edge e);
  /*{\Mop adds a 90$^\circ$ bend to edge |e|. }*/

  void add_concave_bend(leda::edge e);
  /*{\Mop adds a 270$^\circ$ bend to edge |e|. }*/

  leda::face expand(leda::node v);
  /*{\Mop replaces vertex $v$ by face $c$. Returns $c$. }*/

  leda::node collapse(leda::face f);
  /*{\Mop is the undo of expand. Face $f$ will be replaced by its former vertex $v$. }*/

  leda::edge split(leda::edge e);
  /*{\Mop splits egde $e$ and updates the data structure accordingly. }*/
  
  void normalize();
  /*{\Mop transforms the representation into a {\em simple} representation
    by replacing each bend by a vertex and updating the data structure. }*/

  void dissect();
  /*{\Mop dissects all inner faces in subfaces of rectangular shape. }*/

  void undissect();
  /*{\Mop is the inverse operation of dissect. Deletes all previously inserted virtual vertices and edges. }*/
  
  void orientate();
  /*{\Mop assigns consistent directions to the edge segments. }*/

  void orientate(leda::edge e, direction d);
  /*{\Mop assigns consistent directions to the edge segments. The reference
    edge $e$ gets direction $d$. }*/

  void output_level(int o) { output_level_ = o; } 
  /*{\Mop sets the output level to |o|. 
    Reasonable values are $o \in [0 \ldots 10]$. 
    $o = 0$ causes the algorithm to run in silent mode, 
    a value $o \geq 10$ will create the greatest
    possible amount of output. }*/
  
  /*{\Mtext \headerline{Output and Verification Operations}}*/

  bool check(leda::string& msg) const;
  /*{\Mop tests if |\Mvar| is a {\em legal} orthogonal shape desription,
    i.e., if there is an orthogonal grid embedding which realizes the
    given shape. If the test fails, |msg| contains the reason.
    }*/
  
  virtual void print_edge(leda::edge e, ostream& O = cout) const;
  /*{\Mop ouputs edge $e$ and the information associated with $e$. }*/
  


//#line 241 "Lweb/OrthoRep.lw"
protected:

  int ones(leda::edge e) const { return s_[e].length() - zeroes_[e]; }

  void orientate(leda::face f, leda::edge e, direction d);

  int output_level_;


//#line 258 "Lweb/OrthoRep.lw"
private:

  static const leda::string epsilon;
  
  inline char flip(char c) const { 
    if (c == '0') return '1';
    if (c == '1') return '0';
    return -1;
  }

//#line 280 "Lweb/OrthoRep.lw"
 
  leda::face f_0_;
  leda::edge_map<int> a_;
  leda::edge_map<leda::string> s_;

  leda::edge_map<int> zeroes_;
  leda::edge_map<direction> dir_;
  leda::edge_map<leda::node> L_;
  leda::edge_map<bool> scaff_;
  leda::node_map<bool> screw_;
  int number_of_bends_;
};

ostream& operator<< (ostream& O, const OrthogonalRepresentation& H);
/*{\Mbinopfunc writes |H| to output stream $O$.}*/

#endif

