/*******************************************************************************
+
+  AGD 1.2
+
+
+  PertinentGraph.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                          PertinentGraph.h
//
//
//    author: Carsten Gutwenger
//    date:   August 1999
//-------------------------------------------------------------------


#ifndef _AGD_PERT_GRAPH_H
#define _AGD_PERT_GRAPH_H


class SPQRTree;


/*{\Manpage {PertinentGraph} {} {Pertinent Graphs} {P} }*/

class __exportAGD PertinentGraph {

/*{\Mdefinition
The class |\Mname| represents the pertinent graph $G(v_{\calT})$
of a node $v_{\calT}$ in an SPQR-tree $\calT$ with original
graph $G$.

The {\em expansion graph} $E(e)$ of a virtual skeleton edge $e$ is
the skeleton graph of the twin $e'$ of $e$ without $e'$, where
each virtual edge $e_v$ again is replaced by its expansion
graph $E(e_v)$. The {\em pertinent graph} $G(v_{\calT})$ of
a tree node $v_{\calT}$ is obtained from the skeleton $S$ of
$v_{\calT}$, where each edge except for the reference edge
of $S$ is replaced by its expansion graph. Hence, if $v_{\calT}$
is not the root node of $\calT$, all but one edge (and its
reversal) in $G(v_{\calT})$ correspond to real edges, otherwise
all edges correspond to real edges.

The underlying graph, which is returned by get\_graph(), represents a map,
that is, for each edge $(u,v)$ also the reversal $(v,u)$ exists.
If |\Mvar| is the pertinent graph of a |PlanarSPQRTree|, the
underlying graph is a planar map, whose embedding is implied by
the embeddings of the skeletons of $\calT$.
}*/

	friend class SPQRTree;

public:

	/*{\Mcreation}*/
	/*{\Mtext Pertinent graphs are created by the pertinent\_graph()
	function of |SPQRTree|.}*/

	PertinentGraph() : _vT(nil) { }
	/*{\Mcreate creates an empty instance of type |\Mname|.}*/

	void init(leda::node vT) {
		P = leda::graph(1,1);
		_vT = vT;
		_v_edge = _sk_ref_edge = nil;
		origV.use_node_data(P,nil);
		origE.use_edge_data(P,nil);
	}


	/*{\Moperations}*/

	leda::node tree_node() const {
		return _vT;
	}
	/*{\Mop returns the tree node $v_{\calT}$ in $\calT$ whose pertinent graph is
	|\Mvar|.}*/

	const leda::graph &get_graph() const {
		return P;
	}

	leda::graph &get_graph() {
		return P;
	}
	/*{\Mop returns a reference to $G(v_{\calT})$.}*/

	leda::edge reference_edge() const {
		return _v_edge;
	}
	/*{\Mop returns the edge in $G(v_{\calT})$ that corresponds
	to the reference edge in skeleton of $v_{\calT}$ if $v_{\calT}$
	is not the root of $\calT$, |nil| otherwise.}*/

	leda::edge skeleton_reference_edge() const {
		return _sk_ref_edge;
	}
	/*{\Mop returns the reference edge in skeleton of $v_{\calT}$. Notice that
	this edge may differ from the current reference edge in skeleton of $v_{\calT}$
	if $\calT$ has been rerooted after the construction of $\Mvar$.}*/

	leda::node original(leda::node v) const {
		return origV[v];
	}
	/*{\Mop returns the vertex in $G$ that corresponds to $v$. \precond
	$v\in G(v_{\calT})$}*/

	leda::edge original(leda::edge e) const {
		return origE[e];
	}
	/*{\Mop returns the edge in $G$ that corresponds to $e$, or |nil| if
	$e$ is the reference edge (or its reversal). \precond
	$e\in G(v_{\calT})$}*/

protected:
	leda::node _vT;
	leda::graph P;
	leda::edge _v_edge;
	leda::edge _sk_ref_edge;

	leda::node_array<leda::node> origV;
	leda::edge_array<leda::edge> origE;
};


/*{\Mimplementation
All operations take constant time. The data structure uses
$\BigO(\Labs{G(v_{\calT})})$ space.
}*/


#endif
