/*******************************************************************************
+
+  AGD 1.2
+
+
+  PlanarizerModule.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                        PlanarizerModule.h
//
//
//    author: Carsten Gutwenger
//    date:   august 1997
//-------------------------------------------------------------------


#ifndef _AGD_PLANARIZER_MODULE_H
#define _AGD_PLANARIZER_MODULE_H


#include <AGD/AGDModule.h>


class GraphCopy;
class PlaneGraphCopy;


/*{\Manpage {PlanarizerModule} {} {Planarization Modules} {P} }*/

/*{\Mbaseclasses

\begin{hierarchy}{6.5}{2.5}
\hdclass{0}{1.5}{5.5}{AGDModule}
\arhook{5.5}{2}{6.5}
\hdclass{1}{0}{6}{PlanarizerModule}
\end{hierarchy}
}*/

class __exportAGD PlanarizerModule : public AGDModule {

/*{\Mdefinition
The class |\Mname| is the base class for planarization modules. A
planarization module transforms a graph $G$ into a planar graph |PG| by
replacing edge crossings with dummy nodes, and computes a combinatorial
embedding of |PG|.

\headerline{Input and Output Parameters}

\begin{tabular}{lcc}
input parameter: & graph & |in_graph=1| \\
output parameter: & planarized graph & |out_planarized_graph=1|
\end{tabular}

The integer constants |in_graph| and |out_planarized_layout| denote the
positions in the input and output parameter lists. The values can be used to
reference the desired parameter when setting the pre- and postcondition.  

\headerline{Implementation of Planarization Algorithms}

An implementation of a planarization module must override the protected
method do\_call($G$,|PG|),
which gets as input a graph $G=(V,E)$ and produces as output a planar
graph $|PG|=(V\cup V_d,E')$. It replaces a set of edges $E_r\subset E$
by chains of edges leading over dummy nodes $\in V_d$.
Each dummy node is a node of degree $4$, and the output graph |PG|
represents a planar map.

In particular, an edge $e=(v,w)\in E_r$ is replaced with a chain
$e_1=(v_1,v_2),\ldots,e_k=(v_k,v_{k+1})$, where $v_1=v$, $v_{k+1}=w$, and
$v_i\in V_d$ for all $2\leq i\leq k$. The corresponding chain of edges
is stored in |PG|, i.e., $|PG|.|chain|(|original|(e))=e_1,\ldots,e_k$.
Furthermore, tho following must hold:
\begin{itemize}
\item $|degree|(v)=4$ for all $v\in V_d$, and $E_r = E\setminus E'$.
\item Let $E_d=E'\setminus E$ be the set of inserted
new edges.

Then, there must be a mapping $\varphi : E_d \to E_r$ with
$\stackrel{-1}{\varphi}(e)=|chain|(e)$ for all $e\in E_r$.
\end{itemize}
The implementation must also embed the output graph, i.e., the combinatorial
embedding must be given by the clockwise ordering of the edges in all
adjacency lists.
}*/

public:

	// the input and output parameters of a planarization module
	static const int in_graph;
	static const int out_planarized_graph;

	/*{\Minitialization P 4.5}*/

	PlanarizerModule () : AGDModule(1,1) { }

	/*{\Mcreate initializes a planarization module.}*/

	virtual ~PlanarizerModule () { }


	/*{\Moperations 1.5 3.5}*/

	/*{\Mtext
	\headerline{Standard Interface} }*/

	bool check (const GraphCopy& G, AgdKey &p) const;
	/*{\Mop returns |true| if $G$ satisfies the precondition of |\Mvar|.
	Otherwise, |false| is returned and |p| contains a property that
	is not satisfied.}*/

	void call (const leda::graph& G, PlaneGraphCopy& PG);
	/*{\Mop calls the planarization module for graph $PG$. Stores
	the planarization in $PG$, i.e., for every edge $e\in G$
	the ordered list of corresponding edges |chain(e)| in |PG|.}*/

protected:

	/*{\Mtext
	\bigskip
	\headerline{Protected Overridables to Implement Functionality} }*/

	virtual void do_call (const leda::graph& G, PlaneGraphCopy& PG)
	/*{\Moppure implements the planarization algorithm for graph $G$. Stores
	the planarization in |PG|, i.e., for every edge $e\in E_r$
	the ordered list of corresponding edges |chain(e)| in |PG|.}*/
	= 0;

};



#endif
