/*******************************************************************************
+
+  AGD 1.2
+
+
+  PureOrthogonalLayout.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



//#line 85 "Lweb/OrthoLayout.lw"
#ifndef _AGD_PURE_ORTHO_H 
#define _AGD_PURE_ORTHO_H

#include <AGD/Ovid.h>

/*{\Moptions
outfile=PureOrthogonalLayout.man
}*/

/*{\Manpage {PureOrthogonalLayout} {} {The Pure Orthogonal Drawing Algorithm} {P} }*/

/*{\Mbaseclasses
\begin{hierarchy}{9.5}{7}
\hdclass{0}{6}{5}{AGDModule} \arhook{5}{6.5}{6}
\hdclass{1.5}{4.5}{5}{LayoutModule} \arhook{6.5}{5}{7.5}
\hdclass{2}{3}{6}{GridLayoutModule} \arhook{8}{3.5}{9}
\hdclass{6.5}{1.5}{3}{Ovid} \arhook{9.5}{2}{10.5}
\hdclass{6}{0}{8}{PureOrthogonalLayout}
\end{hierarchy}
\hfill\agdeps{PureOrthogonalLayout.eps}{5.5cm} 
}*/


class __exportAGD PureOrthogonalLayout : public Ovid {

/*{\Mdefinition 
  The class |\Mname| is an implementation of 
Tamassia's bend minimizing algorithm for the class of 4--planar graphs \cite{t-eggmn-87}. In addition to planarity, these graphs have the property that no vertex has more than four neighbours. Note that pure orthogonal drawings in which the vertices are mapped to points in the grid are only admissible for 4--graphs. For arbitrary planar graphs we provide two extensions of the orthogonal standard: They are realized by the classes [[QuasiOrthogonalLayout]] and [[OrthogonalLayout]]. 

  The computation of the layout follows the so--called Topology--Shape--Metrics approach \cite{DETT99}. The input is a 4--planar graph with fixed topology. According to this topology the algorithm constructs a network. A minimum--cost flow in this network determines a bend--minimal orthogonal representation for the input graph. In a last phase, one of the available compaction modules assigns lengths to this representation and thus fixes the coordinates of the drawing.

\horizline

\commoninfo{Pure Orthogonal}{Pure Orthogonal Layout}{R. Tamassia}{G.~W.~Klau}{April 1999}{1.1}

  \headerline{Pre- and Postcondition}
  
  \begin{tabular}{lcl}
  precondition & $=$ & \{\,$\emph{planar, simple, no self loops, four--graph}$\,\} \\
  postcondition(|PRE|) & $=$ & $\{\,$\emph{no crossings, orthogonal}$\,\}$
  \end{tabular}\\

  \headerline{Exchangeable Modules}
  Instances of type |\Mname| provide the following module options:
  \begin{options}
  \item {\bf compactor}\\
  a |CompactionModule| option used for the compaction phase.
  
  \begin{tabular}{ll}
  guaranteed precondition: & \emph{no crossings, orthogonal}\\
  required postcondition: &  \emph{no crossings, orthogonal}\\
  initial module: & |FlowCompaction|
  \end{tabular}
  \end{options}

  \headerline{Optional Parameters}
  Instances of |\Mname| provide the following optional parameters:
  \begin{options}
  \item |int| {\bf output\_level} $= 0$\\
  The |output_level| determines the amount of output to |stdout| that
  is produced by the algorithm. The default value, zero, causes |\Mname| to
  run in silent mode, values $\geq 10$ cause the maximal amount of output.
  
  \item |bool| {\bf verify} = |true|\\
  If this flag is set, the checking routines are activated. Especially the 
  orthogonal representation is verified, 
  certificating the correctness of the computation.
  \end{options}
  \horizline
  }*/
               
public: 

    /*{\Mcreation P 5}*/
                
    PureOrthogonalLayout(); 
    /*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/
               
    virtual ~PureOrthogonalLayout() { }
               
    /*{\Moperations 1.5 2.5}*/

    leda::string name() const { return leda::string("PureOrthogonal"); }
               
    leda::string long_name() const { 
      return leda::string("Pure Orthogonal Drawing Algorithm (Tamassia)"); 
    }
  
    leda::string author() const { return leda::string("R.~Tamassia"); }
  
    virtual leda::string impl_author() const { return leda::string("G.~W.~Klau"); }
  
    virtual leda::string impl_version() const { return leda::string("1.1"); }
  
    virtual leda::string impl_date() const { return leda::string("April 1999"); }

    AGDModule* clone() const;

    /*{\Moptions warnings=no}*/
    
    /*{\Mtext       
      \headerline{Standard Interface (Inherited Methods)}
      The detailed description of these methods can be found in the
      manual entries of the base classes (|LayoutModule|, |GridLayoutModule|).
      \stdstart{1.5cm}}*/
    
    /* inherited
       bool check(const graph& G, string& message);
    */
    /*{\Mop}*/
    
    /* inherited
       void call (const graph& G, LayoutInterface& A);
    */
    /*{\Mop}*/
    
    /* inherited
       virtual void call (const graph& G, face f, LayoutInterface& A);
    */
    /*{\Mop}*/
    
    /* inherited
       virtual void call (PlaneGraphCopy& PG, face f, LayoutInterface& A);
    */
    /*{\Mop}*/
    
    /* inherited
       void call_grid (const graph& G, GridLayout &gl);
    */
    /*{\Mop}*/
    
    /* inherited
       void call_grid (const graph& G, face f, GridLayout &gl);
    */
    /*{\Mop}*/
    
    /* inherited
       void call_grid (PlaneGraphCopy& PG, face f, GridLayout &gl);
    */
    /*{\Mop}*/
    
    /* inherited
       IRect bounding_box ();
    */
    /*{\Mop}*/
    
    /* inherited
       void set_mapper (const GridCoordinateMapper& gcm);
    */
    /*{\Mop}*/
    
    /*{\Mtext \stdend}*/
        
protected:

    void preprocessing(OrthogonalRepresentation &H);
    void postprocessing(OrthogonalRepresentation &H, GridLayout &L);
               
};

/*{\Mimplementation The computation of a pure orthogonal grid embedding by
  algorithm [[PureOrthogonalLayout]] takes time $O((n \log n)^2)$ where $n$ is the
  number of nodes of the graph to be drawn.}*/

/*{\Mexample\label{ortho_example}
The following program shows how to use the pure orthogonal layout algorithm in
combination with a [[TopologyModule]] which preserves the eventual planar topology of the input. The user can draw or load a graph in a
[[GraphWin]], pressing the ``done'' button initiates the layout process. If
the input drawing divides the plane into faces, the same faces will emerge in
the output drawing (i.e., the topological representation and the external face
will be preserved during the computations).

{\small \tt
\begin{verbatim}
#include <AGD/TopologyModule.h>
#include <AGD/PureOrthogonalLayout.h>
#include <AGD/GraphWinInterface.h>
#include <LEDA/graphics/graphwin.h>

main() {
  
GraphWin W; W.display();

while (W.edit()) {
  
GraphWinInterface I(W);
graph& G = W.get_graph(); face f_0; string msg;
TopologyModule T; PureOrthogonalLayout P;
T.call(G, f_0, I);
if (P.check(G, msg)) P.call(G, f_0, I);
else W.acknowledge(msg);
W.fill_window();

}

}
\end{verbatim}
}
}*/

#endif

