/*******************************************************************************
+
+  AGD 1.2
+
+
+  ReducedDigraph.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         ReducedDigraph.h
//
//
// author: Carsten Gutwenger
// date:   decembre 1998
//-------------------------------------------------------------------


#ifndef _AGD_REDUCED_DIGRAPH_H
#define _AGD_REDUCED_DIGRAPH_H


#include <AGD/SubgraphModule.h>
#include <LEDA/core/b_stack.h>



/*{\Manpage {ReducedDigraph} {} {Reduced Digraphs} {S} }*/

/*{\Mbaseclasses
\begin{hierarchy}{10}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{1}{1.5}{6}{SubgraphModule}
\arhook{7}{2}{8}
\hdclass{2.5}{0}{6}{ReducedDigraph}
\end{hierarchy}
}*/


class __exportAGD ReducedDigraph : public SubgraphModule {

/*{\Mdefinition
The module |\Mname| computes a reduced subgraph $G'=(V,E)$ of an acyclic
directed input graph $G=(V,E\cup F)$, i.e., $F$ is the set of all transitive
edges in $G$. We say that $(v,w)\in G$ is a transitive edge iff there is
a directed path $v=v_1,\ldots,v_k=w$ in $G$ with $k\geq 3$. Thus, self-loops
are not considered to be transitive edges. The output graph $G'$ may contain
self-loops or multiple edges.

The algorithm performs a depth first search for each node $v$ in order to determine
all nodes $V(v)$ which can be reached from $v$ by a directed path of at least three
nodes. All edges $(v,w)$ with $w\in V(v)$ are transitive edges.

\horizline
\commoninfo{Reduced Digraph}{Reduced Digraph}{}{C. Gutwenger}{Decembre 1998}{1.0}

\begin{preposttab}
precondition & $\{ \emph{directed, acyclic} \} $ \\
postcondition(|PRE|) & $\{ \emph{reduced acyclic} \} $
\end{preposttab}

\horizline
}*/

public:

    /*{\Mcreation S 4}*/
    
    ReducedDigraph();
    /*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/
    
    
    ~ReducedDigraph() { }
    
    /*{\Moperations 1.2 3.5}*/

    leda::string name () const {
	return "Reduced Digraph";
    }

    leda::string impl_author () const {
	return "Carsten Gutwenger";
    }

	leda::string impl_version () const {
		return "1.0";
	}

	leda::string impl_date () const {
		return "Decembre 1998";
	}


    AGDModule *clone() const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|SubgraphModule|).
	}*/

	/* inherited
	bool check (const graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/* inherited
	bool call (const graph& G, list<edge>& L);
	*/
	/*{\Mop}*/

	/* inherited
	bool call (graph& G);
	*/
	/*{\Mop}*/

	/* inherited
	bool call (GraphCopy& G, list<edge>& L);
	*/
	/*{\Mop}*/

	/* inherited
	int num_del_edges() const;
	*/
	/*{\Mop}*/

	/*{\Mtext \stdend}*/
	
protected:
	void do_call (const leda::graph& G, leda::list<leda::edge>& transitives);

private:
	void dfs(leda::node v);

	leda::node_array<int> mark;
	leda::b_stack<leda::node> *visited;
};

/*{\Mimplementation
The algorithm has running time $\BigO(n(n+m))$, where $n$ is the number
of nodes and $m$ is the number of edges of the input graph $G$.
}*/

#endif
