/*******************************************************************************
+
+  AGD 1.2
+
+
+  Rule.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                           Rule.h
//
//
//    author: Carsten Gutwenger
//    date:   january 1999
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_RULE_H
#define _AGD_RULE_H


#include <AGD/AgdKey.h>
#include <LEDA/core/list.h>
#include <LEDA/internal/std/iostream.h>


/*{\Manpage {Rule} {} {Rules} {r} }*/

class __exportAGD Rule {

/*{\Mdefinition
An instance |\Mvar| of the data type |\Mname| is a rule
$p_1 \wedge \ldots \wedge p_k \rightarrow q$,
where $p_1,\ldots p_k,q$ are properties (of type |AgdKey|).
The interpretation of such a rule is, that property $q$
holds if all properties $p_1,\ldots,p_k$ are satisfied,
e.g., $|forest|\wedge |single source|\rightarrow|tree|$ means
that a graph which represents a forest with a single source
is a tree.

If a rule is stored in a rule administrator |admin|, it is
assigned a unique index, which, e.g., can be used for realizing
arrays indexed by rules in |admin|.

We use the following abbreviations: $p\rightarrow q_1\wedge\ldots\wedge q_k$
stands for the $k$ rules $p\rightarrow q_1,\ldots,p\rightarrow q_k$, and
$q\leftrightarrow p_1\wedge\ldots\wedge p_k$ stands for the $k+1$ rules
$p_1\wedge\ldots\wedge p_k\rightarrow q$ and $q\rightarrow p_1,\ldots,
q\rightarrow p_k$.
}*/

	AgdKey _q;
	leda::list<AgdKey> _left;
	int _id;

public:

	/*{\Mcreation}*/

	Rule() { }
	/*{\Mcreate creates an instances |\Mvar| of type |\Mtype|
	initialized to $\rightarrow |nil|$.}*/

	Rule(AgdKey q) : _q(q) { }
	/*{\Mcreate creates an instances |\Mvar| of type |\Mtype|
	initialized to $\rightarrow q$.}*/

	Rule(AgdKey p1, AgdKey q) : _q(q) {
		_left.append(p1);
	}
	/*{\Mcreate creates an instances |\Mvar| of type |\Mtype|
	initialized to $p_1\rightarrow q$.}*/

	Rule(AgdKey p1, AgdKey p2, AgdKey q) : _q(q) {
		_left.append(p1);
		_left.append(p2);
	}
	/*{\Mcreate creates an instances |\Mvar| of type |\Mtype|
	initialized to $p_1\wedge p_2\rightarrow q$.}*/

	Rule(AgdKey p1, AgdKey p2, AgdKey p3, AgdKey q) : _q(q) {
		_left.append(p1);
		_left.append(p2);
		_left.append(p3);
	}
	/*{\Mcreate creates an instances |\Mvar| of type |\Mtype|
	initialized to $p_1\wedge p_2\wedge p_3\rightarrow q$.}*/

	Rule(const Rule &r) : _q(r._q), _left(r._left) { }
	/*{\Mcreate creates an instances |\Mvar| of type |\Mtype|
	identical to $r$.}*/

	~Rule() { }


	/*{\Moperations}*/

	const leda::list<AgdKey> &left() const {
		return _left;
	}
	/*{\Mop returns the left side of |\Mvar|, i.e.,
	$p_1\wedge\ldots\wedge p_k$.}*/

	AgdKey right() const { return _q; }
	/*{\Mop returns the right side of |\Mvar|, i.e., $q$.}*/

	int length() const {
		return _left.length();
	}
	/*{\Mop returns the number $k$ of properties on the left side of |\Mvar|.}*/

	int index() const { return _id; }
	/*{\Mop returns the index of |\Mvar|.}*/

	Rule& And(AgdKey p) {
		_left.append(p);
		return *this;
	}
	/*{\Mop extends |\Mvar| to $p_1\wedge\ldots\wedge
	p_k\wedge p\rightarrow q$ and returns |\Mvar|.}*/

	Rule &implies(AgdKey q) {
		_q = q;
		return *this;
	}
	/*{\Mop sets the right side of |\Mvar| to $q$ and returns |\Mvar|.}*/

	Rule &index(int id) {
		_id = id;
		return *this;
	}
	/*{\Mop sets the index of |\Mvar| to |id| and returns |\Mvar|.}*/	
};


inline istream& operator>> (istream& is, Rule& )
{
	return is;
}

ostream& operator<< (ostream& os, const Rule& r);
/*{\Mbinopfunc writes rule |r| to output stream |os|.}*/


#endif
