/*******************************************************************************
+
+  AGD 1.2
+
+
+  ShapeDescription.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



//#line 846 "Lweb/OrthoRep.lw"
#ifndef _AGD_SHAPE_DESCRIPTION_H
#define _AGD_SHAPE_DESCRIPTION_H

#include <LEDA/graph/graph.h>
#include <LEDA/graph/node_map2.h>

#include <AGD/OrthogonalRepresentation.h>
#include <AGD/GridLayout.h>
#include <LEDA/core/tuple.h>
#include <LEDA/graph/face_array.h>
#include <LEDA/core/set.h>

const bool Green = true;
const bool Red = false;

/*{\Moptions
outfile=ShapeDescription.man
}*/

/*{\Manpage {ShapeDescription} {} {Constraint Graph Shape Description} {S} }*/

class __exportAGD ShapeDescription {

/*{\Mdefinition
The class |\Mname| realizes a pair of \ldots }*/

public:

  /*{\Mcreation}*/

  ShapeDescription(const OrthogonalRepresentation& H);
  /*{\Mcreate creates an instance |\Mvar| of type |\Mname| based on the 
     information given in the orthogonal representation |H|. The constraint graphs will be based on {\bf shape properties}. }*/

  //ShapeDescription(const GridLayout& L);
  /*{\Mcreate creates an instance |\Mvar| of type |\Mname| based on the 
     information given in the grid layout |L|. The constraints graphs will be based on {\bf visibility properties}.  }*/

  /*{\Moperations}*/

  int number_of_segments() const { return D_x_y_.number_of_nodes(); }
  /*{\Mop returns the number of horizontal segments. }*/

  int number_of_arcs() const { return D_x_y_.number_of_edges(); }
  /*{\Mop returns the number of arcs in |\Mname|. }*/

  const leda::graph &get_graphs() const { return D_x_y_; }
  /*{\Mop returns a reference to the constraint graphs. They are coded as two components in the returned graph. }*/

  bool horizontal(leda::node s) const { return D_x_y_[s]; }
  /*{\Mop returns true iff segment |s| is horizontal. }*/

  bool vertical(leda::node s) const { return !D_x_y_[s]; }
  /*{\Mop returns true iff segment |s| is vertical. }*/

  leda::node left(leda::node s) const { return L_[s]; }
  /*{\Mop returns the left limit of |s|. }*/

  leda::node right(leda::node s) const { return R_[s]; }
  /*{\Mop returns the right limit of |s|. }*/

  leda::node bottom(leda::node s) const { return B_[s]; }
  /*{\Mop returns the bottom limit of |s|. }*/

  leda::node top(leda::node s) const { return T_[s]; }
  /*{\Mop returns the top limit of |s|. }*/
  
  leda::node ver(leda::node v) const { return red_[v]; }
  /*{\Mop returns $\mbox{ver}(v)$, the vertical segment |v| lies in. }*/

  leda::node hor(leda::node v) const { return green_[v]; }
  /*{\Mop returns $\mbox{hor}(v)$, the horizontal segment |v| lies in. }*/

  // bool touch(leda::node s_i, leda::node s_j) const;

//   bool adjacent(leda::node s_i, leda::node s_j) const;
//   bool path(leda::node s_i, leda::node s_j) const { return link_(s_i, s_j); }
//   //bool path(int i, int j) const { return link_(i, j); }
//   const leda::graph &hor() const { return D_y_; }
//   const leda::graph &ver() const { return D_x_; }

  bool path(leda::node s_i, leda::node s_j) const { return link_(s_i, s_j); }
  /*{\Mop returns true iff there is a path from |s_i| to |s_j| in one of
    the constraint graphs. }*/

  bool adjacent(leda::node s_i, leda::node s_j) const {
    return !members_[s_i].intersect(members_[s_j]).empty();
  }

  bool distant(leda::node s_i, leda::node s_j) const {
    return faces_[s_i].intersect(faces_[s_j]).empty();
  }

  const leda::set<leda::node> &members(leda::node s) const { return members_[s]; }
  /*{\Mop returns the set of vertices belonging to segment |s|. }*/

  const leda::set<leda::face> &adj_faces(leda::node s) const {
    return faces_[s];
  }
  /*{\Mop returns the set of faces adjacent to vertices belonging to |s|. }*/

  leda::list<leda::list<node_pair> > extend();
  /*{\Mop completes the shape decription as far as possible and returns
    a list with constraints. If the list is empty, compaction is easy.}*/

  void reduce();

  void print() const;
  /*{\Mop sends information about the data structure to stdout. }*/

  void only_opp(bool o) { only_opp_ = o; }

protected:

  void create_nodes(const leda::edge_array<bool>& hor,
                    const leda::edge_array<bool>& ver);

  void initialize_limits();

  void make_links();

private:

  const leda::graph *G_; // pointer to the graph structure 
  // contained in the orthogonal representation 
 
  GRAPH<bool, bool> D_x_y_;
  leda::node_array<leda::node> L_, R_, B_, T_;
  leda::node_array<leda::set<leda::node> > members_;
  leda::node_array<leda::set<leda::face> > faces_;
  node_map2<bool> link_;
  leda::node_array<leda::node> green_, red_;
  face_array<leda::set<leda::node> > adj_segments_;


  bool only_opp_; //temporary. switch; decides whether all pairs of segments or only opposite pairs of segments are taken into consideration of [[extend]].

};

#endif


