/*******************************************************************************
+
+  AGD 1.2
+
+
+  SiftingHeuristic.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                       SiftingHeuristic.h
//
//
//    author: Carsten Gutwenger
//    date:   june 1999
//-------------------------------------------------------------------


#ifndef _AGD_SIFTING_HEURISTIC_H
#define _AGD_SIFTING_HEURISTIC_H


#include <AGD/TwoLayerCrossMin.h>


class CrossingsMatrix;


/*{\Manpage {SiftingHeuristic} {} {Sifting Heuristic} {M} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{0.9}{1.5}{6.1}{TwoLayerCrossMin}
\arhook{7}{2}{8}
\hdclass{2}{0}{6.5}{SiftingHeuristic}
\end{hierarchy}
}*/

class __exportAGD SiftingHeuristic : public TwoLayerCrossMin {

/*{\Mdefinition
The sifting heuristic \cite{Rud93,MSM99} works as follows: First, an order
of the vertices on the changeable layer is determined.
Then, the vertices are sifted in that order. Sifting a
vertex $v$ means to put it on a position which minimizes
the number of crossings, whereby the ordering of all other
vertices remains fixed.

\horizline
\commoninfo{Sifting}{Sifting (Matuszewski et al.)}{C. Matuszewski, R. Sch\"onfeld, P. Molitor}{C. Gutwenger}
{June 1999}{1.0}

\begin{preposttab}
precondition & $\emptyset$ \\
postcondition(|PRE|) & $\emptyset$
\end{preposttab}

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |SiftingHeuristic::Strategy| {\bf strategy} = |desc_degree|\\
determines the order in which vertices are sifted. Possible settings are
|left_to_right|, |desc_degree| (with descending number of neighbors
on the fixed layer) and |random| (in random order).
\end{options}
\horizline
}*/

public:

	/*{\Mtypes}*/ 

	enum Strategy { left_to_right, desc_degree, random };
	/*{\Menum the strategy for sifting vertices.}*/


	/*{\Mcreation M 5}*/

	SiftingHeuristic() : _strategy(desc_degree) { }
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname|.}*/

	~SiftingHeuristic() { }


	/*{\Moperations}*/

	leda::string name () const {
		return "Sifting";
	}

	leda::string long_name () const {
		return "Sifting (Matuszewski et al.)";
	}

	leda::string author () const {
		return "C. Matuszewski, R. Schnfeld, P. Molitor";
	}

	leda::string impl_author () const {
		return "C. Gutwenger";
	}

	leda::string impl_version () const {
		return "1.0";
	}

	leda::string impl_date () const {
		return "June 1999";
	}


	AGDModule *clone() const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|TwoLayerCrossMin|).
	}*/

	/* inherited
	bool check (const leda::graph& G, const Layer& L, AgdKey &p) const;
	*/
	/*{\Mop}*/

	void init (const leda::graph& G, const Hierarchy& H);
	/*{\Mop}*/

	void call (const leda::graph& G, Layer& L);
	/*{\Mop}*/

	void cleanup ();
	/*{\Mop}*/

	
	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	Strategy strategy() const {
		return _strategy;
	}
	/*{\Mop}*/

	void strategy (Strategy s) {
		_strategy = s;
	}
	/*{\Mop}*/

	/*{\Mtext \stdend}*/

private:
	Strategy _strategy;

	CrossingsMatrix *cp;
};


/*{\Mimplementation
Let $n_L$ denote the number of nodes on the changeable layer
and $m_L$ the number of edges between the fixed and the changeable layer.
The algorithm computes in a preprocessing step a so-called
crossing matrix, which takes time $\BigO(n_L m_L)$.
The actual crossing reduction algorithm takes time
$\BigO(n_L^2)$. Both steps are implemented in the
\texttt{call()}-method. The \texttt{init()}-method takes time
$\BigO(n)$, where $n$ is the number of nodes in the input graph.
}*/


#endif
