/*******************************************************************************
+
+  AGD 1.2
+
+
+  SubgraphModule.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                        SubgraphModule.h
//
//
//    author: Carsten Gutwenger
//    date:   august 1997
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_SUBGRAPH_MODULE_H
#define _AGD_SUBGRAPH_MODULE_H


#include <AGD/AGDModule.h>

class GraphCopy;


/*{\Manpage {SubgraphModule} {} {Subgraph Modules} {S} }*/

/*{\Mbaseclasses

\begin{hierarchy}{6.5}{2.5}
\hdclass{0}{1.5}{5.5}{AGDModule}
\arhook{5.5}{2}{6.5}
\hdclass{1}{0}{6}{SubgraphModule}
\end{hierarchy}
}*/

class __exportAGD SubgraphModule : public AGDModule {

/*{\Mdefinition
The class |\Mname| is the base class for subgraph modules. A
subgraph module computes a subgraph $G'=(V,E\setminus E')$ of an input
graph $G=(V,E)$, where $E'\subseteq E$. Usually, $G'$ satisfies
a certain condition, e.g, planar or acyclic. This condition is
expressed by the postcondition of the module.

\headerline{Input and Output Parameters}

\begin{tabular}{lcc}
input parameter: & graph & |in_graph=1| \\
output parameter: & subgraph & |out_subgraph=1|
\end{tabular}

The integer constants |in_graph| and |out_subgraph| denote the
positions in the input and output parameter lists. The values can be used to
reference the desired parameter when setting the pre- and postcondition.  

\headerline{Implementing Subgraph Modules}

An implementation of a subgraph module must override the protected
method do\_call($G$,$L$), which gets as input a const reference to
graph $G$. It determines the set of edges $E'$,
and returns it as a list in $L$.
}*/

public:

	// the input and output parameters of a subgraph module
	static const int in_graph;
	static const int out_subgraph;

	/*{\Minitialization S 4.5}*/

	SubgraphModule () : AGDModule(1,1) { }

	/*{\Mcreate initializes a subgraph module.}*/

	virtual ~SubgraphModule () { }


	/*{\Moperations 1.5 3.5}*/

	/*{\Mtext
	\headerline{Standard Interface} }*/

	bool check (const leda::graph& G, AgdKey &p) const;
	/*{\Mop returns |true| if $G$ satisfies the precondition of |\Mvar|.
	Otherwise, |false| is returned and |p| contains a property that
	is not satisfied.}*/

	void call (const leda::graph& G, leda::list<leda::edge>& L);
	/*{\Mop calls the subgraph module for graph $G$. Returns in $L$ the
	list of edges in $E'$.}*/

	void call (leda::graph& G);
	/*{\Mop calls the subgraph module for graph $G$, and deletes in $G$ all
	edges in $E'$.}*/

	void call (GraphCopy& GC, leda::list<leda::edge>& L);
	/*{\Mop calls the subgraph module for graph copy |GC|. Deletes in |GC| all
	edges in $E'=\{e_1,\ldots,e_k\}$, and returns in $L$ the list of
	corresponding edges $|original|(e_1),\ldots,|original|(e_k)$ in
	|original(GC)|.}*/

	int num_del_edges() const {
		return _num_del_edges;
	}
	/*{\Mop returns the number of edges in $E'$ after a call
	of the algorithm.}*/

protected:

	/*{\Mtext
	\bigskip
	\headerline{Protected Overridables to Implement Functionality} }*/

	virtual void do_call (const leda::graph& G, leda::list<leda::edge>& L)
	/*{\Moppure implements the computation of the subgraph for graph $G$.
	Returns in $L$ the list of edges in $E'$, which are the
	edges that must be deleted in $G$ in order to get the subgraph $G'$.}*/	
	= 0;

private:
	int _num_del_edges;
};



#endif
