/*******************************************************************************
+
+  AGD 1.2
+
+
+  SubgraphPlanarizer.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                        SubgraphPlanarizer.h
//
//
//    author: Carsten Gutwenger
//    date:   august 1997
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_SUBGRAPH_PLANARIZER_H
#define _AGD_SUBGRAPH_PLANARIZER_H


#include <AGD/PlanarizerModule.h>
#include <AGD/SubgraphModule.h>
#include <AGD/EdgeInsertionModule.h>
#include <AGD/AGDModuleOption.h>


/*{\Manpage {SubgraphPlanarizer} {} {Planarization with Planar Subgraph} {P} }*/

/*{\Mbaseclasses

\begin{hierarchy}{7}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{1}{1.5}{6}{PlanarizerModule}
\arhook{7}{2}{8}
\hdclass{2.3}{0}{6.2}{SubgraphPlanarizer}
\end{hierarchy}
}*/

class __exportAGD SubgraphPlanarizer : public PlanarizerModule {

/*{\Mdefinition
The class |\Mname| represents a general framework for planarization using 
planar subgraphs.
It first transforms a general graph into a planar graph (using a subgraph
planarization module) by deleting a set of
edges and then reinserting the edges so that the number of crossings is
small. Finally, the crossings are substituted by artificial crossing vertices 
so that the resulting graph is planar.
The planarization algorithms and the reinsertion algorithms 
are exchangeable modules allowing
to use |\Mname| with different methods for planarization and edge-reinsertion.


\horizline
\commoninfo{Subgraph Planarizer}{Planarization with Planar Subgraph}{}{C. Gutwenger}
{August 1997}{}

\begin{preposttab}
precondition & $\{\emph{ simple, no self loops }\}$ \\
postcondition(|PRE|) & $\{\emph{ planar, simple, no self loops }\}$
\end{preposttab}

\headerline{Exchangeable Modules}
Instances of type |\Mname| provide the following module options:
\begin{options}
\item {\bf subgraph}\\
a |SubgraphModule| option used to compute a planar subgraph of
the input graph.

\begin{tabular}{ll}
guaranteed precondition: & $\{\emph{ simple, no self loops, no reversals }\}$\\
required postcondition: & $\{\emph{ planar, simple, no self loops, no reversals }\}$\\
initial module: & |PlanarSubgraph|
\end{tabular}

\item {\bf inserter}\\
an |EdgeInsertionModule| option used to reinsert the deleted edges.

\begin{tabular}{ll}
guaranteed precondition: & $\emptyset$ \\
required postcondition: & $\emptyset$ \\
initial module: & |ShortestPathInserter|
\end{tabular}
\end{options}\horizline
}*/

	AGD_DECLARE_MODULE(SubgraphModule,subgraph)
	AGD_DECLARE_MODULE(EdgeInsertionModule,inserter)

	int m_nEdges;
	float m_timeSubgraph, m_timeInsertion;

public:

	/*{\Mcreation P 4.5}*/

	SubgraphPlanarizer ();

	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

	virtual ~SubgraphPlanarizer () { }


	/*{\Moperations 1.5 3.5}*/

	leda::string name () const {
		return leda::string ("Subgraph Planarizer");
	}

	leda::string long_name () const {
		return leda::string ("Planarization with Planar Subgraph");
	}

	leda::string impl_author () const {
		return "C. Gutwenger";
	}

	leda::string impl_date () const {
		return "August 1997";
	}
	virtual AGDModule *clone () const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|PlanarizerModule|).
	}*/

	/* inherited
	bool check (const GraphCopy& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/* inherited
	bool call (GraphCopy& G, edge_array<list<edge> >& chain);
	*/
	/*{\Mop}*/

	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	/*
	bool set_subgraph (const SubgraphModule& M);
	*/
	/*{\Mop}*/

	/*
	bool set_inserter (const EdgeInsertionModule& M);
	*/
	/*{\Mop}*/

	/*
	int insertedEdges() const {
		return m_nEdges;
	}
	*/
	/*{\Mop}*/

	/*
	float timeSubgraph() const {
		return m_timeSubgraph;
	}
	*/
	/*{\Mop}*/

	/*
	float timeInsertion() const {
		return m_timeInsertion;
	}
	*/
	/*{\Mop}*/

	/*{\Mtext \stdend}*/


protected:

	virtual void do_call (const leda::graph& G, PlaneGraphCopy& PG);

};



#endif
