/*******************************************************************************
+
+  AGD 1.2
+
+
+  SugiyamaLayout.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                        SugiyamaLayout.h
//
//
//    author: Carsten Gutwenger
//    date:   decembre 1998
//-------------------------------------------------------------------


#ifndef _AGD_SUGIYAMA_LAYOUT1_H
#define _AGD_SUGIYAMA_LAYOUT1_H


#include <AGD/LayoutModuleImpl.h>
#include <AGD/GraphCopyConcept.h>
#include <AGD/AGDModuleOption.h>
#include <AGD/RankAssignment.h>
#include <AGD/HierarchyLayoutModule.h>
#include <AGD/TwoLayerCrossMin.h>


/*{\Manpage {SugiyamaLayout} {} {Sugiyama's Algorithm} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{6}
\hdclass{0}{5}{5}{AGDModule}
\arhook{5}{5.5}{6}
\hdclass{1.5}{3.5}{5}{LayoutModule}
\arhook{6.5}{4}{7.5}
\hdclass{2}{2}{6}{SugiyamaLayout}
\end{hierarchy}
\hfill\agdeps{sugiyama.epsi}{3.7cm}
}*/

class __exportAGD SugiyamaLayout : public LayoutModuleImpl<GraphCopyConcept> {

/*{\Mdefinition
The class |\Mname| represents a general framework for drawing graphs with
the hierarchical drawing method suggested by Sugiyama, Tagawa, and
Toda~\cite{stt-mvuhs-81}. This method works in three phases:
\begin{enumerate}
\item We partition the nodes into levels using a |RankAssignment| module.
Then, we build a proper hierarchy, that is, we reverse and split edges until
each edge is directed downwards and connects two nodes on neighbored layers.
The newly introduced nodes are called {\it dummy nodes} and represent
edge bends in the final drawing.
\item We reorder the nodes on each layer in order to reduce the number
of edge crossings. This is achieved by iteratively calling a |TwoLayerCrossMin|
module. The layers are alternately traversed top--down and bottom-up,
until the total number of crossings does not decrease anymore.
\item We call a |HierarchyLayoutModule| to compute the final layout of the
hierarchy. This module assigns coordinates to nodes and bend points to edges.
\end{enumerate}
The algorithm is modified such that the above process
can be repeated several times. Each repetition (except for the first)
starts with a random order of the nodes on each layer. Since the
total process of $k$-layer crossing minimization used here is a
heuristic, each repetition may stop with a different total number
of crossings. The algorithm chooses the result with the fewest
number of crossings. If option |transpose| is set to |true|, a fine tuning
step is performed after each traversal, which tries to reduce obvious
crossings. In particular, the resulting ordering of the vertices on each
layer is locally optimal with respect to transposition of adjacent vertices
(see also~\cite{GKNV93}).

\horizline
\commoninfo{Sugiyama}{Sugiyama}{K. Sugiyama, S. Tagawa, M. Toda}
{C. Gutwenger, M. Kr\"uger}{March 1998}{2.0}

\begin{preposttab}
precondition & $\{\, \emph{no self loops, directed} \,\}$ \\
postcondition(|PRE|) & $\emptyset$
\end{preposttab}

\headerline{Exchangeable Modules}
Instances of type |\Mname| provide the following module options:
\begin{options}
\item {\bf ranking}\\
a |RankAssignment| option used to compute the ranking.

\begin{tabular}{ll}
guaranteed precondition: & $\{\emph{ no self loops, directed }\}$\\
required postcondition: & $\emptyset$\\
initial module: & |LongestPathRanking|
\end{tabular}

\item {\bf cross\_min}\\
a |TwoLayerCrossMin| option used in the crossing reduction phase.

\begin{tabular}{ll}
guaranteed precondition: & $\{\emph{ no self loops, directed }\}, \emptyset$\\
required postcondition: & $\emptyset$\\
initial module: & |BarycenterHeuristic|
\end{tabular}

\item {\bf comp\_coord}\\
a |HierarchyLayoutModule| option used to compute the final layout.

\begin{tabular}{ll}
guaranteed precondition: & $\emptyset$\\
required postcondition: & $\emptyset$\\
initial module: & |FastHierarchyLayout|
\end{tabular}
\end{options}
\horizline

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |bool| {\bf init\_order} = |true|\\
if set to |true|, an initial order of the nodes on each layer
is determined by a depth-first-search traversal. If the
graph represents a forest the initial order has already no crossings.
\item |int| {\bf runs} $= 3$\\
determines, how many times the crossing minimization is repeated.
Each repetition (except for the first) starts with randomly
permuted nodes on each layer. Deterministic behaviour can be achieved
by setting |runs| to $1$.
\item |int| {\bf fails} $= 4$\\
determines, how many times the total number of crossings after a
complete top down or bottom up traversal may not decrease before
one repetition is stopped.
\item |bool| {\bf transpose} = |true|\\
if set to |true|, an additional fine tuning step is performed after
each traversal, which tries to reduce the total number of crossings
by switching adjacent vertices on the same layer.
\end{options}
\horizline
}*/

	// declare module options for rank assignment
	AGD_DECLARE_MODULE(RankAssignment,ranking)

	// two layer crossing minimization
	AGD_DECLARE_MODULE(TwoLayerCrossMin,cross_min)

	// and computation of coordinates
	AGD_DECLARE_MODULE(HierarchyLayoutModule,comp_coord)

public:

	/*{\Mcreation}*/

	SugiyamaLayout();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

	~SugiyamaLayout() { }


	/*{\Moperations 1.5 4.5}*/

	// common informations
	leda::string name() const {
		return "Sugiyama";
	}
	leda::string author() const {
		return "K. Sugiyama, S. Tagawa, M. Toda";
	}
	leda::string impl_author() const {
		return "C. Gutwenger, M. Kr\201ger";
	}
	leda::string impl_version() const {
		return "2.0";
	}
	leda::string impl_date() const {
		return "March 1998";
	}


	void ext_call (const leda::graph& G, leda::node_array<int>& rank,
		LayoutInterface& A);
	/*{\Mop calls |\Mvar| for graph $G$ with layout interface $A$.
	Uses the level assignment defined by |rank|.}*/

	void reduce_crossings (GraphCopy& GC, Hierarchy& H);
	/*{\Mop calls only the crossing reduction step for graph copy
	|GC| and hierarchy |H|.}*/


	int number_of_crossings() const {
		return best_nr_crossings;
	}
	/*{\Mop returns the number of crossings in the layout produced
	by a preceding call.}*/

	AGDModule *clone() const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext\medskip\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|LayoutModule|).
	}*/

	/* inherited
	bool check (const graph& G, AgdKey &p) const;
	*/
	/*{\Mop}*/

	/* inherited
	void call (const graph& G, LayoutInterface& A);
	*/
	/*{\Mop}*/
	
	/*{\Mtext       
	\smallskip\headerline{Access to Options}}*/

	/*
	bool set_ranking (const RankAssignment& M);
	*/
	/*{\Mop}*/

	/*
	bool set_cross_min (const TwoLayerCrossMin& M);
	*/
	/*{\Mop}*/

	/*
	bool set_comp_coord (const HierarchyLayoutModule& M);
	*/
	/*{\Mop}*/

	bool init_order() const {
		return _init_order;
	}
	/*{\Mop}*/

	void init_order(bool b) {
		_init_order = b;
	}
	/*{\Mop}*/

	int runs() const {
		return _runs;
	}
	/*{\Mop}*/

	void runs(int n) {
		_runs = n;
	}
	/*{\Mop}*/

	int fails() const {
		return _fails;
	}
	/*{\Mop}*/

	void fails(int n) {
		_fails = n;
	}
	/*{\Mop}*/

	bool transpose() const {
		return _transpose;
	}
	/*{\Mop}*/

	void transpose(bool b) {
		_transpose = b;
	}
	/*{\Mop}*/

	/*{\Mtext \stdend}*/

protected:

	void do_call (GraphCopyConcept &C);

private:
	int  _runs, _fails;
	bool _transpose, _init_order;

	int best_nr_crossings;

	leda::array<bool> changed_layer;

	void initial_order_dfs(leda::node v, leda::node_map<int> &pos,
		const Hierarchy &H, leda::array<int> &counter);
	void initial_order(const leda::graph &G, Hierarchy &H);

	void update_d (int p1, int p2, int &d) {
		if (p1 == p2) return;
		if (p1 < p2) d--; else d++;
	}
	int transpose_test (leda::node v, leda::node w, const Hierarchy &H);
	int transpose_test (leda::node v, leda::node x, leda::node w, const Hierarchy &H);

	bool transpose_layer(int i, Hierarchy &H);
	void reset_changed_layer();
	void do_transpose(Hierarchy &H);
	void do_transpose_rev(Hierarchy &H);

	void minimize_crossings(const leda::graph& G, Hierarchy& H);

	void call_cross_min (const leda::graph& G, Layer &L, TwoLayerCrossMin& minimizer);

	int traverse_top_down(const leda::graph& G, Hierarchy& H,
		TwoLayerCrossMin& minimizer);
	int traverse_bottom_up(const leda::graph& G, Hierarchy& H,
		TwoLayerCrossMin& minimizer);

	void rem_dummy_sinks(GraphCopy &GC);
};

/*{\Mexample
{\bf Example 1:} We use only the crossing reduction phase of the
Sugiyama algorithm with the Split heuristic and our own level assignment.
\small\tt
\begin{verbatim}
#include <LEDA/graphics/graphwin.h>
#include <AGD/SugiyamaLayout.h>
#include <AGD/SplitHeuristic.h>
#include <AGD/DfsRanking.h>
#include <AGD/Hierarchy.h>
#include <AGD/GraphCopy.h>

void reduce_crossings(graph& G)
{
  node_array<int> rank(G);

  // assign a non-negative level to each node in G
  // we simply use a rank assignment module
  DfsRanking().call(G,rank);

  // build the hierarchy
  GraphCopy GC(G);
  Hierarchy H(GC,rank);

  // call crossing minimization
  SugiyamaLayout sugiyama;
  sugiyama.set_cross_min(SplitHeuristic());
  sugiyama.reduce_crossings(GC,H);

  // the positions of the nodes (in GC) on each layer
  // can now be accessed via H
}
\end{verbatim}\normalsize\rm

\bigskip
{\bf Example 2:} We compute a layout in a {\it GraphWin} with
our own level assignment.
\small\tt
\begin{verbatim}
#include <LEDA/graphics/graphwin.h>
#include <AGD/SugiyamaLayout.h>
#include <AGD/DfsRanking.h>

void call_with_ranking(GraphWin& GW)
{
  graph& G = GW.get_graph();
  node_array<int> rank(G);

  // assign a non-negative level to each node in G
  // we simply use a rank assignment module
  DfsRanking().call(G,rank);

  // compute layout in GW
  GraphWinInterface A(GW);
  SugiyamaLayout sugiyama;
  sugiyama.ext_call(G,rank,A);
}
\end{verbatim}\normalsize\rm

}*/

#endif
