/*******************************************************************************
+
+  AGD 1.2
+
+
+  TaggedRule.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         TaggedRule.h
//
//
//    author: Carsten Gutwenger
//    date:   july 1997
//-------------------------------------------------------------------


#if defined(_MSC_VER)
#pragma once
#endif

#ifndef _AGD_TAGGED_RULE_H
#define _AGD_TAGGED_RULE_H

#include <AGD/TRLiteral.h>
#include <LEDA/core/list.h>
#include <LEDA/internal/std/iostream.h>


/*{\Manpage {TaggedRule} {} {Tagged Rules} {r} }*/

class __exportAGD TaggedRule {

/*{\Mdefinition
An instance |\Mvar| of the data type |\Mname| is a rule
$(p_1,i_1) \wedge \ldots \wedge (p_k,i_k) \rightarrow q$,
where $(p_j,i_j)$ $(1\leq j\leq k)$ are tagged rule
literals (of type |TRLiteral|), and $q$ is a key.
Tagged rules are used by modules in order to
define postcondition rules.
}*/

private:
	leda::list<TRLiteral> _left;
	AgdKey _right;

public:

	/*{\Mcreation}*/

	TaggedRule () {
		_right = nil;
	}
	/*{\Mcreate creates an instances |\Mvar| of type |\Mtype|
	initialized to $\rightarrow |nil|$.}*/

	TaggedRule (AgdKey q) {
		_right = q;
	}
	/*{\Mcreate creates an instances |\Mvar| of type |\Mtype|
	initialized to $\rightarrow q$.}*/

	TaggedRule (const leda::list<TRLiteral>& L, AgdKey q) {
		_left = L;
		_right = q;
	}
	/*{\Mcreate creates an instances |\Mvar| of type |\Mtype|
	initialized to $l_1 \wedge\ldots\wedge l_k\rightarrow q$ if
	$L = l_1,\ldots,l_k$.}*/


	/*{\Moperations}*/

	const leda::list<TRLiteral>& left () const {
		return _left;
	}
	/*{\Mop returns the left side of |\Mvar|, i.e.,
	$(p_1,i_1),\ldots,(p_k,i_k)$.}*/

	AgdKey right () const {
		return _right;
	}
	/*{\Mop returns the right side of |\Mvar|, i.e., q.}*/

	int length () const {
		return _left.length();
	}
	/*{\Mop returns the number $k$ of literals on the left side of |\Mvar|.}*/

	TaggedRule& And (AgdKey p, int j = 1) {
		_left.append (TRLiteral(p, j));
		return *this;
	}
	/*{\Mop extends |\Mvar| to $(p_1,i_1)\wedge\ldots\wedge
	(p_k,i_k)\wedge (p,j)\rightarrow q$ and returns |\Mvar|.}*/

	TaggedRule &implies(AgdKey q) {
		_right = q;
		return *this;
	}
	/*{\Mop sets the right side of |\Mvar| to $q$ and returns |\Mvar|.}*/
};


inline istream& operator>> (istream& is, TaggedRule& )
{
	return is;
}

ostream& operator<< (ostream& os, const TaggedRule& r);
/*{\Mbinopfunc writes tagged rule |r| to output stream |os|.}*/


/*{\Mexample
We create a tagged rule $(|planar|,1) \wedge (|triconnected|,1)
\wedge (|simple|,1) \wedge (|no_self_loops|,1) \to |convex|$ and set it
as postcondition rule of a layout algorithm.

\small\tt
\begin{verbatim}
SampleLayout::SampleLayout() : LayoutModule (true)
{
  TaggedRule r(convex);

  r.And(planar).And(triconnected).And(simple).And(no_self_loops);

  add_post_rule(r);
  
    ....
}
\end{verbatim}\normalsize\rm
}*/


#endif
