/*******************************************************************************
+
+  AGD 1.2
+
+
+  TopologyModule.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



////#line 61 "Lweb/shape.lw"
#ifndef _AGD_TOPOLOGY_MODULE_H
#define _AGD_TOPOLOGY_MODULE_H

#include <AGD/AGDModule.h> 
#include <AGD/LayoutInterface.h>
#include <AGD/PlaneGraphCopy.h>

/*{\Manpage {TopologyModule} {} {Topology Modules} {T} }*/

/*{\Mbaseclasses

\begin{hierarchy}{6.5}{2.5}
\hdclass{0}{1.5}{5.5}{AGDModule}
\arhook{5.5}{2}{6.5}
\hdclass{1.5}{0}{5.5}{TopologyModule}
\end{hierarchy}
}*/

class __exportAGD TopologyModule : public AGDModule {

  /*{\Mdefinition
    The purpose of class |\Mname| is to relate the geometrical and topological
    structure of a graph.  An instance of type |\Mname| transports the information
    from the actual drawing in the combinatorial structure of the underlying graph
    $G$.  The output is a {\em planar representation} for $G$, i.e., a
    combinatorial embedding defining the faces and an external face $f_0$. 

    \headerline{Input and Output Parameters}

    \begin{tabular}{lrl}
    input parameter: & graph                  & $|in_graph|=1$\\
    & layout                 & $|in_layout|=2$ \\
    output parameter: & planar representation & $|out_plan_rep|=1$
    \end{tabular}

    The integer constants |in_graph|, |in_layout|, and |out_plan_rep| denote the
    positions in the input and output parameter lists. The values can be used to
    reference the desired parameter when setting the pre-- and postcondition.  
    }*/

public:

  static const int in_graph;
  static const int in_layout;
  static const int out_plan_rep;
  
  /*{\Minitialization T 5}*/
  
  TopologyModule();
  
  /*{\Mcreate initializes a topology module.}*/
  
  virtual ~TopologyModule() { }
  
  /*{\Moperations 1.5 2.5}*/

  leda::string name() const { return leda::string("Topology Module"); }
  
  leda::string long_name() const { 
    return leda::string("Topology Module"); 
  }
  
  leda::string author() const { return leda::string("Gunnar W. Klau"); }
  
  virtual leda::string impl_author() const { 
    return leda::string("Gunnar W. Klau"); 
  }
  
  virtual leda::string impl_version() const { return leda::string("1.0"); }
  
  virtual leda::string impl_date() const { return leda::string("June 1998"); }
  
  /*{\Mtext
    \headerline{Standard Interface} }*/

  bool check(const leda::graph& G, AgdKey &p) const;
  /*{\Mop returns |true| if $G$ satisfies the precondition of |\Mvar|.
    Otherwise, |false| is returned and |p| contains a property
    that is not satisfied.}*/ 

  PlaneGraphCopy &call(const leda::graph &G, const LayoutInterface &L, leda::face &f_0,
	  bool preserveMap = true);
  /*{\Mop calls the topology module for graph $G$ and layout interface
  $L$. The method returns a copy of $G$ in which the edges are ordered
  according to the angles in the drawing and thus characterize a combinatorial
  embedding. If the imposed combinatorial embedding is proper, the parameter
  $f_0$ contains the external face, otherwise |nil|. The parameter |preserveMap|
  is passed to the constructor of |PlaneGraphCopy|.}*/
  
  PlaneGraphCopy &call(const leda::graph &G, const LayoutInterface& L,
	  bool preserveMap = true);
  /*{\Mop calls the topology module by only determining the combinatorial
  embedding. It does not compute the face cycles and does not determine
  the external face. The parameter |preserveMap|
  is passed to the constructor of |PlaneGraphCopy|.}*/	

  leda::face call(leda::graph &G, LayoutInterface &L, leda::list<leda::edge> &R);

  AGDModule *clone() const;

protected:

  void get_bends_in_copy(const PlaneGraphCopy &PG,
			 const LayoutInterface &L,
			 leda::edge_array<DPolyline> &bends);

  bool get_planar_representation(leda::graph &PG,
				 const leda::edge_array<DPolyline> &bends);

  leda::face get_outer_face(leda::graph &G, 
			   const leda::edge_array<DPolyline> &bends);

  double angle(DPoint p, DPoint q, DPoint r);

};

/*{\Mtext \headerline{Example}
An example program showing how to use a [[TopologyModule]] in combination with
a planar layout algorithm can be found in Section~\ref{The Orthogonal Drawing Algorithm} (p.~\pageref{ortho_example}).
}*/

#endif

