/*******************************************************************************
+
+  AGD 1.2
+
+
+  WeightedMedianHeuristic.h
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                       WeightedMedianHeuristic.h
//
//
//    author: Carsten Gutwenger
//    date:   decembre 1998
//-------------------------------------------------------------------


#ifndef _AGD_WEIGHTED_MEDIAN_HEURISTIC_H
#define _AGD_WEIGHTED_MEDIAN_HEURISTIC_H


#include <AGD/TwoLayerCrossMin.h>
#include <LEDA/core/array.h>


/*{\Manpage {WeightedMedianHeuristic} {} {Weighted Median Heuristic} {M} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{0.9}{1.5}{6.1}{TwoLayerCrossMin}
\arhook{7}{2}{8}
\hdclass{0.5}{0}{8}{WeightedMedianHeuristic}
\end{hierarchy}
}*/

class __exportAGD WeightedMedianHeuristic : public TwoLayerCrossMin {

/*{\Mdefinition
The class |\Mname| implements the two layer crossing minimization heuristic
{\it Weighted--Median}. It reorders the nodes on the changeable layer according
to the weighted median weight suggested in~\cite{GKNV93}
\[ WeightedMedian(v) = \left\{\begin{array}{cl}
0.0  & \mbox{ if }n=0\\
\frac{1}{2}\left(P(w_1)+P(w_2)\right) & \mbox{ if }n=2\\
P(w_{\lceil \frac{n}{2}\rceil}) & \mbox{ if }n\mbox{ is odd}\\
\frac{1}{\ell+r}\left(r P(w_{\frac{n}{2}}) +  \ell P(w_{\frac{n}{2}+1})\right) & \mbox{ otherwise}
\end{array}\right. \]
where $\ell = P(w_{\frac{n}{2}})-P(w_1)$, $r=P(w_n)-P(w_{\frac{n}{2}+1})$, $v$ is
a node on the changeable layer, and $w_1,\ldots,w_n$
is the sequence of neighbors of $v$ on the fixed layer ordered
according to the position |P|. Thus, the weighted median value of a vertex is
defined as the median position of the adjacent vertices if that is uniquely
defined. Otherwise, it is interpolated between the two median positions
$P(w_{\frac{n}{2}})$ and $P(w_{\frac{n}{2}+1})$
using a measure of tightness. Generally, the weighted median is biased
toward the side where vertices are more closely packed.

\horizline
\commoninfo{Weighted-Median}{Weighted-Median (Gansner et al.)}{E. R. Gansner, E. Koutsofios, S. C. North, K. P. Vo}{C. Gutwenger}
{Decembre 1998}{1.0}

\begin{preposttab}
precondition & $\emptyset$ \\
postcondition(|PRE|) & $\emptyset$
\end{preposttab}

\horizline
}*/

public:

	/*{\Mcreation M 5.5}*/

	WeightedMedianHeuristic() { }
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname|.}*/

	~WeightedMedianHeuristic() { }


	/*{\Moperations}*/



	leda::string name () const {
		return "Weighted-Median";
	}

	leda::string long_name () const {
		return "Weighted-Median (Gansner et al.)";
	}

	leda::string author () const {
		return "E. R. Gansner, E. Koutsofios, S. C. North, K. P. Vo";
	}

	leda::string impl_author () const {
		return "C. Gutwenger";
	}

	leda::string impl_version () const {
		return "1.0";
	}

	leda::string impl_date () const {
		return "Decembre 1998";
	}

	AGDModule *clone() const {
		return new WeightedMedianHeuristic();
	}

	/*{\Moptions warnings=no}*/

	/*{\Mtext\stdstart{1.5cm}
	\headerline{Standard Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|TwoLayerCrossMin|).
	}*/

	/* inherited
	bool check (const leda::graph& G, const Layer& L, AgdKey &p) const;
	*/
	/*{\Mop}*/

	void init (const leda::graph& G, const Hierarchy& H);
	/*{\Mop}*/

	void call (const leda::graph& G, Layer& L);
	/*{\Mop}*/

	void cleanup ();
	/*{\Mop}*/

	/*{\Mtext \stdend}*/

private:

	leda::node_array<double> weight;
	leda::node_array<leda::array<int> > pos;
	leda::node_array<int> size;
};


/*{\Mimplementation
We use a randomized quicksort for sorting the nodes on
the layer according to the median weight.
The init() function takes time $\BigO(n+m)$, and the call()
function takes expected time $\BigO(m_L+n_L\log n_L)$,
where $n$ is the number of nodes in the input graph, $n_L$
is the number of nodes on the changeable layer and $m_L$
is the number of edges between the fixed and the changeable layer.
}*/


#endif
