/*******************************************************************************
+
+  AGD 1.2
+
+
+  augmentation.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         augmentation.cpp
//
//    author: Carsten Gutwenger
//    date:   septembre 1998
//-------------------------------------------------------------------



#include <AGD/PlanAug.h>
#include <LEDA/graph/graph_misc.h>
#include <AGD/LEDAMakeBiconnected.h>

using namespace leda;

#include "l_matcher.h"


const int AugmentationModule::in_graph = 1;
const int AugmentationModule::out_supergraph = 1;



//-------------------------------------------------------------------
//                        AugmentationModule
//-------------------------------------------------------------------

bool AugmentationModule::check (const graph& G, AgdKey &p) const
{
	return check_graph (in_graph, G, p);
}

void AugmentationModule::call (graph& G, list<edge>& L)
{
	do_call (G, L);

	_num_added_edges = L.length();
}


void AugmentationModule::call (graph& G)
{
	list<edge> L;

	call (G, L);
}

void AugmentationModule::hide_self_loops (graph& G, list<edge>& L)
{
	edge e;

	forall_edges (e, G) {
		if (source(e) == target(e)) {
			G.hide_edge(e);
			L.append(e);
		}
	}
}


static node_array<int>* num_ptr;
  
static int source_num(const edge& e) { return (*num_ptr)[source(e)]; }
static int target_num(const edge& e) { return (*num_ptr)[target(e)]; }

void AugmentationModule::hide_multiple_edges (graph& G, list<edge>& L)
{
	list<edge> el = G.all_edges();

	if (el.empty()) return;

	node_array<int> num(G);
	int  n = 0;
	node v;
	forall_nodes(v,G) num[v] = n++;
  
	num_ptr = &num;

	el.bucket_sort(0,n-1,&source_num);
	el.bucket_sort(0,n-1,&target_num);
  
	edge e0 = el.pop();

	edge e;
	forall(e,el) {
		if (source(e0) == source(e) && target(e0) == target(e)) {
			L.append (e);
			G.hide_edge (e);
		} else {
			e0 = e;
		}
	}
 }


//-------------------------------------------------------------------
//                            PlanAug
//-------------------------------------------------------------------

PlanAug::PlanAug() : AugmentationModule()
{
	add_precondition (key::planar);

	add_post_rule (key::planar);
	add_post_rule (key::biconnected);
	add_post_rule (key::simple, in_graph, key::simple);
	add_post_rule (key::no_self_loops, in_graph, key::no_self_loops);
	add_post_rule (key::directed, in_graph, key::directed);
	add_post_rule (key::undirected, in_graph, key::undirected);

	use_blabels (true);
	size_strategy (false);
	_nTests = 0;
}

AGDModule *PlanAug::clone() const
{
	PlanAug *M = new PlanAug();

	M->size_strategy (size_strategy());
	M->use_blabels (use_blabels ());

	return M;
}

void PlanAug::do_call (graph& G, list<edge>& L)
{
        _nTests = 0;
	list<edge> L1;

	bool undirected = G.is_undirected();

	if (undirected) G.make_directed();

	hide_self_loops (G, L1);
	hide_multiple_edges (G, L1);

	l_matcher LM (G, SIZE);
	if ( size_strategy ())
	  L = LM.augment (use_blabels (), SIZE);
	else
	  L = LM.augment (use_blabels (), CONNECT_CONDITION);

	edge e;
	forall(e, L1) G.restore_edge (e);

	if (undirected) G.make_undirected();
	_nTests = LM.nPlanarityTests ();
}


//-------------------------------------------------------------------
//                         LEDAMakeBiconnected
//-------------------------------------------------------------------

LEDAMakeBiconnected::LEDAMakeBiconnected() : AugmentationModule()
{
	add_post_rule (key::planar, in_graph, key::planar);
	add_post_rule (key::simple, in_graph, key::simple);
	add_post_rule (key::no_self_loops, in_graph, key::no_self_loops);
	add_post_rule (key::biconnected);
	add_post_rule (key::directed, in_graph, key::directed);
	add_post_rule (key::undirected, in_graph, key::undirected);
}

void LEDAMakeBiconnected::do_call (graph& G, list<edge>& L)
{
	Make_Biconnected (G, L);
}


AGDModule *LEDAMakeBiconnected::clone() const
{
	return new LEDAMakeBiconnected;
}

