/*******************************************************************************
+
+  AGD 1.2
+
+
+  AGD.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+
*******************************************************************************/


//-------------------------------------------------------------------
//
//                           AGD.cpp
//
//    author: Carsten Gutwenger
//    date:   novembre 1999
//-------------------------------------------------------------------


#include <AGD/AGDModule.h>
#include <AGD/PreCond.h>
#include <AGD/PostCond.h>
#include <AGD/PostCondRule.h>
#include <AGD/DefaultKeymapper.h>
#include <AGD/DefaultRuleAdmin.h>
#include <AGD/GraphChecker.h>
#include <AGD/ClusterGraphChecker.h>
#include <AGD/SubgraphChecker.h>
#include <AGD/PlanAug.h>
#include <AGD/CPlanarEmbed.h>
#include <AGD/TriconnectivityTest.h>
#include <AGD/GraphCopy.h>
#include <AGD/AgdKey.h>

#include <LEDA/graph/graph_alg.h>
#include <LEDA/core/random_source.h>
#include <LEDA/internal/std/iostream.h>

using namespace leda;

//-------------------------------------------------------------------
//                             DLL stuff
//-------------------------------------------------------------------

#ifdef AGD_DLL

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

BOOL APIENTRY DllMain(HANDLE hModule,
	DWORD  ul_reason_for_call, LPVOID lpReserved)
{
    switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
			AGD::init();
			break;

		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
			break;

		case DLL_PROCESS_DETACH:
			AGD::cleanup();
			break;
    }
    return TRUE;
}

#else

//-------------------------------------------------------------------
//                     global non-DLL initialization
//-------------------------------------------------------------------

int AGDInit::count;  //statical int variables are automatically
                     //intialized with 0

AGDInit::AGDInit()
{
	if (count++ == 0)
		AGD::init();
}

AGDInit::~AGDInit()
{
	if (--count == 0)
		AGD::cleanup();
}

#endif


//-------------------------------------------------------------------
//                               AGD
//-------------------------------------------------------------------

DefaultKeymapper *AGD::keymapper;
DefaultRuleAdmin *AGD::admin;
GraphChecker     *AGD::graph_checker;
ClusterGraphChecker *AGD::clustergraph_checker;
SubgraphChecker  *AGD::subgraph_checker;
random_source	 *AGD::Random_Source;


// static library initialization

void AGD::init()
{
	keymapper     = new DefaultKeymapper;
	admin         = new DefaultRuleAdmin(*keymapper);
	graph_checker = new GraphChecker;
	clustergraph_checker = new ClusterGraphChecker(*graph_checker);
	subgraph_checker = new SubgraphChecker(*graph_checker);
	Random_Source = new random_source;
}

void AGD::cleanup()
{
#ifndef AGD_DLL
	delete Random_Source;
	delete subgraph_checker;
	delete clustergraph_checker;
	delete graph_checker;
	delete admin;
	delete keymapper;
#endif
}


// access to the default keymapper

int AGD::number_of_keys()
{
	return keymapper->number_of_keys();
}

AgdKey AGD::ask_key (const leda::string& s)
{
	return keymapper->ask_key(s);
}

AgdKey AGD::get_key (const leda::string& s)
{
	return keymapper->get_key(s);
}

AgdKey AGD::predefined (key::Predefined k)
{
	return keymapper->predefined(k);
}


void AGD::closure(CondBase &Pc, const CondBase &P)
{
	const int n = P.num_param();

	for (int i = P.low(); i <= n; i++)
		Pc[i] = admin->closure(P[i]);
}

PreCond AGD::closure(const PreCond &P)
{
	PreCond Pc(P.num_param());
	closure(Pc,P);
	return Pc;
}

PostCond AGD::closure(const PostCond &P)
{
	PostCond Pc(P.num_param());
	closure(Pc,P);
	return Pc;
}

KeySet AGD::closure(const KeySet &S)
{
	return admin->closure(S);
}


bool AGD::implies(const KeySet &P, const KeySet &Q)
{
	return admin->implies(P,Q);
}

bool AGD::implies(const CondBase &P, const CondBase &Q)
{
	for (int i = P.low(); i <= P.num_param(); i++)
		if (!implies(P[i],Q[i])) return false;

	return true;
}

// testing graph properties

bool AGD::is_cplanar (const ClusterGraph& CG)
{
  CPlanarEmbed CPE;

  return CPE.is_cplanar(CG);
}

bool AGD::is_simple (const graph& G)
{
	return (G.number_of_edges() == 0) ? true : Is_Simple (G);
}

bool AGD::is_connected (const graph& G)
{
	return (G.number_of_nodes() == 0) ? true : Is_Connected (G);
}

bool AGD::is_biconnected (const graph& G)
{
	return Is_Biconnected (G);
}

bool AGD::is_bipartite (const graph& G)
{
	return Is_Bipartite (G);
}

bool AGD::is_acyclic (const graph& G)
{
	if (G.is_directed() == false) return false;
	return Is_Acyclic (G);
}

bool AGD::is_st_graph (const graph& G, node& s, node& t)
{
	return is_acyclic(G) && has_single_source(G,s) && has_single_sink(G,t);
}

bool AGD::is_st_planar (const graph& G, node &s, node &t)
{
	if (!is_st_graph(G,s,t))
		return false;

	if (G.empty())
		return true;

	GraphCopy GC(G);
	GC.new_edge(GC.copy(s),GC.copy(t));

	return PLANAR(GC);
}

bool AGD::is_four_graph (const graph& G)
{
	node v;
	const int four = (G.is_map()) ? 8 : 4;

	forall_nodes (v, G) {
		if (degree(v) > four) return false;
	}
	return true;
}

bool AGD::has_no_self_loops (const graph& G)
{
	edge e;

	forall_edges (e, G) {
		if (source(e) == target(e))
			return false;
	}

	return true;
}

bool AGD::is_planar (const graph& G)
{
	graph directed_G (G);

	directed_G.make_directed();

	return PLANAR (directed_G);
}

bool AGD::is_triconnected (const graph& G)
{
	TriconnectivityTest T;

	return T.call(G);
}

bool AGD::dfs_forest (node v,
	node_array<bool>& visited, node_array<bool>& mark)
{
	node w;
	edge e;
	list<node> N;

	visited [v] = true;

	forall_out_edges (e, v) {
		w = target(e);
		if (!mark[w]) {
			mark[w] = true;
			N.append(w);
		}
	}

	forall(w,N)
		mark[w] = false;

	while(!N.empty()) {
		w = N.pop();

		if (visited [w])
			return false;
		if (dfs_forest (w,visited,mark) == false)
			return false;
	}

	return true;
}

bool AGD::is_forest (const graph& G, int &num_sources)
{
	node_array<bool> visited(G,false), mark(G,false);

	if (G.is_undirected())
		return false;

	if (G.number_of_nodes() == 0)
		return true;

	node v;
	num_sources = 0;
	forall_nodes(v,G)
		if (G.indeg(v) == 0) {
			num_sources++;
			if (dfs_forest(v,visited,mark) == false)
				return false;
		}

	forall_nodes(v,G)
		if (!visited[v]) return false;

	return true;
}

bool AGD::is_forest(const graph &G)
{
	int num_sources;

	return is_forest(G,num_sources);
}

bool AGD::is_tree (const graph& G)
{
	int num_sources;

	if (is_forest(G,num_sources) == false)
		return false;

	return (num_sources == 1);
}

bool AGD::has_single_source (const graph& G, node& s)
{
	node v;
	s = nil;

	if (G.is_directed() == false) return false;

	forall_nodes (v, G) {
		if (G.indeg(v) == 0) {
			if (s != nil)
				return false;
			else
				s = v;
		}
	}
	return (G.empty() || s != nil);
}

bool AGD::has_single_sink (const graph& G, node& t)
{
	node v;
	t = nil;

	if (G.is_directed() == false) return false;

	forall_nodes (v, G) {
		if (G.outdeg(v) == 0) {
			if (t != nil)
				return false;
			else
				t = v;
		}
	}
	return (G.empty() || t != nil);
}

bool AGD::has_no_reversals (const graph& G)
{
	if (G.is_undirected())
		return true;

	edge_array<edge> rev(G);
	edge e;

	Is_Bidirected (G, rev);

	forall_edges (e, G)
		if (rev[e] != nil && source(e) != target(e)) return false;

	return true;
}


// establishing graph properties

list<edge> AGD::triangulate (graph& G)
{
	return TRIANGULATE_PLANAR_MAP (G);
}

list<edge> AGD::make_bidirected (graph& G)
{
	return Make_Bidirected (G);
}

list<edge> AGD::make_biconnected (graph& G)
{
	return Make_Biconnected (G);
}

list<edge> AGD::augment_planar_biconnected (graph& G)
{
	PlanAug augmenter;
	list<edge> L;
	augmenter.call(G,L);
	return L;
}

list<edge> AGD::make_map (graph& G)
{
	list<edge> R;
	G.make_map (R);
	return R;
}

bool AGD::embed_planar_graph (graph& G)
{
	return PLANAR (G, true);
}

bool AGD::embed_cplanar_graph (ClusterGraph& CG, list<edge>& L)
{
        CPlanarEmbed CPE;
	return(CPE.call(CG, L));
}


// compute all standard properties of a clustergraph
KeySet AGD::get_properties (const ClusterGraph& CG)
{
  KeySet all_props;

  Keymapper::const_iterator it;
  for (it = keymapper->begin(); it.valid(); ++it)
    all_props << (*it);

  return clustergraph_checker->get_properties(CG,all_props);

}

// compute all standard properties of a graph

KeySet AGD::get_properties (const graph& G)
{
	KeySet all_props;

	Keymapper::const_iterator it;
	for (it = keymapper->begin(); it.valid(); ++it)
		all_props << (*it);

	return graph_checker->get_properties(G,all_props);
}


//-------------------------------------------------------------------
//                               AGDModule
//-------------------------------------------------------------------

AGDModule::AGDModule (int n, int m) : prePostChanged(true)
{
	_pre  = new PreCond(n);
	_post = new PostCondRule(m);
	_pc   = new PostCond(m);
}

AGDModule::~AGDModule()
{
	delete _pc;
	delete _post;
	delete _pre;
}

int AGDModule::num_in_param () const
{
	return _pre->num_param();
}

int AGDModule::num_out_param () const
{
	return _post->num_param();
}


bool AGDModule::check_graph (int i, const graph& G, AgdKey &p) const
{
	KeySet S = (*_pre)[i];

	get_graph_checker().check(G,S,p);

	if (p.valid()) return false;

	KeySet::const_iterator it;
	for (it = S.begin(); it != S.end(); ++it) {
		p = *it;
		if (check_graph_key (p,G) == false)
			return false;
	}

	p = AgdKey();
	return true;
}

bool AGDModule::check_graph (int i,
			     const ClusterGraph& CG,
			     AgdKey &p) const
{
	KeySet S = (*_pre)[i];

	get_clustergraph_checker().check(CG,S,p);

	if (p.valid()) return false;

	KeySet::const_iterator it;
	for (it = S.begin(); it != S.end(); ++it) {
		p = *it;
		if (check_graph_key (p,CG) == false)
			return false;
	}

	p = AgdKey();
	return true;
}


bool AGDModule::check_graph_key (AgdKey , const graph& ) const
{
	return false;
}

bool AGDModule::check_graph_key (AgdKey , const ClusterGraph& ) const
{
	return false;
}


// maintaining preconditions

void AGDModule::add_precondition  (AgdKey k, int i)
{
	if (i < 1 || i > num_in_param())
		error_handler(1, "AGDModule::add_precondition: index out of range");

	_pre->insert(i,k);
	prePostChanged = true;
}

void AGDModule::add_precondition (AgdKey k, bool status, int i)
{
	if (status)
		add_precondition(k,i);
	else
		rem_precondition(k,i);
	prePostChanged = true;
}

void AGDModule::add_preconditions (const string& ps, char delim, int i)
{
	list<string> L;
	list_item it;

	extract_list (ps, delim, L);
	forall_items (it, L) {
		add_precondition (L[it], i);
	}
	prePostChanged = true;
}

void AGDModule::rem_precondition (AgdKey k, int i)
{
	if (i < 1 || i > num_in_param())
		error_handler(1, "AGDModule::rem_precondition: index out of range");

	(*_pre)[i].del (k);
	prePostChanged = true;
}


// access to the precondition

const KeySet& AGDModule::precondition (int i) const
{
	if (i < 1 || i > num_in_param())
		error_handler(1, "AGDModule::precondition: index out of range");

	return (*_pre)[i];
}

const PreCond& AGDModule::precondition () const
{
	return (*_pre);
}

// computing the postcondition sets

KeySet AGDModule::postcondition (const PreCond& PRE, int i) const
{
	if (i < 0 || i > num_out_param())
		error_handler(1,"AGDModule::postcondition: index out of range");

	return postcondition(PRE)[i];
}

const KeySet &AGDModule::postcondition (int i)
{
	if (prePostChanged) {
		*_pc = postcondition(*_pre);
		prePostChanged = false;
	}

	return (*_pc)[i];
	//return postcondition (*_pre, i);
}

PostCond AGDModule::postcondition (const PreCond& PRE) const
{
	if (PRE.num_param() != num_in_param())
		error_handler(1, "AGDModule::postcondition: PRE has wrong bounds.");

	PreCond PREc = closure(PRE);
	PostCond POST(num_out_param());
	list_item it;
	TRLiteral l;

	for (int i = 0; i <= num_out_param(); i++)
		forall_items (it, (*_post)[i]) {
			bool add_key = true;
			const list<TRLiteral>& L = _post->left(i,it);
			forall (l, L) {
				if (!PREc[l.index].member (l.key)) {
					add_key = false; break;
				}
			}

			if (add_key)
				POST[i].insert(_post->right(i,it));
		}

	return POST;
}


// maintaining postconditions

AGDModule::pr_item AGDModule::add_post_rule (const TaggedRule& r, int i)
{
	if (i < 0 || i > num_out_param())
		error_handler(1,"AGDModule::add_post_rule: index out of range");

	prePostChanged = true;
	return (*_post)[i].append(r);
}

AGDModule::pr_item AGDModule::add_post_rule (AgdKey k, int i)
{
	if (i < 0 || i > num_out_param())
		error_handler(1,"AGDModule::add_post_rule: index out of range");

	prePostChanged = true;
	return (*_post)[i].append(TaggedRule(k));
}

AGDModule::pr_item AGDModule::add_post_rule (AgdKey p1, int i1, AgdKey p, int i)
{
	if (i < 0 || i > num_out_param())
		error_handler(1,"AGDModule::add_post_rule: index out of range");

	prePostChanged = true;

	TaggedRule r(p);
	r.And (p1, i1);
	return (*_post)[i].append (r);
}

void AGDModule::rem_post_rule (pr_item it, int i)
{
	if (i < 0 || i > num_out_param())
		error_handler(1,"AGDModule::rem_post_rule: index out of range");

	_post->del(i,(list_item)it);
	prePostChanged = true;
}


// testing if module is applicable

bool AGDModule::applicable (const PreCond& PRE, const PostCond& POST) const
{
	if (PRE.num_param() != num_in_param())
		error_handler(1, "AGDModule::applicable(PRE,POST): PRE has wrong bounds.");
	if (POST.num_param() != num_out_param())
		error_handler(1, "AGDModule::applicable(PRE,POST): POST has wrong bounds.");

	return (implies(PRE,*_pre) && implies(postcondition(PRE),POST));
}


// auxiliary functions

void AGDModule::extract_list (const string& str, char delim, list<string>& L)
{
	const int len = str.length();
	int start = 0, current = 0;

	L.clear();
	while (current < len) {
		if (str[current] == delim) {
			L.append (str(start,current-1));
			start = current+1;
		}
		current++;
	}
	L.append (str(start,current-1));
}

ostream& operator<< (ostream& os, const AgdKey& k) {
	os << (k.info());
	return os;
}

istream& operator>> (istream& is, AgdKey &) {
	return is;
}

