/*******************************************************************************
+
+  AGD 1.2
+
+
+  GridLayout.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                          GridLayout.cpp
//
//    author: Carsten Gutwenger
//    date:   july 1999
//-------------------------------------------------------------------


#include <AGD/GridLayout.h>
#include <AGD/GraphCopy.h>

#include <LEDA/internal/std/math.h>

using namespace leda;


//-------------------------------------------------------------------
//                           GridLayout
//-------------------------------------------------------------------

GridLayout &GridLayout::operator=(const GridLayout &gl)
{
	_x      = gl._x;
	_y      = gl._y;
	_width  = gl._width;
	_height = gl._height;
	_bends  = gl._bends;
	gp = gl.gp;
	return *this;
}

void GridLayout::init_node_size(int w, int h)
{
	_width .init(*gp,w);
	_height.init(*gp,h);
}


bool GridLayout::in_rect(IPoint ip, node v) const
{
	return (_x[v] <= ip.xcoord() && ip.xcoord() <= _x[v]+_width[v] &&
		_y[v] <= ip.ycoord() && ip.ycoord() <= _y[v]+_height[v]);
}


void GridLayout::get_line (edge e, IPolyline &ipl) const
{
	ipl.clear();

	node v = source(e), w = target(e);
	IPoint pos_v = IPoint(_x[v],_y[v]);
	IPoint pos_w = IPoint(_x[w],_y[w]);

	ipl = _bends[e];

	if (ipl.empty() || !in_rect(ipl.head(),v)) {
		ipl.push_front(_get_intersection(pos_v,
			(ipl.empty() ? pos_w : ipl.head()),
			_width[v], _height[v]));
	}

	if (!in_rect(ipl.tail(),w)) {
		ipl.push_back(_get_intersection(pos_w,ipl.tail(),
			_width[w], _height[w]));
	}
}


IPoint GridLayout::_get_intersection (const IPoint &p, const IPoint &q,
	int w, int h)
{
	int px = p.xcoord(), py = p.ycoord();
	int qx = q.xcoord(), qy = q.ycoord();

	if (qx < px || qy < py) return p;

	if (qy > py) {
		int f = h*(qx-px);
		if (f % (qy-py) == 0) {
			int x = px + f/(qy-py);
			if (x <= px+w) return IPoint(x,py+h);
		}
	}

	if (qx > px) {
		int f = w*(qy-py);
		if (f % (qx-px) == 0) {
			int y = py + f/(qx-px);
			if (y <= py+h) return IPoint(px+w,y);
		}
	}

	return p;
}


int GridLayout::number_of_bends() const
{
	int num = 0;

	edge e;
	forall_edges(e,*gp) {
		const IPolyline &ipl = _bends[e];
		if (ipl.length() >= 2) {
			num += _bends[e].length();
			if (in_rect(ipl.head(),source(e))) --num;
			if (in_rect(ipl.tail(),target(e))) --num;
		} else if (ipl.length() == 1) {
			IPoint ip = ipl.head();
			if (!in_rect(ip,source(e)) && !in_rect(ip,target(e)))
				num++;
		}
	}
	return num;
}


int GridLayout::_manhattan_dist(const IPoint &ip1, const IPoint &ip2)
{
	return abs(ip2.xcoord()-ip1.xcoord()) + abs(ip2.ycoord()-ip1.ycoord());
}


int GridLayout::total_manhattan_edge_length() const
{
	int sum = 0;
	edge e;
	IPolyline ipl;

	forall_edges(e,*gp) {
		get_line(e,ipl);
		IPoint ip, ip1 = ipl.pop_front();

		forall(ip,ipl) {
			sum += _manhattan_dist(ip1,ip);
			ip1 = ip;
		}
	}

	return sum;
}


double GridLayout::_euclidean_dist(const IPoint &ip1, const IPoint &ip2)
{
	double dx = ip2.xcoord()-ip1.xcoord();
	double dy = ip2.ycoord()-ip1.ycoord();
	return sqrt(dx*dx + dy*dy);
}


double GridLayout::total_edge_length() const
{
	double sum = 0;
	edge e;
	IPolyline ipl;

	forall_edges(e,*gp) {
		get_line(e,ipl);
		IPoint ip, ip1 = ipl.pop_front();

		forall(ip,ipl) {
			sum += _euclidean_dist(ip1,ip);
			ip1 = ip;
		}
	}

	return sum;
}


IPolyline GridLayout::get_compact_bends(edge e) const
{
	IPolyline ipl = _bends[e];

	if (ipl.length() == 0) return ipl;
	
	IPoint ip_first = ipl.head();
	IPoint ip_last  = ipl.tail();

	IPoint ip_src(_x[source(e)],_y[source(e)]);
	IPoint ip_tgt(_x[target(e)],_y[target(e)]);

	ipl.push_front(ip_src);
	ipl.push_back (ip_tgt);

	_compact(ipl);

	ipl.pop_front();
	ipl.pop_back();

	if (ip_first != ip_src && (ipl.empty() || ip_first != ipl.head()))
		ipl.push_front(ip_first);
	if (ip_last != ip_tgt && (ipl.empty() || ip_last != ipl.tail()))
		ipl.push_back(ip_last);

	return ipl;
}


void GridLayout::compact_all_bends()
{
	edge e;
	forall_edges(e,*gp)
		_bends[e] = get_compact_bends(e);
}


bool GridLayout::_is_redundant(IPoint &p1, IPoint &p2, IPoint &p3)
{
	int dzy1 = p3.xcoord() - p2.xcoord();
	int dzy2 = p3.ycoord() - p2.ycoord();
	int dyx1 = p2.xcoord() - p1.xcoord();

	if (dzy1 == 0) return (dyx1 == 0 || dzy2 == 0);

	int f = dyx1 * dzy2;

	return (f % dzy1 == 0 && (p2.ycoord() - p1.ycoord()) == f / dzy1);
}


void GridLayout::_compact(IPolyline& ip)
{
	if (ip.length() < 3) return;

	IPoint p;
	list_item it = ip.first();
	p = ip[it]; it = ip.succ(it);
	for (it = ip.succ(it); it != nil; it = ip.succ(it)) {
		list_item it_pred = ip.pred(it);
		if (_is_redundant(p,ip[it_pred],ip[it])) {
			ip.del_item(it_pred);
		} else {
			p = ip[it_pred];
		}
	}
}


IRect GridLayout::bounding_box() const
{
	if (!gp || gp->empty()) return IRect();

	node v = gp->first_node();
	int xmin = _x[v], xmax = _x[v],
		ymin = _y[v], ymax = _y[v];

	forall_nodes (v, *gp) {
		if (_x[v]+_width[v] > xmax) xmax = _x[v]+_width[v];
		if (_x[v] < xmin) xmin = _x[v];
		if (_y[v]+_height[v] > ymax) ymax = _y[v]+_height[v];
		if (_y[v] < ymin) ymin = _y[v];
	}

	edge e;
	forall_edges (e, *gp) {
		const IPolyline& points = _bends[e];

		IPoint ip;
		forall (ip, points) {
			if (ip.xcoord() > xmax) xmax = ip.xcoord();
			if (ip.xcoord() < xmin) xmin = ip.xcoord();
			if (ip.ycoord() > ymax) ymax = ip.ycoord();
			if (ip.ycoord() < ymin) ymin = ip.ycoord();
		}
	}

	return IRect(xmin,xmax,ymin,ymax);
}


void GridLayout::supplement_map_layout()
{
	edge e;
	forall_edges (e, *gp)
		if (_bends[e].empty()) {
			_bends[e] = _bends[gp->reversal(e)];
			_bends[e].reverse_items();
		}
}
