/*******************************************************************************
+
+  AGD 1.2
+
+
+  KeySet.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+
*******************************************************************************/


//-------------------------------------------------------------------
//
//                         KeySet.cpp
//
//
//    author: Carsten Gutwenger
//    date:   octobre 1998
//-------------------------------------------------------------------


#include <AGD/KeySet.h>
#include <AGD/DefaultKeymapper.h>

#include <LEDA/core/int_set.h>
#include <LEDA/internal/std/ctype.h>

using namespace leda;


const int KeySet::min_table_size = 8*sizeof(unsigned long);


KeySet::KeySet()
{
	_table_size = min_table_size;
	_alloc_table(min_table_size);
	_mapper = &AGD::get_keymapper();
}

KeySet::KeySet(const Keymapper &M)
{
	_table_size = min_table_size;
	_alloc_table(min_table_size);
	_mapper = &M;
}

KeySet::KeySet (const KeySet& S)
{
	_table_size = S._table_size;
	_table = new int_set(*S._table);
	_mapper = S._mapper;
}

KeySet::~KeySet()
{
	delete _table;
}


void KeySet::_alloc_table (int sz)
{
	while (_table_size < sz)
		_table_size <<= 1;

	_table = new int_set(_table_size);
}

void KeySet::_expand_table (int sz)
{
	int_set *old_table = _table;
	int i, old_size = _table_size;

	_alloc_table(sz);
	for (i = 0; i < old_size; i++)
		if (old_table->member(i)) _table->insert(i);

	delete old_table;
}

void KeySet::insert (const AgdKey &k)
{
	int id = k.index();

	if (id >= _table_size) _expand_table(id+1);
	_table->insert(id);
}

void KeySet::del (const AgdKey &k)
{
	if (k.index() < _table_size)
		_table->del(k.index());
}

bool KeySet::member (const AgdKey &k) const
{
	int id = k.index();

	return (id < _table_size) ? (_table->member(id) ? true : false) : false;
}

bool KeySet::empty() const
{
	return !(begin().valid());
}


void KeySet::clear()
{
	_table->clear();
}


KeySet& KeySet::operator+= (const KeySet& T)
{
	if (T._table_size > _table_size)
		_expand_table(T._table_size);

	_table->join(*T._table);

	return *this;
}


KeySet& KeySet::operator-= (const KeySet& T)
{
	const_iterator it;

	for (it = T.begin(); it != T.end(); ++it)
		del(*it);

	return *this;
}


KeySet& KeySet::operator&= (const KeySet& T)
{
	if (T._table_size > _table_size)
		_expand_table(T._table_size);

	_table->intersect(*T._table);

	return *this;
}


KeySet& KeySet::operator%= (const KeySet& T)
{
	if (T._table_size > _table_size)
		_expand_table(T._table_size);

	KeySet S1(*this);
	S1 &= T;
	*this += T;
	*this -= S1;

	return *this;
}


bool KeySet::operator<= (const KeySet& T) const
{
	const_iterator it;

	for (it = begin(); it != end(); ++it)
		if (!T.member(*it)) return false;
	return true;
}


bool KeySet::operator== (const KeySet &T) const
{
	Keymapper::const_iterator it;

	for (it = _mapper->begin(); it != _mapper->end(); ++it) {
		bool m1 = member(*it), m2 = T.member(*it);
		if ((m1 && !m2) || (!m1 && m2))
			return false;
	}

	return true;
}


KeySet& KeySet::operator= (const KeySet& T)
{
	delete _table;
	_table = new int_set(*T._table);
	_table_size = T._table_size;
	_mapper = T._mapper;

	return *this;
}

void KeySet::read (istream& is, int delim)
{
	string str;
	char c;

	clear();

	if (delim == EOF) {
		for (; ;) {
			while (is.get(c) && isspace(c)) ;
			if (!is) break;
			is.putback(c);
			is >> str;
			insert (str);
		}
	} else {
		for (; ;) {
			while (is.get(c) && isspace(c) && c != delim) ;
			if (!is || c == delim) break;
			is.putback (c);
			is >> str;
			insert (str);
		}
	}
}


void KeySet::print (ostream& os, char space) const
{
	const_iterator it;
	bool first = true;

	for (it = begin(); it != end(); ++it) {
		if (first) first = false; else os << space;
		os << (it->info());
	}
}

istream& operator>> (istream& is, KeySet& S)
{
	S.read (is);
	return is;
}


ostream& operator<< (ostream& os, const KeySet& S)
{
	S.print (os);
	return os;
}
