/*******************************************************************************
+
+  AGD 1.2
+
+
+  TriconnectivityTest.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



/**********************************************************

                TriconnectivityTest.c

  Test multigraph G = (V,E) if triconnected in
  time O(n+m).

  reference: Hopcroft and Tarjan, "Dividing a Graph into
             Triconnected Components"
             SIAM J. Comput., Vol. 2, No. 3, Sept. 1973
			 pages 135-158

  author:    Carsten Gutwenger
  date:      Novembre 1998

**********************************************************/


#include <AGD/TriconnectivityTest.h>
#include <LEDA/internal/std/stdio.h>

using namespace leda;



/*********************************************************
                    The main procedure

  Returns true <=> G is triconnected.
  
  If false is returned
    s1 == nil, s2 == nil  => G is not connected
    s1 != nil, s2 == nil  => s1 is a cut vertex
    s1 != nil, s2 != nil  => (s1,s2) is a separation pair
**********************************************************/

bool TriconnectivityTest::call (const graph& G, node& s1, node& s2)
{
	const int n = G.number_of_nodes();
	const int m = G.number_of_edges();

	s1 = s2 = nil;

	if (n == 0) return true;

	// initialize arrays
	NUMBER.init(G,0); LOWPT1.init(G);
	LOWPT2.init(G);   FATHER.init(G,nil);
	ND    .init(G);   TYPE  .init(G,unseen);
	NODEAT = array<node>(1,n);

	DEGREE.init(G,0); is_edge.init(G,false);
	build_graph(G);

	num_count = 0;
	s = G.first_node(); // start node

	DFS1(G,s,nil,s1);

	// is graph not connected ?
	if (num_count < n) {
		s1 = nil; return false;
	}

	// is graph not biconnected ?
	if (s1 != nil) {
		return false;
	}

	A.init(G);
	build_acceptable_adj_struct(G);

	DFS2(G);

	TSTACK_h = new int[m];
	TSTACK_a = new int[m];
	TSTACK_b = new int[m];
	top = 0;

	bool result = PathSearch(G,s,s1,s2);

	delete TSTACK_h;
	delete TSTACK_a;
	delete TSTACK_b;

	// free resources
	NUMBER.init(); LOWPT1.init();
	LOWPT2.init(); FATHER.init();
	ND    .init(); TYPE  .init();
	A     .init(); NEWNUM.init();
	HIGHPT.init(); START .init();
	NODEAT = array<node>();
	DEGREE.init();

	is_edge.init(); edges.clear();

	return result;
}


/*********************************************************
             Building the underlying simple graph

  Creates the list edges of all edges of G without
  multiple, reversal and self-loop edges.

  computes edges
           is_edge[e] = true iff e in edges
		   DEGREE [v]
**********************************************************/

static int small_index(const edge& e)
{ return leda::min(index(source(e)),index(target(e))); }

static int large_index(const edge& e)
{ return leda::max(index(source(e)),index(target(e))); }

void TriconnectivityTest::build_graph (const graph &G)
{
	edge e;

	// sort the list of all edges, such that multi- and reversal
	// edges come after each other
	edges = G.all_edges();
	edges.bucket_sort(small_index);
	edges.bucket_sort(large_index);

	list_item it, it_next;
	for (it = edges.first(); it != nil; it = it_next) {
		e = edges[it];
		node v = source(e), w = target(e);

		it_next = edges.succ(it);

		// remove self-loops
		if (v == w) {
			edges.del_item(it);
			continue;
		}

		// e remains an edge in edges
		is_edge[e] = true;
		DEGREE[v]++; DEGREE[w]++;

		// remove multi-edges
		while (it_next) {
			e = edges[it_next];
			if ((source(e) == v && target(e) == w) ||
				(source(e) == w && target(e) == v)) {

				list_item it1 = edges.succ(it_next);
				edges.del_item(it_next);
				it_next = it1;
			} else
				break;
		}
	}
}

		
/*********************************************************
                   The first dfs-search

  computes NUMBER[v], FATHER[v], LOWPT1[v], LOWPT2[v],
           ND[v], TYPE[e]
**********************************************************/

void TriconnectivityTest::DFS1 (const graph& G, node v, node u, node& s1)
{
	node first_son = nil;
	edge e;

	NUMBER[v] = ++num_count;
	FATHER[v] = u;

	LOWPT1[v] = LOWPT2[v] = NUMBER[v];
	ND[v] = 1;

	forall_inout_edges (e,v) {

		if (!is_edge[e] || TYPE[e] != unseen)
			continue;

		node w = G.opposite(v,e);

		if (NUMBER[w] == 0) {
			TYPE[e] = (source(e) == v) ? tree : tree_rev;
			if (first_son == nil) first_son = w;

			DFS1(G,w,v,s1);

			// is v cut vertex ?
	        if (LOWPT1[w] >= NUMBER[v] && (w != first_son || u != nil)) 
				s1 = v;

			if (LOWPT1[w] < LOWPT1[v]) {
				LOWPT2[v] = leda::min(LOWPT1[v],LOWPT2[w]);
				LOWPT1[v] = LOWPT1[w];

			} else if (LOWPT1[w] == LOWPT1[v]) {
				LOWPT2[v] = leda::min(LOWPT2[v],LOWPT2[w]);

			} else {
				LOWPT2[v] = leda::min(LOWPT2[v],LOWPT1[w]);
			}

			ND[v] += ND[w];
		
		} else {

			TYPE[e] = (source(e) == v) ? frond : frond_rev;

			if (NUMBER[w] < LOWPT1[v]) {
				LOWPT2[v] = LOWPT1[v];
				LOWPT1[v] = NUMBER[w];

			} else if (NUMBER[w] > LOWPT1[v]) {
				LOWPT2[v] = leda::min(LOWPT2[v],NUMBER[w]);
			}
		}
	}
}


/*********************************************************
           Construction of ordered adjaceny lists
**********************************************************/

int TriconnectivityTest::Phi (edge e)
{
	node w = tgt(e);
	
	if (TYPE[e] == frond || TYPE[e] == frond_rev)
		return 2*NUMBER[w]+1;

	if (LOWPT2[w] < NUMBER[src(e)])
		return 2*LOWPT1[w];

	return 2*LOWPT1[w]+1;
}

void TriconnectivityTest::build_acceptable_adj_struct(const graph& G)
{
	edge e;
	int max = 2*G.number_of_nodes()+1;
	array<list<edge> > BUCKET(1,max);

	forall(e,edges)
		BUCKET[Phi(e)].append(e);

	for (int i = 1; i <= max; i++)
		forall(e,BUCKET[i])
			A[src(e)].append(e);
}


/*********************************************************
                   The second dfs-search
**********************************************************/

void TriconnectivityTest::path_finder(const graph& G, node v)
{
	edge e;
	node w;

	NEWNUM[v] = num_count - ND[v] + 1;

	forall(e,A[v]) {
		w = G.opposite(v,e);

		if (new_path) {
			new_path = false;
			START[e] = true;
		}

		if (TYPE[e] == tree || TYPE[e] == tree_rev) {
			path_finder(G,w);
			num_count--;

		} else {
			if (HIGHPT[w] == 0)
				HIGHPT[w] = NEWNUM[v];
			new_path = true;
		}
	}
}

void TriconnectivityTest::DFS2 (const graph& G)
{
	NEWNUM.init(G,0);
	HIGHPT.init(G,0);
	START .init(G,false);

	num_count = G.number_of_nodes();
	new_path = true;

	path_finder(G,s);

	node v;
	array<int> old2new(1,G.number_of_nodes());

	forall_nodes(v,G)
		old2new[NUMBER[v]] = NEWNUM[v];

	forall_nodes(v,G) {
		NODEAT[NEWNUM[v]] = v;
		LOWPT1[v] = old2new[LOWPT1[v]];
		LOWPT2[v] = old2new[LOWPT2[v]];
	}
}


bool TriconnectivityTest::PathSearch (const graph& G, node v,
	node& s1, node& s2)
{
	node w;
	edge e;
	int y, vnum = NEWNUM[v], wnum;
	int a, b;

	forall(e,A[v]) {
		w = G.opposite(v,e); wnum = NEWNUM[w];

		if (TYPE[e] == tree || TYPE[e] == tree_rev) {
			if (START[e]) {
				y = 0;
				if (top >= 0 && TSTACK_a[top] > LOWPT1[w]) {
					do {
						y = leda::max(y,TSTACK_h[top]);
						b = TSTACK_b[top--];
					} while (top >= 0 && TSTACK_a[top] > LOWPT1[w]);
					TSTACK_push(leda::max(y,wnum+ND[w]-1),LOWPT1[w],b);
				} else {
					TSTACK_push(wnum+ND[w]-1,LOWPT1[w],vnum);
				}
				TSTACK_push(-1,-1,-1); // end-of-stack marker
			}

			if (PathSearch(G,w,s1,s2) == false)
				return false;

			while (vnum != 1 && ((TSTACK_a[top] == vnum) ||
				(DEGREE[w] == 2 && NEWNUM[tgt(A[w].head())] > wnum)))
			{
				a = TSTACK_a[top];
				b = TSTACK_b[top];

				if (a == vnum && FATHER[NODEAT[b]] == NODEAT[a])
					top--;

				else if (DEGREE[w] == 2 && NEWNUM[tgt(A[w].head())] > wnum) {
					s1 = v;
					s2 = tgt(A[w].head());
					return false;
				} else {
					s1 = NODEAT[a];
					s2 = NODEAT[b];
					return false;
				}
			}

			if (LOWPT2[w] >= vnum && ND[w] <= G.number_of_nodes()-3)
			{
				s1 = NODEAT[LOWPT1[w]];
				s2 = v;
				return false;
			}

			if (START[e]) {
				while (TSTACK_h[top] != -1) {
					top--;
				}
				top--;
			}

			while (top >= 0 && TSTACK_h[top] != -1 && HIGHPT[v] > TSTACK_h[top])
				top--;

		} else { // frond arc
			if (START[e]) {
				y = 0;
				if (top >= 0 && TSTACK_a[top] > wnum) {
					do {
						y = leda::max(y,TSTACK_h[top]);
						b = TSTACK_b[top--];
					} while (top >= 0 && TSTACK_a[top] > wnum);
					TSTACK_push(y,wnum,b);
				} else {
					TSTACK_push(vnum,wnum,vnum);
				}
			}
		}
	}

	return true;
}
