/*******************************************************************************
+
+  AGD 1.2
+
+
+  random_bic_graph.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                     random_bic_graph.cpp
//
//    author: Carsten Gutwenger
//    date:   january 2000
//-------------------------------------------------------------------


#include <AGD/AGD.h>
#include <AGD/generators.h>
#include <LEDA/graph/graph.h>
#include <LEDA/graph/face_map.h>
#include <LEDA/core/array.h>

using namespace leda;


//-------------------------------------------------------------------
//                     planar_biconnected_graph
//-------------------------------------------------------------------

static int faceDeg(face f)
{
	edge e;
	int count = 0;
	forall_face_edges(e,f)
		count++;
	return count;
}

static int src_index(const edge &e)
{
	return index(source(e));
}

static int tgt_index(const edge &e)
{
	return index(target(e));
}

void planar_biconnected_graph(graph &G, int n, int m,
	bool multiEdges, bool planarMap)
{
	if (n < 3) n = 3;
	if (m < n) m = n;
	if (m > 3*n-6) m = 3*n-6;

	int ke = n-3, kf = m-n;

	G.clear();

	array<edge> edges(m);
	array<face> big_faces(m);
	face_map<int> in_bf(G);
	random_source& S = AGD::get_random_source();

	// start with a triangle
	node v1 = G.new_node(), v2 = G.new_node(), v3 = G.new_node();
	edges[0] = G.new_edge(v1,v2);
	edges[1] = G.new_edge(v2,v3);
	edges[2] = G.new_edge(v3,v1);

	int bf_size = 0, e_size = 3;

	list<edge> R;
	G.make_map(R);
	G.make_planar_map();

	while(ke+kf > 0) {
		double p;
		S >> p;

		if (bf_size == 0 || p < double(ke)/(ke+kf)) {
			edge e = edges[S(0,e_size-1)];
			face f = G.face_of(e), fr = G.face_of(G.reversal(e));

			edges[e_size++] = G.split_map_edge(e);

			if (faceDeg(f) == 4) {
				in_bf[f] = bf_size;
				big_faces[bf_size++] = f;
			}
			if (faceDeg(fr) == 4) {
				in_bf[fr] = bf_size;
				big_faces[bf_size++] = fr;
			}

			ke--;

		} else {
			int pos = S(0,bf_size-1);
			face f = big_faces[pos];
			int df = faceDeg(f);
			int i = S(0,df-1), j = S(2,df-2);

			edge e1;
			for (e1 = G.first_face_edge(f); i > 0; e1 = G.face_cycle_succ(e1))
				i--;

			edge e2;
			for (e2 = e1; j > 0; e2 = G.face_cycle_succ(e2))
				j--;

			edge e = G.split_face(e1,e2);
			edges[e_size++] = e;

			face f1 = G.face_of(e), f2 = G.face_of(G.reversal(e));

			big_faces[pos] = f1;
			in_bf[f1] = pos;
			if (faceDeg(f2) >= 4) {
				in_bf[f2] = bf_size;
				big_faces[bf_size++] = f2;
			}
			if (faceDeg(f1) == 3) {
				big_faces[pos] = big_faces[--bf_size];
			}

			kf--;
		}
	}

	if (multiEdges == false) {
		list<edge> E = G.all_edges();

		E.bucket_sort(tgt_index);
		E.bucket_sort(src_index);

		while(!E.empty()) {
			edge e = E.pop();
			node u = source(e), v = target(e);

			while (!E.empty() && source(E.head()) == u && target(E.head()) == v) {
				e = E.pop();
				if (index(source(e)) < index(target(e))) {
					edge er  = G.reversal(e);

					edge eh  = G.face_cycle_succ(e);
					edge etr = G.face_cycle_succ(eh);
					edge ehr = G.face_cycle_succ(er);
					edge et  = G.face_cycle_succ(ehr);

					G.move_edge(e,et,eh,leda::after,leda::before);
					G.move_edge(er,etr,ehr,leda::after,leda::before);
				}
			}
		}
	}

	if (planarMap == false) {
		edge e, e_next;
		for (e = G.first_edge(); e != nil; e = e_next) {
			e_next = G.succ_edge(e);

			if (index(source(e)) > index(target(e)))
				G.del_edge(e);
		}
		G.del_all_faces();
	}
}


//-------------------------------------------------------------------
//                     st_planar_graph
//-------------------------------------------------------------------

void st_planar_graph (graph &G, int n, int m, bool multi_edges, double st_prob)
{
	planar_biconnected_graph(G, n, m, multi_edges);

	if (G.number_of_edges() == 0) return;

	node_array<int> stnum(G);
	list<node> stlist;

	edge e_ref = G.first_edge();
	ST_NUMBERING1(G,stnum,stlist,e_ref);

	edge e;
	forall_edges(e, G) {
		if (stnum[source(e)] > stnum[target(e)])
			G.rev_edge(e);
	}

	random_source& rs = AGD::get_random_source();
	double p; rs >> p;

	if (st_prob == 0 || p > st_prob)
		G.del_edge(e_ref);
}


//-------------------------------------------------------------------
//                     ST_NUMBERING1
//-------------------------------------------------------------------

#if __LEDA__ <= 400

#include <LEDA/graphs/node_list.h>

static void ST_DFS1(const graph& G, node v, int& count, node_array<int>&  dfsnum, 
                                                 node_array<int>&  lownum, 
                                                 node_array<edge>& tree_in,
                                                 node_array<edge>& low_out)
{ dfsnum[v] = count;
  lownum[v] = count;
  count++;
  
  edge e;
  forall_inout_edges(e,v)
  { 
    node w = G.opposite(v,e);

    if (v == w) continue;
  
    if (dfsnum[w] == 0 ) // first visit
      { tree_in[w] = e;
        ST_DFS1(G,w,count,dfsnum,lownum,tree_in,low_out);
        if (lownum[w] < lownum[v]) 
        { lownum[v] = lownum[w]; 
          low_out[v] = e;
         }
       }
    else 
      if (dfsnum[w] < lownum[v]) 
      { lownum[v] = dfsnum[w]; 
        low_out[v] = e;
       }
   }
 }



static void FINDPATH1(const graph& G, node w, node_array<bool>& used_node,
                                        edge_array<bool>& used_edge,
                                        node_array<edge>& pred_edge,
                                        node_list& PATH)
{ while ( !used_node[w] )
  { edge x = pred_edge[w];  
    PATH.push(w);
    used_node[w] = true; 
    used_edge[x] = true;
    w = G.opposite(w,x);
   } 
}
  
#else
#include <LEDA/graph/plane_graph_alg.h>
#endif

node ST_NUMBERING1(const graph& G, node_array<int>& st_num, list<node>& st_list,
	edge e_st) 
{
#if __LEDA__ > 400
	return ST_NUMBERING(G, st_num, st_list, e_st);

#else
  st_list.clear();

  int n = G.number_of_nodes();

  if (n == 0) return nil;

  edge e = e_st;
  node s;
  node t;

  node_array<int>  dfsnum(G,0);
  node_array<int>  lownum(G,0);
  node_array<edge> tree_in(G,nil);
  node_array<edge> low_out(G,nil);

  if ( e != nil ) 
    { s = G.source(e); 
      t = G.target(e); 
     } 
  else
    { t = G.first_node();
      if ( n == 1) 
      { st_num[t] = 1; 
        st_list.append(t);
        return t;
       }

      forall_inout_edges(e,t)
         if (source(e) != target(e)) break;

      s = G.opposite(t,e);
    } 

  int count = 1;

  tree_in[t] = e;  // dummy

  dfsnum[t] = count++;
  tree_in[s] = e_st;
  ST_DFS1(G,s,count,dfsnum,lownum,tree_in,low_out);

  edge_array<bool> used_edge(G,false);
  node_array<bool> used_node(G,false);

  used_node[s] = true;
  used_node[t] = true;
  used_edge[e] = true;
  
  st_list.push(t);
  st_list.push(s);

  int st_count = 1;

  for (list_item top = st_list.first(); top; top = st_list.succ(top))
  { node v = st_list[top];
    st_num[v] = st_count++;

    forall_inout_edges(e,v)
    { 
      if (used_edge[e]) continue;
      used_edge[e] = true;

      node w = G.opposite(v,e);

      if (v == w) continue;

      node_list P;
 
      if (tree_in[w] == e)          // unused tree edge (v,w)
         FINDPATH1(G,w,used_node,used_edge,low_out,P);
      else 
         if (dfsnum[v] < dfsnum[w]) // unused back edge (w,v)
            FINDPATH1(G,w,used_node,used_edge,tree_in,P);

      forall(w,P) st_list.insert(w,top);
    }
  }

 return s;

#endif
}



