//-------------------------------------------------------------------
//
//                          IOPoints.h
//
//    purpose: defines classes InOutPoint and IOPoints
//             implements management of in- and outpoints
//
//    author: Carsten Gutwenger
//    date:   july 1998
//-------------------------------------------------------------------


#ifndef _AGD_IOPOINTS_H
#define _AGD_IOPOINTS_H


#include <AGD/PlaneGraphCopy.h>

using namespace leda;


/********************************************************************
               representation of an in- or outpoint
********************************************************************/

struct InOutPoint {
	int  dx, dy;
	edge e;

	InOutPoint () {
		dx = dy = 0; e = NULL;
	}
	InOutPoint (edge e1) {
		e = e1; dx = dy = 0;
	}
	InOutPoint (const InOutPoint &iop) {
		e = iop.e; dx = iop.dx; dy = iop.dy;
	}
};

ostream& operator<< (ostream& os, const InOutPoint&);
istream& operator>> (istream& is, InOutPoint&);



/********************************************************************
             representation of in- and outpoint lists
********************************************************************/

class IOPoints {
private:
	node_map<leda::list<InOutPoint> > _in, _out;
	edge_map<InOutPoint *> the_point_of;
	edge_map<bool> mark;

public:

	// constructor, destructor
	IOPoints () { }
	IOPoints (const graph &G) : _in(G), _out(G), the_point_of(G,nil), 
                  mark(G,false), depth(G,0), height(G,0) { }

	~IOPoints () { }


	// length of in- or outpoint list
	int out (node v) const {
		return _out[v].length();
	}
	int in (node v) const {
		return _in[v].length();
	}

	// getting a const-reference to in- or outpoint list
	const list<InOutPoint>& inpoints (node v) const {
		return _in [v];
	}
	const list<InOutPoint>& outpoints (node v) const {
		return _out [v];
	}

	// getting a const-reference to an in- or outpoint
	const InOutPoint& inpoints (node v, list_item it) const {
		return _in [v] [it];
	}
	const InOutPoint& outpoints (node v, list_item it) const {
		return _out [v] [it];
	}

	// getting the in-/outpoint belonging to an edge
	const InOutPoint *point_of (edge e) const {
		return the_point_of [e];
	}

	// marking edges
	bool marked(edge e) const {
		return mark[e];
	}

	bool marked (const PlaneGraphCopy &PG,node v) {
		return (PG.outdeg(v) == 1 && marked(PG.first_adj_edge(v)));
	}

	// finding outpoints belonging to non-marked edges
	list_item first_real_out(node v) const {
		return search_real_forward(v,_out[v].first());
	}

	list_item last_real_out(node v) const {
		return search_real_backward(v,_out[v].last());
	}

	list_item next_real_out(node v, list_item it) const {
		return search_real_forward(v,_out[v].succ(it));
	}

	list_item prev_real_out(node v, list_item it) const {
		return search_real_backward(v,_out[v].pred(it));
	}

	// building in-/outpoint lists
	void append_inpoint (edge e) {
		node v = source(e);
		the_point_of [e] = &(_in[v] [_in[v].append (InOutPoint(e))]);
	}
	void append_outpoint (edge e) {
		node v = source(e);
		the_point_of [e] = &(_out[v] [_out[v].append (InOutPoint(e))]);
	}
	void push_inpoint (edge e) {
		node v = source(e);
		the_point_of [e] = &(_in[v] [_in[v].push (InOutPoint(e))]);
	}

	// setting relative coordinates
	void set_out_coord (node v, list_item it, int x, int y) {
		_out[v][it].dx = x;
		_out[v][it].dy = y;
	}
	void set_in_coord (node v, list_item it, int x, int y) {
		_in[v][it].dx = x;
		_in[v][it].dy = y;
	}

	void set_out_dx (node v, list_item it, int x) {
		_out[v][it].dx = x;
	}

	void restore_deg1_nodes(PlaneGraphCopy &PG);

	void change_edge (node v, edge e_new) {
		_out[v].Pop();
		append_inpoint (e_new);
	}

	// belongs v to a chain (= at most to non-marked incoming edges
	bool is_chain(node v) const {
		int i = 0;
		list_item it;
		forall_items(it,_in[v])
			if (!marked((_in[v])[it].e)) i++;
		return i <= 2;
	}

	// width of the left-/right side of an in-/outpoint list
	int out_left(node v) const {
		return (_out[v].empty()) ? 0 : (-_out[v].head().dx);
	}

	int out_right(node v) const {
		return (_out[v].empty()) ? 0 : (_out[v].tail().dx);
	}

	int in_left(node v) const {
		return (_in[v].empty()) ? 0 : (-_in[v].head().dx);
	}

	int in_right(node v) const {
		return (_in[v].empty()) ? 0 : (_in[v].tail().dx);
	}

	int max_left(node v) const {
		return leda_max (in_left(v), out_left(v));
	}

	int max_right(node v) const {
		return leda_max (in_right(v), out_right(v));
	}

	int max_plus_left(node v) const {
		return (in(v) >= 3) ?
			leda_max (in_left(v)+1, out_left(v)) : max_left(v);
	}

	int max_plus_right(node v) const {
		return (in(v) >= 3) ?
			leda_max (in_right(v)+1, out_right(v)) : max_right(v);
	}

	InOutPoint middle_neighbor(node z1) const;

	void num_deg1(node v, int &xl, int &xr,
		bool double_count) const;

	edge switch_begin_in(node v);
	edge switch_end_in  (node v);

	void switch_begin_out(node v);
	void switch_end_out  (node v);


	node_map<int> depth, height;

private:

	list_item search_real_forward (node v, list_item it) const;
	list_item search_real_backward(node v, list_item it) const;

};

#endif
