/*******************************************************************************
+
+  AGD 1.2
+
+
+  MixedModelBase.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                     MixedModelBase.cpp
//
//    author: Carsten Gutwenger
//    date:   september 1998
//-------------------------------------------------------------------


#include "MixedModelBase.h"


/********************************************************************
						  has_left, has_right

  Determine if the kth set in the ordered partition has a "real"
  left or right vertex, respectively.
********************************************************************/

bool MixedModelBase::has_left (int k) const
{
	const OrderSet &V = mmo[k];
	const list<InOutPoint> &L = iops.inpoints(V[1]);

	list_item it = L.first();
	return (it != nil && target(L[it].e) == mmo.left[k]);
}

bool MixedModelBase::has_right(int k) const
{
	const OrderSet &V = mmo[k];
	const list<InOutPoint> &L = iops.inpoints(V[V.len()]);

	list_item it = L.last();
	return (it != nil && target(L[it].e) == mmo.right[k]);
}


/********************************************************************
						  compute_order

  Computes the ordered partition (incl. left[k], right[k]) and
  constructs the in- and outpoint lists.
********************************************************************/

void MixedModelBase::compute_order(AugmentationModule &augmenter,
	bool prefer_nodes, double base_ratio)
{
	// remove (temporary) deg-1-nodes; determine on exterior face
	exterior_face = remove_deg1_nodes();

	// augment PG (temporary) to a biconnected planar graph
	list<edge> augm_edges;
	if (!Is_Biconnected(PG)) {
		augmenter.call (PG,augm_edges);

		// embed PG
		AGD::make_map (PG);
		PG.make_planar_map();
		exterior_face = nil;
	}
	
	if (!exterior_face) exterior_face = PG.max_face();

	// compute ordering of biconnected plane graph
	mmo.init(PG, co_biconnected, exterior_face, prefer_nodes, base_ratio);

	// restore deg1-nodes and mark incident edges
	iops.restore_deg1_nodes(PG);

	// compute in- and outpoint lists
	int k;
	for (k = 1; k <= mmo.length(); k++) {
		const OrderSet &V = mmo[k];
		for (int i = 1; i <= V.len(); i++) {
			node v = V[i];
			node cl = (i == 1)       ? V.left () : V[i-1];
			node cr = (i == V.len()) ? V.right() : V[i+1];

			edge e, er = nil, el = nil;
			forall_out_edges(e,v) {
				if (target(e) == cr) er = e;
				if (target(e) == cl) el = e;
			}
			// one of el and er is not nil by definition of the ordering
			if (!er) er = el;
			if (!el) el = er;

			e = er;
			do {
				if (exists(e))
					iops.push_inpoint(e);
				e = PG.cyclic_adj_pred(e);
			} while (iops.marked(e) || (mmo.rank(target(e)) <= k && e != er));

			for ( ; iops.marked(e) || mmo.rank(target(e)) > k; e = PG.cyclic_adj_pred(e))
				if (exists(e))
					iops.append_outpoint(e);

			// move In-/Outpoints to deg1-nodes to appropriate places
			el = iops.switch_begin_in(v);
			er = iops.switch_end_in  (v);

			// has a left- or right edge ?
			bool has_el = (el != nil);
			bool has_er = (er != nil);
			if (el && el == er) {
				has_el = (target(el) != cr);
				has_er = !has_el;
			}

			// determine left(k) and right(k)
			if (i == 1)       mmo.left [k] = (has_el) ? target(el) : cl;
			if (i == V.len()) mmo.right[k] = (has_er) ? target(er) : cr;

			int xl, xr;
			iops.num_deg1(v,xl,xr,has_el || has_er);

			int x = 0;
			if (!has_el) x += xl;
			if (!has_er) x += xr;

			int x_alpha = leda::max(0,leda::min(x, (iops.in(v)-iops.out(v)+1+2*x-2)/2));
			int x_beta = x - x_alpha;

			if (!has_el)
				for( ; x_beta > 0 && xl > 0; x_beta--, xl--)
					iops.switch_begin_out(v);
			if (!has_er)
				for( ; x_beta > 0 && xr > 0; x_beta--, xr--)
					iops.switch_end_out(v);
		}
	}

	// remove augmented edges
	edge e;
	forall(e,augm_edges) {
		PG.del_edge(PG.reversal(e));
		PG.del_edge(e);
	}
}


/********************************************************************
						remove_deg1_nodes

  Removes deg-1-nodes and store informations for restoring them.
********************************************************************/

face MixedModelBase::remove_deg1_nodes()
{
	node_array<bool> mark(PG,false);
	node v;

	// mark all deg-1-nodes we want to remove
	int n = PG.number_of_nodes();
	forall_nodes(v,PG) {
		if (n <= 3) break;
		if ((mark[v] = (PG.degree(v) == 2)) == true) {
			node w = target(PG.first_adj_edge(v));
			if (mark[w]) mark[w] = false; else n--;
		}
	}

	edge e_exterior = nil, e;
	if (exterior_face)
		forall_face_edges(e,exterior_face)
			if (!mark[source(e)] && !mark[target(e)]) {
				e_exterior = e; break;
			}

	PG.remove_deg1_nodes(mark);

	return (e_exterior) ? PG.face_of(e_exterior) : nil;
}


/********************************************************************
					  assign_iop_coords

  Computes the relative coordinates of the in- and outpoints,
  incl. height(v), depth(v).
********************************************************************/

void MixedModelBase::assign_iop_coords()
{
	for (int k = 1; k <= mmo.length(); k++) {
		const OrderSet &V = mmo[k];

		for (int i = 1; i <= V.len(); i++) {
			node v = V[i];

			bool only_left = (iops.in(v) == 2 && 
				i >= 2 && target(iops.inpoints(v).head().e) == V[i-1] &&
					iops.marked(iops.inpoints(v).tail().e));
			bool only_right = (iops.in(v) == 2 && 
				i < V.len() && target(iops.inpoints(v).tail().e) == V[i+1] &&
					iops.marked(iops.inpoints(v).head().e));

			if (iops.out(v) >= 1) {
				int out_l = 0, out_r = 0;

				int out_plus  = iops.out(v)/2;
				int out_minus = iops.out(v) - 1 - out_plus;
				int delta_l = 0, delta_r = 0;
				out_l = out_minus;

				if (iops.in(v) == 2) {
					delta_l = (only_right) ? 0 : 1;
					delta_r = (only_left)  ? 0 : 1;

				} else if (iops.in(v) >= 3) {
					delta_l = delta_r = 1;

				} else if (iops.in(v) == 1) {
					node vl = (i == 1) ? mmo.left [k] : V[i-1];
					if (target(iops.inpoints(v).head().e) != vl) {
						out_l = out_plus;
						delta_r = 1;

					} else {
						delta_l = 1;
					}
				}

				out_r = iops.out(v) - 1 - out_l;

				const list<InOutPoint> &opl = iops.outpoints(v);
				int j;
				list_item it = opl.first();
				for (j = 0; j < out_l; j++, it = opl.succ(it))
					iops.set_out_coord(v,it,-out_l+j,delta_l+j);
				iops.height[v] = leda::max(out_l+delta_l,out_r+delta_r)-1;
				if (iops.height[v] == 0 && iops.marked(opl[it].e))
					iops.height[v] = 1;
				iops.set_out_coord(v,it,0,iops.height[v]);
				for (j = 1, it = opl.succ(it); j <= out_r;
					j++, it = opl.succ(it))
					iops.set_out_coord(v,it,j,out_r+delta_r-j);
			}

			if (iops.in(v) <= 3) {
				const list<InOutPoint> &ipl = iops.inpoints(v);
				int in_v = iops.in(v);

				if (in_v == 3 || (in_v == 2 && !only_right)) {
					if (iops.marked(ipl.head().e))
						iops.set_in_coord(v,ipl.first(),-1,0);
				}
				if (in_v == 3 || (in_v == 2 && !only_left)) {
					if (iops.marked(ipl.tail().e))
						iops.set_in_coord(v,ipl.last(),  1,0);
				}
				if (in_v != 0 && (in_v != 2 || only_left || only_right)) {
					list_item it = ipl.first();
					if (in_v == 3 || (in_v == 2 && only_left))
						it = ipl.succ(it);
					if (iops.marked(ipl[it].e)) {
						iops.set_in_coord(v,it,0,-1);
						iops.depth[v] = 1;
					}
				}

			} else {
				int in_l = (iops.in(v)-3) / 2;
				int in_r = iops.in(v)-3 - in_l;

				int j;
				const list<InOutPoint> &ipl = iops.inpoints(v);
				list_item it = ipl.first();
				iops.set_in_coord(v,it,(in_l == 0 && iops.marked(ipl[it].e)) ? -1 : -in_l,0);
				for (j = 1, it = ipl.succ(it); j <= in_l; j++, it = ipl.succ(it))
					iops.set_in_coord(v,it,j-in_l-1,-j);
				iops.set_in_coord(v,it,0,-in_r);
				iops.depth[v] = in_r;  // inpoint with smallest y-coordinate
				for (j = 1, it = ipl.succ(it); j <= in_r; j++, it = ipl.succ(it))
					iops.set_in_coord(v,it,j,j-in_r-1);
				iops.set_in_coord(v,it,in_r,0);
			}
		}
	}
}


/********************************************************************
							place_nodes

  Implements the placement step. Computes x[v] and y[v].
********************************************************************/

void MixedModelBase::place_nodes()
{
	dyl = array<int>(2,mmo.length());
	dyr = array<int>(2,mmo.length());

	left_op  = array<list_item>(2,mmo.length());
	right_op = array<list_item>(2,mmo.length());

	next_left .init(PG);
	next_right.init(PG);
	dxla.init(PG,0);
	dxra.init(PG,0);

	compute_x_coords();
	compute_y_coords();
}


/********************************************************************
						  compute_x_coords

  Computes the absolute x-coordinates x[v] of all nodes in the
  ordering, furthermore dyla[k] and dyra[k] (used by compute_y_coordinates)
********************************************************************/

void MixedModelBase::compute_x_coords()
{
	node_map<int> &x = gl.x();

	int  k, i;
	node v;

	// representation of the contour
	node_array<node> prev(PG), next(PG);
	node_array<node> father(PG, nil);

	// maintaining of free space for shifting
	array<int> shift_space(1,mmo.length());
	for (k=1; k <= mmo.length(); k++)
		shift_space[k] = 0;

	node_array<int> comp(PG,0);

	forall_nodes(v,PG) {
		next_left [v] = iops.first_real_out(v);
		next_right[v] = iops.last_real_out (v);
	}

	// last_right[v] = last vertex of highest set with right vertex v
	node_array<node> last_right(PG,nil);
	for (k = 2; k <= mmo.length(); k++) {
		const OrderSet &V = mmo[k];
		last_right[mmo.right[k]] = V[V.len()];
	}

	node_array<int> high(PG,0);
	forall_nodes(v,PG) {
		InOutPoint op;
		forall(op,iops.outpoints(v))
			if (!iops.marked(op.e))
				high[v] = leda::max(mmo.rank(target(op.e)), high[v]);
	}

	// initialization
	const OrderSet &V1 = mmo[1];
	int p = V1.len();
	
	x[V1[1]] = iops.out_left(V1[1]);
	for (i = 2; i <= p; i++) {
		x[V1[i]] = iops.max_right(V1[i-1]) + iops.max_left(V1[i]) + 1;
	}

	for (i = 1; i <= p; i++) {
		if (i < p) next[V1[i]] = V1[i+1];
		if (i > 1) prev[V1[i]] = V1[i-1];
	}
	prev [V1[1]] = next [V1[p]] = nil;

	// main loop
	for (k = 2; k <= mmo.length(); k++) {
		// consider set Vk
		const OrderSet &Vk = mmo[k];
		p = Vk.len();
		node z1 = Vk[1];
		node cl = mmo.left[k];
		node cr = mmo.right[k];

		if (!has_left(k)) {
			while(last_right[cl] && high[cl] < k && !has_right(mmo.rank(last_right[cl])))
				cl = mmo.left[k] = last_right[cl];
		}

		// determine temporarily the x-coordinates of c_l+1,...,c_r relative to cl
		int sum = 0;
		for (v = next[cl]; v != cr; v = next[v]) {
			sum += x[v]; x[v] = sum;
		}
		x[cr] += sum;

		left_op [k] = next_right[cl];
		right_op[k] = next_left [cr];

		// compute dxl, dxr, dyl, dyr
		int dxl, dxr;
		dyl[k] = dyr[k] = 0;

		list_item it;
		if ((it = next_right[cl]) != nil)  {
			dxl = iops.outpoints(cl)[it].dx;
			if (target(iops.outpoints(cl)[it].e) != z1)
				dxl++;
			else
				next_right[cl] = iops.prev_real_out(cl,it);

			if (dxl < 0)
				dyl[k] = iops.height(cl);
			else if ((it = iops.outpoints(cl).succ(it)) != nil)
				dyl[k] = iops.outpoints(cl)[it].dy;
		} else {
			dxl = (iops.out(cl) == 0) ? 0 : -iops.out_left(cl);
		}
		if ((it = next_left[cr]) != nil) {
			dxr = iops.outpoints(cr)[it].dx;
			if (target(iops.outpoints(cr)[it].e) != Vk[p])
				dxr--;
			else
				next_left[cr] = iops.next_real_out(cr,it);

			if (dxr > 0)
				dyr[k] = iops.height(cr);
			else if ((it = iops.outpoints(cr).pred(it)) != nil)
				dyr[k] = iops.outpoints(cr)[it].dy;
		} else {
			dxr = (iops.out(cr) == 0) ? 0 : iops.out_right(cr);
		}

		dxla[Vk[1]] = dxl; dxra[Vk[p]] = dxr;

		int old_x_cr;

		// vertex case
		if (!iops.is_chain(z1)) {
			InOutPoint ip_ct = iops.middle_neighbor(z1);
			InOutPoint op_ct = *iops.point_of(PG.reversal(ip_ct.e));
			node ct = target(ip_ct.e);

			int delta = dxl + iops.max_plus_left(z1) + ip_ct.dx - (x[ct] + op_ct.dx);
			if (delta < 0) delta = 0;
			x[ct] += delta;

			int x_0 = x[ct] + op_ct.dx - ip_ct.dx;

			int sum = 0;
			for (v = prev[ct]; v != cl; v = prev[v]) {
				x[v] -= sum;
				if (next_right[v] && target(iops.outpoints(v,next_right[v]).e) == z1) {
					InOutPoint op_v = iops.outpoints(v,next_right[v]);
					InOutPoint ip_v = *iops.point_of(PG.reversal(op_v.e));
					int diff = x[v] + op_v.dx - x_0 - ip_v.dx;
					if (diff > 0) {
						sum += diff;
						x[v] -= diff;
					}
				}
			}

			for (v = next[cl]; v != next[ct]; v = next[v])
				x[v] += sum;
			x_0 += sum;

			sum += delta;

			for (v = next[ct]; v != next[cr]; v = next[v]) {
				x[v] += sum;
				if (next_right[v] && target(iops.outpoints(v,next_right[v]).e) == z1) {
					InOutPoint op_v = iops.outpoints(v,next_right[v]);
					InOutPoint ip_v = *iops.point_of(PG.reversal(op_v.e));
					int diff = x[v] + op_v.dx - x_0 - ip_v.dx;
					if (diff < 0) {
						sum -= diff;
						x[v] -= diff;
					}
				}
			}

			x[z1] = x_0;

			old_x_cr = x[cr] - x[z1];
			x[cr] = leda::max(old_x_cr, iops.max_plus_right(z1) - dxr);

			for (v = next[cl]; v != cr; v = next[v]) {
				x[v]      = x[v] - x[z1];
				father[v] = z1;
			}

		// chain case
		} else {
			int sum = x[z1] = iops.max_plus_left(z1) + dxl;
			for (i = 2; i <= p; i++)
				sum += (x[Vk[i]] = iops.max_right(Vk[i-1]) + iops.max_left(Vk[i]) + 1);

			old_x_cr = x[cr] - sum;
			int new_x_cr = iops.max_plus_right(Vk[p]) - dxr;
			x[cr] = leda::max(old_x_cr, new_x_cr);
			shift_space[k] = leda::max(0, old_x_cr - new_x_cr);

			for (v = next[cl]; v != cr; v = next[v]) {
				x[v]      = x[v] - x[z1];
				father[v] = z1;
			}
		}

		int need = x[cr] - old_x_cr;
		int k_cr = mmo.rank(cr);
		if (shift_space[k_cr] > 0) {
			int use = leda::min(shift_space[k_cr], need);
			shift_space[k_cr] -= use;
			comp[cr] += use;
			x[mmo.right[k_cr]] -= use;
		}

		// update contour after insertion of z1,...,zp
		for (i = 1; i <= p; i++) {
			if (i < p) next[Vk[i]] = Vk[i+1];
			if (i > 1) prev[Vk[i]] = Vk[i-1];
		}

		next [prev [z1]    = cl] = z1;
		prev [next [Vk[p]] = cr] = Vk[p];

	}

	// compute final x-coordinates for nodes on final contour
	int sum = 0;
	for (v = V1[1]; v != nil; v = next[v])
		x [v] = (sum += x[v]);

	// compute final x-coordinates for inner nodes
	for (k = mmo.length(); k >= 1; k--) {
		for (i = 1; i <= mmo.len(k); i++) {
			v = mmo(k,i);
			if (father[v] != nil) {
				x[v] = x[v] + x[father[v]] - comp[father[v]];
			}
		}
	}
}


/********************************************************************
							compute_y_coords

  Computes the absolute y-coordinates y[v] of all nodes in the
  ordering.
********************************************************************/

class SetYCoords {
public:
	SetYCoords(const graph &GG, const IOPoints &_iops, const MMOrder &_mmo,
		const node_map<int>& _x, const node_map<int>& _y) :
		G(GG), mmo(_mmo), iops(_iops), x(_x), y(_y) {
	}

	void init(int _k);

	void check_y_coord (int xleft, int xright, int ys, bool node_sep);
	void check_y_coord (int xs, int ys, bool node_sep ) {
		check_y_coord (xs, xs, ys, node_sep);
	}

	int get_ymax() const {
		return ymax;
	}

private:
	void get_next_region();
	node z (int j) const {
		return (*V)[j];
	}

	bool marked (edge e) const {
		return iops.marked(e);
	}

	const InOutPoint &inpoint(int j, list_item it) {
		return iops.inpoints(z(j),it);
	}

	const InOutPoint &outpoint(const InOutPoint &ip) {
		return *iops.point_of(G.reversal(ip.e));
	}

	void search_next_inpoint();

	const graph &G;
	const MMOrder &mmo;
	const IOPoints &iops;
	const node_map<int> &x;
	const node_map<int> &y;
	const OrderSet *V;

	int k;
	node cl, cr;

	int ymax, x_next, look_ahead_x, look_ahead_next_x;
	int i, i_next, delta_y, infinity;
	list_item it_ip, it_ip_next, it_look_ahead;
	bool on_base;
};

void SetYCoords::init(int _k)
{
	k = _k; V = &mmo[k];
	ymax = 0;
	look_ahead_x = 0;

	i = 0; it_ip = nil;
	cl = mmo.left[k]; cr = mmo.right[k];

	on_base = true; x_next = -1;
	infinity = x[cr] + iops.out_right(cr) + 1;

	search_next_inpoint();
	it_ip = it_ip_next; i = i_next;

	get_next_region();
}

void SetYCoords::search_next_inpoint()
{
	i_next = i; it_ip_next = it_ip;

	do {
		if (it_ip_next == nil) {
			if (++i_next > V->len()) {
				it_ip_next = nil;
				return;
			}
			it_ip_next = iops.inpoints(z(i_next)).first();
		} else {
			it_ip_next = iops.inpoints(z(i_next)).succ(it_ip_next);
		}
	} while (it_ip_next == nil || inpoint(i_next,it_ip_next).dy == 0);

	if (it_ip_next && iops.marked(inpoint(i_next,it_ip_next).e)) {
		int ip_x = x[z(i_next)] + inpoint(i_next,it_ip_next).dx;

		if (look_ahead_x <= ip_x) {
			for (it_look_ahead = it_ip_next;
				inpoint(i_next,it_look_ahead).dx < 0 && iops.marked(inpoint(i_next,it_look_ahead).e);
				it_look_ahead = iops.inpoints(z(i_next)).succ(it_look_ahead)) ;
			const InOutPoint &ip_look_ahead = inpoint(i_next,it_look_ahead);
			look_ahead_x = x[z(i_next)] + ip_look_ahead.dx;
			if (ip_look_ahead.dx < 0) {
				look_ahead_next_x = x[target(ip_look_ahead.e)] + outpoint(ip_look_ahead).dx;
			} else {
				look_ahead_next_x = look_ahead_x;
			}
		}

		if (look_ahead_next_x <= ip_x) {
			it_ip_next = it_look_ahead;
		}
	}

}

void SetYCoords::get_next_region()
{
	int x_old = x_next;

	do {
		if (on_base) {
			delta_y = 0;
			if (it_ip == nil) {
				x_next = infinity;
			} else {
				const InOutPoint &ip = inpoint(i,it_ip);
				x_next = (marked(ip.e)) ? (x[z(i)] + ip.dx) :
					(x[target(ip.e)] + outpoint(ip).dx);
			}
			on_base = (i_next != i);

		} else {
			const InOutPoint &ip = inpoint(i,it_ip);
			delta_y = -ip.dy;
			search_next_inpoint();
			if (it_ip_next && ip.dx < 0) {
				const InOutPoint &ip_next = inpoint(i_next,it_ip_next);
				x_next = (marked(ip_next.e)) ? (x[z(i)] + ip_next.dx) :
					(x[target(ip_next.e)] + outpoint(ip_next).dx);
			} else {
				x_next = (marked(ip.e)) ? (x[z(i)] + ip.dx + 1) :
					(x[target(ip.e)] + outpoint(ip).dx + 1);
			}

			on_base = (i_next != i);
			i = i_next; it_ip = it_ip_next;
		}
	} while (x_next <= x_old);
}

void SetYCoords::check_y_coord (int xleft, int xright, int ys, bool node_sep)
{
	while (x_next <= xleft)
		get_next_region();

	int max_dy = delta_y;

	while (x_next <= xright) {
		get_next_region();
		if (delta_y > max_dy) max_dy = delta_y;
	}

	if (node_sep && max_dy == 0)
		max_dy = 1;

	if (ys + max_dy > ymax)
		ymax = ys + max_dy;
}


void MixedModelBase::compute_y_coords()
{
	node_map<int> &x = gl.x(), &y = gl.y();

	int  k, i;

	// representation of the contour
	node_array<node> prev(PG), next(PG);

	// initialization
	SetYCoords set_y(PG,iops,mmo,x,y);

	const OrderSet &V1 = mmo[1];
	int p = V1.len();
	
	for (i = 1; i <= p; i++) {
		if (i < p) next[V1[i]] = V1[i+1];
		if (i > 1) prev[V1[i]] = V1[i-1];
	}
	prev [V1[1]] = next [V1[p]] = nil;

	// main loop
	for (k = 2; k <= mmo.length(); k++) {
		// consider set Vk
		const OrderSet &Vk = mmo[k];
		p = Vk.len();
		node cl = mmo.left[k];
		node cr = mmo.right[k];

		set_y.init(k);

		for (node v = cl; v != next[cr]; v = next[v]) {
			list_item it_first, it_last;

			const list<InOutPoint> &out = iops.outpoints(v);

			if (v == cl) {
				if ((it_first = left_op[k]) == nil)
					it_first = out.first();
				else if (target(out[it_first].e) != Vk[1])
					it_first = out.succ(it_first);

				for (it_last = it_first; it_last && (iops.marked(out[it_last].e) ||
					target(out[it_last].e) == Vk[1]); it_last = out.succ(it_last)) ;

			} else if (v == cr) {
				it_last = right_op[k];
				if (it_last && target(out[it_last].e) == Vk[p])
					it_last = out.succ(it_last);
				it_first = (it_last) ? out.pred(it_last) : out.last();

				while (it_first && (iops.marked(out[it_first].e) ||
					target(out[it_first].e) == Vk[p]))
					it_first = out.pred(it_first);
				it_first = (it_first) ? out.succ(it_first) : out.first();

			} else {
				it_first = next_left[v];
				it_first = (it_first) ? out.pred(it_first) : out.last();

				while (it_first && iops.marked(out[it_first].e))
					it_first = out.pred(it_first);

				it_first = (it_first) ? out.succ(it_first) : out.first();

				if (next_right[v] && next_left[v] == next_right[v]) {
					for (it_last = out.succ(next_right[v]); it_last && iops.marked(out[it_last].e);
						it_last = out.succ(it_last)) ;
				} else {
					it_last = next_left[v];
				}
			}
			if (v != cr && it_first != it_last && mmo.rank(next[v]) > mmo.rank(v)) {
				int x_n_v = x[next[v]]-iops.out_left(next[v]);
				list_item it = (it_last) ? out.pred(it_last) : out.last();
				while (true) {
					if (x[v]+out[it].dx >= x_n_v)
						it_last = it;
					else break;
					if (it == it_first) break;
					it = out.pred(it);
				}
			}

			if (v != cl) {
				int xl = x[prev[v]], xr = x[v];
				int r_p_v = mmo.rank(prev[v]), r_v = mmo.rank(v);

				if (r_p_v >= r_v) {
					if (iops.out(prev[v]) > 0)
						xl += 1+iops.out_right(prev[v]);
				} else {
					xl += dxla[v];
				}
				if (r_p_v <= r_v) {
					if (iops.out(v) > 0)
						xr -= 1+iops.out_left(v);
				} else {
					xr += dxra[prev[v]];
				}

				if (xl <= xr)
					set_y.check_y_coord (xl, xr, 1+leda::max(y[prev[v]],y[v]), false);
			}

			for (list_item it = it_first; it != it_last; it = out.succ(it)) {
				const InOutPoint &op = out[it];

				if (iops.marked(op.e))
					set_y.check_y_coord(x[v]+op.dx, y[v]+op.dy+1, false);
				else
					set_y.check_y_coord(x[v]+op.dx, y[v]+op.dy,
						(op.dx == 0 && op.dy == 0 && x[v] == x[target(op.e)]));
			}

		}


		for (i = 1; i <= p; i++) {
			y[Vk[i]] = set_y.get_ymax();
		}

		// update contour after insertion of z1,...,zp
		for (i = 1; i <= p; i++) {
			if (i < p) next[Vk[i]] = Vk[i+1];
			if (i > 1) prev[Vk[i]] = Vk[i-1];
		}

		next [prev [Vk[1]] = cl] = Vk[1];
		prev [next [Vk[p]] = cr] = Vk[p];
	}
}


/********************************************************************
							set_bends

  Assigns polylines to edges of the original graph and computes the
  x- and y-coordinates of deg-1-nodes not in the ordering.
********************************************************************/

void MixedModelBase::set_bends ()
{
	node_map<int> &x = gl.x(), &y = gl.y();
	edge_map<IPolyline> &bends = gl.bends();

	for(int k = 1; k <= mmo.length(); k++)
		for (int i = 1; i <= mmo[k].len(); i++) {
			node v_s = mmo(k,i);
			edge e;
			forall_out_edges(e,v_s) {
				node v_t = target(e);
				const InOutPoint &p_s = *iops.point_of(e);
				if (iops.marked(e)) {
					x[v_t] = x[v_s] + p_s.dx;
					y[v_t] = y[v_s] + p_s.dy;

				} else /*if (PG.original(e))*/ {
					const InOutPoint &p_t = *iops.point_of(PG.reversal(e));
					IPoint p1 (x[v_s] + p_s.dx, y[v_s] + p_s.dy);
					IPoint p2 (x[v_t] + p_t.dx, y[v_t] + p_t.dy);

					bends[e].append (p1);
					if (mmo.rank(v_s) < mmo.rank(v_t)) {
						bends[e].append (IPoint(p1.xcoord(),p2.ycoord()));
					} else {
						bends[e].append (IPoint(p2.xcoord(),p1.ycoord()));
					}
					bends[e].append (p2);
				}
			}
		}
}


/********************************************************************
							postprocessing1

  Tries to reduce the number of bends by changing the outpoints of
  nodes with indeg and outdeg 2.
********************************************************************/

void MixedModelBase::postprocessing1()
{
	node_map<int> &x = gl.x();

	int k;

	for (k = 2; k <= mmo.length(); k++) {
		const OrderSet &V = mmo[k];
		node v = V[V.len()];

		if (iops.in(v) != 2 || iops.out(v) != 2) continue;

		const list<InOutPoint> &in  = iops.inpoints (v);
		const list<InOutPoint> &out = iops.outpoints(v);
		edge el = in[in.first()].e;
		edge er = in[in.last ()].e;

		if (!iops.marked(el) && !iops.marked(er) &&
			x[target(el)] + iops.point_of(PG.reversal(el))->dx < x[v] &&
			x[target(er)] + iops.point_of(PG.reversal(er))->dx == x[v]+1 &&
			gl.y(target(er)) < gl.y(v))
		{
			x[v] += 1;
			iops.set_out_dx(v,out.first(),-1);
			iops.set_out_dx(v,out.last (), 0);
		}
	}
}


/********************************************************************
							postprocessing2

  Tries to reduce the number of bends by moving degree-2 nodes on
  bend points.
********************************************************************/

void MixedModelBase::_first_point(int &x, int &y, edge e)
{
	if (gl.bends(e).empty()) {
		x = gl.x(target(e)); y = gl.y(target(e));
	} else {
		x = gl.bends(e).head().xcoord();
		y = gl.bends(e).head().ycoord();
	}
}

bool MixedModelBase::_is_redundant(int x1, int y1, int x2, int y2, int x3, int y3)
{
	int dzy1 = x3 - x2;
	int dzy2 = y3 - y2;
	int dyx1 = x2 - x1;

	if (dzy1 == 0) return (dyx1 == 0);

	int f = dyx1 * dzy2;

	return (f % dzy1 == 0 && (y2 - y1) == f / dzy1);
}

void MixedModelBase::postprocessing2()
{
	gl.compact_all_bends();

	node v;
	forall_nodes(v,PG) {
		if (PG.outdeg(v) != 2) continue;

		edge e1 = PG.first_adj_edge(v);
		edge e2 = PG. last_adj_edge(v);

		IPolyline &bends1 = gl.bends(e1);
		IPolyline &bends2 = gl.bends(e2);

		if (bends1.empty() && bends2.empty()) continue;

		int x1,y1,x3,y3;
		_first_point(x1,y1,e1);
		_first_point(x3,y3,e2);

		if (_is_redundant(x1,y1,gl.x(v),gl.y(v),x3,y3)) {
			if (!bends1.empty()) {
				gl.x(v) = x1; gl.y(v) = y1;
				bends1.pop_front();
				gl.bends(PG.reversal(e1)).pop_back();
			} else {
				gl.x(v) = x3; gl.y(v) = y3;
				bends2.pop_front();
				gl.bends(PG.reversal(e2)).pop_back();
			}
		}
	}
}


/********************************************************************
						  debugging output
********************************************************************/

void MixedModelBase::printMMOrder (ostream &os)
{
	int k, i;

	os << "left and right:\n\n";
	for (k = 1; k <= mmo.length(); k++) {
		const OrderSet &V = mmo[k];
		
		os << k << ": { ";
		for (i = 1; i <= V.len(); i++)
			os << index(V[i]) << " ";
		os << "};";
		if (k >= 2)
			os << " cl = " << index(mmo.left[k]) <<
				", cr = " << index(mmo.right[k]);
		os << endl;

	}
	os.flush();
}

void MixedModelBase::printInOutPoints (ostream &os)
{
	node v;

	os << "\n\nin- and outpoint lists:\n";
	forall_nodes(v,PG) {
		const list<InOutPoint> &in  = iops.inpoints (v);
		const list<InOutPoint> &out = iops.outpoints(v);

		InOutPoint iop;
		os << "\n" << index(v) << ":\n";
		os << "  outpoints: ";
		forall (iop,out) {
			print(os,iop);
			os << " ";
		}
		os << "\n  inpoints:  ";
		forall (iop,in) {
			print(os,iop);
			os << " ";
		}
	}
	os << endl;
}

void MixedModelBase::print(ostream &os, const InOutPoint &iop)
{
	if (iop.e)
		os << "[(" << index(PG.original(source(iop.e))) << "," <<
			index(PG.original(target(iop.e))) << ")," <<
			iop.dx << "," << iop.dy << "]";
	else
		os << "[ ]";
}

void MixedModelBase::printNodeCoords (ostream &os)
{
	node v;

	os << "\nx- and y-coordinates:\n\n";
	forall_nodes(v,PG)
		os << index(v) << ": (" << gl.x(v) << "," << gl.y(v) << ")\n";
}
