/*******************************************************************************
+
+  AGD 1.2
+
+
+  CompactionPair.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



//#line 123 "Lweb/CompactionPair.lw"
#include <AGD/impl/CompactionPair.h>
#include <LEDA/graph/templates/shortest_path.h>
#include <LEDA/geo/interval_set.h>
#include <LEDA/graph/face_array.h>

#include <LEDA/system/assert.h> // for debug

using namespace leda;

CompactionPair::CompactionPair(const OrthogonalRepresentation &H) {

  PP = &H;

  _dir.init(*PP, unexplored_); 
  edge e; forall_edges(e, *PP) _dir[e] = H.dir(e);
  
  _init();

  create_basic_arcs(N_);
  create_basic_arcs(W_);
 

}
  

void CompactionPair::_init() {
  
  node n;
  
  _hor.init(*PP); _ver.init(*PP);
  _weight.init(DDxy, 0.0);
  _coord.init(DDxy, 0.0);
  left_.init(DDxy); right_.init(DDxy); 
  bottom_.init(DDxy); top_.init(DDxy);

  _n_type.init(DDxy, unknown_node_type);
  _n_dir.init(DDxy, unknown_node_dir);
  _members.init(DDxy); 
  
  create_nodes(vertical);
  create_nodes(horizontal);

  forall_nodes(n, DDxy) {
    left_[n] = _ver[_members[n].head()];
    right_[n] = _ver[_members[n].tail()];
    bottom_[n] = _hor[_members[n].head()];
    top_[n] = _hor[_members[n].tail()];
  }
    
  _a_type.init(DDxy, unknown_arc); 
  
}


//#line 189 "Lweb/CompactionPair.lw"
CompactionPair::CompactionPair(GridLayout &L) {

  PP = &L.get_graph();

  _dir.init(*PP, unexplored_); 

  //obacht: bei knotengroessen und 0-grad-winkeln?
  edge e; forall_edges(e, *PP) {
    node v = source(e), w = target(e);
    if (L.x(v) < L.x(w) && L.y(v) == L.y(w)) { _dir[e] = N_; continue; }
    if (L.x(v) > L.x(w) && L.y(v) == L.y(w)) { _dir[e] = S_; continue; }
    if (L.x(v) == L.x(w) && L.y(v) < L.y(w)) { _dir[e] = W_; continue; }
    if (L.x(v) == L.x(w) && L.y(v) > L.y(w)) { _dir[e] = E_; continue; }
  }

  _init();

  node n;

  forall_nodes(n, DDxy) 
    _coord[n] = ((_n_dir[n] == vertical) ? 
                 L.x(_members[n].head()) :  
                 L.y(_members[n].head()));

  create_basic_arcs(N_);
  create_basic_arcs(W_);

  create_visibility_arcs(L, vertical);
  create_visibility_arcs(L, horizontal);

  _removeDoubleArcs();
  
}


//#line 232 "Lweb/CompactionPair.lw"
void CompactionPair::create_nodes(node_dir d) {
  
  node_array<bool> marked(*PP, false);
  node v; forall_nodes(v, *PP) if (!marked[v]) {
    node n = DDxy.new_node(); 
    _n_type[n] = graph_segment;
    _n_dir[n] = d;
    _members[n].push_back(v);
    add_vertices(n, v, d, marked);
  }
}
 

//#line 251 "Lweb/CompactionPair.lw"
void CompactionPair::add_vertices(node n, node v, node_dir d, 
                                  node_array<bool> &marked) {

  marked[v] = true;

  if (d == horizontal) _hor[v] = n; else _ver[v] = n;
  direction outdir = (d == horizontal) ? N_ : W_;
  direction indir = (d == horizontal) ? S_ : E_;
  
  edge e; 
 
  forall_out_edges(e, v) if (!marked[target(e)]) {
    if (_dir[e] == outdir) {
      _members[n].push_back(target(e));
      add_vertices(n, target(e), d, marked);
    }
    if (_dir[e] == indir) {
      _members[n].push_front(target(e));
      add_vertices(n, target(e), d, marked);
    }
  }

  forall_in_edges(e, v) if (!marked[source(e)]) {
    if (_dir[e] == outdir) {
      _members[n].push_front(source(e));
      add_vertices(n, source(e), d, marked);
    }
    if (_dir[e] == indir) {
      _members[n].push_back(source(e));
      add_vertices(n, source(e), d, marked);
    }
  }

}


//#line 290 "Lweb/CompactionPair.lw"
void CompactionPair::create_basic_arcs(direction d) {

  edge e; forall_edges(e, *PP) if (_dir[e] == d) {

    node i = (d == W_) ? _hor[source(e)] : _ver[source(e)],
      j = (d == W_) ? _hor[target(e)] : _ver[target(e)];
    
    edge a = DDxy.new_edge(i, j);
    _weight[a] = 1.0; 
    _a_type[a] = basic_arc;
  }
    
}


//#line 309 "Lweb/CompactionPair.lw"
void CompactionPair::create_visibility_arcs(const GridLayout &L, 
                                            node_dir d) {

  interval_set<node> I; node n;
  
  node_array<int> fix(DDxy);
  node_array<double> hi(DDxy, -MAXDOUBLE), lo(DDxy, MAXDOUBLE);

  forall_nodes(n, DDxy) {
    if (_n_dir[n] == vertical) {
      fix[n] = L.x(_members[n].head());
      lo[n] = L.y(_members[n].head()); 
      hi[n] = L.y(_members[n].tail());
    } else { 
      fix[n] = L.y(_members[n].head());
      lo[n] = L.x(_members[n].head()); 
      hi[n] = L.x(_members[n].tail());
    }
  }

  DDxy.bucket_sort_nodes(fix);
  
  forall_nodes(n, DDxy) if (_n_dir[n] == d) { 
    
    list<is_item> L_int = I.intersection(lo[n], hi[n]);
    list<triple> L_upd; // to update the data structure after having 
    // processed all segments on one sweep line
    
    is_item it; forall(it, L_int) {
      
      double it_lo = I.left(it), it_hi = I.right(it); 
      node n_it = I.inf(it); I.del_item(it); 
      
      if (it_hi > hi[n]) L_upd.append(triple(hi[n], it_hi, n_it));
      if (it_lo < lo[n]) L_upd.append(triple(it_lo, lo[n], n_it)); 
      
      edge a = DDxy.new_edge(n_it, n);
      _weight[a] = 1.0; 
      _a_type[a] = visibility_arc;
      
    }
    
    triple t; 
    forall(t, L_upd) I.insert(t.first(), t.second(), t.third());
    
    I.insert(lo[n], hi[n], n);
    
  }
}

void CompactionPair::_removeDoubleArcs() {
  list<edge> arcList;
  Is_Simple(DDxy, arcList);
  edge a, a1, a_ref = arcList.head();
  list<edge> zeroes, ones;

  forall(a, arcList) {
    if (source(a_ref) == source(a) && target(a_ref) == target(a)) {
      if (_a_type[a] == visibility_arc) zeroes.push(a);
      if (_a_type[a] == basic_arc) ones.push(a);
    } else {
      forall(a1, ones) if (!zeroes.empty()) 
        DDxy.del_edge(zeroes.pop());

      a_ref = a;
      zeroes.clear(); ones.clear();
      if (_a_type[a] == visibility_arc) zeroes.push(a);
      if (_a_type[a] == basic_arc) ones.push(a);
    }

  }
  
}


//#line 389 "Lweb/CompactionPair.lw"
void CompactionPair::compute_coords(node_dir d) {

  node n, s = DDxy.new_node();
  edge a; 

  forall_nodes(n, DDxy) if (_n_dir[n] == d && indeg(n) == 0) 
    _weight[DDxy.new_edge(s, n)] = 0.0;
  
  node_array<edge> pred(DDxy);
  edge_array<double> w(DDxy);
  forall_edges(a, DDxy) w[a] = -_weight[a];
  node_array<double> l(DDxy, 0.0);

  bool result = SHORTEST_PATH_T(DDxy, s, w, l, pred);
  if (!result) error_handler(1, "negative cycle reachable from s");

  DDxy.del_node(s);

  forall_nodes(n, DDxy) if (_n_dir[n] == d) _coord[n] = -l[n];

  
}

void CompactionPair::compute_coords_properly(node_dir d) {

  edge a, ac; node n, i;

  forall_nodes(n, DDxy) 
    if (_n_dir[n] != d) DDxy.hide_node(n);

  PlaneGraphCopy DC(DDxy); 

  DDxy.restore_all_nodes(); DDxy.restore_all_edges();

  graph N; node_map<face> N2D(N); face_array<node> D2N(DC);

  face f; forall_faces(f, DC) {
    i = N.new_node();
    N2D[i] = f; D2N[f] = i;
  }

  edge_map<edge> A2E(N);

  forall_edges(ac, DC) if (DC.original(ac)) {
    a = N.new_edge(D2N[DC.face_of(ac)], D2N[DC.face_of(DC.reversal(ac))]);
    A2E[a] = ac;
  }

  node_array<int> b(N, 0);
  edge_array<int> cost(N), u(N), l(N), flow(N);
  forall_edges(a, N) { // sollten eigentlich doubles sein
    edge o = DC.original(A2E[a]);
    cost[a] = (_a_type[o] == basic_arc) ? 1 : 0;
    l[a] = (int) rint(_weight[o]); 
    //u[a] = MAXINT;   // obacht: muss nicht so gross sein
    u[a] = 1000;   // obacht: muss nicht so gross sein
  }

 /* modification done by Thomas Ziegler, date: 13-Nov-2000
    problem: MIN_COST_FLOW() does not work for loops with 
	lower bounds.
	We set the flow on loops manually to one and hide the 
	loops befor calling MIN_COST_FLOW()*/
  list<edge> loops;
  forall_edges(a,N) {
	  if (source(a)==target(a)) {
		  loops.append(a); /* flow[a] = 1; */ N.hide_edge(a);
	  }
  }
  //flow on loops is one, and loops are hidden

  if (!N.empty()) {
    bool result = MIN_COST_FLOW(N, l, u, cost, b, flow);
    if (!result) error_handler(1, "CompactionPair::compute_coords_properly. "
                             "No feasible flow.");
  }

  N.restore_all_edges();  //make loops visible again

/*
	modification by Sven Thiel (14-Jun-2004):
	It seems that the call to MIN_COST_FLOW sets the flow of the hidden edges
	to zero. So we have to move the statement "flow[a]=1" in the modification
	by T. Ziegler AFTER the call to MIN_COST_FLOW!
*/
  forall(a, loops) flow[a] = 1;

  edge_array<double> length(DC);
  forall_edges(a, N)
	  length[A2E[a]] = (double) flow[a];

  node_array<bool> processed(DC, false);
  node s = DC.first_node();

  process_node(s, 0.0, DC, length, processed);

}


//#line 469 "Lweb/CompactionPair.lw"
void CompactionPair::process_node(node n, double c, 
                                  const PlaneGraphCopy &DC,
                                  const edge_array<double> &length,
                                  node_array<bool> &processed) {
  if (processed[n]) return;

  processed[n] = true; _coord[DC.original(n)] = c;
  
  edge a; 
  forall_out_edges(a, n) if (DC.original(a)) 
    process_node(target(a), c + length[a], DC, length, processed);
  forall_in_edges(a, n) if (DC.original(a))
    process_node(source(a), c - length[a], DC, length, processed);
}


//#line 486 "Lweb/CompactionPair.lw"
void CompactionPair::print(ostream &o) const {
  
  node v, n; edge a;

  o << "pertinent graph" << endl;
  ((graph)(*PP)).print();
  forall_nodes(v, *PP) {
    PP->print_node(v, o); o << "\t";
    DDxy.print_node(_hor[v]); o << "\t";
    DDxy.print_node(_ver[v]); cout << endl;
  }

  o << "D_x_y" << endl;

  forall_nodes(n, DDxy) {
    forall_out_edges(a, n) {
      DDxy.print_edge(a); 
      cout << "\t" << _weight[a] << endl;
    }
    cout << endl;
  }

  forall_nodes(n, DDxy) {
    DDxy.print_node(n, o); o << "\t";
    DDxy.print_node(left_[n], o); o << " ";
    DDxy.print_node(right_[n], o); o << " ";
    DDxy.print_node(bottom_[n], o); o << " ";
    DDxy.print_node(top_[n], o); o << "\t";
    node v; cout << " ("; forall(v, _members[n]) PP->print_node(v); 
    cout << ")" << endl;
  }
}

