/*******************************************************************************
+
+  AGD 1.2
+
+
+  TopologyModule.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



//#line 178 "Lweb/shape.lw"
#include <AGD/TopologyModule.h>
#include <LEDA/graph/plane_graph_alg.h>

using namespace leda;

const int TopologyModule::in_graph = 1;
const int TopologyModule::in_layout = 2;
const int TopologyModule::out_plan_rep = 1;

TopologyModule::TopologyModule() : AGDModule(2, 1) {
  add_precondition(key::planar, in_graph);
  add_precondition(key::connected, in_graph);
}


//#line 199 "Lweb/shape.lw"
double TopologyModule::angle(DPoint p, DPoint q, DPoint r) {

  //if (q == r) return 2 * LEDA_PI;

  double dx1 = q.xcoord() - p.xcoord(), dy1 = q.ycoord() - p.ycoord();
  double dx2 = r.xcoord() - p.xcoord(), dy2 = r.ycoord() - p.ycoord();

  if ((dx1 == 0 && dy1 == 0) || (dx2 == 0 && dy2 == 0)) 
    error_handler(1, string("Two vertices on the same place: (%.2f, %.2f)", 
			   p.xcoord(), p.ycoord()));
  
  double norm  = (dx1*dx1+dy1*dy1)*(dx2*dx2+dy2*dy2);
  
  double cosphi = (dx1*dx2+dy1*dy2) / sqrt(norm);

  if (cosphi >=  1.0 ) return 0;
  if (cosphi <= -1.0 ) return LEDA_PI;
  
  double phi = acos(cosphi);

  if (dx1*dy2 < dy1*dx2) phi = -phi;

  if (phi < 0) phi += 2*LEDA_PI;

  return phi;
}

bool TopologyModule::check(const graph& G, AgdKey &p) const {
  return check_graph(in_graph, G, p);
}

PlaneGraphCopy &TopologyModule::call(const graph &G, const LayoutInterface &L,
	face &f_0, bool preserveMap) {
  PlaneGraphCopy *PG = new PlaneGraphCopy(G,preserveMap);
  edge_array<DPolyline> bends(*PG); get_bends_in_copy(*PG, L, bends);
  f_0 = (get_planar_representation(*PG, bends)) ? 
    get_outer_face(*PG, bends) : nil;
  if (f_0 == nil) PG->make_planar_map();
  return *PG;
}

PlaneGraphCopy &TopologyModule::call(const graph &G, 
	const LayoutInterface &L, bool preserveMap) {

  PlaneGraphCopy *PG = new PlaneGraphCopy(G,preserveMap);
  edge_array<DPolyline> bends(*PG); get_bends_in_copy(*PG, L, bends);
  if (get_planar_representation(*PG, bends) == false)
	  PG->make_planar_map();
  return *PG;  

}

AGDModule *TopologyModule::clone() const {
  TopologyModule *T = new TopologyModule();
  return T;
}


//#line 261 "Lweb/shape.lw"
bool TopologyModule::get_planar_representation(graph &PG, 
                       const edge_array<DPolyline> &bends) {
				     

  // graph &G = L.get_graph();
  //G.make_map(M);
  
  //L.refresh();

  //edge e; forall(e, M) {
  //DPolyline l = L.get_bend_points(G.reversal(e));
  //l.reverse();
  //L.set_bend_points(e, l);
  //}

  edge_array<double> dx(PG), dy(PG);

  edge e; forall_edges(e, PG) {
    DPolyline B = bends[e];
    DPoint ps = B[B.get_item(0)];
    DPoint pt = B[B.get_item(1)];
    dx[e] = pt.xcoord() - ps.xcoord();
    dy[e] = pt.ycoord() - ps.ycoord();
  }
//     DPoint ps = L.get_position(PG.original(source(e)));
//     DPoint pt = L.get_position(PG.original(target(e)));
//     if (!L.get_bend_points(e).empty()) pt = L.get_bend_points(e).head();
//     dx[PG.chain(e).head()] = pt.xcoord() - ps.xcoord();
//     dy[PG.chain(e).head()] = pt.ycoord() - ps.ycoord();
//   }

  bool result = SORT_EDGES(PG, dx, dy);

  return result;
  
}


//#line 305 "Lweb/shape.lw"
face TopologyModule::get_outer_face(graph &PG, 
				    const edge_array<DPolyline>& bends) {

  PG.compute_faces();

  face f; forall_faces(f, PG) {
    
    DPolyline l; double rho = 0;
    edge e; DPolyline B; 
    forall_face_edges(e, f) {
      B = bends[e]; B.Pop(); // we do not want points twice  
      l.conc(B);
    }
    
    list_item it; forall_items(it, l) {
      DPoint p = l[it], r = l[l.cyclic_pred(it)], q = l[l.cyclic_succ(it)];
      rho += angle(p, r, q) - LEDA_PI;
    }

    //    if (rho < 0) return f;
    //cout << "rho: " << rho << endl;
    if (rho < 0) return f;
  
  }

  return nil;

}



void TopologyModule::get_bends_in_copy(const PlaneGraphCopy &PG,
				       const LayoutInterface &L,
				       edge_array<DPolyline>& bends) {

  edge e; forall_edges (e, PG) {
    
    if (!PG.original(e)) continue;
    bends[e] += L.get_position(PG.original(source(e)));
    DPolyline l = L.get_bend_points(PG.original(e));
    bends[e].conc(l);
    bends[e] += L.get_position(PG.original(target(e)));
  }
  
  forall_edges (e, PG) {
    if (PG.original(e)) continue;
    DPolyline l = bends[PG.reversal(e)];
    l.reverse();
    bends[e] = l;
  }
} 
  

face TopologyModule::call(graph &G, LayoutInterface &L,
	list<edge> &R) {

	list<edge> edges = G.all_edges();
	R.clear(); G.make_map(R); L.refresh();

	edge_array<DPolyline> bends(G);

	edge e;
	forall(e, edges) {
		bends[e] += L.get_position(source(e));
		DPolyline l = L.get_bend_points(e);
		bends[e].conc(l);
		bends[e] += L.get_position(target(e));
	}

	forall(e,R) {
		bends[e] = bends[G.reversal(e)];
		bends[e].reverse();
	}

	if (get_planar_representation(G, bends))
		return get_outer_face(G, bends);
	else {
		G.make_planar_map();
		return nil;
	}
}
