/*******************************************************************************
+
+  AGD 1.2
+
+
+  SPQRTree.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/



#include <AGD/AGD.h>
#include <AGD/impl/misc.h>
#include <AGD/PlanarSPQRTree.h>
#include "TricComp.h"
#include <AGD/GraphCopy.h>
#include <LEDA/core/tuple.h>


//-------------------------------------------------------------------
//                            Skeleton
//-------------------------------------------------------------------

Skeleton::Skeleton(const SPQRTree *T, node vT) : _owner(T), _tree_node(vT),
	M(1,2)
{
	_orig.use_node_data(M,nil);
	_real.use_edge_data(M,nil);
	_tree_edge.use_edge_data(M,nil);
}


edge Skeleton::twin_edge (edge e) const
{
	edge et = _tree_edge[e];
	if (et == nil)
		return nil;

	edge e_src = _owner->sk_edge_src[et];
	edge e_tgt = _owner->sk_edge_tgt[et];

	if (source(et) == _tree_node)
		return (e == e_src) ? e_tgt : _owner->sk[target(et)]->M.reversal(e_tgt);
	else
		return (e == e_tgt) ? e_src : _owner->sk[source(et)]->M.reversal(e_src);
}


node Skeleton::twin_tree_node (edge e) const
{
	edge et = _tree_edge[e];
	if (et == nil)
		return nil;
	return (source(et) == _tree_node) ? target(et) : source(et);
}


//-------------------------------------------------------------------
//                            SPQRTree
//-------------------------------------------------------------------

void SPQRTree::init(const graph &G, edge e_ref, bool is_map)
{
	_cpV =nil; _cpE = nil;

	TricComp tricComp(G,is_map);

	const GraphCopy &GC = *tricComp.gcp;

	_type.use_node_data(_tree,SNode);
	sk.use_node_data(_tree,nil);

	sk_edge_src.use_edge_data(_tree,nil);
	sk_edge_tgt.use_edge_data(_tree,nil);

	node_array<node> mapV(GC);
	b_stack<node> inMapV(GC.number_of_nodes());

	edge_array<node> partnerNode(GC,nil);
	edge_array<edge> partnerEdge(GC,nil);

	num_S = num_P = num_R = 0;

	for (int i = 0; i < tricComp.num_comp; i++) {
		const TricComp_comp_struct &C = tricComp.component[i];

		if (C.edges.empty()) continue;

		node vT = _tree.new_node();

		switch(C.type) {
		case TricComp_bond:
			_type[vT] = PNode;
			num_P++; break;

		case TricComp_polygon:
			_type[vT] = SNode;
			num_S++; break;

		case TricComp_triconnected:
			_type[vT] = RNode;
			num_R++; break;
		}

		sk[vT] = new Skeleton(this,vT);
		Skeleton &S = *sk[vT];

		edge e;
		forall(e,C.edges) {

			node uGC = source(e), vGC = target(e);
			node uM = mapV[uGC], vM = mapV[vGC];

			if (uM == nil) {
				uM = mapV[uGC] = S.M.new_node();
				inMapV.push(uGC);
				S._orig[uM] = GC.original(uGC);
			}
			if (vM == nil) {
				vM = mapV[vGC] = S.M.new_node();
				inMapV.push(vGC);
				S._orig[vM] = GC.original(vGC);
			}

			edge eM  = S.M.new_edge(uM,vM);
			edge eMr = S.M.new_edge(vM,uM);
			S.M.set_reversal(eM,eMr);

			edge eG  = GC.original(e);

			if (eG == nil) {
				if (partnerNode[e] == nil) {
					partnerNode[e] = vT;
					partnerEdge[e] = eM;

				} else {
					edge eT = _tree.new_edge(partnerNode[e],vT);
					Skeleton &pS = *sk[partnerNode[e]];
					pS._tree_edge[partnerEdge[e]] =
						pS._tree_edge[pS.M.reversal(partnerEdge[e])] = eT;
					S._tree_edge[eM] = S._tree_edge[eMr] = eT;
					sk_edge_src[eT] = partnerEdge[e];
					sk_edge_tgt[eT] = eM;
				}

			} else {
				if (is_map) {
					if (source(eG) != S.original(source(eM)))
						leda::swap(eM,eMr);
					copy_of[S._real[eM ] = eG            ] = eM;
					copy_of[S._real[eMr] = G.reversal(eG)] = eMr;
					sk_of[eG] = sk_of[G.reversal(eG)] = &S;


				} else {
					S._real[eM] = S._real[eMr] = eG;
					copy_of[eG] = eM;
					sk_of  [eG] = &S;
				}
			}
		}

		while(!inMapV.empty())
			mapV[inMapV.pop()] = nil;
	}

	root_tree_at(e_ref);
}


SPQRTree::~SPQRTree()
{
	node vT;

	forall_nodes(vT,_tree)
		delete sk[vT];

	delete _cpV; delete _cpE;
}


list<node> SPQRTree::nodes_of_type(NodeType t) const
{
	list<node> L;
	node v;
	forall_nodes(v,_tree)
		if (_type[v] == t) L.append(v);

	return L;
}

node SPQRTree::root_tree_at(edge e)
{
	_root_edge = e;
	_root_node = sk_of[e]->tree_node();

	sk[_root_node]->ref_edge = copy_of[e];
	root_rec(_root_node,nil);

	return _root_node;
}

void SPQRTree::root_rec(node v, edge ef)
{
	edge e, e_next;
	for(e = _tree.first_in_edge(v); e != nil; e = e_next) {
		e_next = _tree.in_succ(e);

		if (e == ef) continue;

		_tree.rev_edge(e);
		leda::swap(sk_edge_src[e],sk_edge_tgt[e]);
	}

	forall_out_edges(e,v) {
		node w = target(e);
		sk[w]->ref_edge = sk_edge_tgt[e];
		root_rec(w,e);
	}
}


void SPQRTree::expand_virtual(edge eT, node src, PertinentGraph &Gp)
{
	const Skeleton &S = skeleton(target(eT));
	const graph    &M = S.M;
	graph &P = Gp.P;

	edge eS = skeleton_edge_tgt(eT);
	if (Gp.original(src) != S.original(source(eS)))
		eS = M.reversal(eS);

	edge e;
	for (e = M.cyclic_adj_succ(eS); e != eS; e = M.cyclic_adj_succ(e)) {
		node w_orig = S.original(target(e));
		node wP = cpV_add_node(w_orig,Gp);

		edge e_orig = S.real_edge(e);

		if (e_orig != nil)
			Gp.origE[P.new_edge(src,wP)] = e_orig;
		else
			expand_virtual(S.tree_edge(e),src,Gp);
	}
}


void SPQRTree::create_inner_vertices(node vT, PertinentGraph &Gp)
{
	const Skeleton &S = skeleton(vT);
	const graph& M = S.get_graph();
	graph &P = Gp.P;

	node src = source(S.reference_edge()), tgt = target(S.reference_edge());
	node v;
	forall_nodes(v,M) {
		if (v == src || v == tgt) continue;

		node v_orig = S.original(v);
		node vP = cpV_add_node(v_orig,Gp);

		edge e;
		forall_out_edges(e,v) {
			node w_orig = S.original(target(e));
			node wP = cpV_add_node(w_orig,Gp);

			edge e_orig = S.real_edge(e);

			if (e_orig != nil) {
				Gp.origE[P.new_edge(vP,wP)] = e_orig;
			} else {
				expand_virtual(S.tree_edge(e),vP,Gp);
			}
		}
	}

	node wT;
	forall_adj_nodes(wT,vT)
		create_inner_vertices(wT,Gp);
}


void SPQRTree::pertinent_graph(node vT, PertinentGraph &Gp)
{
	if (_cpV == nil) {
		_cpV = new node_array<node>(*cgp,nil);
		_cpE = new edge_array<edge>(*cgp,nil);
	}

	Gp.init(vT);
	graph &P = Gp.P;

	node_array<node> &cpV = *_cpV;

	const Skeleton &S = skeleton(vT);
	const graph    &M = S.get_graph();

	node v;
	forall_nodes(v,M) {
		cpV_add_node(S.original(v),Gp);
	}

	edge e;
	forall_nodes(v,M)
		forall_out_edges(e,v) {
			edge eT = S.tree_edge(e);

			if (eT == nil || source(eT) != vT) {
				edge eP = P.new_edge(cpV[S.original(source(e))],
					cpV[S.original(target(e))]);
				if (eT) {
					if (Gp._v_edge == nil) {
						Gp._v_edge = eP;
						Gp._sk_ref_edge = e;
					} else
						P.set_reversal(Gp._v_edge,eP);
				} else
					Gp.origE[eP] = S.real_edge(e);

			} else
				expand_virtual(eT,cpV[S.original(source(e))],Gp);
		}

	node wT;
	forall_adj_nodes(wT,vT)
		create_inner_vertices(wT,Gp);

	slist<edge> cpE_added;
	edge_array<edge> &cpE = *_cpE;
	forall_edges(e,P) {
		edge e_orig = Gp.original(e);
		if (e_orig != nil) {
			if (cpE[e_orig] != nil)
				P.set_reversal(cpE[e_orig],e);
			else {
				if (_is_map) {
					cpE_added.append(cgp->reversal(e_orig));
					cpE[cgp->reversal(e_orig)] = e;
				} else {
					cpE_added.append(e_orig);
					cpE[e_orig] = e;
				}
			}
		}
	}

	while (!cpE_added.empty())
		cpE[cpE_added.pop()] = nil;

	while (!cpV_added.empty())
		cpV[cpV_added.pop()] = nil;

}


//-------------------------------------------------------------------
//                           PlanarSPQRTree
//-------------------------------------------------------------------

//random_source PlanarSPQRTree::rs;


void PlanarSPQRTree::embed(graph &G)
{
	list<edge> edges;

	const Skeleton &S = skeleton(_root_node);
	const graph &M = S.get_graph();

	node v;
	edge e;
	forall_nodes(v,M)
		forall_out_edges(e,v) {
			edge e_orig = S.real_edge(e);

			if (e_orig != nil)
				edges.append(e_orig);
			else
				expand_virtual_embed(S.tree_edge(e),S.original(source(e)),edges);
		}

	node vT;
	forall_adj_nodes(vT,_root_node)
		create_inner_vertices_embed(vT,edges);

	G.sort_edges(edges);
}


void PlanarSPQRTree::expand_virtual_embed(edge eT, node v_orig, list<edge> &edges)
{
	const Skeleton &S = skeleton(target(eT));
	const graph &M = S.M;

	edge eS = skeleton_edge_tgt(eT);
	if (v_orig != S.original(source(eS)))
		eS = M.reversal(eS);

	edge e;
	for (e = M.cyclic_adj_succ(eS); e != eS; e = M.cyclic_adj_succ(e)) {
		edge e_orig = S.real_edge(e);

		if (e_orig != nil)
			edges.append(e_orig);
		else
			expand_virtual_embed(S.tree_edge(e),v_orig,edges);
	}
}

void PlanarSPQRTree::create_inner_vertices_embed(node vT, list<edge> &edges)
{
	const Skeleton &S = skeleton(vT);
	const graph& M = S.get_graph();

	node src = source(S.reference_edge()), tgt = target(S.reference_edge());
	node v;
	forall_nodes(v,M) {
		if (v == src || v == tgt) continue;

		edge e;
		forall_out_edges(e,v) {
			edge e_orig = S.real_edge(e);
			if (e_orig != nil) {
				edges.append(e_orig);
			} else {
				expand_virtual_embed(S.tree_edge(e),S.original(v),edges);
			}
		}
	}

	node wT;
	forall_adj_nodes(wT,vT)
		create_inner_vertices_embed(wT,edges);
}


void PlanarSPQRTree::rec_init_embed(node vT,
	list<two_tuple<edge,node> > &edges, node_array<list<list_item> > &adj,
	const edge_array<list_item> &pos_in_adj)
{
	node wT;
	forall_adj_nodes(wT,vT)
		rec_init_embed(wT,edges,adj,pos_in_adj);

	if (vT == _root_node) return;

	const Skeleton &S = *sk[vT];
	const graph &M = S.M;
	
	edge e;
	forall_edges(e,M) {
		edge er = S.real_edge(e);
		if (er) adj[source(er)].move_to_back(pos_in_adj[er]);
	}

	e           = S.reference_edge();
	edge er     = M.reversal(e);
	edge e_twin = S.twin_edge(e);

	wT = S.twin_tree_node(e);
	const Skeleton &S_wT = *sk[wT];

	node v_orig = S.original(source(e)), w_orig = S.original(source(er));

	edges.insert(two_tuple<edge,node>(e     ,vT),adj[v_orig].head());
	edges.insert(two_tuple<edge,node>(e_twin,wT),adj[v_orig].tail());
	edges.insert(two_tuple<edge,node>(er    ,vT),adj[w_orig].head());
	edges.insert(two_tuple<edge,node>(S_wT.M.reversal(e_twin),wT),
		adj[w_orig].tail());
}


void PlanarSPQRTree::init(const graph &G, bool is_planar_map = false)
{
	if (is_planar_map) {
		list<two_tuple<edge,node> > edges;

		node_array<list<list_item> > adj(G);
		edge_array<list_item> pos_in_adj(G);

		node v;
		forall_nodes(v,G) {
			edge e;
			forall_adj_edges(e,v) {
				const Skeleton &S = skeleton_of_real(e);
				pos_in_adj[e] = adj[v].append(edges.append(
					two_tuple<edge,node>(copy_of_real(e),S.tree_node())));
			}
		}

		rec_init_embed(_root_node,edges,adj,pos_in_adj);

		node_array<list<edge> > ES(_tree);
		while(!edges.empty()) {
			two_tuple<edge,node> pair = edges.pop();
			ES[pair.second()].append(pair.first());
		}

		forall_nodes(v,_tree)
			sk[v]->M.sort_edges(ES[v]);

	} else {
		node v;
		forall_nodes(v,_tree) {
			if (type(v) == RNode)
				sk[v]->M.make_planar_map();
		}
	}
}


void PlanarSPQRTree::reverse(node vT)
{
	if (type(vT) != RNode)
		error_handler(1,"PlanarSPQRTree::reverse only on R-node allowed.");

	graph &M = sk[vT]->M;

	list<edge> edges;

	node v;
	forall_nodes(v,M) {
		edge e;
		for(e = M.last_adj_edge(v); e != nil; e = M.adj_pred(e))
			edges.append(e);
	}

	M.sort_edges(edges);
}


void PlanarSPQRTree::swap_out(graph &M, edge e1, edge e2)
{
	node v = source(e1), w = target(e1);

	edge e_succ = M.adj_succ(e2);

	if (e_succ == e1) {
		M.move_edge(e2,e1,w,leda::after);

	} else {
		M.move_edge(e2,e1,w,leda::before);
		if (e_succ == nil)
			M.move_edge(e1,M.last_adj_edge(v),w,leda::after);
		else
			M.move_edge(e1,e_succ,w,leda::before);
	}
}

void PlanarSPQRTree::swap(node v, edge e1, edge e2)
{
	if (type(v) != PNode)
		error_handler(1,"PlanarSPQRTree::swap only on P-node allowed.");

	graph &M = sk[v]->M;

	if (source(e1) == source(e2)) {
		swap_out(M,e1,e2);
		swap_out(M,M.reversal(e1),M.reversal(e2));
	} else {
		swap_out(M,M.reversal(e1),e2);
		swap_out(M,e1,M.reversal(e2));
	}
}


integer PlanarSPQRTree::number_of_embeddings(node vT) const
{
	integer num = 1;

	switch(_type[vT]) {
	case RNode:
		num = 2; break;
	case PNode:
		{ const graph &M = sk[vT]->M;
		  for (integer i = M.outdeg(M.first_node())-1; i >= 2; i--)
			num *= i;
		  break;
                 }
        case SNode:
                break;
	}

	node wT;
	forall_adj_nodes(wT,vT)
		num *= number_of_embeddings(wT);

	return num;
}


void PlanarSPQRTree::random_embed()
{
	random_source& rs = AGD::get_random_source();
	node v;
	forall_nodes(v,_tree) {
		if (type(v) == RNode) {
			bool rev;
			rs >> rev;

			if (rev) reverse(v);

		} else if (type(v) == PNode) {
			const Skeleton &S = skeleton(v);
			const graph &M = S.get_graph();
			edge eRef = S.reference_edge();

			list<edge> edges;
			edge e;
			for (e = M.cyclic_adj_succ(eRef); e != eRef; e = M.cyclic_adj_succ(e))
				edges.append(e);

			permute(edges);

			edge eN;
			e = M.cyclic_adj_succ(eRef);
			forall (eN, edges) {
				if (eN != e) {
					swap(v,e,eN);
					e = eN;
				}
				e = M.cyclic_adj_succ(e);
			}
		}
	}
}
