/*******************************************************************************
+
+  AGD 1.2
+
+
+  TricComp.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


#include "TricComp.h"

#include <AGD/GraphCopy.h>
#include <LEDA/graph/node_set.h>


//#define OUTPUT

#ifdef OUTPUT
extern ofstream os;
#endif


/*********************************************************
                      Destructor

  deletes GraphCopy
**********************************************************/

TricComp::~TricComp()
{
	delete gcp;
}


/*********************************************************
                      Constructor

  Divides G into triconnected components.
**********************************************************/

TricComp::TricComp (const graph& G, bool is_map) :
	ESTACK(G.number_of_edges())
{
	gp = &G;
	gcp = new GraphCopy(G);
	GraphCopy &GC = *gcp;

	if (is_map) remove_reversals();

	const int n = GC.number_of_nodes();
	const int m = GC.number_of_edges();

#ifdef OUTPUT
	os << "Dividing G into triconnected components.\n" << endl;
	os << "n = " << n << ", m = " << m << endl << endl;
#endif

	component = array<TricComp_comp_struct>(3*m-6);
	num_comp = 0;

	// special cases
	if (n <= 2) {
		if (n < 2) {
			error_handler(1,"TricComp::Ticomp: graph has too few nodes!");
			return;
		}
		if (m < 3) {
			error_handler(1,"TricComp::Ticomp: graph with two nodes has too few edges!");
			return;
		}
		TricComp_comp_struct &C = new_comp();
		edge e;
		forall_edges(e,GC)
			C << e;
		C.type = bond;
		return;
	}

	TYPE.init(GC,3*m-6,unseen);
	split_multi_edges();

	// initialize arrays
	NUMBER.init(GC,0); LOWPT1.init(GC);
	LOWPT2.init(GC);   FATHER.init(GC,nil);
	ND    .init(GC);   DEGREE.init(GC);
	TREE_ARC.init(GC,nil);
	NODEAT = array<node>(1,n);

	num_count = 0;
	start = GC.first_node();
	DFS1(GC,start,nil);

	edge e;
	forall_edges(e,GC) {
		bool up = (NUMBER[target(e)] - NUMBER[source(e)] > 0);
		if ((up && TYPE[e] == frond) || (!up && TYPE[e] == tree))
			GC.rev_edge(e);
	}

#ifdef OUTPUT
	os << "\nnode\tNUMBER\tFATHER\tLOWPT1\tLOWPT2\tND" << endl;
	node v;
	forall_nodes(v,GC) {
		os << index(GC.original(v)) << ":  \t" << NUMBER[v] << "   \t";
		if (FATHER[v] == nil) os << "nil \t";
		else os << index(GC.original(FATHER[v])) << "   \t";
		os << LOWPT1[v] << "   \t" << LOWPT2[v] << "   \t" << ND[v] << endl;
	}
#endif

	A.init(GC);
	IN_ADJ.init(GC,3*m-6,nil);
	build_acceptable_adj_struct(GC);

#ifdef OUTPUT
	os << "\nadjaceny lists:" << endl;
	forall_nodes(v,GC) {
		os << index(v) << "\t";
		forall(e,A[v]) {
			print_os(e);
		}
		os << endl;
	}
#endif

	DFS2(GC);

#ifdef OUTPUT
	os << "\nnode\tNEWNUM\tLOWPT1\tLOWPT2\tHIGHPT" << endl;
	forall_nodes(v,GC) {
		os << index(GC.original(v)) << ":  \t" << NEWNUM[v] << "   \t";
		os << LOWPT1[v] << "   \t" << LOWPT2[v] << "   \t";
		int h;
		forall(h,HIGHPT[v])
			os << h << " ";
		os << endl;
	}

	os << "\nedges starting a path:" << endl;
	forall_edges(e,GC) {
		if (START[e]) {
			print_os(e);
		}
	}
#endif


	TSTACK_h = new int[2*m+1];
	TSTACK_a = new int[2*m+1];
	TSTACK_b = new int[2*m+1];
	//top = -1;
	TSTACK_a[top = 0] = -1; // start with EOS

	PathSearch(G,start);

	// last split component
	TricComp_comp_struct &C = new_comp();
	while(!ESTACK.empty()) {
		C << ESTACK.pop();
	}
	C.type = (C.edges.length() > 4) ? triconnected : polygon;

#ifdef OUTPUT
	print_stacks();
#endif

	delete [] TSTACK_h;
	delete [] TSTACK_a;
	delete [] TSTACK_b;

	// free resources
	NUMBER.init(); LOWPT1.init();
	LOWPT2.init(); FATHER.init();
	ND    .init(); TYPE  .init();
	A     .init(); NEWNUM.init();
	HIGHPT.init(); START .init();
	DEGREE.init(); TREE_ARC.init();
	NODEAT = array<node>();
	ESTACK.clear();

	assemble_triconnected_components();

#ifdef OUTPUT
	os << "\n\nTriconnected components:\n";
	for (int i = 0; i < num_comp; i++) {
		const list<edge> &L = component[i].edges;
		if (L.length() == 0) continue;
		os << "[" << i << "] ";
		switch(component[i].type) {
		case bond: os << "bond "; break;
		case polygon: os << "polygon "; break;
		case triconnected: os << "triconnected "; break;
		}

		edge e;
		forall(e,L)
			print_os(e);
		os << "\n";
	}
#endif
}


/*********************************************************
                     remove_reversals

  Removes reversal edges (in G) from GC.
**********************************************************/

void TricComp::remove_reversals()
{
	const graph &G  = *gp;
	GraphCopy   &GC = *gcp;

	edge e, r;
	forall_edges(e,G) {
		r = G.reversal(e);
		if (index(e) > index(r))
			GC.del_edge(GC.chain(e).head());
	}
}


/*********************************************************
                     split_multi_edges

  Splits bundles of multi-edges into bonds and creates
  a new virtual edge in GC.
**********************************************************/

static edge_array<int> sm_index, gr_index;

int smaller_index(const edge &e)
{
	return sm_index[e];
}

int greater_index(const edge &e)
{
	return gr_index[e];
}

void TricComp::split_multi_edges()
{
	GraphCopy &GC = *gcp;
	sm_index.init(GC); gr_index.init(GC);

	edge e, e1;
	forall_edges(e,GC) {
		int is = index(source(e)), it = index(target(e));
		if (is < it) {
			sm_index[e] = is; gr_index[e] = it;
		} else {
			sm_index[e] = it; gr_index[e] = is;
		}
	}

	GC.bucket_sort_edges(0,GC.max_node_index(),smaller_index);
	GC.bucket_sort_edges(0,GC.max_node_index(),greater_index);

	for (e = GC.first_edge(); e  && GC.original(e); e = e1) {
		int sm = sm_index[e], gr = gr_index[e];
		e1 = GC.succ_edge(e);
		if (e1 && sm == sm_index[e1] && gr == gr_index[e1]) {
			TricComp_comp_struct &C = new_comp(bond);
			C << GC.new_edge(source(e),target(e)) << e << e1;
			TYPE[e] = TYPE[e1] = removed;

			for (e1 = GC.succ_edge(e1);
				e1 && sm == sm_index[e1] && gr == gr_index[e1];
				e1 = GC.succ_edge(e1))
			{ C << e1; TYPE[e1] = removed; }
		}
	}

	sm_index.init(); gr_index.init();
}


/*********************************************************
                      check_comp

  Checks if computed triconnected components are correct.
**********************************************************/

bool TricComp::check_sep_pair(edge e_virt)
{
	GraphCopy G(*gp);

	G.del_node(G.copy(gcp->original(source(e_virt))));
	G.del_node(G.copy(gcp->original(target(e_virt))));

	return !Is_Connected(G);
}

bool TricComp::check_comp()
{
	bool ok = true;

#ifdef OUTPUT

	GraphCopy &GC = *gcp;
	GraphCopy GTest(GC.original());

	if (!Is_Loopfree(GC)) {
		ok = false;
		os << "GC contains loops!" << endl;
	}

	int i;
	edge e;
	node v;

	edge_array<int> count(GC,0);
	for (i = 0; i < num_comp; i++)
		forall(e,component[i].edges)
			count[e]++;

	forall_edges(e,GC) {
		if (GC.original(e) == nil) {
			if (count[e] != 2) {
				ok = false;
				os << "virtual edge contained " << count[e];
				print_os(e); os << endl;
			}
			if (check_sep_pair(e) == false) {
				ok = false;
				os << "virtual edge"; print_os(e);
				os << " does not correspond to a sep. pair." << endl;
			}

		} else {
			if (count[e] != 1) {
				ok = false;
				os << "real edge contained " << count[e];
				print_os(e); os << endl;
			}
		}
	}

	node_set S(GC);
	node_array<node> map(GC);

	for(i = 0; i < num_comp; i++) {
		TricComp::TricComp_comp_struct &C = component[i];
		const list<edge> &L = C.edges;
		if (L.length() == 0) continue;

		S.clear();

		forall(e,L) {
			S.insert(source(e));
			S.insert(target(e));
		}

		const int n = S.size();

		switch(C.type) {
		case bond:
			if (n != 2) {
				ok = false;
				os << "bond [" << i << "] with " << n << " nodes!" << endl;
			}
			break;

		case polygon:
			if (n < 3) {
				ok = false;
				os << "polygon [" << i << "] with " << n << " nodes!" << endl;
			}

			if (L.length() != n) {
				ok = false;
				os << "polygon [" << i << "] with " << n << " vertices and " << L.length() << " edges!" << endl;

			} else {
				graph Gp;
				forall(v,S)
					map[v] = Gp.new_node();
				forall(e,L)
					Gp.new_edge(map[source(e)],map[target(e)]);

				forall_nodes(v,Gp)
					if (degree(v) != 2) {
						ok = false;
						os << "polygon [" << i << "] contains node with degree " << degree(v) << endl;
					}
					if (!Is_Connected(Gp)) {
						ok = false;
						os << "polygon [" << i << "] not connected." << endl;
					}
			}
			break;

		case triconnected:
			if (n < 4) {
				ok = false;
				os << "triconnected component [" << i << "] with " << n << " nodes!" << endl;
			}

			{
			graph Gp;
			forall(v,S)
				map[v] = Gp.new_node();
			forall(e,L)
				Gp.new_edge(map[source(e)],map[target(e)]);

			if (!Is_Triconnected(Gp)) {
				ok = false;
				os << "component [" << i << "] not triconnected!" << endl;
			}
			if (!Is_Simple(Gp)) {
				ok = false;
				os << "triconnected component [" << i << "] not simple!" << endl;
			}
			}
			break;

		default:
			ok = false;
			os << "component [" << i << "] with undefined type!" << endl;
		}
	}

	if (ok)
		os << "ok" << endl;
#endif
	return ok;
}


/*********************************************************
               assemble_triconnected_components

  joins bonds and polygons with common virtual edge in
  order to build the triconnected components.
**********************************************************/

void TricComp::assemble_triconnected_components()
{
	GraphCopy &GC = *gcp;

	edge_array<int>       comp1(GC), comp2(GC);
	edge_array<list_item> item1(GC,nil), item2(GC);
	bool *visited = new bool[num_comp];

	int i;
	for(i = 0; i < num_comp; i++) {
		visited[i] = false;
		const list<edge> &L = component[i].edges;

		list_item it;
		forall_items(it,L) {
			edge e = L[it];
			if (item1[e] == nil) {
				comp1[e] = i; item1[e] = it;
			} else {
				comp2[e] = i; item2[e] = it;
			}
		}
	}

	for(i = 0; i < num_comp; i++) {
		TricComp_comp_struct &C1 = component[i];
		list<edge> &L1 = C1.edges;
		visited[i] = true;

		if (L1.length() == 0) continue;

		if (C1.type == polygon || C1.type == bond) {
			list_item it, it_next;
			for(it = L1.first(); it != nil; it = it_next) {
				it_next = L1.succ(it);
				edge e  = L1[it];

				if (GC.original(e) != nil) continue;

				int j = comp1[e];
				list_item it2;
				if (visited[j]) {
					j = comp2[e];
					if (visited[j]) continue;
					it2 = item2[e];
				} else
					it2 = item1[e];

				TricComp_comp_struct &C2 = component[j];

				if (C2.type != C1.type) continue;

				visited[j] = true;
				list<edge> &L2 = C2.edges;

				L2.del_item(it2);
				L1.conc(L2);
				if (it_next == nil)
					it_next = L1.succ(it);
				L1.del_item(it);

				GC.del_edge(e);
			}
		}
	}

	delete [] visited;
}



/*********************************************************
                   The first dfs-search

  computes NUMBER[v], FATHER[v], LOWPT1[v], LOWPT2[v],
           ND[v], TYPE[e], DEGREE[v]
**********************************************************/

void TricComp::DFS1 (const graph& G, node v, node u)
{
	node first_son = nil;
	edge e;

	NUMBER[v] = ++num_count;
	FATHER[v] = u;
	DEGREE[v] = G.degree(v);

	LOWPT1[v] = LOWPT2[v] = NUMBER[v];
	ND[v] = 1;

	forall_inout_edges (e,v) {

		if (TYPE[e] != unseen)
			continue;

		node w = G.opposite(v,e);

		if (NUMBER[w] == 0) {
			TYPE[e] = tree;

			TREE_ARC[w] = e;
			if (first_son == nil) first_son = w;

			DFS1(G,w,v);

			if (LOWPT1[w] < LOWPT1[v]) {
				LOWPT2[v] = leda::min(LOWPT1[v],LOWPT2[w]);
				LOWPT1[v] = LOWPT1[w];

			} else if (LOWPT1[w] == LOWPT1[v]) {
				LOWPT2[v] = leda::min(LOWPT2[v],LOWPT2[w]);

			} else {
				LOWPT2[v] = leda::min(LOWPT2[v],LOWPT1[w]);
			}

			ND[v] += ND[w];
		
		} else {

			TYPE[e] = frond;

			if (NUMBER[w] < LOWPT1[v]) {
				LOWPT2[v] = LOWPT1[v];
				LOWPT1[v] = NUMBER[w];

			} else if (NUMBER[w] > LOWPT1[v]) {
				LOWPT2[v] = leda::min(LOWPT2[v],NUMBER[w]);
			}
		}
	}
}


/*********************************************************
           Construction of ordered adjaceny lists
**********************************************************/

void TricComp::build_acceptable_adj_struct(const graph& G)
{
	edge e;
	int max = 3*G.number_of_nodes()+2;
	array<list<edge> > BUCKET(1,max);

	forall_edges(e,G) {
		edge_type t = TYPE[e];
		if (t == removed) continue;

		node w = target(e);
		int phi = (t == frond) ? 3*NUMBER[w]+1 : (
			(LOWPT2[w] < NUMBER[source(e)]) ? 3*LOWPT1[w] :
			3*LOWPT1[w]+2);
		BUCKET[phi].append(e);
	}

	for (int i = 1; i <= max; i++)
		forall(e,BUCKET[i])
			IN_ADJ[e] = A[source(e)].append(e);
}


/*********************************************************
                   The second dfs-search
**********************************************************/

void TricComp::path_finder(const graph& G, node v)
{
	edge e;
	node w;

	NEWNUM[v] = num_count - ND[v] + 1;

	forall(e,A[v]) {
		w = G.opposite(v,e);

		if (new_path) {
			new_path = false;
			START[e] = true;
		}

		if (TYPE[e] == tree) {
			path_finder(G,w);
			num_count--;

		} else {
			IN_HIGH[e] = HIGHPT[w].append(NEWNUM[v]);
			new_path = true;
		}
	}
}

void TricComp::DFS2 (const graph& G)
{
	NEWNUM.init(G,0);
	HIGHPT.init(G);
	IN_HIGH.init(G,3*G.number_of_edges()-6,nil);
	START .init(G,3*G.number_of_edges()-6,false);

	num_count = G.number_of_nodes();
	new_path = true;

	path_finder(G,start);

	node v;
	array<int> old2new(1,G.number_of_nodes());

	forall_nodes(v,G)
		old2new[NUMBER[v]] = NEWNUM[v];

	forall_nodes(v,G) {
		NODEAT[NEWNUM[v]] = v;
		LOWPT1[v] = old2new[LOWPT1[v]];
		LOWPT2[v] = old2new[LOWPT2[v]];
	}
}


void TricComp::PathSearch (const graph& G, node v)
{
	node w;
	edge e;
	int y, vnum = NEWNUM[v], wnum;
	int a, b;

	list<edge> &Adj = A[v];
	int outv = Adj.length();

	list_item it;
	forall_items(it,Adj) {
		e = Adj[it];
		//w = G.opposite(v,e); wnum = NEWNUM[w];
		w = target(e); wnum = NEWNUM[w];

		if (TYPE[e] == tree) {

			if (START[e]) {
				y = 0;
				if (TSTACK_a[top] > LOWPT1[w]) {
					do {
						y = leda::max(y,TSTACK_h[top]);
						b = TSTACK_b[top--];
					} while (TSTACK_a[top] > LOWPT1[w]);
					TSTACK_push(leda::max(y,wnum+ND[w]-1),LOWPT1[w],b);
				} else {
					TSTACK_push(wnum+ND[w]-1,LOWPT1[w],vnum);
				}
				TSTACK_pushEOS();
			}

			PathSearch(G,w);

			ESTACK.push(TREE_ARC[w]);  // add (v,w) to ESTACK (can differ from e!)

			//bool FLAG = false;
			//edge savedEdge;
			node x;

			while (vnum != 1 && ((TSTACK_a[top] == vnum) ||
				(DEGREE[w] == 2 && NEWNUM[target(A[w].head())] > wnum)))
			{
				a = TSTACK_a[top];
				b = TSTACK_b[top];

				edge e_virt;

				if (a == vnum && FATHER[NODEAT[b]] == NODEAT[a]) {
					top--;
				}

				else {
					edge e_ab = nil;

					if (DEGREE[w] == 2 && NEWNUM[target(A[w].head())] > wnum) {
#ifdef OUTPUT
						os << endl << "\nfound type-2 separation pair " <<
							index(gcp->original(v)) << ", " <<
							index(gcp->original(target(A[w].head())));
#endif
						
						edge e1 = ESTACK.pop();
						edge e2 = ESTACK.pop();
						A[w].del_item(IN_ADJ[e2]);

						x = target(e2);

						e_virt = gcp->new_edge(v,x);
						DEGREE[x]--; DEGREE[v]--;

#ifdef OUTPUT
						if (source(e2) != w)
							os << endl << "oops, this is not w ..." << endl;
#endif

						TricComp_comp_struct &C = new_comp(polygon);
						C << e1 << e2 << e_virt;

						if (!ESTACK.empty()) {
							e1 = ESTACK.top();
							if (source(e1) == x && target(e1) == v) {
								//FLAG = true;
								e_ab = ESTACK.pop();
								A[x].del_item(IN_ADJ[e_ab]);
								del_high(e_ab);
							}
						}

					} else {
#ifdef OUTPUT
						os << "\nfound type-2 separation pair " <<
							index(gcp->original(NODEAT[a])) << ", " <<
							index(gcp->original(NODEAT[b]));
#endif

						int h = TSTACK_h[top--];

						TricComp_comp_struct &C = new_comp();
						while(true) {
							edge xy = ESTACK.top();
							x = source(xy); node y = target(xy);
							if (!(a <= NEWNUM[x] && NEWNUM[x] <= h &&
								a <= NEWNUM[y] && NEWNUM[y] <= h)) break;

							if ((NEWNUM[x] == a && NEWNUM[y] == b) ||
								(NEWNUM[y] == a && NEWNUM[x] == b))
							{
								//FLAG = true;
								e_ab = ESTACK.pop();
								A[source(e_ab)].del_item(IN_ADJ[e_ab]);
								del_high(e_ab);

							} else {
								edge eh = ESTACK.pop();
								if (it != IN_ADJ[eh]) {
									A[source(eh)].del_item(IN_ADJ[eh]);
									del_high(eh);
								}
								C << eh;
								DEGREE[x]--; DEGREE[y]--;
							}
						}

						e_virt = gcp->new_edge(NODEAT[a],NODEAT[b]);
						C.finishTricOrPoly(e_virt);
						x = NODEAT[b];
					}

					if (e_ab != nil) {
						//FLAG = false;
						TricComp_comp_struct &C = new_comp(bond);
						C << e_ab << e_virt;

						e_virt = gcp->new_edge(v,x);
						C << e_virt;

						DEGREE[x]--; DEGREE[v]--;
					}

					ESTACK.push(e_virt);
					Adj[it] = e_virt;
					IN_ADJ[e_virt] = it;

					DEGREE[x]++; DEGREE[v]++;
					FATHER[x] = v;
					TREE_ARC[x] = e_virt;
					TYPE[e_virt] = tree;

					w = x; wnum = NEWNUM[w];
				}
			}

			if (LOWPT2[w] >= vnum && LOWPT1[w] < vnum && (FATHER[v] != start || outv >= 2))
			{
#ifdef OUTPUT
				os << "\nfound type-1 separation pair " <<
					index(gcp->original(NODEAT[LOWPT1[w]])) << ", " <<
					index(gcp->original(v));
#endif

				TricComp_comp_struct &C = new_comp();
				edge xy;
				int x = 0;
				while (!ESTACK.empty()) {
					xy = ESTACK.top();
					x = NEWNUM[source(xy)];
					y = NEWNUM[target(xy)];
	
					if (!((wnum <= x && x < wnum+ND[w]) || (wnum <= y && y < wnum+ND[w])))
						break;

					C << ESTACK.pop();
					del_high(xy);
					DEGREE[NODEAT[x]]--; DEGREE[NODEAT[y]]--;
				}

				edge e_virt = gcp->new_edge(v,NODEAT[LOWPT1[w]]);
				C.finishTricOrPoly(e_virt);

				if ((x == vnum && y == LOWPT1[w]) || (y == vnum && x == LOWPT1[w])) {
					TricComp_comp_struct &C = new_comp(bond);
					edge eh = ESTACK.pop();
					if (IN_ADJ[eh] != it) {
						A[source(eh)].del_item(IN_ADJ[eh]);
					}
					C << eh << e_virt;
					e_virt = gcp->new_edge(v,NODEAT[LOWPT1[w]]);
					C << e_virt;
					IN_HIGH[e_virt] = IN_HIGH[eh];
					DEGREE[v]--;
					DEGREE[NODEAT[LOWPT1[w]]]--;
				}

				if (NODEAT[LOWPT1[w]] != FATHER[v]) {
					ESTACK.push(e_virt);
					Adj[it] = e_virt;
					IN_ADJ[e_virt] = it;
					if (IN_HIGH[e_virt] == nil && high(NODEAT[LOWPT1[w]]) < vnum)
						IN_HIGH[e_virt] = HIGHPT[NODEAT[LOWPT1[w]]].push_front(vnum);

					DEGREE[v]++;
					DEGREE[NODEAT[LOWPT1[w]]]++;

				} else {
					Adj.del_item(it);

					TricComp_comp_struct &C = new_comp(bond);
					C << e_virt;
					e_virt = gcp->new_edge(NODEAT[LOWPT1[w]],v);
					C << e_virt;

					edge eh = TREE_ARC[v];

					C << TREE_ARC[v];

					TREE_ARC[v] = e_virt;
					TYPE[e_virt] = tree;

					IN_ADJ[e_virt] = IN_ADJ[eh];
					(A[source(eh)])[IN_ADJ[eh]] = e_virt;
				}
			}

			if (START[e]) {
				while (TSTACK_notEOS()) {
					top--;
				}
				top--;
			}

			while (TSTACK_notEOS() && TSTACK_a[top] != vnum &&
				TSTACK_b[top] != vnum && high(v) > TSTACK_h[top]) {
				top--;
			}

			outv--;

		} else { // frond arc
			if (START[e]) {
				y = 0;
				if (TSTACK_a[top] > wnum) {
					do {
						y = leda::max(y,TSTACK_h[top]);
						b = TSTACK_b[top--];
					} while (TSTACK_a[top] > wnum);
					TSTACK_push(y,wnum,b);
				} else {
					TSTACK_push(vnum,wnum,vnum);
				}
			}

			if (w == FATHER[v]) {
				TricComp_comp_struct &C = new_comp(bond);
				C << e << TREE_ARC[v];

				TREE_ARC[v] = gcp->new_edge(w,v); // v und w getauscht
				TYPE[TREE_ARC[v]] = tree;
				DEGREE[v]--; DEGREE[w]--;  // ++ -> --
				
				C << TREE_ARC[v];

			} else {
				ESTACK.push(e);  // add (v,w) to ESTACK
			}
		}
	}
}


void TricComp::print_os(edge e)
{
#ifdef OUTPUT
	os << " (" << index(gcp->original(source(e))) << "," <<
		index(gcp->original(target(e))) << "," << index(e) << ")";
	if (gcp->original(e) == nil) os << "v";
#endif
}

void TricComp::print_stacks()
{
#ifdef OUTPUT
	os << "\n\nTSTACK:" << endl;

	for (int i = top; i >= 0; i--)
		os << "(" << TSTACK_h[i] << "," << TSTACK_a[i] << "," << TSTACK_b[i] << ")\n";

	os << "\nESTACK\n";
	while(!ESTACK.empty()) {
		print_os(ESTACK.pop()); os << "\n";
	}
#endif
}
