/*******************************************************************************
+
+  AGD 1.2
+
+
+  CoffmanGrahamRanking.c
+
+
+  Copyright (c) 1998-2003
+  by  Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


//-------------------------------------------------------------------
//
//                       CoffmanGraham.cpp
//
//    author: Carsten Gutwenger
//    date:   decembre 1998
//-------------------------------------------------------------------


#include <AGD/CoffmanGrahamRanking.h>
#include <AGD/LEDAMakeAcyclic.h>
#include <AGD/GraphCopy.h>
#include <AGD/ReducedDigraph.h>

using namespace leda;

CoffmanGrahamRanking::CoffmanGrahamRanking() : W(3)
{
	add_precondition (key::directed);

	PreCond  PRE;
	PostCond POST;
	PRE [1] << key::directed;
	POST[1] << key::directed << key::maximal_acyclic;
	subgraph.init (PRE,POST,LEDAMakeAcyclic());
}


AGDModule *CoffmanGrahamRanking::clone() const
{
	CoffmanGrahamRanking *M = new CoffmanGrahamRanking;

	M->width(width());
	M->set_subgraph(subgraph.get());

	return M;
}


void CoffmanGrahamRanking::call (const graph& G, node_array<int>& rank)
{
	GraphCopy gc(G);
	ReducedDigraph rd;

	subgraph.get().call(gc);
	rd.call(gc);

	list<two_tuple<node,int> > ready_nodes;
	node_array<int> deg(gc);
	node_array<int> pi(gc);
	S.init(gc);

	node v;
	forall_nodes(v,gc) {
		if ((deg[v] = gc.indeg(v)) == 0)
			ready_nodes.append(two_tuple<node,int>(v,0));
		S[v].init(deg[v]);
	}

	int i = 1;
	while(!ready_nodes.empty()) {
		v = ready_nodes.pop_front().first();
		pi[v] = i++;

		node u;
		forall_adj_nodes(u,v) {
			S[u].insert(pi[v]);
			if (--deg[u] == 0)
				insert(u,ready_nodes);
		}
	}


	list<node> ready, waiting;

	forall_nodes(v,gc) {
		if ((deg[v] = gc.outdeg(v)) == 0)
			ready.append(v);
	}

	int k;
	for (k = 1; !ready.empty(); k++) {

		for (i = 1; i <= W && !ready.empty(); i++) {
			node u = ready.pop_front();
			rank[gc.original(u)] = k;

			edge e;
			forall_in_edges(e,u) {
				if (--deg[source(e)] == 0)
					waiting.append(source(e));
			}
		}

		while (!waiting.empty())
			insert (waiting.pop(), ready, pi);
	}

	k--;
	forall_nodes(v,G)
		rank[v] = k - rank[v];

	S.init();
}


void CoffmanGrahamRanking::insert (node u, list<two_tuple<node,int> > &ready_nodes)
{
	list_item it;
	int j = 0;

	for (it = ready_nodes.last(); it != nil; it = ready_nodes.pred(it)) {

		node v     = ready_nodes[it].first();
		int  sigma = ready_nodes[it].second();

		if (sigma < j) {
			ready_nodes.insert(two_tuple<node,int>(u,j), it, leda::after);
			return;
		}

		if (sigma > j) continue;

		const _int_set &x = S[u], &y = S[v];
		int k = leda::min(x.length(),y.length());

		while (j < k && x[j] == y[j])
			j++;

		if (j == k) {
			if (x.length() < y.length()) continue;

			ready_nodes[it].second() = k;
			ready_nodes.insert(two_tuple<node,int>(u,sigma), it, leda::after);
			return;
		}

		if (x[j] < y[j]) continue;

		ready_nodes[it].second() = j;
		ready_nodes.insert(two_tuple<node,int>(u,sigma), it, leda::after);
		return;
	}

	ready_nodes.push_front(two_tuple<node,int>(u,j));
}


void CoffmanGrahamRanking::insert (node v, list<node> &ready,
	const node_array<int> &pi)
{
	for (list_item it = ready.last(); it != nil; it = ready.pred(it)) {

		if (pi[v] <= pi[ready[it]]) {
			ready.insert (v, it, leda::after);
			return;
		}
	}

	ready.push_front(v);
}
