/****************************************************************************
 *  This file is part of PPMd project                                       *
 *  Written and distributed to public domain by Dmitry Shkarin 1997,        *
 *  1999-2001                                                               *
 *  Contents: memory allocation routines                                    *
 ****************************************************************************/

// constraint: UNIT_SIZE >= max(sizeof(MEM_BLK), sizeof(PPM_CONTEXT))
// This is checked in EncodeStart (PPMII_Model.h)

#define STD_UNIT_SZ 12

#ifndef PPMII_SZ_OF_UNIT
#define PPMII_SZ_OF_UNIT STD_UNIT_SZ
#define PPMII_UNSAFE_MEM_COPY
#endif

/*
	If the size of a unit differs from the standard unit size, we have to be
	careful: Since the model is restored whenever the coder runs out of memory,
	we have to make sure that this event occurs independent of the unit size.
	This means that the memory layouts have to be equivalent no matter which
	unit size we choose. Hence, if the actual unit size differs from the 
	standard size, we have to simulate the layout with the standard size.
	(Note: This may force us to report memory exhaustion as the standard layout
	run out of memory although we actually have some memory left!)
	This is a little bit tricky, but the changes below should do the trick!
*/
#if PPMII_SZ_OF_UNIT == STD_UNIT_SZ
#define STD_MEM_LAYOUT
#else
#undef STD_MEM_LAYOUT
#endif

#if !defined(STD_MEM_LAYOUT) || !defined(DWORD_HAS_32_BITS)
#undef PPMII_UNSAFE_MEM_COPY
#endif

enum { UNIT_SIZE=PPMII_SZ_OF_UNIT, 
       N1=4, N2=4, N3=4, N4=(128+3-1*N1-2*N2-3*N3)/4,
       N_INDEXES=N1+N2+N3+N4 };

#if !defined(__HP_aCC)
#pragma pack(1)
#else
#pragma pack 1
#endif
struct BLK_NODE {
    DWORD Stamp;
    BLK_NODE* next;
    BOOL   avail()      const { return (next != NULL); }
    void    link(BLK_NODE* p) { p->next=next; next=p; }
    void  unlink()            { next=next->next; }
    void* remove()            {
        BLK_NODE* p=next;                   unlink();
        Stamp--;                            return p;
    }
    inline void insert(void* pv,int NU);
} BList[N_INDEXES];
struct MEM_BLK: public BLK_NODE { DWORD NU; } _PACK_ATTR;
#if !defined(__HP_aCC)
#pragma pack()
#else
#pragma pack
#endif

// special MARKER used in BLK_NODE::Stamp
#define MARKER (unsigned int)(0-1)

static BYTE Indx2Units[N_INDEXES], Units2Indx[128]; // constants
static DWORD GlueCount, SubAllocatorSize=0;
static BYTE* HeapStart, * HeapEnd, * pText, * UnitsStart, * LoUnit, * HiUnit;

#ifdef STD_MEM_LAYOUT
#define RealSubAllocatorSize SubAllocatorSize
#else
static DWORD RealSubAllocatorSize=0;
#endif

inline BYTE PrefetchData(void* Addr)
{
#if defined(_USE_PREFETCHING)
    volatile BYTE PrefetchByte = *(volatile BYTE*) Addr;
    return PrefetchByte;
#endif /* defined(_USE_PREFETCHING) */
}

inline void BLK_NODE::insert(void* pv,int NU) {
    MEM_BLK* p=(MEM_BLK*) pv;               link(p);
    p->Stamp=MARKER;                        p->NU=NU;
    Stamp++;
}

inline UINT U2B(UINT NU)
{
#if PPMII_SZ_OF_UNIT == 12
	return 8*NU+4*NU;
#else
	return UNIT_SIZE * NU;
#endif
}
inline void SplitBlock(void* pv,UINT OldIndx,UINT NewIndx)
{
    UINT i, k, UDiff=Indx2Units[OldIndx]-Indx2Units[NewIndx];
    BYTE* p=((BYTE*) pv)+U2B(Indx2Units[NewIndx]);
    if (Indx2Units[i=Units2Indx[UDiff-1]] != UDiff) {
        k=Indx2Units[--i];                  BList[i].insert(p,k);
        p += U2B(k);                        UDiff -= k;
    }
    BList[Units2Indx[UDiff-1]].insert(p,UDiff);
}
DWORD _STDCALL GetUsedMemory()
{
#ifdef STD_MEM_LAYOUT
    DWORD i, RetVal=SubAllocatorSize-(HiUnit-LoUnit)-(UnitsStart-pText);
    for (i=0;i < N_INDEXES;i++)
            RetVal -= UNIT_SIZE*Indx2Units[i]*BList[i].Stamp;
    return RetVal;
#else
	DWORD allocated_units = ( (HeapEnd-HiUnit) + (LoUnit-UnitsStart) ) / UNIT_SIZE; 
    for (DWORD i=0;i < N_INDEXES;i++)
            allocated_units -= Indx2Units[i]*BList[i].Stamp;
	return (pText - HeapStart) + allocated_units * STD_UNIT_SZ;
#endif
}
void _STDCALL StopSubAllocator() {
    if ( SubAllocatorSize ) {
        SubAllocatorSize=0;                 delete[] HeapStart;
    }
}
BOOL _STDCALL StartSubAllocator(UINT SASize)
{
	DWORD t = SASize == 0 ? (32 * 1024) : (SASize << 20U);
    if (SubAllocatorSize == t)              return TRUE;
    StopSubAllocator();                     SubAllocatorSize=t;
#ifndef STD_MEM_LAYOUT
	// make t large enough to hold ceil(t / STD_UNIT_SZ) units
	t = RealSubAllocatorSize = UNIT_SIZE * ((t + STD_UNIT_SZ - 1) / STD_UNIT_SZ);
#endif
	if ((HeapStart=new BYTE[t]) == NULL)    { SubAllocatorSize=0; return FALSE; }
    return TRUE;
}
static inline void InitSubAllocator()
{
    memset(BList,0,sizeof(BList));
    HeapEnd=HiUnit=(pText=HeapStart)+RealSubAllocatorSize;
    UINT Diff=UNIT_SIZE*(SubAllocatorSize/8/STD_UNIT_SZ*7);
    LoUnit=UnitsStart=HiUnit-Diff;          GlueCount=0;
}

#ifdef STD_MEM_LAYOUT
#define EnoughRoomBtwTextAndUnits(bytes, index) UnitsStart-pText > int(bytes)
#define BelowUnits(ptr) ptr < UnitsStart
#define NotBelowUnits(ptr) ptr >= UnitsStart
#define AddUnitsToPtr(PTR, NU) PTR+NU
#else
inline bool EnoughRoomBtwTextAndUnits(UINT, UINT index)
{
	return SubAllocatorSize > (pText-HeapStart) 
		                    + ((HeapEnd - UnitsStart) / UNIT_SIZE + Indx2Units[index]) * STD_UNIT_SZ;
}
inline bool BelowUnits(BYTE* ptr)
{ return (ptr < HeapStart) || (SubAllocatorSize > (ptr-HeapStart) + ((HeapEnd - UnitsStart) / UNIT_SIZE) * STD_UNIT_SZ); }
inline bool NotBelowUnits(BYTE* ptr)
{ return (ptr >= HeapStart) && (SubAllocatorSize <= (ptr-HeapStart) + ((HeapEnd - UnitsStart) / UNIT_SIZE) * STD_UNIT_SZ); }
#define AddUnitsToPtr(PTR, NU) (MEM_BLK*)(((BYTE*) PTR)+U2B(NU))
#endif

static void GlueFreeBlocks()
{
    UINT i, k, sz;
    MEM_BLK s0, * p, * p0, * p1;
    if (LoUnit != HiUnit)                   *LoUnit=0;
    for (i=0, (p0=&s0)->next=NULL;i < N_INDEXES;i++)
            while ( BList[i].avail() ) {
                p=(MEM_BLK*) BList[i].remove();
                if ( !p->NU )               continue;
                while ((p1=AddUnitsToPtr(p,p->NU))->Stamp == MARKER) {
                    p->NU += p1->NU;        p1->NU=0;
                }
                p0->link(p);                p0=p;
            }
    while ( s0.avail() ) {
        p=(MEM_BLK*) s0.remove();           sz=p->NU;
        if ( !sz )                          continue;
        for ( ;sz > 128;sz -= 128, p = AddUnitsToPtr(p,128))
                BList[N_INDEXES-1].insert(p,128);
        if (Indx2Units[i=Units2Indx[sz-1]] != sz) {
            k=sz-Indx2Units[--i];           BList[k-1].insert(AddUnitsToPtr(p,(sz-k)),k);
        }
        BList[i].insert(p,Indx2Units[i]);
    }
    GlueCount=1 << 13;
}

static void* _STDCALL AllocUnitsRare(UINT indx)
{
    UINT i=indx;
    if ( !GlueCount ) {
        GlueFreeBlocks();
        if ( BList[i].avail() )             return BList[i].remove();
    }
    do {
        if (++i == N_INDEXES) {
            GlueCount--;                    i=U2B(Indx2Units[indx]);
            return (EnoughRoomBtwTextAndUnits(i, indx))?(UnitsStart -= i):(NULL);
        }
    } while ( !BList[i].avail() );
    void* RetVal=BList[i].remove();         SplitBlock(RetVal,i,indx);
    return RetVal;
}
inline void* AllocUnits(UINT NU)
{
    UINT indx=Units2Indx[NU-1];
    if ( BList[indx].avail() )              return BList[indx].remove();
    void* RetVal=LoUnit;                    LoUnit += U2B(Indx2Units[indx]);
    if (LoUnit <= HiUnit)                   return RetVal;
    LoUnit -= U2B(Indx2Units[indx]);        return AllocUnitsRare(indx);
}
inline void* AllocContext()
{
    if (HiUnit != LoUnit)                   return (HiUnit -= UNIT_SIZE);
    else if ( BList->avail() )              return BList->remove();
    else                                    return AllocUnitsRare(0);
}

inline void UnitsCpy(void* Dest,void* Src,UINT NU)
{
#ifdef PPMII_UNSAFE_MEM_COPY
    DWORD* p1=(DWORD*) Dest, * p2=(DWORD*) Src;
    do {
        p1[0]=p2[0];                        p1[1]=p2[1];
        p1[2]=p2[2];
        p1 += 3;                            p2 += 3;
    } while ( --NU );
#else
	memcpy(Dest, Src, size_t(NU) * UNIT_SIZE);
#endif
}
inline void* ExpandUnits(void* OldPtr,UINT OldNU)
{
    UINT i0=Units2Indx[OldNU-1], i1=Units2Indx[OldNU-1+1];
    if (i0 == i1)                           return OldPtr;
    void* ptr=AllocUnits(OldNU+1);
    if ( ptr ) {
        UnitsCpy(ptr,OldPtr,OldNU);         BList[i0].insert(OldPtr,OldNU);
    }
    return ptr;
}
inline void* ShrinkUnits(void* OldPtr,UINT OldNU,UINT NewNU)
{
    UINT i0=Units2Indx[OldNU-1], i1=Units2Indx[NewNU-1];
    if (i0 == i1)                           return OldPtr;
    if ( BList[i1].avail() ) {
        void* ptr=BList[i1].remove();       UnitsCpy(ptr,OldPtr,NewNU);
        BList[i0].insert(OldPtr,Indx2Units[i0]);
        return ptr;
    } else {
        SplitBlock(OldPtr,i0,i1);           return OldPtr;
    }
}
inline void FreeUnits(void* ptr,UINT NU) {
    UINT indx=Units2Indx[NU-1];
    BList[indx].insert(ptr,Indx2Units[indx]);
}
inline void SpecialFreeUnit(void* ptr)
{
    if ((BYTE*) ptr != UnitsStart)          BList->insert(ptr,1);
    else { *(DWORD*) ptr=MARKER;            UnitsStart += UNIT_SIZE; }
}
inline void* MoveUnitsUp(void* OldPtr,UINT NU)
{
    UINT indx=Units2Indx[NU-1];
    if ((BYTE*) OldPtr > UnitsStart+1365*UNIT_SIZE || (BLK_NODE*) OldPtr > BList[indx].next)
            return OldPtr;
    void* ptr=BList[indx].remove();
    UnitsCpy(ptr,OldPtr,NU);                NU=Indx2Units[indx];
    if ((BYTE*) OldPtr != UnitsStart)       BList[indx].insert(OldPtr,NU);
    else                                    UnitsStart += U2B(NU);
    return ptr;
}
static inline void ExpandTextArea()
{
    BLK_NODE* p;
    UINT Count[N_INDEXES];                  memset(Count,0,sizeof(Count));
    while ((p=(BLK_NODE*) UnitsStart)->Stamp == MARKER) {
        MEM_BLK* pm=(MEM_BLK*) p;           UnitsStart=(BYTE*) AddUnitsToPtr(pm,pm->NU);
        Count[Units2Indx[pm->NU-1]]++;      pm->Stamp=0;
    }
    for (UINT i=0;i < N_INDEXES;i++)
        for (p=BList+i;Count[i] != 0;p=p->next)
            while ( !p->next->Stamp ) {
                p->unlink();                BList[i].Stamp--;
                if ( !--Count[i] )          break;
            }
}

/* DEBUG stuff
static void DumpBList(UINT indx)
{
	cout << "BList[" << indx << "]" << endl;
	BLK_NODE* p = BList + indx;
	do {
		if (p->Stamp == MARKER) cout << "M!"; else cout << p->Stamp;
		MEM_BLK* pb = (MEM_BLK*) p;
		cout << "/" << pb->NU;
		cout << "\t";
	} while (p = p->next);
	cout << endl;
}

static void DebugCheck()
{
	static unsigned cnt = 0; ++cnt;

	if (cnt < 2493 || cnt > 2494 || cnt % 1 != 0) return;

	DWORD used_mem = GetUsedMemory();
	DWORD unit_chk1 = 0, unit_chk2 = 0;
	for (DWORD i=0;i < N_INDEXES;i++) {
            unit_chk1 += Indx2Units[i]*BList[i].Stamp;
            if (i % 2 == 0) unit_chk2 += Indx2Units[i]*BList[i].Stamp;
	}

	cout << cnt << ": " << used_mem << " / " << unit_chk1 << " / " << unit_chk2 << endl;

	for (DWORD i=0;i < 20 && i < N_INDEXES;i++)
            cout << BList[i].Stamp << " ";
	cout << endl;

	if (cnt >= 2493) {
		cout << "now\n";
		//for (UINT i = 0; i <= 5; ++i) DumpBList(i);
	}
}
*/
