// compression module includes
#include <LEDA/coding/PPMII.h>

/*
	The PPMII code may access data in a misaligned away, because the data 
	structures are packed as much as possible in order to reduce memory 
	consumption.
	For some compilers like MSVC and g++ we achieve this packing by compiler 
	directives. Other compilers like CC need a command-line switch to generate 
	code that can deal with these misaligned memory accesses.
	For the latter compilers the config script must contain a line like 
	"MISALGN_FLAG = -misalign" which specifies the correct command line switch.
*/

#if defined(LEDA_DEBUG)
#include <LEDA/system/assert.h>
#else
#undef assert
#define	assert(cond) ((void)0)
#endif

#if defined(_MSC_VER)
#pragma warning(push)
#pragma warning(disable:4018 4146 4244)
#endif

#if defined(WORD_LENGTH_32)
// do nothing (standard configuration)
#elif defined(WORD_LENGTH_64)
#define PPMII_SZ_OF_UNIT 20
#else
#error Either WORD_LENGTH_32 or WORD_LENGTH_64 must be defined
#endif

#include "PPMII_Model.h"

#if defined(_MSC_VER)
#pragma warning(pop)
#endif

LEDA_BEGIN_NAMESPACE

typedef coder_types::uint32 uint32;

/*
	The PPMII code by Shkarin and Subbotin makes use of static global variables.
	In order to protect this data in a multithreading environment, we use a
	mutex!
*/
static leda_lock PPMII_mutex;

/// PPMIICoder /////////////////////////////////////////////////////////////////////////////

PPMIICoder::PPMIICoder(streambuf* src_stream, streambuf* tgt_stream, bool own_streams)
: base(src_stream, tgt_stream, own_streams),
  MaxMemory(def_max_memory), MaxModelOrder(def_order), 
  ModelRestorationMethod(mr_method(def_mrm))
{}

PPMIICoder::PPMIICoder(const char* src_file_name, const char* tgt_file_name)
: base(src_file_name, tgt_file_name),
  MaxMemory(def_max_memory), MaxModelOrder(def_order), 
  ModelRestorationMethod(mr_method(def_mrm))
{}

PPMIICoder::~PPMIICoder()
{
	StopSubAllocator();
}

void 
PPMIICoder::do_reset(reset_type t)
{
	base::do_reset(t);
	if (t == r_hard) {
		StopSubAllocator();
			// NOTE: Since StartSubAllocator is reentrant, it is not necessary to call this ...

		MaxMemory = def_max_memory;
		MaxModelOrder = def_order;
		ModelRestorationMethod = mr_method(def_mrm);
	}
}

void 
PPMIICoder::write_header_to_tgt_stream()
{
	assert(MAX_O < 32 && ModelRestorationMethod < 4);
	byte mode;
	mode = MaxModelOrder | ((byte(ModelRestorationMethod)&3) << 5);
		// Bits 0-4: MaxModelOrder, Bits 5-6: MRM, Bit 7: reserved(=0)
	if (write_byte(mode) < 0) { error(); return; }
	write_uint16(MaxMemory);
}

void 
PPMIICoder::read_header_from_src_stream()
{
	int c = read_byte();
	if (c < 0) { error(); return; }
	byte mode = (byte) c;
	MaxModelOrder = mode & 31;
	ModelRestorationMethod = mr_method((mode >> 5) & 3);
	read_uint16(MaxMemory);
}

uint32 
PPMIICoder::get_header_length(bool& is_exact) const
{
	return base::get_header_length(is_exact) + 2 /* Order/Restoration/MaxMem */;
}


void 
PPMIICoder::init_encoding()
{
	base::init_encoding();

	PPMII_mutex.lock();
	StartSubAllocator(MaxMemory);
	EncodeStart(MaxModelOrder, MR_METHOD(ModelRestorationMethod));
}

void 
PPMIICoder::encode_stream()
{
	if (EncodeGoOn(get_tgt_stream(), get_src_stream()) != 0) return;

	StopSubAllocator();
	PPMII_mutex.unlock();

	encoding_finished();
}

void 
PPMIICoder::init_decoding()
{
	base::init_decoding();

	PPMII_mutex.lock();
	StartSubAllocator(MaxMemory);
	DecodeStart(MaxModelOrder, MR_METHOD(ModelRestorationMethod), get_src_stream());
}

void 
PPMIICoder::decode_stream()
{
	if (DecodeGoOn(get_tgt_stream(), get_src_stream()) != 0) return;

	StopSubAllocator();
	PPMII_mutex.unlock();
	decoding_finished();
}

string 
PPMIICoder::get_description() const
{
	string descr = base::get_description();
	if (MaxMemory != def_max_memory || MaxModelOrder != def_order)
		descr += string("[%d,%dMB]", MaxModelOrder, MaxMemory);
	return descr;
}

void
PPMIICoder::set_model_memory_bound(uint16 max_mem)
{
	if (can_change_parameters()) {
		uint16 new_bound = max_mem <= 1024 ? max_mem : 1024;
		if (new_bound == 0) new_bound = def_max_memory;
		if (max_mem == uint16(-1)) new_bound = 0; // DebugMode
		if (new_bound != MaxMemory) {
			MaxMemory = new_bound;
			StopSubAllocator(); // NOTE: StopSubAllocator is reentrant
		}
	}
	else error();
}

void
PPMIICoder::set_model_order(byte order)
{
	if (can_change_parameters()) {
		MaxModelOrder = order <= MAX_O ? order : MAX_O;
		if (MaxModelOrder <= 2) MaxModelOrder = 2;
	}
	else error();
}

void
PPMIICoder::set_model_restoration_method(mr_method method)
{
	if (can_change_parameters()) {
		ModelRestorationMethod = method;
		if ((ModelRestorationMethod & 3) != ModelRestorationMethod) error();
	}
	else error();
}

LEDA_END_NAMESPACE
