// compression module includes
#include <LEDA/coding/arithmetic_coding.h>

// core LEDA includes
#include <string.h> // -> we need the function memset

#if defined(LEDA_DEBUG)
#include <LEDA/system/assert.h>
#else
#undef assert
#define	assert(cond) ((void)0)
#endif

LEDA_BEGIN_NAMESPACE

/// arithmetic_coder_base ///////////////////////////////////////////////////////////////////

arithmetic_coder_base::arithmetic_coder_base(streambuf* src_stream, streambuf* tgt_stream, bool own_streams)
: base(src_stream, tgt_stream, own_streams)
{}

arithmetic_coder_base::arithmetic_coder_base(const char* src_file_name, const char* tgt_file_name)
: base(src_file_name, tgt_file_name)
{}

void 
arithmetic_coder_base::ResetCoder()
{
	BitBuffer = 0;
	BitsInBuffer = 0;

	L = 0;
	H = TopValue;

	UnderflowBits = 0;
	DecodeBuffer = 0;
}

void 
arithmetic_coder_base::GenOutputBit(byte bit)
{
	BitBuffer = (BitBuffer << 1) | bit;
	BitsInBuffer++;
	if (BitsInBuffer == 8) {
		BitsInBuffer = 0;
		write_byte(BitBuffer);
	}
}

arithmetic_coder_base::uint32
arithmetic_coder_base::GetNextBitFromBuf()
{
	if (BitsInBuffer == 0) {
		int c = read_byte();
		if (c != EOF) {
			BitBuffer = (byte) c; BitsInBuffer = 8;
		}
		else {
			BitBuffer = 0; BitsInBuffer = 0; // return only zeros from now
		}
	}

	// can be speed up by assembler rotation!
	uint32 bit = BitBuffer >> 7;
	BitBuffer <<= 1;
	--BitsInBuffer;
	return bit;
}

void 
arithmetic_coder_base::FlushBitBuffer()
{
	while(BitsInBuffer != 0) GenOutputBit(0);
}

void 
arithmetic_coder_base::UpdateEncoder(uint32 lower_range, uint32 upper_range, uint32 total_range)
{
/*
	We encode a number n between >=0 and <1 to the base 2:
	. Out_0 Out_1 ... Out_k H_CodeBits 1^UnderflowBits H_{CodeBits-1} ... H_0 1 1 1 ...
	. Out_0 Out_1 ... Out_k L_CodeBits 0^UnderflowBits L_{CodeBits-1} ... L_0 0 0 0 ...

	where Out_0 ... Out_k are the bits that have been output by GenOutputBit so far.
*/

	uint32 step = (H - L + 1) / total_range;
	assert(step >= 1);
	assert(total_range < FirstQuarter);
	assert(upper_range > lower_range);

	H = L + upper_range * step - 1;
	L = L + lower_range * step;

	// while L and R have the same rightmost bit ...
	for(;;) {
		if (H < Half) // i.e. leftmost bit is 0 in both L and R
		{
			// write the leftmost bit
			GenOutputBit(0);

			// shift L and R to the left, shift a 0 into the rightmost bit of L 
			// and a 1 into the rightmost bit of R
			L = 2 * L;
			H = 2 * H + 1;

			// write underflow bits if any
			OutputUnderflowBits(1);
		}
		else if (L >= Half) // i.e. leftmost bit is 1 in both L and R
		{
			// write the leftmost bit
			GenOutputBit(1);

			// shift L and R to the right, shift a 0 into the rightmost bit of L 
			// and a 1 into the rightmost bit of R
			L = 2 * (L - Half);
			H = 2 * (H - Half) + 1;

			// write underflow bits if any
			OutputUnderflowBits(0);
		}
		else break;
	}

	// Now the rightmost bit of L is 0 and the righmost bit of R is 1 

	// while L=01... and H=10...
	while ((L >= FirstQuarter) && (H < ThirdQuarter)) {
		// remember there was an underflow
		++UnderflowBits;

		// remove the bit right to the leftmost bit of L and R to the right, 
		// shift a 0 into the leftmost bit of L and a 1 into the leftmost bit of R
		L = 2 * (L - FirstQuarter);
		H = 2 * (H - FirstQuarter) + 1;
	}

	// Now: L=00... and H=1..., or L=01... and H=11... => H-L >= FirstQuarter
	assert(H-L >= FirstQuarter && L < Half && H >= Half);
}

void 
arithmetic_coder_base::FinishEncoding()
{
	++UnderflowBits;
	if (L < FirstQuarter) {
		GenOutputBit(0);
		OutputUnderflowBits(1);
	}
	else {
		GenOutputBit(1);
		//zeros at the end of the input do not have to be written (see GetNextBitFromBuf)
		//OutputUnderflowBits(0);
	}
	FlushBitBuffer();
}

void 
arithmetic_coder_base::StartDecoding()
{
	// initial fill of decode buffer
	for(int i=0; i<CodeBits; i++)
	{
		uint32 b = GetNextBitFromBuf();		
		DecodeBuffer = (DecodeBuffer << 1) | b;
	}
}

void 
arithmetic_coder_base::UpdateDecoder(uint32 lower_range, uint32 upper_range, uint32 total_range)
{
	uint32 step = (H - L + 1) / total_range;
	assert(step >= 1);
	assert(total_range < FirstQuarter);
	assert(upper_range > lower_range);

	H = L + upper_range * step - 1;
	L = L + lower_range * step;

	for (;;) {
		if (H < Half) {
			L = 2 * L;
			H = 2 * H + 1;
			DecodeBuffer = 2*DecodeBuffer + GetNextBitFromBuf();
		}
		else if (L >= Half) {
			L = 2 * (L - Half);
			H = 2 * (H - Half) + 1;
			DecodeBuffer = 2*(DecodeBuffer - Half) + GetNextBitFromBuf();
		}
		else break;
	}

	while((L >= FirstQuarter) && (H < ThirdQuarter)) {
		L = 2 * (L - FirstQuarter);
		H = 2 * (H - FirstQuarter) + 1;
		DecodeBuffer = 2*(DecodeBuffer - FirstQuarter) + GetNextBitFromBuf();
	}

	assert(H-L >= FirstQuarter && L < Half && H >= Half);
}

arithmetic_coder_base::uint32
arithmetic_coder_base::DecodeValueInRange(uint32 total_range)
{
	uint32 step = (H - L + 1) / total_range;
	assert(total_range < FirstQuarter);
	assert(step >= 1);
	return (DecodeBuffer - L) / step;
}

// must be at least 0x100 (unbiased)
#define ZERO_BIAS_FOR_COUNTS 0x200
/*
	Model for writing/reading counts:
	0x01 <= byte <= 0xFF: lower range = byte-1, upper range = byte
	byte == 0           : lower range = 0xFF,   upper range = ZERO_BIAS_FOR_COUNTS

	So if ZERO_BIAS_FOR_COUNTS > 0x100, then a zero byte in the counts is encoded 
	with less than 8 bits!
 */

void 
arithmetic_coder_base::WriteBiasedUInt32(uint32 val)
{
	for (int bits=0; bits<32; bits+=8) {
		uint32 _byte = (val >> bits) & 0xFF;
		if (_byte == 0)
			UpdateEncoder(0xFF, ZERO_BIAS_FOR_COUNTS, ZERO_BIAS_FOR_COUNTS);
		else
			UpdateEncoder(_byte-1, _byte, ZERO_BIAS_FOR_COUNTS);
	}
}

void 
arithmetic_coder_base::ReadBiasedUInt32(uint32& val)
{
	val = 0;
	for (int bits=0; bits<32; bits+=8) {
		uint32 _byte = DecodeValueInRange(ZERO_BIAS_FOR_COUNTS);
		if (_byte >= 0xFF) {
			_byte = 0;
			UpdateDecoder(0xFF, ZERO_BIAS_FOR_COUNTS, ZERO_BIAS_FOR_COUNTS);
		}
		else {
			++_byte;
			UpdateDecoder(_byte-1, _byte, ZERO_BIAS_FOR_COUNTS);
		}
		val |= _byte << bits;
	}
}


/// Order0Model ////////////////////////////////////////////////////////////////////////////

/*
void 
Order0Model::get_range(byte symbol, uint32& low, uint32& high)
{
	int idx = symbol;
	high = get_cumulative_count(idx);
	low  = high - Counts[idx];
}
*/

Order0Model::uint32
Order0Model::get_symbol_and_range(uint32 value, uint32& low, uint32& high) const
{
	int base_idx = 0;
	uint32 remainder = value;
	if (remainder >= Tree[0]) {
		remainder -= Tree[0];
		int half = NumSymbols >> 1;
		do {
			int probe_idx = base_idx + half;
			if (remainder >= Tree[probe_idx]) {
				base_idx = probe_idx;
				remainder -= Tree[base_idx];
			}
			half >>= 1;
		} while (half > 0);
		++base_idx;
	}

	low  = value - remainder;
	high = low + Counts[base_idx]; // NOTE: This is wrong if base_idx == EOFSym!
	assert( base_idx == EOFSym || high == get_cumulative_count(base_idx) );

	return uint32(base_idx);
}

void 
Order0Model::update(byte symbol)
{
	int idx = symbol;
	++Counts[idx];
	if (idx == 0) { ++Tree[0]; return; }
	do {
		++Tree[idx];
		idx += idx & (-idx);
	} while (idx < NumSymbols);
}

void 
Order0Model::reset(uint32 initial_count)
{
	if (initial_count == 0) {
		memset(&Counts[0], 0, (NumSymbols-1) * sizeof(uint32));
		memset(&Tree[0],   0, NumSymbols * sizeof(uint32));
		Tree[EOFSym] = 1;
	}
	else {
		for (int idx = 0; idx < EOFSym; ++idx)
			Counts[idx] = initial_count;
		rebuild_from_raw_counts();
	}
}

void 
Order0Model::scale(uint32 scale_divisor)
{
	if (scale_divisor > 1) {
		for (int idx = 0; idx < EOFSym; ++idx) {
			Counts[idx] = (Counts[idx] + scale_divisor-1) / scale_divisor;
				// = ceil(double(old_count) / double(scale_divisor)), 
			    // i.e. old_count stays > 0 if it was before the scaling
		}
		rebuild_from_raw_counts();
	}
}

void 
Order0Model::rebuild_from_raw_counts()
{
	memset(&Tree[0], 0, NumSymbols * sizeof(uint32));
	Tree[0] = Counts[0]; Tree[EOFSym] = 1;
	for (int idx = 1; idx < EOFSym; ++idx) {
		int t_idx = idx; uint32 val = Counts[idx];
		do {
			Tree[t_idx] += val;
			t_idx += t_idx & (-t_idx);
		} while (t_idx < NumSymbols);
	}
}

Order0Model::uint32 
Order0Model::get_cumulative_count(int idx) const
{
	uint32 sum = Tree[0];
	while (idx > 0) {
		sum += Tree[idx];
		idx = idx & (idx-1);
	}
	return sum;
}

/*
void 
Order0Model::add_to_tree(int idx, uint32 val)
{
	if (idx == 0) { Tree[0] += val; return; }
	do {
		Tree[idx] += val;
		idx += idx & (-idx);
	} while (idx < NumSymbols);
}
*/

/// Order0StaticModel //////////////////////////////////////////////////////////////////////

Order0StaticModel::uint32
Order0StaticModel::get_symbol_and_range(uint32 value, uint32& low, uint32& high) const
{
	// we use binary seach to find the symbol ...
	uint32 base = 0;
	if (value >= Counts[1]) {
		++base;
		uint32 range = NumSymbols >> 1;
		do {
			uint32 m = base + range;
			if (value >= Counts[m]) base = m;
			range >>= 1;
		} while (range > 0);
	}
	low = Counts[base]; high = Counts[base+1];
	assert(value >= low || value < high);
	return base;
}

void 
Order0StaticModel::reset(uint32 initial_count)
{
	if (initial_count == 0) {
		memset(&Counts[0], 0, NumSymbols * sizeof(uint32));
		Counts[NumSymbols] = 1;
	}
	else {
		uint32 cum_count = 0;
		for (int idx = 0; idx < EOFSym; ++idx) {
			Counts[idx] = cum_count;
			cum_count += initial_count;
		}
		Counts[EOFSym] = cum_count; Counts[NumSymbols] = ++cum_count;
	}
}

void 
Order0StaticModel::scale(uint32 scale_divisor)
{
	if (scale_divisor > 1) {
		uint32 cum_count = 0;
		for (int idx = 0; idx < EOFSym; ++idx) {
			uint32 old_count = Counts[idx+1] - Counts[idx];
			Counts[idx] = cum_count;
			cum_count += (old_count + scale_divisor-1) / scale_divisor;
				// = ceil(double(old_count) / double(scale_divisor)), 
			    // i.e. old_count stays > 0 if it was before the scaling
		}
		Counts[EOFSym] = cum_count; Counts[NumSymbols] = ++cum_count;
	}
}

void 
Order0StaticModel::rebuild_from_raw_counts()
{
	uint32 cum_count = 0;
	for (int idx = 0; idx < EOFSym; ++idx) {
		uint32 cnt = Counts[idx];
		Counts[idx] = cum_count;
		cum_count += cnt;
	}
	Counts[EOFSym] = cum_count; Counts[NumSymbols] = ++cum_count;
}

/// A0Coder ////////////////////////////////////////////////////////////////////////////////

A0Coder::A0Coder(streambuf* src_stream, streambuf* tgt_stream, bool own_streams)
: base(src_stream, tgt_stream, own_streams)
{
	set_reset_threshold(DefaultResetThreshold);
	set_scale_threshold(DefaultScaleThreshold);
}

A0Coder::A0Coder(const char* src_file_name, const char* tgt_file_name)
: base(src_file_name, tgt_file_name)
{
	set_reset_threshold(DefaultResetThreshold);
	set_scale_threshold(DefaultScaleThreshold);
}

void
A0Coder::set_reset_threshold(uint32 t)
{
	if (can_change_parameters()) {
		ResetThreshold = t;
	}
	else error("A0Coder::set_reset_threshold: wrong state");
}

void 
A0Coder::set_scale_threshold(uint32 t)
{
	if (can_change_parameters()) {
		if (t < 512) t = 512;
		else if (t > get_max_scale_threshold()) t = get_max_scale_threshold();
		ScaleThreshold = t;
	}
	else error("A0Coder::set_scale_threshold: wrong state");
}

void 
A0Coder::do_reset(reset_type t)
{
	base::do_reset(t);

	if (t == r_hard) {
		set_reset_threshold(DefaultResetThreshold);
		set_scale_threshold(DefaultScaleThreshold);
	}
}

void 
A0Coder::UpdateModel(byte symbol)
{
	Model.update(symbol);
	++BytesSinceLastReset;

	if (BytesSinceLastReset == ResetThreshold && ResetThreshold != 0) {
		BytesSinceLastReset = 0; Model.reset(1);
	}
	else if (Model.get_sum_of_non_EOFs() == ScaleThreshold) {
		Model.scale(2);
	}
}

void 
A0Coder::init_encoding()
{
	base::init_encoding();

	ResetCoder();

	WriteBiasedUInt32(ResetThreshold);
	WriteBiasedUInt32(ScaleThreshold);

	BytesSinceLastReset = 0; Model.reset(1);
}

void 
A0Coder::encode_stream()
{
	int c;
	while ((c = read_byte()) >= 0 && c != EOF) {
		byte symbol = (byte) c;

		// find the range for the symbol
		uint32 lc = 0, hc = 0;
		Model.get_range(symbol, lc, hc);

		// encode the range
		UpdateEncoder(lc, hc, Model.get_total());

		// update the model
		UpdateModel(symbol);
	}

	if (c == EOF) {
		uint32 lc = 0, hc = 0;
		Model.get_range_EOF(lc, hc);
		UpdateEncoder(lc, hc, Model.get_total());
		FinishEncoding();

		encoding_finished();
	}
}


void 
A0Coder::init_decoding()
{
	base::init_decoding();

	ResetCoder();
	StartDecoding();

	if (BitsInBuffer + CodeBits == 0)
		error("A0Coder::init_decoding: bad src stream");

	ReadBiasedUInt32(ResetThreshold);
	ReadBiasedUInt32(ScaleThreshold);

	BytesSinceLastReset = 0; Model.reset(1);
}

void
A0Coder::decode_stream()
{
	for(;;) {
		// find a value in the range of the next symbol
		uint32 value = DecodeValueInRange(Model.get_total());

		assert(value <= Model.get_sum_of_non_EOFs());

		// find the symbol that corresponds to the value and its range
		uint32 lc = 0, hc = 0;
		uint32 symbol = Model.get_symbol_and_range(value, lc, hc);

		if (symbol == EOFSymbol) { decoding_finished(); return; }

		// update according to the range
		UpdateDecoder(lc, hc, Model.get_total());

		// update the model
		UpdateModel((byte) symbol);

		if (write_byte((byte) symbol) < 0) break;
	}
}



/// A0sCoder ///////////////////////////////////////////////////////////////////////////////

A0sCoder::A0sCoder(streambuf* src_stream, streambuf* tgt_stream, bool own_streams)
: base(src_stream, tgt_stream, own_streams)
{}

A0sCoder::A0sCoder(const char* src_file_name, const char* tgt_file_name)
: base(src_file_name, tgt_file_name)
{}

void 
A0sCoder::ComputeModelFromStaticCounts()
{
	// compute cumulative counts
	Model.rebuild_from_raw_counts();
	uint32 sum_of_counts = Model.get_total(); // include EOF

	// scale model if necessary
	if (sum_of_counts >= FirstQuarter) {
		uint32 scale_divisor = sum_of_counts / FirstQuarter + 1;

		// below we round up, so we might get too high in some rare cases ...
		if ((sum_of_counts/scale_divisor)*scale_divisor + (EOFSymbol+1) >= FirstQuarter)
			++scale_divisor;

		//assert(scale_factor >= 2);
		Model.scale(scale_divisor);
	}
}

void 
A0sCoder::init_encoding()
{
	base::init_encoding();

	Model.reset(0);
}

void 
A0sCoder::encode_stream()
{
	int c;
	while ((c = read_byte()) >= 0 && c != EOF) {
		byte symbol = (byte) c;
		Model.raw_increment_count(symbol);
	}
	if (c != EOF) return;

	ResetCoder();

	// write static counts for all the chars (using Encoder!)
	for (unsigned i=0; i<EOFSymbol; ++i) {
		WriteBiasedUInt32(Model.raw_get_count((leda::coder_types::byte)i));
	}

	ComputeModelFromStaticCounts();

	// start the actual encoding
	rewind_src_stream();
	if (failed()) { encoding_finished(); return; }

	uint32 sum_of_counts = Model.get_total();
	while ((c = read_byte()) != EOF) {
		byte symbol = (byte) c;
		uint32 lc = 0, hc = 0;
		Model.get_range(symbol, lc, hc);
		UpdateEncoder(lc, hc, sum_of_counts);
	}

	// encode EOF and finish
	uint32 lc = 0, hc = 0;
	Model.get_range_EOF(lc, hc);
	UpdateEncoder(lc, hc, sum_of_counts);
	FinishEncoding();

	encoding_finished();
}


void 
A0sCoder::init_decoding()
{
	base::init_decoding();

	ResetCoder();
	StartDecoding();

	// read static counts for all the chars (using Decoder!)
	for (unsigned s=0; s<EOFSymbol; ++s) {
		uint32 count;
		ReadBiasedUInt32(count);
		Model.raw_set_count((leda::coder_types::byte)s, count);
	}

	ComputeModelFromStaticCounts();
}

void
A0sCoder::decode_stream()
{
	uint32 sum_of_counts = Model.get_total(); // includes EOF
	for(;;) {
		uint32 value = DecodeValueInRange(sum_of_counts);
		assert(value < sum_of_counts);

		uint32 lc = 0, hc = 0;
		uint32 symbol = Model.get_symbol_and_range(value, lc, hc);

		if (symbol == EOFSymbol) break;

		UpdateDecoder(lc, hc, sum_of_counts);

		if (write_byte((byte) symbol) < 0) return;
	}

	decoding_finished();
}

LEDA_END_NAMESPACE
