// compression module includes
#include <LEDA/coding/block_coder.h>

// core LEDA includes
#include <LEDA/system/file.h>

#if defined(LEDA_DEBUG)
#include <LEDA/system/assert.h>
#else
#undef assert
#define	assert(cond) ((void)0)
#endif


#define CHK_SEEK(Exp) if ((Exp) < streamoff(0)) { error(); return; }


LEDA_BEGIN_NAMESPACE

typedef coder_types::uint32 uint32;

namespace block_coder_helpers {

/// block_table_cache ///////////////////////////////////////////////////////////////////////

bool 
block_table_cache::write_to_stream(streambuf* stream)
{
	for (uint16 i = 0; i < NumberOfEntries; ++i)
		if (! coder_base::write_uint32(stream, Entries[i])) return false;
	return true;
}

bool 
block_table_cache::read_from_stream(streambuf* stream, uint32 size)
{
	NumberOfEntries = (leda::coder_types::uint16)size;
	for (uint32 i = 0; i < size; ++i)
		if (! coder_base::read_uint32(stream, Entries[i])) return false;
	return true;
}

/// block_table /////////////////////////////////////////////////////////////////////////////

bool 
block_table::append_entry(uint32 entry)
{
	if (Cache.is_full()) {
		// open a stream for buffering
		if (!pStream) {
			TmpFileName = tmp_file_name();
			filebuf* file_stream = new filebuf();
			if (! file_stream->open(TmpFileName, ios::openmode(ios::out | ios::in | ios::trunc | coder_base::binary))) {
				delete file_stream; return false;
			}
			pStream = file_stream;
		}

		Cache.write_to_stream(pStream);
		Cache.clear( Cache.next_index() );
	}

	++Size;
	Cache.append_entry(entry);

	return true;
}

bool 
block_table::write_to_stream(streambuf* stream)
{
	const int buf_sz = 1024;

	if (pStream) {
		bool ok = true;
		if (pStream->pubseekpos(0) != streampos(0)) ok = false;

		// copy pStream -> stream
		char buffer[buf_sz];
		int bytes_read, bytes_written;
		do {
			bytes_read    = (int)pStream->sgetn(buffer, buf_sz);
			bytes_written = (int)stream->sputn(buffer, bytes_read);
			if (bytes_read != bytes_written) { ok = false; break; }
		} while (bytes_read == buf_sz);

		delete pStream; pStream = 0;
		delete_file(TmpFileName);

		if (!ok) return false;
	}

	return Cache.write_to_stream(stream);
}

bool 
block_table::read_from_stream(streambuf* stream, uint32 size)
{
	clear();
	
	Size = size;
	if (block_table_cache::MaxNumEntries >= size)
		return Cache.read_from_stream(stream, size);
	else {
		pStream = stream;
		StartInStream = pStream->pubseekoff(0, ios::cur, ios::in);
		if (StartInStream < 0) return false;
		return Cache.read_from_stream(stream, block_table_cache::MaxNumEntries);
	}
}

block_table::uint32 
block_table::get_entry(uint32 idx, bool cache_successors)
{
	if (!Cache.contains_index(idx)) {
		assert(pStream);
		streampos current_pos = pStream->pubseekoff(0, ios::cur, ios::in);
		streampos entry_pos = StartInStream + streamoff(4 * idx);
		pStream->pubseekpos(entry_pos);

		if (!cache_successors) {
			uint32 entry;
			coder_base::read_uint32(pStream, entry);
			pStream->pubseekpos(current_pos);
			return entry;
		}

		Cache.clear(idx);
		uint32 to_read = block_table_cache::MaxNumEntries;
		if (idx + to_read > Size) to_read = Size - idx;
		Cache.read_from_stream(pStream, to_read);
		pStream->pubseekpos(current_pos);
	}

	return Cache.get_entry(idx);
}

}

/// block_coder ////////////////////////////////////////////////////////////////////////////

block_coder::block_coder(streambuf* src_stream, streambuf* tgt_stream, bool own_streams,
						 uint32 block_size)
: base(src_stream, tgt_stream, own_streams), 
  pCoder(0), OwnCoder(false)
{
	set_block_size(block_size);
}

block_coder::block_coder(const char* src_file_name, const char* tgt_file_name,
						 uint32 block_size)
: base(src_file_name, tgt_file_name), 
  pCoder(0), OwnCoder(false)
{
	set_block_size(block_size);
}

block_coder::~block_coder()
{
	set_state(s_finished); // clears errors
	// clear_parameters_lock();
	set_coder(0);
}

void 
block_coder::do_reset(reset_type t)
{
	base::do_reset(t);
	if (get_coder()) get_coder()->do_reset(t);

	if (t == r_hard) {
		BlockSize = DefaultBlockSize;
	}
}

string 
block_coder::get_description() const
{
	string result = base::get_description() 
	              + string('<')
	              + (get_coder() ? get_coder()->get_description() : string("none"))
	              + string('>');

	return result;
}

void
block_coder::set_coder(coder_base* coder, bool own_coder)
{
	if (can_change_parameters()) {
		if (OwnCoder) delete pCoder;
		pCoder = coder; OwnCoder = own_coder;
	}
	else error("block_coder::set_coder: wrong state");
}

void
block_coder::set_block_size(uint32 block_size)
{
	if (can_change_parameters()) {
		BlockSize = block_size > 16 ? block_size : 16;
	}
	else error("block_coder::set_block_size: wrong state");
}


void 
block_coder::write_id_to_tgt_stream()
{
	base::write_id_to_tgt_stream();

	streambuf* save_tgt_stream = pCoder->change_tgt_stream_temporarily( get_tgt_stream() );
	pCoder->write_id_to_tgt_stream();
	pCoder->change_tgt_stream_temporarily( save_tgt_stream );
}

void 
block_coder::read_and_check_id_in_src_stream()
{
	base::read_and_check_id_in_src_stream();

	streambuf* save_src_stream = pCoder->change_src_stream_temporarily( get_src_stream() );
	pCoder->read_and_check_id_in_src_stream();
	pCoder->change_src_stream_temporarily( save_src_stream );
}

uint32 
block_coder::get_id_length(bool& is_exact) const
{
	uint32 len = base::get_id_length(is_exact);
	if (! pCoder) { len += coder_base::IdLength; is_exact = false; return len; }
	bool coder_exact;
	len += pCoder->get_id_length(coder_exact);
	is_exact &= coder_exact;
	return len;
}

void 
block_coder::finish_construction_from_stream(id_type id_in_stream, streambuf* stream, streambuf* additional_data)
{
	base::finish_construction_from_stream(id_in_stream, stream, additional_data);

	leda_assert(pCoder == 0, "coder_pipe::finish_construction_from_stream");

	coder_base* coder = construct_from_stream(stream, additional_data);
	if (! coder) { 
		error("block_coder::finish_construction_from_stream: could not construct coder"); 
		return;
	}
	set_coder(coder, true);
}

void 
block_coder::write_header_to_tgt_stream()
{
	base::write_header_to_tgt_stream();

	write_uint32(BlockSize);

	streambuf* save_tgt_stream = pCoder->change_tgt_stream_temporarily( get_tgt_stream() );
	pCoder->write_header_to_tgt_stream();
	pCoder->change_tgt_stream_temporarily( save_tgt_stream );
}

void 
block_coder::read_header_from_src_stream()
{
	base::read_header_from_src_stream();

	read_uint32(BlockSize);

	streambuf* save_src_stream = pCoder->change_src_stream_temporarily( get_src_stream() );
	pCoder->read_header_from_src_stream();
	pCoder->change_src_stream_temporarily( save_src_stream );
}

uint32 
block_coder::get_header_length(bool& is_exact) const
{
	uint32 len = base::get_header_length(is_exact) + 4 /* BlockSize */;
	if (! pCoder) { is_exact = false; return len; }
	bool coder_exact;
	len += pCoder->get_header_length(coder_exact);
	is_exact &= coder_exact;
	return len;
}


void 
block_coder::init_encoding()
{
	if (! pCoder) {
		error("block_coder::init_encoding: coder missing");
		return;
	}

	// set the streams for the coder: 
	//   it reads from MemoryBuf and writes immediately to tgt stream
	pCoder->set_src_stream(&MemoryBuf, false);
	pCoder->set_tgt_stream(get_tgt_stream(), false);

	base::init_encoding();

	if (get_pos_behind_header() < 0) {
		error("block_coder::init_encoding: seek failed"); return;
	}

	// intialize data structures
	MemoryBuf.set_size(streamsize(BlockSize));
	BlockTable.clear();
	BlockTable.append_entry(0); // start of first encoded block
	SizeOriginalStream = 0;
}

void 
block_coder::encode_stream()
{
	// main loop: read block from src stream into MemoryBuf; when it is full, encode it ...
	int c;
	while ((c = read_byte()) >= 0 && c != EOF) {
		if (MemoryBuf.sputc((char) c) == EOF) {
			SizeOriginalStream += BlockSize;
			encode_memory_buffer();
			MemoryBuf.reset();
			MemoryBuf.sputc((char) c);
		}
	}

	if (c != EOF) return;

	// handle the last (possibly not completely full) block
	if (MemoryBuf.out_count() > 0) {
		SizeOriginalStream += uint32(MemoryBuf.out_count());
		MemoryBuf.truncate(MemoryBuf.out_count());
		encode_memory_buffer();
	}

	// write BlockTable
	BlockTable.write_to_stream(get_tgt_stream());

	// write epilog: size of BlockTable and SizeOriginalStream
	write_uint32(BlockTable.get_size());
	write_uint32(SizeOriginalStream); // NOTE: last entry in BlockTable = size encoded stream

	encoding_finished();
}

void 
block_coder::encode_memory_buffer()
{
	pCoder->do_reset(r_block);
	pCoder->do_encode();
	if (!pCoder->finished() || pCoder->failed()) error();

	streampos end_pos;
	CHK_SEEK(end_pos = get_pos_tgt_stream());
	BlockTable.append_entry( uint32(end_pos) );
}

void 
block_coder::init_decoding()
{
	if (! pCoder) {
		error("block_coder::init_decoding: coder missing");
		return;
	}

	pCoder->set_src_stream(&MemoryBuf, false);
	pCoder->set_tgt_stream(get_tgt_stream(), false);

	base::init_decoding();

	// NOTE: This check is not subsumed by the checks below!
	if (get_pos_behind_header() < 0) {
		error("block_coder::init_decoding: seek failed"); return;
	}

	// read epilog
	uint32 block_table_sz;
	streampos epilog_pos;
	CHK_SEEK(epilog_pos = seek_off_src_stream(-2*4, ios::end));
	read_uint32(block_table_sz);
	read_uint32(SizeOriginalStream);
	assert(block_table_sz > 0 || SizeOriginalStream == 0);

	// read BlockTable
	streampos blk_tbl_pos = epilog_pos - streamoff(block_table_sz * 4);
	CHK_SEEK(seek_pos_src_stream(blk_tbl_pos));
	BlockTable.read_from_stream(get_src_stream(), block_table_sz);

	// now seek to beginning again
	CHK_SEEK(seek_pos_src_stream(0));

	IdxCurrentBlock = (leda::coder_types::uint32)0-1;
	if (!fetch_next_encoded_block()) { decoding_finished(); return; } // maybe empty file
	pCoder->do_reset(r_block);
}

void 
block_coder::decode_stream()
{
	// NOTE: It is important that we start with pCoder->do_decode()
	//       because there might have been a SreamBlocked ...
	do {
		pCoder->do_decode();
		if (pCoder->failed()) { error(); return; }
		if (!pCoder->finished()) return; // Coder received StreamBlocked
		pCoder->do_reset(r_block);
	} while (fetch_next_encoded_block());

	decoding_finished();
}

bool 
block_coder::fetch_next_encoded_block()
{
	if (++IdxCurrentBlock+1 >= BlockTable.get_size()) return false;

	uint32 start_pos = BlockTable.get_entry(IdxCurrentBlock, true);
	uint32 end_pos   = BlockTable.get_entry(IdxCurrentBlock+1, true);

	int enc_size_next_blk = int(end_pos - start_pos);
	MemoryBuf.set_size(enc_size_next_blk);
	if (get_src_stream()->sgetn(MemoryBuf.get_buffer(), enc_size_next_blk) != enc_size_next_blk) { error(); return false; }

	return true;
}

bool 
block_coder::fetch_encoded_block(uint32 idx_blk)
{
	// check if we already have the current block in the MemoryBuf
	if (IdxCurrentBlock == idx_blk) {
		MemoryBuf.reset();
		return true;
	}

	IdxCurrentBlock = idx_blk;
	uint32 start_pos = BlockTable.get_entry(IdxCurrentBlock);
	uint32 end_pos   = BlockTable.get_entry(IdxCurrentBlock+1);

	if (seek_pos_src_stream(streamoff(start_pos)) < 0) { error(); return false; }
	assert(IdxCurrentBlock < BlockTable.get_size());

	int enc_size_next_blk = int(end_pos - start_pos);
	MemoryBuf.set_size(enc_size_next_blk);
	if (get_src_stream()->sgetn(MemoryBuf.get_buffer(), enc_size_next_blk) != enc_size_next_blk) { error(); return false; }

	return true;
}

streambuf* 
block_coder::change_tgt_stream_temporarily(streambuf* tgt_stream)
{
	if (running() && get_coder()) get_coder()->change_tgt_stream_temporarily(tgt_stream);
	return base::change_tgt_stream_temporarily(tgt_stream);
}

streampos 
block_coder::decode_seek(streampos current_pos, streampos new_pos)
{
	if (!decoding() || failed()) return base::decode_seek(current_pos, new_pos);
	if (new_pos < streamoff(0)) {
		// seek to the end of the stream
		set_state(s_decoding_finished);
		return streampos(SizeOriginalStream);
	}

	set_state(s_decoding); // make it running ...
	uint32 idx_new_block = uint32(streamoff(new_pos)) / BlockSize;
	if (idx_new_block != IdxCurrentBlock || current_pos > new_pos) {
		// we fetch the correct block and decode it from the beginning
		if (!fetch_encoded_block(idx_new_block)) return streampos(-1);
		current_pos = streamoff(idx_new_block * BlockSize);
		pCoder->do_reset(r_block);
	}
	// ow.: we keep on decoding the current block

	return base::decode_seek(current_pos, new_pos);
}

LEDA_END_NAMESPACE
