// compression module includes
#include <LEDA/coding/coder_base.h>
#include <LEDA/coding/coder_factory.h>
#include <LEDA/coding/internal/strbuf_helpers.h>
#include <LEDA/coding/memory_streambuf.h>

// core LEDA includes

#if defined(LEDA_DEBUG)
#include <LEDA/system/assert.h>
#else
#undef assert
#define	assert(cond) ((void)0)
#endif

#undef coder_base

/*
	Work-arounds for compression:
	- LEDA_COMPRESSION_WA1 (defined in LEDA/internal/PREAMBLE.h)
	  iostream library is not standard compliant: 
	  * the symbols openmode and seekdir have an "_" in their name
	  * pubsync/pubseek... do not exist in class std::streambuf
	  (The fix involves also LEDA/internal/POSTAMBLE.h and 
	  LEDA/coding/coder_strbuf_adapter.h)

	- LEDA_COMPRESSION_WA2 (defined in LEDA/internal/PREAMBLE.h)
	  implementation of std::filebuf is buggy: after reading an EOF one cannot
	  seek anymore (in particular rewind to the beginning)
	  (The fix includes LEDA/internal/std/sgi_filebuf.h)

	- LEDA_COMPRESSION_WA3 (defined in LEDA/coding/coder_factory.h)
	  problem with template instantiation
	  (The fix also involves _coder_factory.cpp)

	- LEDA_COMPRESSION_WA4 (OBSOLETE)
	  implementation of std:streambuf::xsgetn(char* buf, int n) is buggy
	  (If there are underflows in the streambuf object, the buf pointer is 
	  incremented too much.)

	- LEDA_COMPRESSION_WA5 (defined in _coder_strbuf_adapter.cpp)
	  bugs in the implementation of std::streambuf:
	  * underflow may be called altough the internal get-buffer is not empty
	  * underflow is called where uflow should have been called

	- LEDA_COMPRESSION_WA6 (defined in LEDA/coding/coder_base.h)
	  non-standard version of std::streambuf::sputc(int c) (instead of char c)

	- LEDA_COMPRESSION_WA7 (defined in LEDA/coding/internal/streambuf_workaround.h)
	  strange behaviour with __exportC: if we derive a class immediately from 
	  streambuf, this does not work. (We get a bunch of multiply defined symbols.)
	  (The fix affects coder_strbuf_adapter.h, memory_streambuf.h, 
	  socket_streambuf.h, secure_socket_streambuf.h)

	- LEDA_COMPRESSION_WA8 (defined in LEDA/coding/coder_base.h)
	  buggy implementation of streambuf, where uflow is not present
	  => StreamBlocked cannot be signaled easily
	  In the work-around we return EOF instead of StreamBlocked in the 
	  functions overflow/underflow. In order distinguish a real EOF from
	  a StreamBlocked, we use the functions check_eof and pubseekoff 
	  (see _coder_base.cpp and _coder_strbuf_adapter.cpp).

	- LEDA_COMPRESSION_WA9 (defined in LEDA/coding/coder_base.h)
	  buggy implementation of streambuf::sbumpc: even if underflow is called,
	  the return value (if != EOF) is *gptr() and NOT simply the value returned
	  by underflow.
	  (The fix also involves _coder_strbuf_adapter.cpp)
*/

LEDA_BEGIN_NAMESPACE

typedef coder_types::uint32 uint32;

/// coder_base //////////////////////////////////////////////////////////////////////////////

coder_base::coder_base(streambuf* src_stream, streambuf* tgt_stream, bool own_streams)
: State(s_init), 
  pSrcStream(src_stream), pTgtStream(tgt_stream), 
  OwnSrcStream(own_streams), OwnTgtStream(own_streams),
  ProcessIdInStreams(true), ProcessHeaderInStreams(true)
{}

coder_base::coder_base(const char* src_file_name, const char* tgt_file_name)
: State(s_init), 
  pSrcStream(0), pTgtStream(0), 
  OwnSrcStream(false), OwnTgtStream(false),
  ProcessIdInStreams(true), ProcessHeaderInStreams(true)
{
	set_src_file(src_file_name);
	set_tgt_file(tgt_file_name);
}

coder_base::~coder_base()
{
	set_state(s_finished); // also clears errors!
	//clear_parameters_lock();
	set_src_stream(0); set_tgt_stream(0);
}

#ifdef _MSC_VER
coder_base::coder_base(const coder_base&) 
{
	LEDA_EXCEPTION(1, "copy ctor of coder_base must not be called");
}

const coder_base& coder_base::operator=(const coder_base&)
{ 
	LEDA_EXCEPTION(1, "operator= of coder_base must not be called");
	return *this;
}
#endif

void 
coder_base::encode()
{
	do_encode();

	// NOTE: It is crucial to avoid the following check in do_encode!
	if (!finished() && !failed()) error("coder_base::encode: not finished");
}

void 
coder_base::decode()
{
	do_decode();

	// NOTE: It is crucial to avoid the following check in do_decode!
	if (!finished() && !failed()) error("coder_base::decode: not finished");
}

void 
coder_base::set_src_stream(streambuf* src_stream, bool own_stream)
{
	if (can_change_parameters()) {
		if (OwnSrcStream) delete pSrcStream;
		pSrcStream = src_stream; OwnSrcStream = own_stream;
	}
	else error("coder_base::set_src_stream: wrong state");
}

void 
coder_base::set_src_file(const char* file_name, ios::openmode mode)
{
	if (file_name == 0) { set_src_stream(0, false); return; }

	filebuf* stream = new filebuf;
	if (! stream->open(file_name, mode)) {
		delete stream; stream = 0;
		error("coder_base::set_src_file: open failed");
	}

	set_src_stream(stream, true); // own it and destroy it
}

void 
coder_base::set_tgt_stream(streambuf* tgt_stream, bool own_stream)
{
	if (can_change_parameters()) {
		if (OwnTgtStream) delete pTgtStream;
		pTgtStream = tgt_stream; OwnTgtStream = own_stream;
	}
	else error("coder_base::set_tgt_stream: wrong state");
}

void 
coder_base::set_tgt_file(const char* file_name, ios::openmode mode)
{
	if (file_name == 0) { set_tgt_stream(0, false); return; }

	filebuf* stream = new filebuf;
	if (! stream->open(file_name, mode)) {
		delete stream; stream = 0;
		error("coder_base::set_tgt_file: open failed");
	}

	set_tgt_stream(stream, true); // own it and destroy it
}


void 
coder_base::do_reset(reset_type t)
// NOTE: Every kind of reset puts the coder in the init-state again! 
{
	switch (t) {
	case r_hard: // put coder in exactly the same state as the default ctor
	case r_soft: // put coder in the same state as default ctor but keep parameters like buffer sizes
		set_state(s_init); // also clears errors
		//clear_parameters_lock();
		set_src_stream(0); set_tgt_stream(0);
		ProcessIdInStreams = true; ProcessHeaderInStreams = true;
		break;
	case r_seek: // reset coder for seek from beginning (rewinds source stream)
		if (pSrcStream->pubseekpos(PosBehindHeader, ios::in) != streampos(PosBehindHeader)) error();
		// fall-through
	case r_seek_without_rewind: // reset coder for seek (do not read Id or header again)
				// needed by coder_pipe and by block_coder (as r_block)
		// we keep the info if we were encoding or decoding
		set_state( coder_state( s_init | (State & (s_encoding | s_decoding))) );
		ProcessIdInStreams = false; ProcessHeaderInStreams = false;
		break;
	default:
		error(); return;
	}
}

void 
coder_base::do_encode()
{
	switch (State) {
	case s_init:
	case s_init_encoding:
						init_encoding();
						if (finished() || failed()) return;
						// else fall-through
	case s_encoding:	encode_stream(); break;
	default:			error("coder_base::encode: wrong state");
	}
}

void 
coder_base::init_encoding()
{
	if (State == s_init || State == s_init_encoding) {
		if (!pSrcStream || !pTgtStream) { 
			error("coder_base::init_encoding: missing stream"); return; 
		}

		if (ProcessIdInStreams)     write_id_to_tgt_stream();
		if (ProcessHeaderInStreams) write_header_to_tgt_stream();

		if (State == s_init) 
			PosBehindHeader = pTgtStream->pubseekoff(0, ios::cur, ios::out); // get absolute pos.

		State = s_encoding;
	}
}

void 
coder_base::encoding_finished()
{
	assert(encoding() || failed());
	if (State == s_encoding) {
		State = s_encoding_finished;
		pTgtStream->pubsync();
	}
}

void 
coder_base::do_decode()
{
	switch (State) {
	case s_init:
	case s_init_decoding:
						init_decoding();
						if (finished() || failed()) return;
						// else fall-through
	case s_decoding:	decode_stream(); break;
	default:			error("coder_base::decode: wrong state");
	}
}

void 
coder_base::init_decoding()
{
	if (State == s_init || State == s_init_decoding) {
		if (!pSrcStream || !pTgtStream) { 
			error("coder_base::init_decoding: missing stream"); return; 
		}

		if (ProcessIdInStreams)     read_and_check_id_in_src_stream();
		if (ProcessHeaderInStreams) read_header_from_src_stream();

		if (State == s_init) 
			PosBehindHeader = pSrcStream->pubseekoff(0, ios::cur, ios::in); // get absolute pos.

		State = s_decoding;
	}
}

void 
coder_base::decoding_finished()
{
	assert(decoding() || failed());
	if (State == s_decoding) {
		State = s_decoding_finished;
		pTgtStream->pubsync();
	}
}

/// encoding / decoding in memory >>>

coder_base::uint32 
coder_base::encode_memory_chunk(const char* in_buf,  uint32 in_len, 
								      char* out_buf, uint32 out_len)
{
	if (State != s_init) {
		error("coder_base::encode_memory_chunk: wrong state");
		return 0;
	}

	memory_streambuf in_mem_strbuf((char*) in_buf, streamsize(in_len),  false, false);
	memory_streambuf out_mem_strbuf(      out_buf, streamsize(out_len), false, false);

	streambuf* save_src_stream = change_src_stream_temporarily(&in_mem_strbuf);
	streambuf* save_tgt_stream = change_tgt_stream_temporarily(&out_mem_strbuf);

	init_encoding();
	if (!finished() && !failed()) encode_memory(in_mem_strbuf, out_mem_strbuf);

	change_src_stream_temporarily(save_src_stream); 
	change_tgt_stream_temporarily(save_tgt_stream);

	return uint32(out_mem_strbuf.out_count());
}

void 
coder_base::encode_memory(memory_streambuf& in, memory_streambuf& out)
{
	assert(get_src_stream() == &in && get_tgt_stream() == &out);
	encode_stream();
	if (!finished() || !in.had_underflow() || out.had_overflow()) error();
}

coder_base::uint32 
coder_base::decode_memory_chunk(const char* in_buf,  uint32 in_len, 
								      char* out_buf, uint32 out_len)
{
	if (State != s_init) {
		error("coder_base::encode_memory_chunk: wrong state");
		return 0;
	}

	memory_streambuf in_mem_strbuf((char*) in_buf, streamsize(in_len),  false, false);
	memory_streambuf out_mem_strbuf(      out_buf, streamsize(out_len), false, false);

	streambuf* save_src_stream = change_src_stream_temporarily(&in_mem_strbuf);
	streambuf* save_tgt_stream = change_tgt_stream_temporarily(&out_mem_strbuf);

	init_decoding();
	if (!finished() && !failed()) decode_memory(in_mem_strbuf, out_mem_strbuf);

	change_src_stream_temporarily(save_src_stream); 
	change_tgt_stream_temporarily(save_tgt_stream);

	return uint32(out_mem_strbuf.out_count());
}

void 
coder_base::decode_memory(memory_streambuf& in, memory_streambuf& out)
{
	assert(get_src_stream() == &in && get_tgt_stream() == &out);
	decode_stream();
	if (!finished() || in.in_avail()>0 || out.had_overflow()) error();
}

/// encoding / decoding in memory <<<

void
coder_base::error(const char* err_msg)
{
	State = s_error;
}


void 
coder_base::write_id_to_tgt_stream()
{
	ProcessIdInStreams = false;

	write_id(get_id());
}

void 
coder_base::read_and_check_id_in_src_stream()
{
	ProcessIdInStreams = false;

	id_type id_in_stream = 0;
	read_id(id_in_stream);
	check_id(id_in_stream);
}

void 
coder_base::check_id(id_type id_to_check)
{
	if (id_to_check != get_id()) {
		error("coder_base: bad header in stream");
	}
}

uint32 
coder_base::get_id_length(bool& is_exact) const
{
	is_exact = true; return IdLength;
}

void
coder_base::write_header_to_tgt_stream()
{
	ProcessHeaderInStreams = false;
}

void
coder_base::read_header_from_src_stream()
{
	ProcessHeaderInStreams = false;
}

uint32 
coder_base::get_header_length(bool& is_exact) const
{
	is_exact = true; return 0;
}

uint32 
coder_base::get_prolog_length(bool& is_exact) const
{
	uint32 len = 0;
	bool id_exact = true, header_exact = true;
	if (ProcessIdInStreams) len += get_id_length(id_exact);
	if (ProcessHeaderInStreams) len += get_header_length(header_exact);
	is_exact = id_exact && header_exact;
	return len;
}


void
coder_base::read_uint16(uint16& val)
{
#if defined(LEDA_CODER_EXACT_SIZE_INT16) && defined(LITTLE_ENDIAN_MACHINE)
	bool ok = read_bytes((byte*) &val, sizeof(val)) == sizeof(val);
	if (! ok) error("coder_base::read_uint16");
#else
	byte bytes[2];
	bool ok = read_bytes(bytes, 2) == 2;
	if (! ok) error("coder_base::read_uint16");
	val = uint16(bytes[0]) | (uint16(bytes[1]) << 8);
#endif
}

void 
coder_base::write_uint16(uint16 val)
{
#if defined(LEDA_CODER_EXACT_SIZE_INT16) && defined(LITTLE_ENDIAN_MACHINE)
	bool ok = write_bytes((byte*) &val, sizeof(val)) == sizeof(val);
	if (! ok) error("coder_base::write_uint16");
#else
	bool ok = (write_byte((byte)  (val & 0xFF)) != EOF)
		   && (write_byte((byte) ((val & 0xFF00) >> 8)) != EOF);
	if (! ok) error("coder_base::write_uint16");
#endif
}

void
coder_base::read_uint32(uint32& val)
{
#if defined(LEDA_CODER_EXACT_SIZE_INT32) && defined(LITTLE_ENDIAN_MACHINE)
	bool ok = read_bytes((byte*) &val, sizeof(val)) == sizeof(val);
	if (! ok) error("coder_base::read_uint32");
#else
	byte bytes[4];
	bool ok = read_bytes(bytes, 4) == 4;
	if (! ok) error("coder_base::read_uint32");
	val =  uint32(bytes[0])        | (uint32(bytes[1]) << 8) 
		| (uint32(bytes[2]) << 16) | (uint32(bytes[3]) << 24);
#endif
}

void 
coder_base::write_uint32(uint32 val)
{
#if defined(LEDA_CODER_EXACT_SIZE_INT32) && defined(LITTLE_ENDIAN_MACHINE)
	bool ok = write_bytes((byte*) &val, sizeof(val)) == sizeof(val);
	if (! ok) error("coder_base::write_uint32");
#else
	bool ok = (write_byte((byte)  (val & 0xFF)) != EOF)
		   && (write_byte((byte) ((val & 0xFF00) >> 8)) != EOF)
		   && (write_byte((byte) ((val & 0xFF0000) >> 16)) != EOF)
		   && (write_byte((byte) ((val & 0xFF000000) >> 24)) != EOF);
	if (! ok) error("coder_base::write_uint32");
#endif
}

bool 
coder_base::read_id(id_type& id, streambuf* stream)
{
	byte bytes[4];
	bool ok = stream->sgetn((char*)bytes, 4) == 4;
	id =  id_type(bytes[0])        | (id_type(bytes[1]) << 8) 
	   | (id_type(bytes[2]) << 16) | (id_type(bytes[3]) << 24);
	return ok;
}

bool
coder_base::write_id(id_type id, streambuf* stream)
{
	byte bytes[4];
	bytes[0] = (byte)  (id & 0xFF);
	bytes[1] = (byte) ((id & 0xFF00) >> 8);
	bytes[2] = (byte) ((id & 0xFF0000) >> 16);
	bytes[3] = (byte) ((id & 0xFF000000) >> 24);
	return stream->sputn((char*)bytes, 4) == 4;
}

coder_base* 
coder_base::construct_from_stream(streambuf* stream, streambuf* additional_data)
{
	id_type id_in_stream;
	bool ok = read_id(id_in_stream, stream);
	if (! ok) return 0;

	coder_base* coder = CoderFactory.construct_coder_by_id(id_in_stream);

	if (coder) coder->finish_construction_from_stream(id_in_stream, stream, additional_data);

	return coder;
}

void 
coder_base::finish_construction_from_stream(id_type id_in_stream, streambuf* stream, streambuf* additional_data)
{
	check_id(id_in_stream);
	process_no_id_in_streams();
}


#ifdef LEDA_COMPRESSION_WA8
int 
coder_base::check_eof(int c, streambuf* stream)
{ 
	if (c != EOF) return c; 
	int query = int(stream->pubseekoff(streamoff(1), ios::end, ios::in));
	return query == StreamBlocked ? StreamBlocked : EOF;
}
#endif


/// seek >>>>

streambuf* 
coder_base::change_src_stream_temporarily(streambuf* src_stream)
{
	streambuf* old_src_stream = pSrcStream;
	pSrcStream = src_stream;
	return old_src_stream;
}

streampos
coder_base::encode_seek(streampos current_pos, streampos new_pos)
{
	switch (State) {
	case s_init:
	case s_init_encoding:
		assert(current_pos == streampos(0));
		init_encoding();
		if (failed()) return streampos(-1);
		// else fall-through

	case s_encoding:
		{
			streamoff to_feed = new_pos - current_pos;
			if (to_feed == 0) return new_pos;
			if (to_feed < 0) { error(); return streamoff(-1); }

			get_feeder feeder(to_feed);
			streambuf* save_src_stream = change_src_stream_temporarily(&feeder);
			do_encode();
			change_src_stream_temporarily(save_src_stream);

			if (failed()) return streamoff(-1);

			assert( feeder.fed_everything() );
			return new_pos;
		}

	default:
		error("coder_base::encode_seek: wrong state");
		return streampos(-1);
	}
}

streambuf* 
coder_base::change_tgt_stream_temporarily(streambuf* tgt_stream)
{
	streambuf* old_tgt_stream = pTgtStream;
	pTgtStream = tgt_stream;
	return old_tgt_stream;
}

streampos
coder_base::decode_seek(streampos current_pos, streampos new_pos)
{
	switch (State) {
	case s_init:
	case s_init_decoding:
		assert(current_pos == streampos(0));
		init_decoding();
		if (failed()) return streampos(-1);
		assert(State != s_init && State != s_init_decoding);
		return decode_seek(current_pos, new_pos);
		// NOTE: this is NOT equivalent to falling-through because decode_seek is virtual

	case s_decoding:
	case s_decoding_finished: // in this case: to_eat will be negative!
		{
			streamoff to_eat = new_pos - current_pos;
			if (streamoff(new_pos) < streamoff(0)) to_eat = -1; // skip till EOF
			else if (to_eat < 0) {
				do_reset(r_seek); current_pos = 0;
				to_eat = new_pos;
			}
			if (to_eat == 0) return new_pos;

			put_eater eater(to_eat);
			streambuf* save_tgt_stream = change_tgt_stream_temporarily(&eater);
			do_decode();
			change_tgt_stream_temporarily(save_tgt_stream);

			if (to_eat != -1) {
				if (eater.get_number_of_puts() != to_eat) error();
			}
			else new_pos = current_pos + eater.get_number_of_puts();

			if (failed()) return streampos(-1);
			return new_pos;
		}

	default:
		error("coder_base::decode_seek: wrong state");
		return streampos(-1);
	}
}

bool 
coder_base::read_uint16(streambuf* stream, uint16& val)
{
#if defined(LEDA_CODER_EXACT_SIZE_INT16) && defined(LITTLE_ENDIAN_MACHINE)
	return stream->sgetn((char*) &val, sizeof(val)) == sizeof(val);
#else
	byte bytes[2];
	bool ok = stream->sgetn((char*) bytes, 2) == 2;
	val = uint16(bytes[0]) | (uint16(bytes[1]) << 8);
	return ok;
#endif
}

bool 
coder_base::read_uint32(streambuf* stream, uint32& val)
{
#if defined(LEDA_CODER_EXACT_SIZE_INT32) && defined(LITTLE_ENDIAN_MACHINE)
	return stream->sgetn((char*) &val, sizeof(val)) == sizeof(val);
#else
	byte bytes[4];
	bool ok = stream->sgetn((char*) bytes, 4) == 4;
	val =  uint32(bytes[0])        | (uint32(bytes[1]) << 8) 
		| (uint32(bytes[2]) << 16) | (uint32(bytes[3]) << 24);
	return ok;
#endif
}

bool 
coder_base::write_uint16(streambuf* stream, uint16 val)
{
#if defined(LEDA_CODER_EXACT_SIZE_INT16) && defined(LITTLE_ENDIAN_MACHINE)
	return stream->sputn((char*) &val, sizeof(val)) == sizeof(val);
#else
	bool ok = (stream->sputc((char)  (val & 0xFF)) != EOF)
		   && (stream->sputc((char) ((val & 0xFF00) >> 8)) != EOF);
	return ok;
#endif
}

bool 
coder_base::write_uint32(streambuf* stream, uint32 val)
{
#if defined(LEDA_CODER_EXACT_SIZE_INT32) && defined(LITTLE_ENDIAN_MACHINE)
	return stream->sputn((char*) &val, sizeof(val)) == sizeof(val);
#else
	bool ok = (stream->sputc((char)  (val & 0xFF)) != EOF)
		   && (stream->sputc((char) ((val & 0xFF00) >> 8)) != EOF)
		   && (stream->sputc((char) ((val & 0xFF0000) >> 16)) != EOF)
		   && (stream->sputc((char) ((val & 0xFF000000) >> 24)) != EOF);
	return ok;
#endif
}

/// coder_base_debug ////////////////////////////////////////////////////////////////////////

#ifdef LEDA_CODER_DEBUG
#if !defined(DEBUG) && !defined(_DEBUG)
#error LEDA_CODER_DEBUG should be switched off in release version
#endif

coder_base_debug::coder_base_debug(streambuf* src_stream, streambuf* tgt_stream, bool own_streams)
: base(src_stream, tgt_stream, own_streams),
  ReadLogCount(0), WriteLogCount(0)
{}

coder_base_debug::coder_base_debug(const char* src_file_name, const char* tgt_file_name)
: base(src_file_name, tgt_file_name),
  ReadLogCount(0), WriteLogCount(0)
{}

coder_base_debug* 
coder_base_debug::construct_from_stream(streambuf* stream, streambuf* additional_data)
{
	return static_cast<coder_base_debug*>(base::construct_from_stream(stream, additional_data));
}

int
coder_base_debug::log_read(int v)
{
	if (v != StreamBlocked && ReadLogCount < 2048) ReadLog[ReadLogCount++] = v;
	return v;
}

int
coder_base_debug::log_write(int v)
{
	if (v != StreamBlocked && WriteLogCount < 2048) WriteLog[WriteLogCount++] = v;
	return v;
}

int 
coder_base_debug::read_byte()
{
	return log_read(base::read_byte());
}

int 
coder_base_debug::read_bytes(byte* buffer, int buf_size)
{
	int res = base::read_bytes(buffer, buf_size);
	int copy = res; byte* p = buffer;
	while (copy > 0) { --copy; log_read((int) *(p++)); }
	return res;
}

int 
coder_base_debug::write_byte(byte b)
{
	log_write((int) b);
	return base::write_byte(b);
}

int 
coder_base_debug::write_bytes(byte* buffer, int buf_size)
{
	int res = base::write_bytes(buffer, buf_size);
	int copy = res; byte* p = buffer;
	while (copy > 0) { --copy; log_write((int) *(p++)); }
	return res;
}

#endif

LEDA_END_NAMESPACE
