// crypt module includes
#include <LEDA/coding/block_cipher_modes.h>
#include <LEDA/coding/block_cipher_factory.h>
#include <LEDA/coding/memory_streambuf.h>
#include <LEDA/coding/internal/strbuf_helpers.h>

// core library includes
#include <time.h> // for generating IVs

#if defined(LEDA_DEBUG)
#include <LEDA/system/assert.h>
#else
#undef assert
#define	assert(cond) ((void)0)
#endif

LEDA_BEGIN_NAMESPACE

typedef coder_types::uint32 uint32;

/// block_cipher_mode_base //////////////////////////////////////////////////////////////////

block_cipher_mode_base::block_cipher_mode_base(streambuf* src_stream, streambuf* tgt_stream, 
											   bool own_streams, 
											   block_cipher_base* cipher, bool own_cipher)
 : base(src_stream, tgt_stream, own_streams), 
   pBlockCipher(0)
{
	set_block_cipher(cipher, own_cipher);
}

block_cipher_mode_base::block_cipher_mode_base(const char* src_file_name, const char* tgt_file_name, 
											   block_cipher_base* cipher, bool own_cipher)
 : base(src_file_name, tgt_file_name), 
   pBlockCipher(0)
{
	set_block_cipher(cipher, own_cipher);
}

block_cipher_mode_base::~block_cipher_mode_base()
{
	set_state(s_finished); // clears errors
	//clear_parameters_lock();
	set_block_cipher(0, false); // free the block cipher if we own it
}

string 
block_cipher_mode_base::get_description() const
{
	string blk_cipher_descr 
		= get_block_cipher() ? get_block_cipher()->get_description() : string("none");

	return base::get_description() + string("[") + blk_cipher_descr + string("]");
}

block_cipher_mode_base::uint16 
block_cipher_mode_base::get_accepted_key_size(uint16& min, uint16& max) const
{
	if (!get_block_cipher()) {
		LEDA_EXCEPTION(1, "get_accepted_key_size: no block cipher"); return 0;
	}

	return get_block_cipher()->get_accepted_key_size(min, max);
}

void 
block_cipher_mode_base::set_key(const CryptKey& key)
{
	if (can_change_parameters()) Key = key;
	else error();
}


void 
block_cipher_mode_base::encoding_finished()
{
	Block.wipe();
	base::encoding_finished();
}

void 
block_cipher_mode_base::decoding_finished()
{
	Block.wipe();
	base::decoding_finished();
}

void 
block_cipher_mode_base::do_reset(reset_type t)
{
	base::do_reset(t);
	Block.wipe();
	if (t == r_hard) set_key(CryptKey());
}

void 
block_cipher_mode_base::write_id_to_tgt_stream()
{
	base::write_id_to_tgt_stream();

	if (!get_block_cipher()) { error(); return; }
	write_id(get_block_cipher()->get_id());
}

void 
block_cipher_mode_base::read_and_check_id_in_src_stream()
{
	base::read_and_check_id_in_src_stream();

	if (!get_block_cipher()) { error(); return; }
	id_type id_cipher;
	read_id(id_cipher);
	if (id_cipher != get_block_cipher()->get_id()) { error(); return; }
}

uint32 
block_cipher_mode_base::get_id_length(bool& is_exact) const
{
	return base::get_id_length(is_exact) + coder_base::IdLength /* cipher id*/;
}

void 
block_cipher_mode_base::finish_construction_from_stream(id_type id_in_stream, streambuf* stream, streambuf* additional_data)
{
	base::finish_construction_from_stream(id_in_stream, stream, additional_data);

	// construct the block_cipher from the id in the stream
	id_type id_cipher;
	read_id(id_cipher, stream);
	block_cipher_base* cipher = BlockCipherFactory.construct_cipher_by_id(id_cipher);
	if (!cipher || !additional_data) { error(); return; }
	set_block_cipher(cipher, true);

	// now we reed the key from additional_data (i.e. the stream of secrets)
	istream key_stream(additional_data);
	key_stream >> Key;
}

void 
block_cipher_mode_base::set_block_cipher(block_cipher_base* cipher, bool own_cipher)
{
	if (can_change_parameters()) {
		if (OwnBlockCipher) delete pBlockCipher;
		pBlockCipher = cipher; OwnBlockCipher = own_cipher;
		Block.set_size(cipher ? cipher->get_block_size() : 0);
	}
	else error();
}

uint32 
block_cipher_mode_base::calculate_length_of_encoded_data(uint32 input_length) const
{
	bool is_exact;
	uint32 output_length = 0;
	output_length += get_prolog_length(is_exact); assert(is_exact);
	output_length += calculate_length_of_encoded_data_without_prolog(input_length, is_exact); assert(is_exact);
	return output_length;
}

void 
block_cipher_mode_base::memxor(byte* tgt, const byte* src, size_t sz)
{
	      unsigned long* tgt_ul  =       (unsigned long*) tgt;
	const unsigned long* src_ul  = (const unsigned long*) src;
	const unsigned long* stop_ul = src_ul + sz / sizeof(unsigned long);
	while (src_ul != stop_ul)
		*(tgt_ul++) ^= *(src_ul++);

	tgt = (byte*) tgt_ul; src = (const byte*) src_ul;
	for (size_t remainder = sz % sizeof(unsigned long); remainder > 0; --remainder)
		*(tgt++) ^= *(src++);
}

/// bcm_stream_cipher_base //////////////////////////////////////////////////////////////////

bcm_stream_cipher_base::bcm_stream_cipher_base(streambuf* src_stream, streambuf* tgt_stream, 
											   bool own_streams, 
											   block_cipher_base* cipher, bool own_cipher)
 : base(src_stream, tgt_stream, own_streams, cipher, own_cipher)
{}

bcm_stream_cipher_base::bcm_stream_cipher_base(const char* src_file_name, const char* tgt_file_name, 
											   block_cipher_base* cipher, bool own_cipher)
 : base(src_file_name, tgt_file_name, cipher, own_cipher)
{}

void 
bcm_stream_cipher_base::init_encoding()
{
	base::init_encoding();
	if (failed()) return;

	get_block_cipher()->set_key(get_key(), block_cipher_dir_for_enciphering());
	pCurrent = get_block(); pStop = get_block() + get_block_cipher()->get_block_size();
}

void 
bcm_stream_cipher_base::encode_stream()
{
	const int blk_sz = get_block_cipher()->get_block_size();

	int c;
	while ((c = read_byte()) >= 0 && c != EOF) {
		*pCurrent = (byte) c;
		if (++pCurrent == pStop) {
			encipher_block();

			write_bytes(get_block(), blk_sz);
			pCurrent = get_block();
		}
	}

	if (c != EOF) return;

	// encode the last block with ciphertext stealing
	byte bytes_used = byte(pCurrent - get_block());
	assert(bytes_used < get_block_cipher()->get_block_size());
	get_block()[blk_sz - 1] = bytes_used;

	encipher_block();
	if (write_bytes(get_block(), blk_sz) != blk_sz) { error(); return; }
	
	encoding_finished();
}

void 
bcm_stream_cipher_base::init_decoding()
{
	base::init_decoding();
	if (failed()) return;

	get_block_cipher()->set_key(get_key(), block_cipher_dir_for_deciphering());
	prepare_resume_decoding();
}

void 
bcm_stream_cipher_base::prepare_resume_decoding()
{
	pStop = get_block() + get_block_cipher()->get_block_size(); pCurrent = pStop;
	LastBlockFetched = false;
}

void 
bcm_stream_cipher_base::decode_stream()
{
	const int blk_sz = get_block_cipher()->get_block_size();

	while (pCurrent != pStop) if (write_byte(*(pCurrent++)) < 0) return;

	while (!LastBlockFetched) {
		pCurrent = get_block();
		if (read_bytes(pCurrent, blk_sz) != blk_sz) { error(); break; }
		decipher_block();

		int c = read_byte();
		if (c != EOF)
			get_src_stream()->sputbackc((char)c);
		else {
			// treat last block
			LastBlockFetched = true;
			byte bytes_used = get_block()[blk_sz - 1];
			if (bytes_used >= blk_sz) { error(); break; } // some integrity check
			pStop = get_block() + bytes_used;
		}

		while (pCurrent != pStop) if (write_byte(*(pCurrent++)) < 0) return;
	}

	decoding_finished();
}

uint32 
bcm_stream_cipher_base::calculate_length_of_encoded_data_without_prolog(uint32 input_length, bool& is_exact) const
{
	block_cipher_base* cipher = get_block_cipher();
	if (!cipher) { is_exact = false; return 0; }
	uint32 blk_sz = cipher->get_block_size();
	/*
		If input_sz is a multiple of blk_sz, we add one whole block of padding.
		Otherwise, we pad the final block (with cipher text stealing).
	*/
	is_exact = true;
	return (input_length / blk_sz + 1) * blk_sz;
}

/// bcm_stream_cipher_with_iv_base //////////////////////////////////////////////////////////

bcm_stream_cipher_with_iv_base::bcm_stream_cipher_with_iv_base(
	streambuf* src_stream, streambuf* tgt_stream, bool own_streams,
	block_cipher_base* cipher, bool own_cipher)
 : base(src_stream, tgt_stream, own_streams, cipher, own_cipher),
   ExternalIV(0)
{}

bcm_stream_cipher_with_iv_base::bcm_stream_cipher_with_iv_base(
	const char* src_file_name, const char* tgt_file_name,
	block_cipher_base* cipher, bool own_cipher)
 : base(src_file_name, tgt_file_name, cipher, own_cipher),
   ExternalIV(0)
{}

void 
bcm_stream_cipher_with_iv_base::init_encoding()
{
	base::init_encoding();
	if (failed()) return;

	const int blk_sz = get_block_cipher()->get_block_size();
	FeedbackBlock.set_size(blk_sz); TempBlock.set_size(blk_sz);

	byte* iv = FeedbackBlock;
	if (! ExternalIV) {
		// generate IV
		time_t seed1; time(&seed1);
		clock_t seed2 = clock(); while (seed2 == clock());

		size_t s1 = sizeof(seed1), s2 = sizeof(seed2);
		if (s1 + s2 > size_t(blk_sz)) {
			s1 = size_t(blk_sz) / 2; s2 = size_t(blk_sz) - s1;
			if (s1 < 4) /* we need at least 4 bytes of each seed1 */ {
				LEDA_EXCEPTION(0, "block size too small, initialization vector might not be unique!");
			}
		}

		size_t b;
		for (b = 0; b < s1 && b < size_t(blk_sz); ++b) {
			iv[b] = (byte) seed1; seed1 >>= 8;
		}
		for (; b < (s1 + s2) && b < size_t(blk_sz); ++b) {
			iv[b] = (byte) seed2; seed2 >>= 8;
		}
	}
	else {
		memcpy(iv, ExternalIV, blk_sz);
	}
	ExternalIV = 0; // do not use it again

	// allow the actual block cipher mode to change the iv before its written/used
	finish_initialization_vector(iv);

	// write IV
	if (write_bytes(iv, blk_sz) != blk_sz) { error(); }
}

void 
bcm_stream_cipher_with_iv_base::encoding_finished()
{
	FeedbackBlock.wipe(); TempBlock.wipe();
	base::encoding_finished();
}

void 
bcm_stream_cipher_with_iv_base::init_decoding()
{
	base::init_decoding();
	if (failed()) return;

	// read IV
	const int blk_sz = get_block_cipher()->get_block_size();
	FeedbackBlock.set_size(blk_sz); TempBlock.set_size(blk_sz);
	if (read_bytes(FeedbackBlock, blk_sz) != blk_sz) { error(); }
}

void 
bcm_stream_cipher_with_iv_base::decoding_finished()
{
	FeedbackBlock.wipe(); TempBlock.wipe();
	base::decoding_finished();
}


void 
bcm_stream_cipher_with_iv_base::do_reset(reset_type t)
{
	base::do_reset(t);
	ExternalIV = 0;
	FeedbackBlock.wipe(); TempBlock.wipe(); // safety first ...
}


uint32 
bcm_stream_cipher_with_iv_base::calculate_length_of_encoded_data_without_prolog(uint32 input_length, bool& is_exact) const
{
	block_cipher_base* cipher = get_block_cipher();
	if (!cipher) { is_exact = false; return 0; }
	uint32 blk_sz = cipher->get_block_size();
	return base::calculate_length_of_encoded_data_without_prolog(input_length, is_exact) 
	     + blk_sz /* IV */;
}

/// bcm_str_cipher_with_cipher_feedb_base ///////////////////////////////////////////////////

bcm_str_cipher_with_cipher_feedb_base::bcm_str_cipher_with_cipher_feedb_base(
	streambuf* src_stream, streambuf* tgt_stream, bool own_streams,
	block_cipher_base* cipher, bool own_cipher)
 : base(src_stream, tgt_stream, own_streams, cipher, own_cipher)
{}

bcm_str_cipher_with_cipher_feedb_base::bcm_str_cipher_with_cipher_feedb_base(
	const char* src_file_name, const char* tgt_file_name,
	block_cipher_base* cipher, bool own_cipher)
 : base(src_file_name, tgt_file_name, cipher, own_cipher)
{}

streampos 
bcm_str_cipher_with_cipher_feedb_base::decode_seek(streampos current_pos, streampos new_pos)
{
	if (!decoding() || failed()) return base::decode_seek(current_pos, new_pos);

	const streamoff blk_sz(get_block_cipher()->get_block_size());
	prepare_resume_decoding();

	streampos pos_to_seek_to_in_dec_stream; // rel. to the beginning of the decoded(!) stream
	if (new_pos < streamoff(0)) {
		// seek to the end
		//   -> we have to decode the last block to determine 
		//      the number of used bytes in this block.
		pos_to_seek_to_in_dec_stream = seek_end_src_stream_relative_to_header() - 2*blk_sz; 
			// subtract last block and IV (because the encoded stream begins with it
			// and we want the pos. in the decoded stream)
	}
	else {
		// regular seek
		set_state(s_decoding); // make it running ...
		streamoff remainder = new_pos % blk_sz;
		pos_to_seek_to_in_dec_stream = new_pos - remainder; 
			// start of block that contains current_pos (rel. to beg. of decoded stream)
	}

	current_pos = seek_pos_src_stream_relative_to_header(pos_to_seek_to_in_dec_stream);
	if (read_bytes(get_feedback_blk(), (int)blk_sz) != blk_sz) {
		error(); return streampos(-1);
	}
	// current_pos is correct because the IV is at rel. position zero in the
	// encoded(!) src_stream and we have just read one block.

	return base::decode_seek(current_pos, new_pos);
}


/// ecb_mode ////////////////////////////////////////////////////////////////////////////////

streampos 
ecb_mode::decode_seek(streampos current_pos, streampos new_pos)
{
	if (!decoding() || failed()) return base::decode_seek(current_pos, new_pos);

	const streamoff blk_sz(get_block_cipher()->get_block_size());
	prepare_resume_decoding();

	if (new_pos < streamoff(0)) {
		// seek to the end
		streampos size = seek_end_src_stream_relative_to_header();
		current_pos = seek_pos_src_stream_relative_to_header(size - blk_sz);
		return base::decode_seek(current_pos, new_pos); // only the last block
	}

	// regular seek
	set_state(s_decoding); // make it running ...
	streamoff remainder = new_pos % blk_sz;
	current_pos = seek_pos_src_stream_relative_to_header(new_pos - remainder);
	return base::decode_seek(current_pos, new_pos);
}

/// cbc_mode ////////////////////////////////////////////////////////////////////////////////

void
cbc_mode::finish_initialization_vector(byte* iv)
// NOTE: follows NIST document sp800-38a.pdf appendix C
{
	get_block_cipher()->encipher(iv);
	base::finish_initialization_vector(iv);
}

void 
cbc_mode::encipher_block()
{
	const size_t blk_sz = get_block_cipher()->get_block_size();
	byte* p_blk = get_block();
	byte* p_prev_cipher_blk = get_feedback_blk();

	memxor(p_blk, p_prev_cipher_blk, blk_sz); // p_blk = P_i ^ C_{i-1}
	get_block_cipher()->encipher(p_blk);      // p_blk = E_k(P_i ^ C_{i-1}) = C_i
	memcpy(p_prev_cipher_blk, p_blk, blk_sz); // C_i -> p_prev_cipher_blk
}

void 
cbc_mode::decipher_block()
{
	const size_t blk_sz = get_block_cipher()->get_block_size();
	byte* p_blk = get_block();
	byte* p_prev_cipher_blk = get_feedback_blk();
	byte* p_tmp_blk = get_tmp_blk();

	memcpy(p_tmp_blk, p_prev_cipher_blk, blk_sz); // C_{i-1} -> p_tmp_blk
	memcpy(p_prev_cipher_blk, p_blk, blk_sz);     // C_i -> p_prev_cipher_blk
	get_block_cipher()->decipher(p_blk);          // p_blk = D_k(C_i) = P_i ^ C_{i-1}
	memxor(p_blk, p_tmp_blk, blk_sz);             // p_blk = P_i ^ C_{i-1} ^ C_{i-1} = P_i
}


/// cfb_mode ////////////////////////////////////////////////////////////////////////////////

void
cfb_mode::finish_initialization_vector(byte* iv)
// NOTE: follows NIST document sp800-38a.pdf appendix C
{
	get_block_cipher()->encipher(iv);
	base::finish_initialization_vector(iv);
}

void 
cfb_mode::encipher_block()
{
	const size_t blk_sz = get_block_cipher()->get_block_size();
	byte* p_blk = get_block();
	byte* p_prev_cipher_blk = get_feedback_blk();

	get_block_cipher()->encipher(p_prev_cipher_blk); // p_prev_cipher_blk = E_k(C_{i-1})
	memxor(p_blk, p_prev_cipher_blk, blk_sz);        // p_blk = P_i ^ E_k(C_{i-1}) = C_i
	memcpy(p_prev_cipher_blk, p_blk, blk_sz);        // C_i -> p_prev_cipher_blk
}

void 
cfb_mode::decipher_block()
{
	const size_t blk_sz = get_block_cipher()->get_block_size();
	byte* p_blk = get_block();
	byte* p_prev_cipher_blk = get_feedback_blk();
	byte* p_tmp_blk = get_tmp_blk();

	memcpy(p_tmp_blk, p_prev_cipher_blk, blk_sz); // C_{i-1} -> p_tmp_blk
	memcpy(p_prev_cipher_blk, p_blk, blk_sz);     // C_i -> p_prev_cipher_blk
	get_block_cipher()->encipher(p_tmp_blk);      // p_tmp_blk = E_k(C_{i-1}
	memxor(p_blk, p_tmp_blk, blk_sz);             // p_blk = C_i ^ E_k(C_{i-1} = P_i
}

/// ofb_mode ////////////////////////////////////////////////////////////////////////////////

void 
ofb_mode::encipher_block()
{
	const size_t blk_sz = get_block_cipher()->get_block_size();
	byte* p_blk = get_block();
	byte* p_prev_cipher_blk = get_feedback_blk();

	get_block_cipher()->encipher(p_prev_cipher_blk); // p_prev_cipher_blk = E_k(S_{i-1}) = S_i
	memxor(p_blk, p_prev_cipher_blk, blk_sz);        // p_blk = P_i ^ S_i = C_i
}

void 
ofb_mode::decipher_block()
{
	const size_t blk_sz = get_block_cipher()->get_block_size();
	byte* p_blk = get_block();
	byte* p_prev_cipher_blk = get_feedback_blk();

	get_block_cipher()->encipher(p_prev_cipher_blk); // p_prev_cipher_blk = E_k(S_{i-1}) = S_i
	memxor(p_blk, p_prev_cipher_blk, blk_sz);        // p_blk = C_i ^ S_i = P_i
}


/// omac_mode ///////////////////////////////////////////////////////////////////////////////

omac_mode::omac_mode(streambuf* src_stream, streambuf* tgt_stream, bool own_streams,
					 block_cipher_base* cipher, bool own_cipher)
 : base(src_stream, tgt_stream, own_streams),
   MACInStream(true), MACState(mac_invalid), ReadAheadBuffer(MaxBlkSz + 1)
{
	set_block_cipher(cipher, own_cipher); // virtual function
}

omac_mode::omac_mode(const char* src_file_name, const char* tgt_file_name,
					 block_cipher_base* cipher, bool own_cipher)
 : base(src_file_name, tgt_file_name),
   MACInStream(true), MACState(mac_invalid), ReadAheadBuffer(MaxBlkSz + 1)
{
	set_block_cipher(cipher, own_cipher); // virtual function
}

const omac_mode::byte*
omac_mode::check()
{
	if (! get_tgt_stream()) {
		// we use a dummy streambuf!
		set_tgt_stream(new eat_all_streambuf(), true);
		decode();
		set_tgt_stream(0);
	}
	else decode();

	return (MACState == mac_okay) ? (byte*) MAC : (byte*)0;
}

const omac_mode::byte*
omac_mode::compute_MAC(string str)
{
	memory_streambuf mbuf(str.cstring(), str.length());
	set_src_stream(&mbuf, false); set_tgt_stream(0);
	bool sav_flag = get_MAC_in_stream_flag();
	set_MAC_in_stream_flag(false);
	decode();
	set_MAC_in_stream_flag(sav_flag);
	set_src_stream(0);
	return (MACState == mac_okay) ? (byte*) MAC : (byte*)0;
}

string 
omac_mode::get_MAC_as_hex_string() const
{
	string result;
	if (MACState != mac_okay || !get_block_cipher()) return result;

	const uint32 blk_sz = get_block_cipher()->get_block_size();
	for (uint32 i = 0; i < blk_sz; ++i) {
		result += string("%02x", MAC[i]);
	}
	return result;
}

void 
omac_mode::set_MAC_in_stream_flag(bool mac_in_stream)
{
	if (can_change_parameters()) {
		MACInStream = mac_in_stream;
	}
	else error();
}

void 
omac_mode::set_block_cipher(block_cipher_base* cipher, bool own_cipher)
{
	if (cipher) {
		const uint16 blk_sz = cipher->get_block_size();
		if ((blk_sz != 128/8 && blk_sz != 64/8 && blk_sz != 256/8) || blk_sz > MaxBlkSz) {
			LEDA_EXCEPTION(1, "omac_mode: block size not supported"); return;
		}
	}
	base::set_block_cipher(cipher, own_cipher);
	MACState = mac_invalid;
}

void 
omac_mode::init_encoding()
{
	if (! MACInStream) process_no_id_in_streams();
	base::init_encoding();

	MACState = mac_computing;
	init_processing(false); // false -> no read ahead during encoding
}

void 
omac_mode::encode_stream()
{
	if (! process_stream()) return;
	
	// write the MAC (if the user wants this)
	if (MACInStream) {
		const int blk_sz = get_block_cipher()->get_block_size();
		if (write_bytes(MAC, blk_sz) != blk_sz) { error(); return; }
	}

	encoding_finished();
}

void 
omac_mode::init_decoding()
{
	if (! MACInStream) process_no_id_in_streams();
	base::init_decoding();

	MACState = mac_computing;
	init_processing(MACInStream);
}

void 
omac_mode::decode_stream()
{
	if (! process_stream()) return;

	if (MACInStream && MAC_is_valid()) {
		// compare ReadAheadBuffer with the computed MAC
		pCurrent = MAC;
		while (! ReadAheadBuffer.empty())
			if (*(pCurrent++) != ReadAheadBuffer.pop()) {
				error(); return; // but the MAC is still okay! (-> get_MAC_as_hex_string())
			}
	}

	decoding_finished();
}

void 
omac_mode::do_reset(reset_type t)
{
	base::do_reset(t);

	MACState = mac_invalid;
	if (t == r_hard) {
		MACInStream = true;
	}
}

streampos 
omac_mode::decode_seek(streampos current_pos, streampos new_pos)
{
	if (!decoding() || failed()) return base::decode_seek(current_pos, new_pos);

	MACState = mac_invalid; // no MAC will be computed!

	if (new_pos < streamoff(0)) {
		// seek to the end of the stream
		if (!finished()) decoding_finished();
		streampos res = seek_end_src_stream_relative_to_header();
		if (MACInStream) res -= streamoff(get_block_cipher()->get_block_size());
		if (res < streamoff(0)) error();
		return res;
	}

	// regular seek
	set_state(s_decoding); // make it running ...
	streampos res = seek_pos_src_stream_relative_to_header(new_pos);
	if (res < streamoff(0)) error();

	init_processing(MACInStream); // fill ReadAheadBuffer ...

	return res;
}

void 
omac_mode::init_processing(bool read_ahead)
{
	if (failed()) return;

	const int blk_sz = get_block_cipher()->get_block_size();

	ReadAheadBuffer.clear();
	if (read_ahead) {
		while (ReadAheadBuffer.size() < blk_sz) {
			int c = read_byte();
			if (c < 0 || c == EOF) { error(); return; }
			ReadAheadBuffer.append((byte) c);
		}
	}

	MAC.set_size(blk_sz);
	if (MACState != mac_invalid) { // (if the MAC is invalid anyway, we can save these computations)
		get_block_cipher()->set_key(get_key(), block_cipher_base::d_encipher);
		pCurrent = get_block(); pStop = get_block() + blk_sz;
		memset(MAC, 0, blk_sz);
	}
}

bool 
omac_mode::process_stream()
{
	const int blk_sz = get_block_cipher()->get_block_size();

	// if the MAC is invalid anyway, we can save some computations
	if (MACState == mac_invalid) {
		int c;
		while ((c = read_byte()) >= 0 && c != EOF) {
			ReadAheadBuffer.append((byte) c);
			byte curr_sym = ReadAheadBuffer.pop();
			if (write_byte((byte) curr_sym) < 0) return false;
		}
		return c == EOF;
	}

	// read a byte, do a CBC update when the block is full, and write the byte
	int c;
	while ((c = read_byte()) >= 0 && c != EOF) {
		if (pCurrent == pStop) {
			memxor(MAC, get_block(), blk_sz);
			get_block_cipher()->encipher(MAC);

			pCurrent = get_block();
		}

		ReadAheadBuffer.append((byte) c);
		byte curr_sym = *(pCurrent++) = ReadAheadBuffer.pop();

		if (write_byte((byte) curr_sym) < 0) return false;
	}
	if (c != EOF) return false;

	// EOF reached in src stream -> finish MAC computation

	// pad the last block, if necessary
	int bytes_used = int(pCurrent - get_block());
	bool needs_pading = bytes_used != blk_sz;
	if (needs_pading) {
		*pCurrent = 0x80;
		while (++pCurrent != pStop) *pCurrent = 0x00;
	}
	memxor(MAC, get_block(), blk_sz);

	// compute H_L = L*u (resp. L*u*u) in block
	memset(get_block(), 0, blk_sz);
	get_block_cipher()->encipher(get_block());
	multiply_blk_by_u();
	if (needs_pading) multiply_blk_by_u();

	// add H_L to MAC and encrypt
	memxor(MAC, get_block(), blk_sz);
	get_block_cipher()->encipher(MAC);

	MACState = (MACState == mac_computing) ? mac_okay : mac_invalid;
	return true;
}

void 
omac_mode::multiply_blk_by_u()
{
   const int blk_sz = get_block_cipher()->get_block_size();

   // big endian format 
   bool most_sig_bit_set = (get_block()[0] & 0x80) != 0;

   // shift right to multiply by u (most sig. bit gets lost)
   for(int i = 0; i < blk_sz - 1; ++i)
      get_block()[i] = (get_block()[i] << 1) | (get_block()[i + 1] >> 7);

   get_block()[blk_sz - 1] <<= 1;

   // if the old most sig. bit was set we have to add u^blk_sz
   if (most_sig_bit_set) {
		switch (blk_sz) {
		case 128/8:	get_block()[blk_sz - 1] ^= 0x87; break; 
					// u^128 = u^7+u^2+u^1+u^0
		case 64/8:	get_block()[blk_sz - 1] ^= 0x1B; break; 
					// u^64  = u^4+u^3+u^1+u^0
		case 256/8:	get_block()[blk_sz - 2] ^= 0x04; get_block()[blk_sz - 1] ^= 0x25; break; 
					// u^256 = u^10+u^5+u^2+u^0
		default:	error(); MACState = mac_invalid;
		}
    }
}


uint32 
omac_mode::get_prolog_length(bool& is_exact) const
{
	uint32 len = 0;
	bool id_exact = true, header_exact = true;
	if (look_for_id_in_streams() && MACInStream) len += get_id_length(id_exact);
	if (look_for_header_in_streams()) len += get_header_length(header_exact);
	is_exact = id_exact && header_exact;
	return len;
}

uint32 
omac_mode::calculate_length_of_encoded_data_without_prolog(uint32 input_length, bool& is_exact) const
{
	block_cipher_base* cipher = get_block_cipher();
	if (!cipher) { is_exact = false; return 0; }
	uint32 blk_sz = cipher->get_block_size();
	// The output is simply the input with (possibly) the MAC appended.
	is_exact = true;
	return input_length + (MACInStream ? blk_sz : 0);
}

LEDA_END_NAMESPACE
