// crypt module includes
#include <LEDA/coding/crypt_key.h>
#include <LEDA/coding/checksum.h> // for MD5SumCoder
#include <LEDA/coding/memory_streambuf.h>
#include <LEDA/coding/block_cipher_modes.h> // for memxor

// core LEDA includes
#include <LEDA/core/random_source.h> // for salt generation
#include <LEDA/system/date.h> // for salt generation
#include <LEDA/system/error.h>
#include <stdio.h> // for read_phrase below

#if defined(LEDA_DEBUG)
#include <LEDA/system/assert.h>
#else
#undef assert
#define	assert(cond) ((void)0)
#endif

#ifdef __win32__
#undef MAXINT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

LEDA_BEGIN_NAMESPACE

typedef coder_types::byte   byte;
typedef coder_types::uint16 uint16;
typedef coder_types::uint32 uint32;
typedef coder_types::sint32 sint32;

#define b0(x) ((byte)(x))
#define b1(x) ((byte)((x) >> 8))
#define b2(x) ((byte)((x) >> 16))
#define b3(x) ((byte)((x) >> 24))

/// secure_byte_array ///////////////////////////////////////////////////////////////////////

secure_byte_array::secure_byte_array(const byte* bytes, uint32 sz)
 : Size(0), Bytes(0)
{
	set_size(sz);
	memcpy(Bytes, bytes, Size);
}

secure_byte_array::secure_byte_array(uint32 sz)
 : Size(0), Bytes(0)
{
	set_size(sz);
	memset(Bytes, 0, Size);
}

void 
secure_byte_array::set_size(uint32 sz)
{
	if (sz != Size) {
		if (Size > 0) {
			wipe();
#ifdef __win32__
			if (Size <= MaxSizeLockAgainstSwap) VirtualUnlock(Bytes, Size);
#endif
			if (Size % (32/8) == 0 && Size <= (256/8))
				std_memory.deallocate_bytes(Bytes, Size);
			else
				std_memory.deallocate_vector(Bytes);
			Bytes = 0;
		}
		Size = sz;
		if (Size > 0) {
			if (Size % (32/8) == 0 && Size <= (256/8))
				Bytes = (byte*) std_memory.allocate_bytes(Size);
			else
				Bytes = (byte*) std_memory.allocate_vector(Size);
#ifdef __win32__
			if (Size <= MaxSizeLockAgainstSwap) {
				if (! VirtualLock(Bytes, Size))
					LEDA_EXCEPTION(0, "secure_byte_array: could not memory!");
			}
#endif
		}
	}
}

/// CryptByteString /////////////////////////////////////////////////////////////////////////

CryptByteString::CryptByteString(const char* str)
{
	PTR = new rep();
	operator=( from_string(str) );
}

byte
CryptByteString::operator[](uint16 idx) const
{
	if (/* idx < 0 || */ idx >= get_size()) {
		LEDA_EXCEPTION(1, "CryptByteString::operator[]: index out of range"); return 0;
	}
	return get_bytes()[idx];
}

byte&
CryptByteString::operator[](uint16 idx)
{
	if (/* idx < 0 || */ idx >= get_size()) {
		LEDA_EXCEPTION(1, "CryptByteString::operator[]: index out of range"); idx = 0;
	}

	make_single_ref();
	return ptr()->get_bytes()[idx];
}

CryptByteString
CryptByteString::operator()(uint16 idx_start, uint16 idx_end) const
{
	if (/* idx_start < 0 || */ idx_start > idx_end || idx_end >= get_size())
		return CryptByteString();

	uint16 sz = idx_end + 1 - idx_start;
	CryptByteString result(sz);
	memcpy(result.get_bytes(), get_bytes() + idx_start, size_t(sz));
	return result;
}

CryptByteString 
CryptByteString::hash(uint16 num_iterations) const
{
	SHACoder hasher;
	CryptByteString result(hasher.get_byte_length_of_checksum());

	const byte* input_bytes = get_bytes(); uint16 input_sz = get_size();
	do {
		memory_streambuf buffer((char*) input_bytes, (streamsize) input_sz);
		hasher.reset();
		hasher.set_checksums_in_stream_flag(false);
		hasher.set_src_stream(&buffer);
		hasher.check(); assert(!hasher.failed());
		memcpy(result.get_bytes(), hasher.get_digest(), result.get_size());
		input_bytes = result.get_bytes(); input_sz = result.get_size();
	} while (--num_iterations > 0);

	return result;
}

string 
CryptByteString::to_hex_string() const
{
	string result;
	for (uint16 i = 0; i < get_size(); ++i)
		result += string("%02x", get_bytes()[i]); // slow!
	return result;
}

string 
CryptByteString::to_string() const
{
	string result;
	for (uint16 i = 0; i < get_size(); ++i)
		result += string((char) get_bytes()[i]); // slow!
	return result;
}

CryptByteString 
CryptByteString::from_hex_string(const char* hex_str)
{
	uint16 len = (uint16) strlen(hex_str);
	if (len % 2 != 0) {
		LEDA_EXCEPTION(1, "CryptByteString::CryptByteString: invalid string");
	}

	CryptByteString result(len/2);
	int pos = 0;
	for (uint16 i = 0; i < result.get_size(); ++i) {
		byte lo_nibble = 0;
		if (hex_str[pos] >= '0' && hex_str[pos] <= '9') lo_nibble = hex_str[pos] - '0';
		else if (hex_str[pos] >= 'a' && hex_str[pos] <= 'f') lo_nibble = hex_str[pos] - 'a' + 10;
		else if (hex_str[pos] >= 'A' && hex_str[pos] <= 'F') lo_nibble = hex_str[pos] - 'A' + 10;
		else LEDA_EXCEPTION(1, "CryptByteString::CryptByteString: invalid string");
		byte hi_nibble = 0;
		if (hex_str[pos] >= '0' && hex_str[pos] <= '9') hi_nibble = hex_str[pos] - '0';
		else if (hex_str[pos] >= 'a' && hex_str[pos] <= 'f') hi_nibble = hex_str[pos] - 'a' + 10;
		else if (hex_str[pos] >= 'A' && hex_str[pos] <= 'F') hi_nibble = hex_str[pos] - 'A' + 10;
		else LEDA_EXCEPTION(1, "CryptByteString::CryptByteString: invalid string");
		result[i] = lo_nibble + (hi_nibble << 4);
	}
	return result;
}

CryptByteString 
CryptByteString::from_string(const char* str)
{
	CryptByteString result((uint16) strlen(str));
	for (uint16 i = 0; i < result.get_size(); ++i)
		result[i] = str[i];
	return result;
}

CryptByteString operator+(const CryptByteString& x, const CryptByteString& y)
{
	if (y.get_size() == 0) return x;

	CryptByteString res(x.get_size() + y.get_size());
	byte* bytes = res.get_bytes();
	memcpy(bytes, x.get_bytes(), x.get_size());
	bytes += x.get_size();
	memcpy(bytes, y.get_bytes(), y.get_size());
	return res;
}


streamsize
CryptByteString::get_size_in_stream() const
{
	return 2 + get_size(); // 2 = 16 bits for the key size
}

istream& operator>>(istream& i, CryptByteString& str)
{
	// read key size
	byte sz_lo = (byte) i.get();
	byte sz_hi = (byte) i.get();
	uint16 sz = sz_lo + (sz_hi << 8);
	// read key bytes
	CryptByteString new_str(sz);
	i.read((char*) new_str.get_bytes(), sz);
	str = i.fail() ? CryptByteString() : new_str;

	return i;
}

ostream& operator<<(ostream& o, const CryptByteString& str)
{
	// write key size
	uint16 sz = str.get_size();
	o.put((char) b0(sz)); // write low byte of sz
	o.put((char) b1(sz)); // write high byte of sz
	// write key bytes
	o.write((const char*) str.get_bytes(), sz);

	return o;
}

/// CryptKey ////////////////////////////////////////////////////////////////////////////////

CryptKey 
CryptKey::generate_key(uint16 key_size, const CryptByteString& seed, 
					   uint32 num_iterations)
// follows PKCS#5 V2.0 PBKDF2 (see www.rsasecurity.com)
// we assume that seed is concatenation of passphrase and salt
{
	SHACoder hash;
	uint32 hash_size = uint32(hash.get_byte_length_of_checksum());

	// step 1: check key_size (key_size <= (2^32-1)*hash_size)
#ifndef LEDA_CODER_EXACT_SIZE_INT16
	if (key_size > 0xFFFFFFFF) { // even more restricted
		LEDA_EXCEPTION(1, "CryptKey::generate_key: key_size too large");
		return CryptKey();
	}
#endif

	// step 2: compute l and r (left out, will be computed implicitely
	//uint32 l = (key_size + (hash_size - 1)) / hash_size; // = ceil(key_size / hash_size)
	//uint32 r = key_size % hash_size;

	// step 3: compute portions of key
	uint32 buffer_size = seed.get_size() + 6;
	if (buffer_size < hash_size) buffer_size = hash_size;

	secure_byte_array buffer_memory(buffer_size);
	memory_streambuf buffer(buffer_memory, streamsize(buffer_size), false, true);

	CryptKey key(key_size);
	byte* key_portion = key.get_bytes();
	byte* key_stop = key_portion + key_size;
	memset(key_portion, 0, key_size); 
		// it is important that all key bytes are 0 because of the xor below!

	uint32 p = 0;
	while (key_portion != key_stop) {
		size_t key_portion_sz 
			= (key_portion+hash_size<=key_stop) ? size_t(hash_size) : size_t(key_stop-key_portion);

		// first iteration: starts with "seed || key_size || p"
		uint32 seed_size = seed.get_size();
		buffer.set_size(streamsize(seed_size+6));
		memcpy(buffer_memory, seed.get_bytes(), seed_size);
		++p;
		buffer_memory[seed_size + 0] = b1(key_size);
		buffer_memory[seed_size + 1] = b0(key_size);
		buffer_memory[seed_size + 2] = b3(p);
		buffer_memory[seed_size + 3] = b2(p);
		buffer_memory[seed_size + 4] = b1(p);
		buffer_memory[seed_size + 5] = b0(p);

		uint32 cnt = num_iterations;
		do {
			// apply hash function to buffer and use the output as new input
			hash.reset();
			hash.set_checksums_in_stream_flag(false);
			hash.set_src_stream(&buffer);
			hash.check(); assert(!hash.failed());
			if (! hash.checksum_is_valid()) 
				LEDA_EXCEPTION(1, "CryptKey::generate_key: hash invalid!");
			memcpy(buffer_memory, hash.get_digest(), hash_size);
			buffer.set_size(streamsize(hash_size));
			block_cipher_mode_base::memxor(key_portion, hash.get_digest(), key_portion_sz);
		} while (--cnt > 0);

		key_portion += key_portion_sz;
	}

	return key;
}

CryptKey 
CryptKey::generate_key(uint16 key_size, 
					   const CryptByteString& passphrase, 
					   const CryptByteString& salt, 
					   uint32 num_iterations)
{
	CryptByteString seed = passphrase + salt;
	return generate_key(key_size, seed, num_iterations);
}

CryptByteString 
CryptKey::generate_nonce(uint16 nonce_size, 
						 const CryptByteString& salt2, 
						 uint32 num_iterations)
{
	CryptByteString salt1 = generate_salt(32);
	return generate_key(nonce_size, salt1+salt2, num_iterations);
}

CryptByteString 
CryptKey::generate_salt(uint16 salt_size)
{
	CryptByteString salt(salt_size);

	random_source rs;
	for (uint16 i = 0; i < salt_size; ++i) rs >> salt[i];

	// if we have enough space, we also put the current date into the salt (at the end)
	sint32 offset = sint32(salt_size) - sizeof(date);
	if (offset >= 4) {
		date d; // today
		memcpy(salt.get_bytes()+offset, &d, sizeof(d));
	}

	return salt;
}

CryptKey 
CryptKey::generate_key_and_salt(uint16 key_size, uint16 salt_size, 
								const CryptByteString& passphrase, 
								CryptByteString& salt, 
								uint32 num_iterations)
{
	salt = generate_salt(salt_size);
	return generate_key(key_size, passphrase, salt, num_iterations);
}

CryptByteString
CryptKey::read_passphrase(const string& prompt, uint16 min_length)
{
	fflush(stdin);
	bool ok = setvbuf(stdin, NULL, _IONBF, BUFSIZ) == 0; // make it unbuffered!
	if (!ok) LEDA_EXCEPTION(0, "could not make stdin unbuffered!");

	CryptByteString phrase;
	do {
		if (prompt.length() > 0) { cout << prompt; cout.flush(); }

		phrase.clear();
		CryptByteString phrase_buf(1024);
		uint16 l = 0; char ch;
		while ( (ch = fgetc( stdin )) != '\n' ) {
			phrase_buf[l] = (byte) ch;
			if (++l == phrase_buf.get_size()) {
				phrase += phrase_buf; l = 0;
			}
		}
		if (l > 0) phrase += phrase_buf(0, l-1);
	} while (phrase.get_size() < min_length);

	fflush(stdin);
	setvbuf(stdin, NULL, _IOFBF, BUFSIZ); // make it buffered again

	return phrase;
}

LEDA_END_NAMESPACE
