/*******************************************************************************
+
+  LEDA 7.0  
+
+
+  cin_cout_coder.cpp
+
+
+  Copyright (c) 1995-2023
+  by Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/


/*
	This program demonstrates how the LEDA coders can be used to process 
	input/output from stdin/stdout.
	It also shows how to manipulate cin and cout so that these streams use a 
	LEDA coder. (This is not possible for MSVC 6.0, because its iostream 
	implementation is not standard-compliant!)
 */

#include <LEDA/core/string.h>

#include <fstream>

#if !defined( __unix__ )
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#endif

#include <LEDA/coding/compress.h>

typedef leda::A0Coder Coder;

using namespace leda;

#if defined(LEDA_STD_IO_HEADERS)
using std::cout;
using std::cerr;
using std::cin;
using std::endl;
using std::streambuf;
using std::ostream;
using std::istream;
#endif

#ifdef LEDA_COMPRESSION_WA1
#define openmode open_mode
#endif

#define PROPER_IOSTREAM_LIB
#if (defined(_MSC_VER) && _MSC_VER < 1300) || defined(__DECCXX) || defined(__HP_aCC)
#undef PROPER_IOSTREAM_LIB
#endif

void usage(const char* my_name)
{
	cerr << "usage: " << my_name << " -e|-d < in_file > out_file" << endl;
	cerr << "  reads input from stdin, processes it and writes it to stdout" << endl;
	cerr << "  -e: encoding mode" << endl;
	cerr << "  -d: encoding mode" << endl;
	cerr << endl;
}

int copy_istream_to_ostream(istream& is = cin, ostream& os = cout)
{
	while (is.good() && os.good()) {
		char ch;
		is.get(ch);
		if (is) os.put(ch);
	}

	return os.fail() ? 1 : 0;
}

int encode() 
{
	// change cout such that all data sent to cout is encoded by Coder
	// before it is written to stdout
	streambuf* cur_cout_buf = cout.rdbuf();
	CoderAdapter<Coder> new_cout_buf(cur_cout_buf);

#ifdef PROPER_IOSTREAM_LIB
	cout.rdbuf(&new_cout_buf);
	int result = copy_istream_to_ostream(cin, cout);
	cout.rdbuf(cur_cout_buf);
#else
	ostream out(&new_cout_buf);
	int result = copy_istream_to_ostream(cin, out);
#endif

	return result;
}

int decode() 
{
	// change cin such that all data obtained from cin is decoded by the
	// AutoDecoder before it is returned to the user
	streambuf* cur_cin_buf = cin.rdbuf();
	CoderAdapter<AutoDecoder> new_cin_buf(cur_cin_buf);

#ifdef PROPER_IOSTREAM_LIB
	cin.rdbuf(&new_cin_buf);
	int result = copy_istream_to_ostream(cin, cout);
	cin.rdbuf(cur_cin_buf);
#else
	istream in(&new_cin_buf);
	int result = copy_istream_to_ostream(in, cout);
#endif

	return result;
}

int main(int argc, char* argv[]) 
{
#if !defined( __unix__ )
	// set stdin and stdout to binary mode
	setmode(_fileno(stdin), _O_BINARY);
	setmode(_fileno(stdout), _O_BINARY);
#endif

	if (argc > 1 && argv[1][0] == '-')
		switch (argv[1][1]) {
		case 'e': return encode();
		case 'd': return decode();
		}

	usage(argv[0]);

	return 1;
}
