#ifndef _LEDA_TESTER_H_
#define _LEDA_TESTER_H_

#include "DirectoryTree.h"

/// TestTarget /////////////////////////////////////////////////////////////////////////////

class TestTarget {
public:
	enum status { 
		S_None = 0, S_MakeError = 1, S_MakeWarning = 2, S_MakeOk = 4,
		S_RunError = 8, S_RunOk = 16, S_RunSkip = 32
	};

public:
	TestTarget(string name, DirectoryTree& tree, DirectoryTreeNode* dir, bool uses_GUI = false);

	string get_directory() const { return MyTree.get_absolute_path(MyDirectory); }

	string get_base_name() const { return BaseName; }
	string get_source_name() const { return BaseName + SourceExtension; }
	string get_obj_name() const { return BaseName + ObjExtension; }
	string get_exe_name() const { return BaseName + ExeExtension; }
	string get_make_out_name() const { return BaseName + MakeOutExtension; }
	string get_make_err_name() const { return BaseName + MakeErrExtension; }
	string get_run_in_name() const { return BaseName + RunInExtension; }
	string get_run_out_name() const { return BaseName + RunOutExtension; }
	string get_run_err_name() const { return BaseName + RunErrExtension; }
	string get_run_param_name() const { return BaseName + RunParamExtension; }

	string get_source_path() const { return get_directory() + get_directory_delimiter() + get_source_name(); }
	string get_obj_path() const { return get_directory() + get_directory_delimiter() + get_obj_name(); }
	string get_exe_path() const { return get_directory() + get_directory_delimiter() + get_exe_name(); }
	string get_make_out_path() const { return get_directory() + get_directory_delimiter() + get_make_out_name(); }
	string get_make_err_path() const { return get_directory() + get_directory_delimiter() + get_make_err_name(); }
	string get_run_in_path() const { return get_directory() + get_directory_delimiter() + get_run_in_name(); }
	string get_run_out_path() const { return get_directory() + get_directory_delimiter() + get_run_out_name(); }
	string get_run_err_path() const { return get_directory() + get_directory_delimiter() + get_run_err_name(); }
	string get_run_param_path() const { return get_directory() + get_directory_delimiter() + get_run_param_name(); }

	bool uses_GUI() const { return UsesGUI; }
	void set_GUI_flag(bool flag) { UsesGUI = flag; }

	bool make_ok() const { return (Status & S_MakeOk) != 0; }
	bool make_error() const { return (Status & S_MakeError) != 0; }
	bool make_warning() const { return (Status & S_MakeWarning) != 0; }
	bool run_ok() const { return (Status & S_RunOk) != 0; }
	bool run_error() const { return (Status & S_RunError) != 0; }
	bool run_skip() const { return make_error() || ((Status & S_RunSkip) != 0); }

	void set_make_ok() { if (! make_error()) Status |= S_MakeOk; }
	void set_make_error() { Status |= S_MakeError; Status &= ~S_MakeOk; }
	void set_make_warning() { Status |= S_MakeWarning; }
	void set_run_ok() { if (! run_error()) Status |= S_RunOk; }
	void set_run_error() { Status |= S_RunError; Status &= ~S_RunOk; }
	void set_run_skip() { Status |= S_RunSkip; }

	string get_skip_comment() const { return SkipComment; }
	void   set_skip_comment(string comment) { SkipComment = comment; }

	static bool delete_file_with_check(string path);

	bool delete_obj_file() { return delete_file_with_check(get_obj_path()); }
	bool delete_exe_file() { return delete_file_with_check(get_exe_path()); }
	bool delete_make_out_file() { return delete_file_with_check(get_make_out_path()); }
	bool delete_make_err_file() { return delete_file_with_check(get_make_out_path()); }
	bool delete_run_out_file() { return delete_file_with_check(get_run_out_path()); }
	bool delete_run_err_file() { return delete_file_with_check(get_run_out_path()); }

	string find_run_in_file() const;
	string find_run_param_file() const;
	string read_run_parameters() const;

public:
	static string ObjExtension;
	static string ExeExtension;
	static string MakeOutExtension;
	static string MakeErrExtension;
	static string RunInExtension;
	static string RunOutExtension;
	static string RunErrExtension;
	static string RunParamExtension;

	static string get_default_run_in_name() { return "default" + RunInExtension; }
	static string get_default_run_param_name() { return "default" + RunParamExtension; }

private:
	unsigned long      Status;
	string             BaseName;
	string             SourceExtension;
	DirectoryTree&     MyTree;
	DirectoryTreeNode* MyDirectory;
	bool               UsesGUI;
	string             SkipComment;
};

/// LedaTester /////////////////////////////////////////////////////////////////////////////

class LedaTester {
public:
	LedaTester(string leda_directory);
	~LedaTester();

	void add_path_to_ignore(string relative_path); // skipped during make & run

	void add_file_to_skip(string relative_path, string comment = string()); // skipped during
	void add_noGUI_exception(string relative_path); // do not skip during run
		// e.g.: graph\flow\amo (no ".exe" !)

	void set_show_only_err_flag(bool show_only_err) { ShowOnlyErrorsAndWarnings = show_only_err; }

	string get_test_directory() const { return Tree.get_path_to_root(); }

	void init();

	void make_clean();
	void make_all();
	void run_all();

	void generate_html_report();
	void generate_html_report(string html_path);

public:
	static string MakeCmd;
	static string ErrRedirection;
	static string RunPrefix;

private:
	void init_streams();
	enum report_level { L_LOW, L_MED, L_HIGH };
	ostream& report(report_level l = L_HIGH);

	void error(string msg);
	void copy_file_to_report_stream(string file_name, report_level l = L_HIGH);

	void make(TestTarget& tgt);
	void run(TestTarget& tgt);

	static void parse_make_output(string file_name, TestTarget& tgt);

private:
	bool                   Verbose;
	string                 OriginalWorkingDirectory;
	string                 ReportPath, DummyReportPath;
	ostream*               ReportLowStream;
	ostream*               ReportHighStream;
	string                 LedaDirectory;
	DirectoryTree          Tree;
	list<TestTarget>       Targets;
	d_array<string,string> FilesToSkip;
	set<string>            NoGUIExceptions;
	set<string>            PathsToIgnore;
	bool                   ShowOnlyErrorsAndWarnings;
};

#endif
