#include <LEDA/core/d_array.h>
#include <LEDA/core/set.h>
#include <LEDA/system/file.h>

#define IGNORE_GEOWIN_MSC
#define SKIP_STATIC_GRAPHS

#ifdef __win32__
#define IGNORE_OPEN_GL
#define IGNORE_SYSTEM_SOCKET_CC
#endif

using namespace leda;

#if defined(LEDA_STD_IO_HEADERS)
using std::cout;
using std::cerr;
using std::endl;
using std::flush;
using std::ostream;
using std::istream;
using std::ofstream;
using std::ifstream;
using std::streambuf;
using std::filebuf;
using std::ios;
#endif

#include "LedaTester.h"

#ifndef _MSC_IDE
#include "DirectoryTree.cpp"
#include "LedaTester.cpp"
#include "HtmlReport.cpp"
#endif // ifdef _MSC_IDE

const string std_leda_directory = "D:\\Projekte\\leda5_sources";
string leda_directory;

const bool def_do_make_clean = false;
const bool def_do_make_all   = true;
const bool def_do_run        = true;
const bool def_show_only_err = false;
const bool def_generate_html = true;

void usage()
{
	cout << "usage: ltest [options]" << endl;
	cout << " options:" << endl;
	cout << "   -c: make clean (delete all obj and exe files)" << endl;
	cout << "   -C: NO make clean" << endl;
	cout << "   -m: make all (generate all exe files)" << endl;
	cout << "   -M: NO make all" << endl;
	cout << "   -r: run all (run all non-GUI exe files)" << endl;
	cout << "   -R: NO run all" << endl;
	cout << "   -e: show only errors (during make and run)" << endl;
	cout << "   -E: show all messages" << endl;
	cout << "   -h: generate html report" << endl;
	cout << "   -H: NO html report" << endl;
	cout << endl;
	cout << " default options:";
	if (def_do_make_clean) cout << " -c"; else cout << " -C";
	if (def_do_make_all)   cout << " -m"; else cout << " -M";
	if (def_do_run)        cout << " -r"; else cout << " -R";
	if (def_show_only_err) cout << " -e"; else cout << " -E";
	if (def_generate_html) cout << " -h"; else cout << " -H";
	cout << endl;
	cout << endl;
	cout << " leda directory: " << leda_directory << endl;
	cout << "   (The report file(s) will be stored in the subdirectory 'test'.)" << endl;
}

int main(int argc, char* argv[])
{
	const char* leda_root = getenv("LEDAROOT");
	leda_directory = leda_root != 0 ? string(leda_root) : std_leda_directory;
	remove_trailing_directory_delimiter(leda_directory);

	string test_dir = leda_directory + get_directory_delimiter() + string("test");
	if (!is_directory(leda_directory) || !is_directory(test_dir)) {
		cerr << "Could not find a valid leda-directory: " << leda_directory << endl;
		cerr << "Have you defined LEDAROOT (correctly)?" << endl;
		return 1;
	}

	bool do_make_clean = def_do_make_clean;
	bool do_make_all   = def_do_make_all;
	bool do_run        = def_do_run;
	bool show_only_err = def_show_only_err;
	bool generate_html = def_generate_html;

	while (argc > 1) {
		if (argv[1][0] != '-' && argv[1][0] != '/') { usage(); return 1; }
		switch (argv[1][1]) {
		case 'c': do_make_clean = true; break;
		case 'C': do_make_clean = false; break;
		case 'm': do_make_all = true; break;
		case 'M': do_make_all = false; break;
		case 'r': do_run = true; break;
		case 'R': do_run = false; break;
		case 'e': show_only_err = true; break;
		case 'E': show_only_err = false; break;
		case 'h': generate_html = true; break;
		case 'H': generate_html = false; break;
		default:  usage(); return 1;
		}
		--argc; ++argv;
	}


	LedaTester test(leda_directory);

	test.set_show_only_err_flag(show_only_err);

	if (leda_root == 0) {
		// the following test(s) need(s) LEDAROOT:
		test.add_file_to_skip("numbers/residual_test", "LEDAROOT undefined");
	}

#ifdef IGNORE_GEOWIN_MSC
	test.add_path_to_ignore("graphics/geowin/msc");
#endif
#ifdef IGNORE_OPEN_GL
	test.add_path_to_ignore("graphics/opengl");
#endif
#ifdef IGNORE_SYSTEM_SOCKET_CC
	test.add_path_to_ignore("system/socket/cc");
#endif

	// skip always because it is designed to produce an error ;-)
	test.add_file_to_skip("core/misc/trace", "meant to produce error");

	// skip because it is subsumed by another test
	test.add_file_to_skip("graph/misc/gml2leda", "done by graph_io");
	test.add_file_to_skip("graph/misc/leda2gml", "done by graph_io");

	// todo
	test.add_file_to_skip("graph/misc/join", "have no input yet");
	test.add_file_to_skip("graph/misc/kura_test", "have no input yet");
	test.add_file_to_skip("graph/misc/match_test", "have no input yet");
	test.add_file_to_skip("graph/misc/map_test", "have no input yet");
	test.add_file_to_skip("graph/misc/mc_flow_test", "have no input yet");
	test.add_file_to_skip("graph/misc/mwm_test", "have no input yet");
	test.add_file_to_skip("graph/misc/plan_test", "have no input yet");
	test.add_file_to_skip("graph/misc/planar_map", "have no input yet");
	test.add_file_to_skip("graph/misc/short", "have no input yet");
	test.add_file_to_skip("graph/misc/short_test", "have no input yet");
	test.add_file_to_skip("graph/misc/span_test", "have no input yet");
	test.add_file_to_skip("graph/misc/speed_test", "have no input yet");
	test.add_file_to_skip("graph/misc/testall", "have no input yet");

	// todo
	test.add_file_to_skip("graph/flow/ffs", "no dimacs input yet");
	test.add_file_to_skip("graph/flow/mf0", "no dimacs input yet");
	test.add_file_to_skip("graph/flow/mfg", "no dimacs input yet");
	test.add_file_to_skip("graph/flow/mfs", "no dimacs input yet");
	test.add_file_to_skip("graph/flow/mcf_rev", "no dimacs input yet");
	test.add_file_to_skip("graph/flow/mcf_scale", "no dimacs input yet");
	test.add_file_to_skip("graph/flow/mcf_sort", "no dimacs input yet");
	test.add_file_to_skip("graph/flow/mcfs", "no dimacs input yet");
	test.add_file_to_skip("graph/flow/mcfs_csd", "no dimacs input yet");
	test.add_file_to_skip("graph/flow/mcfs_csd18", "no dimacs input yet");
	test.add_file_to_skip("graph/flow/sp0", "no dimacs input yet");
	test.add_file_to_skip("graph/flow/spg", "no dimacs input yet");
	test.add_file_to_skip("graph/flow/sps", "no dimacs input yet");

	// todo
#ifdef SKIP_STATIC_GRAPHS
	test.add_file_to_skip("graph/mcf/lgraph_dynamic", "no dimacs input yet");
	test.add_file_to_skip("graph/mcf/lgraph_external", "no dimacs input yet");
	test.add_file_to_skip("graph/mcf/lgraph_static", "no dimacs input yet");
	test.add_file_to_skip("graph/mcf/mcf_cap_scaling", "no dimacs input yet");
	test.add_file_to_skip("graph/mcf/sgraph_dynamic", "no dimacs input yet");
	test.add_file_to_skip("graph/mcf/sgraph_external", "no dimacs input yet");
	test.add_file_to_skip("graph/mcf/sgraph_mixed", "no dimacs input yet");
	test.add_file_to_skip("graph/mcf/sgraph_static", "no dimacs input yet");

	test.add_file_to_skip("graph/netalg/mcf_cap_scaling", "have no input yet");
	test.add_file_to_skip("graph/netalg/mf_preflow_push", "have no input yet");
	test.add_file_to_skip("graph/netalg/sp_dijkstra", "have no input yet");
#endif

	test.add_noGUI_exception("geo/d2_geo/gen_polygon_test");

	test.init();

	if (do_make_clean) test.make_clean();
	if (do_make_all)   test.make_all();
	if (do_run)        test.run_all();
	if (generate_html) test.generate_html_report();
	
	return 0;
}

