#if !defined(LEDA_ROOT_INCL_ID)
#define LEDA_ROOT_INCL_ID 430971
#include <LEDA/PREAMBLE.h>
#endif

//-----------------------------------------------------------------------------
//
// MAX_FLOW
//
// preflow-push + highest level + gap heuristic 
//
//-----------------------------------------------------------------------------


#include <LEDA/graph.h>
#include <LEDA/std/assert.h>

LEDA_BEGIN_NAMESPACE

#if defined(LEDA_GRAPH_ARRAY1)
typedef node_slot<node,0,graph> succ_array;
#else
typedef node_array<node,graph> succ_array;
#endif


template <class graph_t>
class node_level_queue {

typedef graph_t graph;
typedef typename graph_t::node node;
typedef typename graph_t::edge edge;

#if defined(LEDA_GRAPH_ARRAY1)
typedef node_slot<node,0,graph_t> succ_array;
#else
typedef node_array<node,graph_t> succ_array;
#endif

node*  T;
node*  head;
node*  max;

succ_array& NEXT;

public:

node_level_queue(const graph& G, succ_array& succ) : NEXT(succ)
{ 
  int sz = G.number_of_nodes();
  T = new node[sz+1];
  for(node* p = T+sz; p >= T;  p--) *p = nil; 
  head = T+1;
  max = T;
 }

~node_level_queue() { delete[] T; }

void insert(node v, int i)
{ // insert on level i
  node* h = head + i;
  if (h > max) max = h;
  NEXT[v] = *h;
  *h = v;
 }


void insert_non_max(node v, int i)
{ // do not update max pointer
  node* h = head + i;
  NEXT[v] = *h;
  *h = v;
 }


void insert_max(node v, int i)
{ // new max at level i
  max = head + i;
  *max = v;
  NEXT[v] = 0;
 }

node del_max() 
{ while (*max == nil) max--;
  node v = *max;
  *max = NEXT[v];
  return v; 
}


void clear() { while (max >= T) *max-- =  nil; }

};




template<class NT, class graph_t>
class max_flow  
{
  typedef graph_t graph;
  typedef typename graph_t::node node;
  typedef typename graph_t::edge edge;

#if defined(LEDA_GRAPH_ARRAY1)
  typedef node_slot<node,0,graph> succ_array;
  typedef node_slot<int,1,graph> excess_array;
  typedef node_slot<int,2,graph> dist_array;
  typedef edge_slot<int,0,graph> cap_array;
  typedef edge_slot<int,1,graph> flow_array;
#else
  typedef node_array<int,graph> excess_array;
  typedef node_array<int,graph> dist_array;
  typedef edge_array<int,graph> cap_array;
  typedef edge_array<int,graph> flow_array;
  typedef node_array<node,graph> succ_array;
#endif



  float h;

  NT fval;

  // statistics (phase 1, 2, and sum)

  int num_pushes[3];
  int num_relabels[3];
  int num_updates[3];
  int num_gap_nodes[3]; 
  int num_inspections[3];

  float cputime[3];
  float update_time;
  float gap_time;


  const graph* g_p; 
  node s_p; 
  node t_p;
  const cap_array*  cap_p; 
  const flow_array* flow_p;

public:

int pushes(int i=0)      { return num_pushes[i]; }
int relabels(int i=0)    { return num_relabels[i]; }
int updates(int i=0)     { return num_updates[i]; }
int gap_nodes(int i=0)   { return num_gap_nodes[i]; }
int inspections(int i=0) { return num_inspections[i]; }

float cpu_time(int i=0) { return cputime[i]; }

void set_heuristic(float hh) { h = hh; }


NT flow() { return fval; }


void reset_counters()
{ for(int i=0; i<3; i++) 
  { num_pushes[i] = 0;
    num_relabels[i] = 0;
    num_updates[i] = 0;
    num_gap_nodes[i] = 0;
    num_inspections[i] = 0;
    cputime[i] = 0;
    update_time = 0;
    gap_time = 0;
   }
}


max_flow(float heur = 5.0)
{ reset_counters();
  fval = 0;
  h  = heur;
}



void compute_dist01(const graph& G, node s, node t,
                    excess_array& excess,
                    succ_array& succ,
                    dist_array& dist,
                    node_level_queue<graph_t>& U, int* count)
{ 
  // compute exact distance values by a "backward" bfs in the
  // residual network starting at t

  float T = used_time();

  num_updates[1]++;

  int n = G.number_of_nodes();

  for(int i=0; i<n; i++) count[i] = 0;

  if (h < 0)
  { U.insert(t,0);
    node v;
    forall_nodes(v,G)
    { dist[v] = 0;
      if (v != t && excess[v] > 0) U.insert(v,0);
     }
    count[0] = n-1;
    dist[s] = n;
    return;
   }


  dist.init(G,n);

  excess[t] += 1;

  dist[t] = 0;
  dist[s] = 0;

  node v = t;
  node level_last = t;
  node queue_last = t;


  int d = 0;
  int level_sz = 0;

  for(;;)
  { 
    level_sz++;

    edge e;
    forall_in_edges(e,v)
    { node u = G.source(e,v); 
      int& du = dist[u];
      if (du == n)
      { du = d+1;
        succ[queue_last] = u;
        queue_last = u;
       }
    }

    node next = succ[v];

    if (excess[v] > 0) U.insert(v,d);

    if (v == level_last)             
    { // finish level d
      count[d++] = level_sz;
      level_sz = 0;
      if (queue_last == level_last) break; // no level d+1: stop
      level_last = queue_last;
     }

    v = next;
  }

 dist[s] = n;

 excess[t] -= 1;

 float tt = used_time(T);

//cout << string("dist01: time = %.2f",tt) <<endl;

 update_time += tt;
}




void compute_dist1(const graph& G, node s, node t,
                   const cap_array& cap,
                   const flow_array& flow,
                   excess_array& excess,
                   succ_array& succ,
                   dist_array& dist,
                   node_level_queue<graph_t>& U, int* count)
{ 
  // compute exact distance values by a "backward" bfs in the
  // residual network starting at t

  float T = used_time();

  num_updates[1]++;

  int n = G.number_of_nodes();
  dist.init(G,n);

  dist[t] = 0;
  excess[t] += 1; // we want t to be in U (the only node on level 0)

  node v = t;
  node level_last = t;
  node queue_last = t;


  int d = 0;
  int level_sz = 0;

  for(;;)
  { 
    level_sz++;

    edge e;
    forall_out_edges(e,v)
    { if (flow[e] == 0) continue;
      node u = G.target(e,v); 
      int& du = dist[u];
      if (du == n)
      { du = d+1;
        succ[queue_last] = u;
        queue_last = u;
       }
     }

    forall_in_edges(e,v)
    { if (cap[e] == flow[e]) continue;
      node u = G.source(e,v); 
      int& du = dist[u];
      if (du == n)
      { du = d+1;
        succ[queue_last] = u;
        queue_last = u;
       }
    }

    node next = succ[v];

    if (excess[v] > 0) U.insert(v,d);

    if (v == level_last)             
    { // finish level d
      count[d++] = level_sz;
      level_sz = 0;
      if (queue_last == level_last) break; // no level d+1: stop
      level_last = queue_last;
     }

    v = next;
  }

 while (d < n) count[d++] = 0;

 assert(dist[s] == n);

 excess[t] -= 1;

 float tt = used_time(T);

 update_time += tt;
}



void compute_dist2(const graph& G, node s, node t,
                   const flow_array& flow,
                   const excess_array& excess,
                   succ_array& succ,
                   dist_array& dist,
                   node_level_queue<graph_t>& U)
{ 
  float T = used_time();

  num_updates[2]++;

  int n = G.number_of_nodes();

  dist.init(G,n);

  node v = s;
  node level_last = s;
  node queue_last = s;
  dist[s] = 0;

  U.insert(t,0);

  int d = 0;

  for(;;)
  { 
    edge e;
    forall_out_edges(e,v)
    { if (flow[e] == 0) continue;
      node u = G.target(e,v); 
      int& du = dist[u];
      if (du == n)
      { du = d+1;
        succ[queue_last] = u;
        queue_last = u;
       }
     }

    node next = succ[v];

    if (excess[v] > 0 && v != t) U.insert(v,d);

    if (v == level_last)             
    { if (queue_last == level_last) break;
      level_last = queue_last;
      d++;
     }

    v = next;
  }

 float tt = used_time(T);

 update_time += tt;
}



void handle_gap(const graph& G, node v, dist_array& dist, succ_array& succ, int* count)
{

  float T = used_time();

  int gap_nodes = 0;

  int n  = G.number_of_nodes();
  int dv = dist[v];
  for (int i = dv; i < n && count[i] > 0; i++) count[i] = 0;

  node queue_last = v;
  succ[v] = v;

  do { gap_nodes++;

       v = succ[v];

       edge e;
       forall_out_edges(e,v)
       { node u = G.target(e,v);
         int& du = dist[u];
         if (du < n && du > dv)
         { succ[queue_last] = u;
           queue_last = u; 
           du = n;
          }
       }

       forall_in_edges(e,v)
       { node u = G.source(e,v);
         int& du = dist[u];
         if (du < n && du > dv)
         { succ[queue_last] = u;
           queue_last = u;
           du = n;
          }
        }

   } while (v != queue_last);

 num_gap_nodes[1] += gap_nodes;

 gap_time += used_time(T);
}



void phase1(const graph& G, node s, node t,
            const cap_array& cap,
            flow_array& flow,
            excess_array& excess,
            succ_array& succ)
{ 
  int n = G.number_of_nodes();
  int m = G.number_of_edges();

  int heuristic  = (h > 0) ? int(h*m) : int(-h*m);
  int limit_heur = heuristic;

  int relabel_count = 0;
  int push_count = 0;
  int insp_count = 0;

  edge e;
  forall_out_edges(e,s) 
  { node u = G.target(e,s);
    NT c = cap[e];
    flow[e] = c;
    excess[s] -= c;
    excess[u] += c;
  }

  dist_array dist(G);


  node_level_queue<graph_t>  U(G,succ);
  int* count = new int[n];

  compute_dist01(G,s,t,excess,succ,dist,U,count);

  excess[t] += 1;  // prevents t from beeing inserted again


  node v;
  while ((v = U.del_max()) != t)
  {
    NT  ev = excess[v]; 
    int dv = dist[v]; 

    int  dmin = n;    
    edge emin = 0;
    int  fmin = 0;
  
 
    edge e;
    forall_out_edges(e,v)
    { insp_count++;
      NT& fe = flow[e];
      NT  rc = cap[e] - fe;
      if (rc == 0) continue;
      node w = G.target(e,v);
      int dw = dist[w];
      if (dw == dv-1)
       { push_count++;
         NT& ew = excess[w];
         if (ew == 0) U.insert_non_max(w,dw);
         if (ev <= rc) 
         { ew += ev;
           fe += ev;
           ev = 0;   // excess[v] exhausted (stop)
           break;
           //excess[v] = 0;
           //goto end_loop;
          }
         else
         { ew += rc;
           fe += rc;
           ev -= rc;
          }
        }
      else
        if (v != w && dw < dmin) 
        { dmin = dw; emin = e; fmin = 1;  }
    }

    if (ev == 0)
    { excess[v] = 0;
      continue;
     }


    forall_in_edges(e,v)
    { insp_count++;
      NT& fe = flow[e];
      if (fe == 0) continue;
      node w = G.source(e,v);
      int dw = dist[w];
      if (dw == dv-1)
       { push_count++;
         NT& ew = excess[w];
         if (ew == 0) U.insert_non_max(w,dw);
         if (ev <= fe) 
         { fe -= ev; 
           ew += ev;
           ev = 0;   // excess[v] exhausted
           break;
           //excess[v] = 0;
           //goto end_loop;
          }
         else
         { ew += fe;
           ev -= fe;
           fe = 0;
          }
        }
      else
        if (v != w && dw < dmin) { dmin = dw; emin = e; fmin = -1; }
    } 

    excess[v] = ev;
    if (ev == 0) continue;

    if (dmin < n-1) 
    { 
      NT rmin = flow[emin];
      if (fmin > 0) rmin = cap[emin] - rmin;

      if (ev <= rmin)
      { if (--count[dv] == 0) // gap
        {  handle_gap(G,v,dist,succ,count);
           continue;
         }
        dv = dmin + 1;
        dist[v] = dv;
        count[dv]++;
        node w = G.opposite(emin,v);
        if (excess[w] == 0) U.insert(w,dmin);
        excess[w] += ev;
        flow[emin] += fmin*ev;
        excess[v] = 0;
        continue;
       }
    }
  

    // remaining excess at v
    // relabel vertex v (i.e. update dist[v]) because all
    // admissible edges in the residual graph have been saturated 

    if (insp_count >= limit_heur)
     { U.clear();
       compute_dist1(G,s,t,cap,flow,excess,succ,dist,U,count);
       limit_heur += heuristic;
      }
    else
     { relabel_count++;
       if (--count[dv] == 0) // gap
          handle_gap(G,v,dist,succ,count);
       else
        { dv = dmin+1;
          dist[v] = dv;
          if (dv < n) 
          { count[dv]++;
            U.insert_max(v,dv);
           }
         }
      }

//end_loop:;
  }

  excess[t] -= 1;

  num_pushes[1] += push_count;
  num_relabels[1] += relabel_count;
  num_inspections[1] += insp_count;

  delete[] count;

  fval = excess[t];
}



void phase2(const graph& G, node s, node t,
           const cap_array& cap,
           flow_array& flow,
           excess_array& excess,
           succ_array& succ)
{ 
  int n = G.number_of_nodes();
  int m = G.number_of_edges();
  
  int heuristic  = (h > 0) ? int(h*m) : int(-h*m);
  int limit_heur = heuristic;

  int relabel_count = 0;
  int push_count = 0;
  int insp_count = 0;

  node_level_queue<graph_t>  U(G,succ);

  dist_array dist(G);
  compute_dist2(G,s,t,flow,excess,succ,dist,U);

  node v;
  while ((v = U.del_max()) != t)
  {
    NT  ev = excess[v]; 
    int dv = dist[v]; 
    int dmin = n;

    edge e;
    forall_in_edges(e,v)
    { insp_count++;
      NT& fe = flow[e];
      if (fe == 0) continue;
      node w = G.source(e,v);
      int dw = dist[w];
      if (dw == dv-1)
       { push_count++;
         NT& ew = excess[w];
         if (ew == 0) U.insert_non_max(w,dw);
         if (ev <= fe) 
         { fe -= ev; 
           ew += ev;
           ev = 0; 
           break;
          }
         else
         { ew += fe;
           ev -= fe;
           fe = 0;
          }
        }
      else
        if (dw < dmin) dmin = dw;
    } 
    
    excess[v] = ev;
    if (ev == 0) continue;

    if (insp_count >= limit_heur)
      { U.clear();
        compute_dist2(G,s,t,flow,excess,succ,dist,U);
        limit_heur += heuristic;
       }
    else
      { relabel_count++;
        dv = dmin+1;
        dist[v] = dv;
        U.insert_max(v,dv);
       }
  }

  num_pushes[2] += push_count;
  num_relabels[2] += relabel_count;
  num_inspections[2] += insp_count;

}




NT run(const graph& G, node s, node t, 
       const cap_array& cap, 
       flow_array& flow)
{ 
  if (s == t) error_handler(1,"MAXFLOW: source == sink");

  float T = used_time();

  g_p = &G;
  s_p = s;
  t_p = t;
  cap_p = &cap;
  flow_p = &flow;

  reset_counters();

  excess_array excess;
  assert(excess.use_node_data(G,0));

  succ_array succ;
  assert(succ.use_node_data(G));

  flow.init(G,0);

  phase1(G,s,t,cap,flow,excess,succ);
  cputime[1] = used_time(T);

  phase2(G,s,t,cap,flow,excess,succ);
  cputime[2] = used_time(T);

  num_pushes[0]      = num_pushes[1]      + num_pushes[2];
  num_relabels[0]    = num_relabels[1]    + num_relabels[2];
  num_updates[0]     = num_updates[1]     + num_updates[2];
  num_gap_nodes[0]   = num_gap_nodes[1]   + num_gap_nodes[2];
  num_inspections[0] = num_inspections[1] + num_inspections[2];

  cputime[0] = cputime[1] + cputime[2];

  return fval;
}


NT operator()(const graph& G, node s, node t, 
              const cap_array& cap, flow_array& flow) 
{ 
  return run(G,s,t,cap,flow); 
 }




void statistics(ostream& out)
{ 
  out << endl;

  for(int i=1; i<=2; i++)
  { out << 
    string("%8d pushes %7d relabels %3d updates %9d insp %5d gaps  %.2f s",
            pushes(i), relabels(i), updates(i), inspections(i), gap_nodes(i),
            cputime[i]) << endl;
  }

  cout << string("update: %.2f   gaps: %.2f",update_time,gap_time) << endl;

  out << endl;
}




bool check()
{
  const graph& G = *g_p;
  node s = s_p; 
  node t = t_p;

  const cap_array& cap  = *cap_p;
  const flow_array& flow = *flow_p;


  string loc = "max_flow::check: ";
  bool res = true;

  edge e;
  forall_edges(e,G) 
    res = res && leda_assert(flow[e] >= 0 && flow[e] <= cap[e], 
              loc + string("illegal flow value for edge %d",G.index(e)),1);
  
  excess_array excess(G);

  node v;
  forall_nodes(v,G) excess[v] = 0;

  forall_nodes(v,G) 
  { edge e;
    forall_out_edges(e,v) 
    { node w = G.target(e,v);
      excess[v] -= flow[e]; 
      excess[w] += flow[e];
     }
  }

  leda_assert(fval == excess[t],"fval != excess[t]");
  
  node_array<bool,graph_t> reached(G,false);
  reached[s] = true;

  list<node> Q;
  Q.append(s); 

  while ( !Q.empty() )
  { node v = Q.pop(); 
    forall_out_edges(e,v) 
    { node w = G.target(e,v);
      if ( flow[e] < cap[e] && !reached[w] ) 
      { reached[w] = true; 
        Q.append(w); 
       }
    }
    forall_in_edges(e,v) 
    { node w = G.source(e,v);
      if ( flow[e] > 0 && !reached[w] ) 
      { reached[w] = true; 
        Q.append(w); 
       }
    }
  }
  res = res && leda_assert(!reached[t],"t is reachable in G_f",1);

  forall_nodes(v,G)
    res = res && leda_assert(v == s || v == t || excess[v] == 0,
                 loc + string("node %d has non-zero excess",G.index(v)),1);

  return res;
}


void print()
{
  const graph& G = *g_p;
  node s = s_p; 
  node t = t_p;
  const cap_array& cap  = *cap_p;
  const flow_array& flow = *flow_p;

  cout << "s = " << G.index(s) << endl;
  cout << "t = " << G.index(t) << endl;
  cout << endl;

  node v;
  forall_nodes(v,G)
  { cout << string("%2d :",G.index(v));
    edge e;
    forall_out_edges(e,v)
       cout << string(" --%d/%d-->%d",flow[e],cap[e],G.index(G.target(e,v)));
    cout << endl;
   }
  cout << endl;
}
  
 
};

LEDA_END_NAMESPACE


#if LEDA_ROOT_INCL_ID == 430971
#undef LEDA_ROOT_INCL_ID
#include <LEDA/POSTAMBLE.h>
#endif
