
#include<LEDA/window.h>
#include<LEDA/file_panel.h>

static int* matrix = 0;
static int   width = 0;
static int  height = 0;

static string fname = "none";
static string dname = ".";

static bool show_grid = true;
static int anim_steps = 1;

static list<color> ColorTable;
int compare( const color &c1, const color &c2 ) { return c1 != c2; }

static window W(500,500,"LEDA XPM-Editor");



void draw_matrix(window& W, int* mat, int w, int h)
{ W.start_buffering();
  W.clear();
  int pix_d = W.real_to_pix(1);
  double d = 0.5;
  for(int x = 0; x < w; x++)
    for(int y = 0; y < h; y++) { 
      int c = *(mat + (h-1-y)*w + x);
      pix_d < 2 ? W.draw_pixel(x,y,color(c)) :
W.draw_box(x,y,x+2*d,y+2*d, color(c));
    }

  if (show_grid && pix_d > 3)
  { for(int x = 0; x <= w; x++) W.draw_segment(x,0,x,h,black);
    for(int y = 0; y <= h; y++) W.draw_segment(0,y,w,y,black);
  }
  W.stop_buffering();
  W.flush_buffer();
  W.delete_buffer();
}


void redraw(window* wp, double x0, double y0, double x1, double y1)
{ window& W = *wp;
  if (x0 == W.xmin() && y0 == W.ymin() && x1 == W.xmax() && y1 ==
W.ymax()) 
    draw_matrix(W,matrix,width,height);
  else 
    W.flush_buffer(x0,y0,x1,y1);
}


void read_file( string fname )
{ char* pmap = W.create_pixrect( fname );
  width  = W.get_width( pmap );
  height = W.get_height( pmap );
  if( matrix ) delete[] matrix;
  matrix = new int[ width*height ];
  W.pixrect_to_matrix( pmap, matrix );
  draw_matrix( W, matrix, width, height );
  ColorTable.clear();
  for(int i=0; i<width*height; i++ )
    if( !ColorTable.search( color( matrix[i] ) ) )
      ColorTable.append( color( matrix[i] ) );
}

void write_file( string fname ) 
{ ofstream os( fname );
  if( os.fail() ) {
    error_handler( 0, "Cant write " + fname + "!" );
    return;
  }
  
  if( width == 0 || height == 0 || ColorTable.empty() || !matrix )
    return;
  
  string n = fname.replace_all( ".", "_" ); 
  int c    = ColorTable.length();
  int cpp  = int(string( "%d", c ).length());

  os << "/* XPM */" << endl << "static char * " << n << "[] = {" <<
endl;
  os << "/* w h n_c cpp */" << endl << "\"";
  os << string( "%d", width ) << string( " %d", height ) << string( "
%d", c );
  os << string( " %d", cpp ) << "\"," << endl << "/* colors */" << endl;

  list_item it;
  forall_items( it, ColorTable ) {
    color c =  ColorTable.contents( it );
    int r,g,b; c.get_rgb( r,g,b );   
    string n = string( "%d", ColorTable.rank( c ) ); /* rank is the key
*/
    while( n.length() < cpp ) n = n.insert( 0, "0" );
    os << "\"" << n << " c #"<< hex << r << g << b << "\"," << endl;
  }

  os << "/* pixels */" << endl;
  for(int y=0; y<height; y++) {
    os << "\"";
    for(int x=0; x<width; x++) {    
      string col_idx = string( "%d", ColorTable.rank( matrix[ (y*width)
+ x] ) );
      while( col_idx.length() < cpp ) col_idx=col_idx.insert( 0, "0" );
      os << col_idx;
    }
    if( y < height-1 ) os << "\"," << endl; else os << "\"" << endl;
  }
  os << "};";
}


void file_handler(int what)
{ file_panel FP(W, fname, dname);
  switch (what) {
  case 0: exit(0);
          break;
  case 1: FP.set_load_handler(read_file);
          break;
  case 2: FP.set_save_handler(write_file);
          break;
  }
  FP.set_load_handler(read_file);
  FP.set_pattern("*.xpm");
  FP.open();
}



void zoom_handler(int what)
{ switch (what) {
  case 1: W.zoom(2.0,anim_steps);
          break;
  case 2: W.zoom(0.5,anim_steps);
          break;
  case 3: W.zoom_area(0,0,width,height,anim_steps);
          break;
  }
}


void option_handler(int)
{ panel P;
  P.text_item("\\bf\\blue Options");
  P.text_item("");
  P.bool_item("show grid", show_grid);
  P.int_item("animation",anim_steps,0,16);
  P.open(W);
}


void show_color_table( point &p )
{ int x = int( p.xcoord() ); int y = int( p.ycoord() );
  if( y > height || x > width || y < 0 || x < 0) return;
  menu M( "Color table" );
  color col = color( *( matrix + ( height-1-y ) * width + x ) );    
  M.color_item( "Pxl", col, ColorTable );
  M.make_menu_bar( 2 );
  M.open( W, W.xpix(x), W.ypix(y) );
  *(matrix + (height-1-y)*width + x ) = col;  
  draw_matrix( W, matrix, width, height );	          
}


int main(int argc, char** argv)
{
  menu file_menu;
  file_menu.button("load",1,file_handler);
  file_menu.button("save",2,file_handler);
  file_menu.button("exit",0,file_handler);

  menu zoom_menu;
  zoom_menu.button("zoom up", 1, zoom_handler);
  zoom_menu.button("zoom down", 2, zoom_handler);
  zoom_menu.button("fill window", 3, zoom_handler);

  W.button("&File", file_menu);
  W.button("&Zoom", zoom_menu);
  W.button("&Options", option_handler);
  W.make_menu_bar();

  W.init(-10,120,-10);
  W.set_redraw( redraw );
  W.display( window::center, window::center );

  if (argc > 1) read_file( fname = argv[1] );    

  for(;;) {
    point p;
    int but = W.read_mouse(p); 
    
    if (but == MOUSE_BUTTON(2)) show_color_table(p);
 
    bool dblclick,drag;
    W.read_mouse(p,300,300,dblclick,drag);
  
    if (dblclick)
    { point a,b;
      W.read_zoom_rect(p,a,b);
      W.zoom_area(a,b);
      continue;
     }
  
     if (drag)
     { W.scroll_window(p);
       continue;
      }
   }
  return 0;
}

