#define LEDA_CHECKING_OFF

#include <LEDA/graph/graph.h>
#include <LEDA/graph/static_fgraph.h>
#include <LEDA/graph/dimacs_graph.h>
#include <LEDA/graph/node_slot1.h>
#include <LEDA/graph/edge_slot1.h>

#include <LEDA/core/b_queue.h>



#define out_cap(e)       (cap[e]-flow[e])
#define in_cap(e)        flow[e]
#define residual_cap(e)  (e.out() ? out_cap(e) : in_cap(e))
#define residual_cost(e) (e.out() ? cost[e]    : -cost[e])

#define push_flow(e,x)   flow[e] += (e.out() ? x : -x)



using namespace leda;

using std::cin;
using std::cout;
using std::cerr;
using std::flush;
using std::endl;


template<class graph, class node, class pred_array> 
int tree_scan(const graph& G, node v, const pred_array& pred)
{
  typedef typename graph::incident_edge incident_edge;

  int n = G.number_of_nodes();

  node* Q = new node[n];
  node* q = Q;
  node* q_last = Q;

  *q = v;

  while (q <= q_last)
  { node v = *q++;
    incident_edge e;
    forall_incident_edges(e,v)
    {  node w = G.opposite(e,v);
       if (pred[w] != e) continue; // non-tree edge
       *++q_last = w;
    }
  }

  int sz = q_last - Q + 1;
  delete[] Q;
  return sz;
}
       
    

  
template<class graph, class lcap_array, class ucap_array, 
                      class cost_array, class supply_array, class flow_array>

double mcf(const graph& G, const lcap_array& lcap, const ucap_array& ucap,
                           const cost_array& cost, const supply_array& supply,
                           flow_array& flow)
{
  typedef typename graph::node node;
  typedef typename graph::edge edge;
  typedef typename graph::incident_edge incident_edge;

  int n = G.number_of_nodes();
      
  node_slot1<incident_edge,graph,0> pred(G);
  node_slot1<int,graph,1>  excess(G);
  node_slot1<int,graph,2>  dist(G);
  node_slot1<int,graph,3>  mark(G);

  ucap_array& cap = (ucap_array&)ucap;

  G.hide_all_in_edges();

  node v;
  forall_nodes(v,G) 
  { excess[v] = supply[v];
    mark[v]   = 0; 
    dist[v]   = 0;     
   }      
     
      
  forall_nodes(v,G) 
  { edge e;
    forall_out_edges(e,v)
    { node w = G.opposite(e,v);
        
      int lc = lcap[e];

      excess[v] -= lc;
      excess[w] += lc;
      cap[e]    -= lc;

      int ce = cap[e];
        
      if (cost[e] < 0)
        { flow[e] = ce;
          excess[v] -= ce;
          excess[w] += ce;
          G.restore_in_edge(e,w);
         }
      else
         flow[e] = 0;
    }
  }


  int count = 0;
  
  node s;
  forall_nodes(s,G) 
  {
    if (excess[s] <= 0) continue;

    int augment_count = 0;
    int t_count_max = 0;

    while (excess[s] > 0)
    { 
      cout << string("excess = %6d  t_max = %3d", excess[s],t_count_max);
      cout  << endl;

      augment_count++;

      node t = 0;
      int t_count = 0;

      count += 2;        
    
      b_queue<node> Q(n);      
      Q.append(s);

      dist[s] = 0;    
      mark[s] = count;
                   
      while (!Q.empty())
      { node v = Q.pop();
        mark[v] = count+1;
        int dv = dist[v];

        if (v == t) t_count++;
              
        incident_edge e;
        forall_incident_edges(e,v)
        { if (residual_cap(e) == 0) continue;
          int c = residual_cost(e) + dv;              
          node w = G.opposite(e,v);            

          if (mark[w] >= count && c >= dist[w]) continue;

          dist[w] = c;        
          pred[w] = e;             
          if (mark[w] != count) 
          { Q.append(w);
            mark[w] = count;
           }
         
         if (t == 0 && excess[w] < 0) t = w;        
        }       
      }
       
      if (t_count > t_count_max) t_count_max = t_count;

      int df = min(excess[s],-excess[t]);
 
      //node root = 0;
 
      for(node u = t; u != s; )
      { incident_edge e = pred[u];
        int rc = residual_cap(e);            
        if (rc <= df) { df = rc; /* root = u; */}
        u = G.opposite(e,u);                    
       }
     
      for(node u = t; u != s; )
      { incident_edge e = pred[u];       
        push_flow(e,df);
        if (e.out()) G.restore_in_edge(e,u);
        u = G.opposite(e,u);              
        if (e.in() && flow[e] == 0) G.hide_in_edge(e,u);
      }         
                
      excess[s] -= df;
      excess[t] += df;
 
   } // while (excess[s] > 0)

   cout << string("augmentations = %3d   t_count_max = %2d", 
                        augment_count,t_count_max) << endl;

 } 
    

 double total_cost = 0;

 edge e;
 forall_edges(e,G)
 { int lc = lcap[e];
   flow[e] += lc;
   cap[e] += lc;
   total_cost += double(flow[e])*cost[e];
  }

 return total_cost;
      
}
   
   
int main()
{

  typedef static_fgraph<opposite_graph,data_slots<4>,data_slots<3> > st_graph;
  //typedef st_graph::node node;
  //typedef st_graph::edge edge;


  st_graph G;
  node_array<int,st_graph>  supply(G);
  edge_array<int,st_graph>  lcap(G);

  edge_slot1<int,st_graph,0> flow(G);
  edge_slot1<int,st_graph,1> ucap(G);
  edge_slot1<int,st_graph,2> cost(G);


  dimacs_mcf D(cin);
  D.translate(G,supply,lcap,ucap,cost);

  double n = G.number_of_nodes();
  double m = G.number_of_edges();

  cout << endl;
  cout << string("%6.0f nodes",n) << endl;
  cout << string("%6.0f edges",m) << endl;
  cout << endl;



  float T = used_time();

  double c = mcf(G,lcap,ucap,cost,supply,flow);

  cout << string("time: %6.2f cost: %16.0f", used_time(T), c) << endl;

  return 1;
}

