#define LEDA_CHECKING_OFF

#include <LEDA/graph/graph.h>
#include <LEDA/graph/static_fgraph.h>
#include <LEDA/graph/dimacs_graph.h>
#include <LEDA/graph/node_slot1.h>
#include <LEDA/graph/edge_slot1.h>

#include <LEDA/core/b_queue.h>



#define out_cap(e)       (cap[e]-flow[e])
#define in_cap(e)        flow[e]
#define residual_cap(e)  (e.out() ? out_cap(e) : in_cap(e))
#define residual_cost(e) (e.out() ? cost[e]    : -cost[e])

#define push_flow(e,x)   flow[e] += (e.out() ? x : -x)



using namespace leda;

using std::cin;
using std::cout;
using std::cerr;
using std::flush;
using std::endl;


template<class graph, class node, class pred_array> 
int tree_scan(const graph& G, node v, const pred_array& pred, node* Q)
{
  typedef typename graph::incident_edge incident_edge;

  node* q = Q;
  node* q_last = Q;

  *q = v;

  while (q <= q_last)
  { node v = *q++;
    incident_edge e;
    forall_incident_edges(e,v)
    {  node w = G.opposite(e,v);
       if (pred[w] != e) continue; // non-tree edge
       *++q_last = w;
    }
  }

  return q_last - Q + 1;
}
       
    

  
template<class graph, class lcap_array, class ucap_array, 
                      class cost_array, class supply_array, class flow_array>

double mcf(const graph& G, const lcap_array& lcap, const ucap_array& ucap,
                           const cost_array& cost, const supply_array& supply,
                           flow_array& flow)
{
  typedef typename graph::node node;
  typedef typename graph::edge edge;
  typedef typename graph::incident_edge incident_edge;

  int n = G.number_of_nodes();
      
  node_slot1<incident_edge,graph,0> pred(G);
  node_slot1<int,graph,1>  excess(G);
  node_slot1<int,graph,2>  dist(G);
  node_slot1<int,graph,3>  mark(G);

  ucap_array& cap = (ucap_array&)ucap;

  G.hide_all_in_edges();

  node v;
  forall_nodes(v,G) 
  { excess[v] = supply[v];
    mark[v]   = 0; 
    dist[v]   = 0;     
   }      
     
      
  forall_nodes(v,G) 
  { edge e;
    forall_out_edges(e,v)
    { node w = G.opposite(e,v);
        
      int lc = lcap[e];

      excess[v] -= lc;
      excess[w] += lc;
      cap[e]    -= lc;

      int ce = cap[e];
        
      if (cost[e] < 0)
        { flow[e] = ce;
          excess[v] -= ce;
          excess[w] += ce;
          G.restore_in_edge(e,w);
         }
      else
         flow[e] = 0;
    }
  }

  node* V = new node[n];
  
  int t_max = 0;
  int p_max = 0;

  node s;
  forall_nodes(s,G) 
  {
    if (excess[s] <= 0) continue;

    node v;
    forall_nodes(v,G)  
    { dist[v] = MAXINT;
      pred[v] = 0;
     }

    dist[s] = 0;

    int augment_count = 0;
    int count = 0;

    while (excess[s] > 0)
    { 
      count += 2;

      if (++augment_count % 1000 == 0)
      { cout << string("%5d: excess = %6d  t_max = %3d  p_max = %3d", 
                             augment_count,excess[s], t_max, p_max);
         cout << endl;
       }

      node t = 0;
      int t_count = 0;

      b_queue<node> Q(n);      

      int k = tree_scan(G,s,pred,V);
      for(int i=0; i<k; i++)
      { node v  = V[i];
        Q.append(v);
        mark[v] = count;
       }

      //cout << "k = " << k << endl;

      while (!Q.empty())
      { node v = Q.pop();
        mark[v] = count+1;
        int dv = dist[v];

        if (v == t) t_count++;
              
        edge e;
        forall_out_edges(e,v)
        { if (flow[e] == cap[e]) continue;
          int c = dv + cost[e];              
          node w = G.opposite(e,v);            
          if (mark[w] >= count && c >= dist[w]) continue;
          dist[w] = c;        
          pred[w] = incident_edge(e,true);          
          if (mark[w] != count) 
          { Q.append(w);
            mark[w] = count;
           }
         if (t == 0 && excess[w] < 0) t = w;        
        }       

        forall_in_edges(e,v)
        { if (flow[e] == 0) continue;
          int c = dv - cost[e];              
          node w = G.opposite(e,v);            
          if (mark[w] >= count && c >= dist[w]) continue;
          dist[w] = c;        
          pred[w] = incident_edge(e,false);          
          if (mark[w] != count) 
          { Q.append(w);
            mark[w] = count;
           }
         if (t == 0 && excess[w] < 0) t = w;        
        }       
/*
        incident_edge e;
        forall_incident_edges(e,v)
        { if (residual_cap(e) == 0) continue;
          int c = dv + residual_cost(e);              
          node w = G.opposite(e,v);            
          if (c >= dist[w]) continue;
          dist[w] = c;        
          pred[w] = e;             
          if (mark[w] == 0) 
          { Q.append(w);
            mark[w] = 1;
           }
         if (t == 0 && excess[w] < 0) t = w;        
        }       
*/
      } 

     assert(t != 0);

     if (t_count > t_max) t_max = t_count;
       
     int df = min(excess[s],-excess[t]);

     node r = 0;
     int  p_len = 0;

     for(node u = t; u != s; )
     { incident_edge e = pred[u];
       int rc = residual_cap(e);            
       if (rc <= df) { df = rc; r = u; }
       u = G.opposite(e,u);                    
       p_len++;
     }

     if (p_len > p_max) p_max = p_len;

/*
     if (r) 
     { int count = 0;
       int k = tree_scan(G,r,pred,V);
       for(int i=0; i<k; i++)
       { node v  = V[i];
         dist[v] = MAXINT;
         if (v == t) count++;
        }
       assert(count == 1);
      }
*/

     excess[s] -= df;
     excess[t] += df;

     for(node u = t; u != s; )
     { incident_edge e = pred[u];       
       if (e.out() && flow[e] == 0) G.restore_in_edge(e,u);
       push_flow(e,df);
       u = G.opposite(e,u);              
       if (e.in() && flow[e] == 0) G.hide_in_edge(e,u);
     }         

     if (r) pred[r] = 0;

   } // while (excess[s] > 0)

   cout << string("augmentations = %3d", augment_count) << endl;

  }

  delete[] V;
    

  double total_cost = 0;

  edge e;
  forall_edges(e,G)
  { int lc = lcap[e];
    flow[e] += lc;
    cap[e] += lc;
    total_cost += double(flow[e])*cost[e];
   }
 
  return total_cost;
}
   
   
int main()
{

  typedef static_fgraph<opposite_graph,data_slots<4>,data_slots<3> > st_graph;
  //typedef st_graph::node node;
  //typedef st_graph::edge edge;


  st_graph G;
  node_array<int,st_graph>  supply(G);
  edge_array<int,st_graph>  lcap(G);

  edge_slot1<int,st_graph,0> flow(G);
  edge_slot1<int,st_graph,1> ucap(G);
  edge_slot1<int,st_graph,2> cost(G);


  dimacs_mcf D(cin);
  D.translate(G,supply,lcap,ucap,cost);

  double n = G.number_of_nodes();
  double m = G.number_of_edges();

  cout << endl;
  cout << string("%6.0f nodes",n) << endl;
  cout << string("%6.0f edges",m) << endl;
  cout << endl;



  float T = used_time();

  double c = mcf(G,lcap,ucap,cost,supply,flow);

  cout << string("time: %6.2f cost: %16.0f", used_time(T), c) << endl;

  return 1;
}

