/*******************************************************************************
+
+  LEDA 7.0  
+
+
+  _ws.c
+
+
+  Copyright (c) 1995-2023
+  by Algorithmic Solutions Software GmbH
+  All rights reserved.
+ 
*******************************************************************************/

#include <LEDA/core/string.h>
#include <LEDA/core/list.h>
#include <LEDA/core/h_array.h>

#include <LEDA/system/websocket.h>

#include <LEDA/coding/zlib.h>
#include <LEDA/coding/png.h>

#include <LEDA/graphics/x_window.h>


#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <assert.h>

#include <iostream>

using std::cout;
using std::endl;

/*
extern
unsigned lodepng_encode24(unsigned char** out, size_t* outsize,
                          const unsigned char* image, unsigned w, unsigned h);
*/


LEDA_BEGIN_NAMESPACE

//-----------------------------------------------------------------------------

extern int INIT_GRAPHICS(int mode,int& width,int& height,int& depth,int& dpi);

extern void UPDATE_SCREEN_AREA(unsigned int* pixels, int w, int h, 
                             int x0, int y0, int x1, int y1, int x, int y);

extern int  CHECK_NEXT_EVENT(int& val,int& x,int& y,unsigned long& t,int msec);

extern void SET_MOUSE_CURSOR(int id);

//-----------------------------------------------------------------------------

enum { RLE=0, RGB=1, RGBA=2, PNG=3 };

const int COMP_LIMIT = 4096;

const int HEADER_SZ = 21; // 1 byte for format + 4 int's (w,h,x,y)

static websocket ws;
static int PORT = 8081;

static int format = RLE; 
static int compression = 0;  // off(0) on(1) auto(2)

static unsigned long bytes_total = 0;
static unsigned long bytes_sent = 0;


inline string format_string() {
  string s = "";
  switch (format) {
    case RLE:  s = "RLE";
               break;
    case RGB:  s = "RGB";
               break;
    case RGBA: s = "RGBA";
               break;
    case PNG:  s = "PNG";
               break;
  }

  s += string("-%d",compression);

  return "";
}


inline int compress_buffer(unsigned char* buf, int buf_sz)
{ buf[0] |= 0x10;
  return zlib().deflate(buf+1,buf_sz-1) + 1;
 }


int INIT_GRAPHICS(int mode, int& width, int& height, int& depth, int& dpi)
{
  // return socket fd or 0  (-1 on error)

  ws.set_port(PORT);
  ws.set_trace(false);

  cout << endl;
  cout << "INIT_GRAPHICS: mode = " << mode << endl;
  cout << endl;


  if (mode == 0) {
   cout << "WS: CLOSE DISPLAY" << endl;
   ws.disconnect();
   return 0;
  }

  bytes_total = 0;
  bytes_sent = 0;

//string prog_name = program_invocation_short_name;
  string prog_name = program_invocation_name;

  cout << "program: " << prog_name << endl;

  char* cgi_client = getenv("REMOTE_ADDR");
  if (cgi_client) cout << "cgi client: " << cgi_client << endl; 

  cout << endl;


  cout << "ws: listening on port " << PORT << endl;

  if (!ws.listen()) {
    cout << "WEBSOCKET: listen failed" << endl; 
    cout << ws.get_error() << endl;
    return -1;
  }

  cout << "ws: waiting for connection" << endl;


  if (!ws.accept())  {
    cout << "WEBSOCKET: accept failed" << endl; 
    cout << ws.get_error() << endl;
    return -1;
  }

  string ws_client = ws.get_socket().client_ip();
  int    fd = ws.sock_fd();
  string msg = ws.receive_text();
  cout << string("ws: connected to %s  fd = %d  msg = %s",
                                       ~ws_client, fd, ~msg) << endl;

  cout << "ws: detach from socket endpoint" << endl;
  ws.detach();

  if (cgi_client && (ws_client != cgi_client)) {
    cout << endl;
    cout << "WARNING: ws_client != cgi_client." << endl;
  }


  // default values

  width  = 1200;
  height = 1000;
  depth = 24;
  dpi = 92;

/*
  int pid = getpid();
  ws.send_text(string("open: %s %d", ~prog_name,pid));
*/

  ws.send_text("open: " + prog_name + " " + format_string());

  // wait for display event (888)

  cout << endl;
  cout << "waiting for display_event" << endl;

  int e,val,x,y;
  unsigned long t;
  while ((e = CHECK_NEXT_EVENT(val,x,y,t,0)) != display_event) {
    cout << "SKIP EVENT: " << e << "  " << event_name[e] << endl;
  }

  width = x;
  height = y;
  dpi = val;

  return fd;
}


static unsigned char* write_header(unsigned char* q, unsigned char format,
                                   int xpos, int ypos, int w, int h, int t_sz)
{ *q = format;
  q += 1;
  *(unsigned int*)q = xpos;
  q += 4;
  *(unsigned int*)q = ypos;
  q += 4;
  *(unsigned int*)q = w;
  q += 4;
  *(unsigned int*)q = h;
  q += 4;
  *(unsigned int*)q = t_sz;
  q += 4;
  return q;
}



void UPDATE_SCREEN_AREA(unsigned int* pixels, int w, int h, 
                        int x0, int y0, int x1, int y1,
                        int xpos, int ypos)
{
  // copy rectangle [x0,y0,x1,y1] from (w x h) pixel buffer
  // to canvas to position (xpos,ypos)
  
  float t = used_time();

  if (x1 > w-1) x1 = w-1;
  if (y1 > h-1) y1 = h-1;
  if (x0 < 0) { xpos -= x0; x0 = 0; }
  if (y0 < 0) { ypos -= y0; y0 = 0; }

  if (x0 > x1 || y0 > y1) return;

  int width = x1-x0+1;
  int height = y1-y0+1;

  int sz = width*height;

  bool deflate = (compression == 1 || (compression == 2 && 4*sz > COMP_LIMIT)); 

/*
  int f = format;
  if (deflate) f |= 0x10;
  ws.send_text(string("image: %d %d %d %d %d",f,width,height,xpos,ypos));
*/

  if (format == RGBA)
  { // extract rgba pixels from (x0,y0,x1,y1) into char buffer

    int buf_sz = 4*sz + HEADER_SZ;
    unsigned char* buf = new unsigned char[buf_sz];
    unsigned char* q = write_header(buf,RGBA,xpos,ypos,width,height,0); 

    for (int j = y0; j <= y1; j++)
    { for(int i = x0; i <= x1; i++)
      { unsigned int pix = pixels[j*w + i];
         *q++ = (pix >> 16) & 0xff; //red
         *q++ = (pix >>  8) & 0xff; //green
         *q++ = (pix >>  0) & 0xff; //blue
         *q++ = 255;
      }
    }

    int n = buf_sz;
    if (deflate) n = compress_buffer(buf,buf_sz);

    ws.send_data(buf,n);
    delete[] buf;

    bytes_total += 4*sz;
    bytes_sent += n;
    ws.send_text(string("statistics: %lu %d",bytes_total,bytes_sent));

    return;
  }


  if (format == RGB)
  { // extract rgb pixels from (x0,y0,x1,y1) into char buffer
    
    int buf_sz = 3*sz + HEADER_SZ;
    unsigned char* buf = new unsigned char[buf_sz];
    unsigned char* q = write_header(buf,RGB,xpos,ypos,width,height,0); 
  
    for (int j = y0; j <= y1; j++)
    { for(int i = x0; i <= x1; i++)
      { unsigned int pix = pixels[j*w + i];
         *q++ = (pix >> 16) & 0xff; //red
         *q++ = (pix >>  8) & 0xff; //green
         *q++ = (pix >>  0) & 0xff; //blue
      }
    }

    int n = buf_sz;
    if (deflate) n = compress_buffer(buf,buf_sz);

    ws.send_data(buf,n+1);
    delete[] buf;

    bytes_total += 4*sz;
    bytes_sent += n;
    ws.send_text(string("statistics: %lu %d",bytes_total,bytes_sent));

    return;
  }



  if (format == PNG)
  { // extract pixels from (x0,y0,x1,y1) into char buffer and 
    // convert from bgra to rgb

    unsigned char* rgb = new unsigned char[3*sz];
    unsigned char* q = rgb;
  
    for (int j = y0; j <= y1; j++)
    { for(int i = x0; i <= x1; i++)
      { unsigned int pix = pixels[j*w + i];
         *q++ = (pix >> 16) & 0xff; //red
         *q++ = (pix >>  8) & 0xff; //green
         *q++ = (pix >>  0) & 0xff; //blue
      }
    }

/*
    unsigned err = lodepng_encode24(&png,&p_sz,rgb,width,height);
*/

    size_t p_sz = 0;
    unsigned char* png = png_encode_rgb(p_sz,rgb,width,height);

    if (png == 0)
    { error_handler(1,"ERROR: PNG DECODING");
      return;
    }

    int buf_sz = p_sz + HEADER_SZ;
    unsigned char* buf = new unsigned char[buf_sz];

    unsigned char* p = write_header(buf,PNG,xpos,ypos,width,height,0); 
    for(int i=0; i<p_sz; i++) *p++ = png[i];
    delete[] png;

    int png_sz = p-buf;

    bytes_total += 4*sz;
    bytes_sent += png_sz;
    ws.send_text(string("statistics: %lu %d",bytes_total,bytes_sent));

/*
    float percent = (100.0*png_sz)/(4*sz);
    cout << endl;
    cout << string("%4d x%4d  %8d --->%6d bytes %5.2f %%  time: %.2f sec", 
                              width,height,4*sz,png_sz,percent, used_time(t));
*/

    int n = png_sz;
/*
    if (deflate) n = compress_buffer(buf,png_sz);
*/

    ws.send_data(buf,n);
    delete[] buf;

    return;
  }


  assert(format == RLE);

  unsigned char* buf = 0;
  int numPixels = 0;

// RLE (run lengh encoding)
// use alpha-byte for encoding number of consecutive pixels of the same color

  int buf_sz = 0;
  int bytes = 0;

  h_array<int,int> color_table(0);

  for (int j = y0; j <= y1; j++)
  { for(int i = x0; i <= x1; i++)
    { unsigned int clr = pixels[j*w + i] & 0xffffff;
      color_table[int(clr)]++;
      //if (color_table.size() >= 256) break;
     }
   }

  int numColors = color_table.size();


  if (numColors < 256)
  { 
    // use color table and represent colors by a 1-byte index
    int table_sz = numColors;

    numPixels = 0;

    // 2 bytes per pixel, 4 bytes for each table entry, header 
    buf_sz = 2*sz + 4*table_sz + HEADER_SZ;
    buf = new unsigned char[buf_sz];

    unsigned char* q = write_header(buf,RLE,xpos,ypos,width,height,table_sz); 

    // write color table

    int index = 0;
    int clr;
    forall_defined(clr,color_table) {
       color_table[clr] = index++;
       *(unsigned int*)q = clr;
       q += 4;
    }

    assert(index == numColors);

    unsigned int current_clr = pixels[y0*w+x0] & 0xffffff;
    int count = 0;

    for (int j = y0; j <= y1; j++)
      for(int i = x0; i <= x1; i++)
      { unsigned int clr = pixels[j*w + i] & 0xffffff;
        if (clr == current_clr && count < 255)
           count++;
        else
        { *q++ = (unsigned char)count;
          int i = color_table[current_clr];
          assert(i < numColors);
          *q++ = (unsigned char)i;
          numPixels++;
          current_clr = clr;
          count = 1;
         }
      }

    *q++ = (unsigned char)count;
    *q++ = (unsigned char)color_table[current_clr];
    numPixels++;

    bytes = q-buf;
    assert(bytes == 4*table_sz + 2*numPixels + HEADER_SZ);
  }
  else
  { // more than 255 colors

    buf_sz = 4*sz + HEADER_SZ; // pixels + format + header
    buf = new unsigned char[buf_sz];

    unsigned char* q = write_header(buf,RLE,xpos,ypos,width,height,0); 

    numPixels = 0;

    unsigned int current_clr = pixels[y0*w+x0] & 0xffffff;
    int count = 0;

    for (int j = y0; j <= y1; j++)
      for(int i = x0; i <= x1; i++)
      { unsigned int clr = pixels[j*w + i] & 0xffffff;
        if (clr == current_clr && count < 255)
          count++;
        else
        { *(unsigned int*)q = current_clr | (count << 24);
          q += 4;
          numPixels++;
          current_clr = clr;
          count = 1;
         }
      }

    *(unsigned int*)q = current_clr | (count << 24);
    q += 4;

    numPixels++;
/*
    int sum_alpha = 0;
    for(unsigned char* p = buf; p < q; p+=4) sum_alpha += p[3];
    assert(sum_alpha = sz);
*/
    bytes = q-buf;
  }


/*
  float p1 = (100.0*numPixels)/sz;
  float p2 = (100.0*bytes)/(4*sz);
  cout << string("%4d|%4d|%4d %8d --->%6d %4.1f %% --->%6d %4.1f %% ",
                  width,height,numColors,sz,numPixels,p1,bytes,p2);
*/

  assert(bytes <= buf_sz);

  int n = bytes;
  if (deflate) {
    n = compress_buffer(buf,bytes);
    //cout << string(" ---> %5d %4.1f %%", n,(100.0*n)/(4*sz));
  }

  ws.send_data(buf,n);
  delete[] buf;

  bytes_total += 4*sz;
  bytes_sent += n;
  ws.send_text(string("statistics: %lu %d",bytes_total,bytes_sent));

}



//------------------------------------------------------------------------------
// event handling
//------------------------------------------------------------------------------

int CHECK_NEXT_EVENT(int& val, int& x, int& y, unsigned long& t, int msec)
{
  // msec > 0: blocking / timeout
  // msec = 0: blocking / no timeout
  // msec < 0: non-blocking (check)

//cout << "CHECK_NEXT_EVENT" << endl;

  if (!ws.connected()) 
  { sleep(1);
    cout << "WS: SOCKET NOT CONNECTED --> EXIT" << endl;
    exit(0);
    val = 0;
    return display_event;
  }

  int e = no_event;

  if (msec < 0) {
    // non blocking (set timeout to 1 msec)
    msec = 1;
  }


  if (msec > 0 && !ws.poll(msec))
  { //cout << "TIMEOUT: msec = " << msec << endl;
    return no_event;
   }


  string msg = ws.receive_text();

  //cout << msg << endl;

  if (!msg.starts_with("event:")) 
  { cout << "msg = " << msg << endl;
    if (msg == "")
    { cout << "EMPTY MESSAGE: DISCONNECT" << endl;
      ws.disconnect();
      val = 0;
      return display_event;
     }
    else
     return no_event;
  }

  //cout << "EVENT: " << msg << endl;

  msg = msg.replace("event: ","");

  string A[5];
  int num = msg.break_into_words(A,5);

  assert(num == 5);

  string event = A[0];
  val = atoi(A[1]);
  x = atoi(A[2]);
  y = atoi(A[3]);
  t = atol(A[4]);

  if (event == "exit") {
     cout << "WS EXIT EVENT: (DISCONNECT)" << endl;
     ws.disconnect();
     val = 0;
     return display_event;
  }

  if (event == "display")  {
   int dpi = val;
   int w = x;
   int h = y;
   int f = t;

   cout << string("WS DISPLAY EVENT: %d x %d  dpi: %d  format: %d",w,h,dpi,f);
   cout << endl;

   format = f & 0x0f;
   compression = (f >> 4)& 0x0f;

   cout << "compression = " << compression << endl;

   bytes_total = 0;
   bytes_sent = 0;

   e = display_event;
  }

  if (event == "mousedown") e = button_press_event;
  if (event == "mouseup")   e = button_release_event;
  if (event == "mousemove") e = motion_event;

  if (event == "keypress") {
     e = key_press_event;
     if (val ==  8) val = KEY_BACKSPACE;
     if (val == 13) val = KEY_RETURN;
     if (val == 27) val = KEY_ESCAPE;
     if (val == 37) val = KEY_LEFT;
     if (val == 38) val = KEY_UP;
     if (val == 39) val = KEY_RIGHT;
     if (val == 40) val = KEY_DOWN;
  }

/*
  cout << event_name[e];
  cout << " val: " << val << " x: " << x << " y: " << y << " t: " << t << endl;
*/

  return e;
}


void SET_MOUSE_CURSOR(int id) 
{ static int cursor_id = -1;
  //cout << "SET_MOUSE_CURSOR: id = " << id << endl;
  if (id == cursor_id) return;
  cursor_id = id;
  ws.send_text(string("cursor: %d",id));
}


LEDA_END_NAMESPACE
